/*
Andreas Kuehlmann, (kuehl)
bck.c        : create circuit structure using AND/INVERTER 
created      : 13/04/98
last changes : 
 09/25/98 Andreas Kuehlmann: update for work with Adnan & Malay
 11/20/98 Malay K. Ganai   : various bug fixes
 11/22/98 Andreas Kuehlmann: major overhaul
 11/23/98 Andreas Kuehlmann: added access to children and oututs
 11/29/98 Andreas Kuehlmann: added service routines for bck_alloc_aux and 
                             bck_release_aux
 01/25/99 Malay K. Ganai   : added PIs bck_get_index, bck_get_first_index, bck_get_last_index
                             bck_create_null, bck_is_null
                             changed unsigned short to unsigned
                             fix a bug in bck_alloc_aux
                             bck_is_flag_set and bck_is_inverted is made to return 1/0
 01/26/99 Malay K. Ganai   : updated bck_quit to avoid memory leaks
 02/13/99 Malay K. Ganai   : added bck_set_all_flag & bck_reset_all_flag routine
 02/17/99 Malay K. Ganai   : arg of bck_aux_alloc_service made (void **) from (void *)
                             and also of bck_alloc_aux and bck_release_aux
 02/22/99 Andreas Kuehlmann: fixed bug related to aux_service
 04/28/99 Andreas Kuehlmann: added service routine for compose function
 05/02/99 Andreas Kuehlmann: register individual aux_init routines
 08/26/99 Andreas Kuehlmann: quite some clean-up
 08/27/99 Andreas Kuehlmann: changed the setup for the through node 
 08/27/99 Andreas Kuehlmann: get_through -> get_canonical
 08/27/99 Andreas Kuehlmann: is_through -> is_canonical
 08/30/99 Andreas Kuehlmann: changed the way through nodes are stored
 08/30/99 Andreas Kuehlmann: bck_do_for_all_merged_nodes added
 09/02/99 Andreas Kuehlmann: added service routine for merge
 09/02/99 Andreas Kuehlmann: added service routine for var
 11/01/00 Malay K. Ganai   : # flag field increased from 32->64. changed all related routines
 12/01/00 Malay K. Ganai   : added bck_cofactor
 12/05/00 Malay K. Ganai   : added bck_release_node for garbage collection
*/

#include "bck.h"

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <pthread.h>

#define BCK_NODE_HASH_SIZE      1023 
#define BCK_NODE_TABLE_SIZE     8192

#define BCK_NULL                0
#define BCK_FIRST               1
#define BCK_INVALID             UINT_MAX

#define BCK_INVERT_BIT          ((unsigned)0x80000000) /* inverted edge      */

#define BCK_PLAY_BITS           (BCK_INVERT_BIT)

                                                    /* to access node index        */
#define BCK_IDX(x)              ((unsigned) (x & ~BCK_PLAY_BITS))
#define BCK_PTR(x)              (&bck_nodes[BCK_IDX(x)])

#define BCK_VALID(x)            ((BCK_IDX(x) != BCK_NULL) && (BCK_IDX(x) < bck_num_nodes) && \
				 (!BCK_IS_PASS_THROUGH(x) || (BCK_PTR(x)->out[0] != BCK_IDX(x))))

                                    /* to invert a node and to check for inversion */
#define BCK_INVERT(x)           ((unsigned) (x ^ BCK_INVERT_BIT))
#define BCK_IS_INVERTED(x)      ((x & BCK_INVERT_BIT) != 0 )

#define BCK_OUT_OVERFLOW(x)     ((BCK_OUT_COUNTER_BITS - 1) >= x)
#define BCK_ID_OVERFLOW(x)      ((BCK_NR_ID_BITS - 1) >= x)

#define BCK_AND                 0                   /* various types of nodes      */
#define BCK_CONSTANT_0          1
#define BCK_VARIABLE            2
#define BCK_PASS_THROUGH        3

#define BCK_LEVEL(x)            (BCK_PTR(x)->level)  
#define BCK_TYPE(x)             (BCK_PTR(x)->type)   /* access macros              */
#define BCK_IS_CONSTANT(x)      (BCK_TYPE(x) == BCK_CONSTANT_0)
#define BCK_IS_ZERO(x)          ((BCK_TYPE(x) == BCK_CONSTANT_0) && !BCK_IS_INVERTED(x))
#define BCK_IS_ONE(x)           ((BCK_TYPE(x) == BCK_CONSTANT_0) && BCK_IS_INVERTED(x))
#define BCK_IS_AND(x)           (BCK_TYPE(x) == BCK_AND)
#define BCK_IS_VAR(x)           (BCK_TYPE(x) == BCK_VARIABLE)
#define BCK_IS_PASS_THROUGH(x)  (BCK_TYPE(x) == BCK_PASS_THROUGH)


typedef union {                             /* for clean type conversion */
  void           *ptr;
  BCK             bck;
} bck_conv;


/* global globals */
/* local globals */
static unsigned        num_flag = 0;
static bck_conv        **aux_fields = NULL;
static unsigned         num_aux = 0;
static alloc_srv       *bck_aux_alloc_service;
static unsigned         num_and_service = 0;
static alloc_and_srv   *bck_and_service;
static unsigned         num_var_service = 0;
static alloc_var_srv   *bck_var_service;
static unsigned         num_merge_service = 0;
static alloc_merge_srv *bck_merge_service;
static int             *aux_allocated = NULL;
static unsigned        flags_allocated[2];
static BCK              bck_zero = BCK_FIRST;
static BCK_NODE        *bck_nodes = NULL;
static unsigned         bck_num_allocated = 0;
static unsigned         bck_num_nodes = 0;
static unsigned         bck_nr_variables;
static BCK              bck_hash_table[BCK_NODE_HASH_SIZE]; 
static BCK              bck_constraint;
static unsigned long    bck_equal_calls;
static unsigned long    bck_total_gates;
static unsigned long    bck_normal_vars;
static unsigned long    bck_hash_lookups;
static unsigned long    bck_hash_hits;

/* protos */
static BCK        bck_create_and1(BCK mem, BCK pp1, BCK pp2);
static BCK        bck_create_learn_and1(BCK mem, BCK pp1, BCK pp2);
static BCK        bck_get_or_create_and(BCK mem, BCK  p1, BCK p2);
static BCK        bck_create_node(void);
static BCK        bck_make_zero(void);
static void       bck_connect(BCK from, BCK to);
static void       bck_unconnect(BCK from, BCK to);
static void       bck_update_levels(BCK node);
static BCK        bck_get_and_node1(BCK left, BCK right);
static void       bck_init_hash_tables(void);

static unsigned   bck_hash(const BCK left, const BCK right);
static void       bck_remove_hash(BCK pp);
static void       bck_insert_hash(BCK new);
static void       bck_pre(BCK f, unsigned comp_aux);
static void       bck_post(BCK f, unsigned comp_aux);
static BCK        bck_comp(BCK f, unsigned comp_aux, void (*service)(BCK orig, BCK new));
static int        bck_print_aux(BCK bck, void *parameter);
static void       bck_print_node(BCK bck, unsigned print_flag);
static void       bck_traverse_pre_aux(BCK bck, unsigned visited_flag, void *parameter, void (*service)(BCK node, void *parameter));
static void       bck_traverse_post_aux(BCK bck, unsigned visited_flag, void *parameter, void (*service)(BCK node, void *parameter));
static BCK        bck_cofactor1(BCK f, BCK var, unsigned cofactor_aux);
static void       bck_initialize_nodes(int start, int end);


/* ------------------------------------------------------------------------------- */
void
bck_init(void)
{
  unsigned i;

  aux_fields              = NULL;
  aux_allocated           = NULL;
  num_aux                 = 0;
  bck_aux_alloc_service   = NULL;
  num_merge_service       = 0;
  bck_merge_service       = NULL;
  num_and_service         = 0;
  bck_and_service         = NULL;
  num_var_service         = 0;
  bck_var_service         = NULL;
  flags_allocated[0]      = 0;
  flags_allocated[1]       = 0;
  num_flag                = 0;
  bck_nr_variables        = 0;
  bck_equal_calls         = 0;
  bck_total_gates         = 0;
  bck_normal_vars         = 0;
  bck_hash_lookups        = 0;
  bck_hash_hits           = 0;
  bck_constraint          = bck_create_one();

  bck_num_allocated       = BCK_NODE_TABLE_SIZE;
  bck_num_nodes           = 1;                      /* 0 index is left open as NULL node */
  bck_nodes               = bck_realloc_nodes(NULL, bck_num_allocated);
  bck_initialize_nodes(0,bck_num_allocated);
  
  bck_zero                = bck_make_zero();


  bck_init_hash_tables();


}

/* ------------------------------------------------------------------------------- */
void
bck_quit(void)
{
  unsigned i;
  BCK      bck;

  for(bck = BCK_FIRST; bck < bck_num_nodes; bck++){
    BCK *bck_out = BCK_PTR(bck)->out;
    if(bck_out != NULL) bck_free_bcks(bck_out);
  }
  bck_free_nodes(bck_nodes);

  for(i = 0; i < num_aux; i++) {
    bck_free_convs(aux_fields[i]);
  }
  if(aux_fields    != NULL) bck_free_convptrs(aux_fields);
  if(aux_allocated != NULL) bck_free_auxs(aux_allocated);

  if(bck_aux_alloc_service != NULL) bck_free_aux_srv(bck_aux_alloc_service);
  if(bck_and_service       != NULL) bck_free_and_srv(bck_and_service);
  if(bck_var_service       != NULL) bck_free_var_srv(bck_var_service);
  if(bck_merge_service     != NULL) bck_free_merge_srv(bck_merge_service);

  aux_allocated     = NULL;
  num_aux           = 0;
  num_and_service   = 0;
  num_merge_service = 0;
  flags_allocated[0]   = 0;
  flags_allocated[1]   = 0;

  bck_num_allocated = 0;
  bck_num_nodes     = 0;
  bck_nodes         = NULL;

  bck_aux_alloc_service = NULL;
  bck_and_service       = NULL;
  bck_var_service       = NULL;
  bck_merge_service     = NULL;

  bck_nr_variables = 0;
}

/*------------------------------------------------------------------------------- */
unsigned
bck_register_and_service(void (*service)(BCK p))
{
  int i;
  for(i=0;i<num_and_service;i++){
    if (bck_and_service[i] == NULL){
      bck_and_service[i] = service;
      return i;
    }
  }
  num_and_service++;  
  bck_and_service  = bck_realloc(bck_and_service,
                                 sizeof(alloc_and_srv) * num_and_service);
  bck_and_service[num_and_service-1] = service;
  return num_and_service-1;
}

/* ------------------------------------------------------------------------------- */
void
bck_unregister_and_service(unsigned id)
{
  int i;
  
  for(i=0;i<num_and_service;i++){
    if (i == id){
      bck_and_service[i] = NULL;
      return;
    }
  }
  assert(0);
}

/* ------------------------------------------------------------------------------- */
unsigned
bck_register_var_service(void (*var_service)(BCK node))
{
  unsigned i;

  num_var_service++;
  bck_var_service = bck_realloc_var_srv(bck_var_service, num_var_service);
  bck_var_service[num_var_service-1] = var_service;
  if(bck_var_service[num_var_service-1] != NULL ) {
    for(i = BCK_FIRST; i < bck_num_nodes; i++) {
      if(bck_is_var(i) || bck_is_constant(i)) bck_var_service[num_var_service-1](i);
    }
  }
  return num_var_service-1;
}

/* ------------------------------------------------------------------------------- */
void
bck_unregister_var_service(unsigned id)
{
  int i;
  
  for(i=0;i<num_var_service;i++){
    if (i == id){
      bck_var_service[i] = NULL;
      return;
    }
  }
  assert(0);
}

/* ------------------------------------------------------------------------------- */
unsigned
bck_register_merge_service(void (*merge_service)(BCK canonical_node, BCK merged_node))
{
  unsigned i;

  num_merge_service++;
  bck_merge_service = bck_realloc_merge_srv(bck_merge_service, num_merge_service);
  bck_merge_service[num_merge_service-1] = merge_service;
  return num_merge_service-1;
}

/* ------------------------------------------------------------------------------- */
void
bck_unregister_merge_service(unsigned id)
{
  int i;
  
  for(i=0;i<num_merge_service;i++){
    if (i == id){
      bck_merge_service[i] = NULL;
      return;
    }
  }
  assert(0);
}


/*
  -------------------------------------------------------------------------------
  Note: The code for the "and" alogrithm as presented in the paper
  "On-the-fly Compression of Logical Circuit",IWLS 2000 could not
  be released as it was developed while at IBM, Austin.
  However, implementing the presented pseudo-code of "and" is
  very straight forward. If anyone is interested in knowing specific details, please
  contact me at malay@ece.utexas.edu. 
  -------------------------------------------------------------------------------
*/
BCK
bck_create_and(BCK pp1, BCK pp2)
{
  return bck_create_learn_and1(BCK_NULL, pp1, pp2);
}

/* ------------------------------------------------------------------------------- */
static BCK
bck_create_learn_and1(BCK mem, BCK pp1, BCK pp2)
{
  BCK p1 = bck_get_canonical(pp1);      /* if merged vertex -> get the "real" one */
  BCK p2 = bck_get_canonical(pp2);

  ASSERT(BCK_VALID(pp1));
  ASSERT(BCK_VALID(pp2));

  if(BCK_IS_ZERO(p1))      return p1;
  if(BCK_IS_ZERO(p2))      return p2;
  if(BCK_IS_ONE(p1))       return p2;
  if(BCK_IS_ONE(p2))       return p1;
  if(p1 == p2)             return p1;
  if(p1 == BCK_INVERT(p2)) return bck_zero;

  return bck_get_or_create_and(mem, p1, p2);
}


/* ------------------------------------------------------------------------------- */
static BCK
bck_get_or_create_and(BCK mem, BCK  p1, BCK p2)
{
  BCK      new;
  unsigned i;

  if(BCK_IDX(p1) > BCK_IDX(p2)) {       /* make the operands canonical             */
    BCK tmp = p1; p1 = p2; p2 = tmp;
  }

  if((new = bck_get_and_node(p1, p2)) == BCK_NULL) {
    if(mem != BCK_NULL) { /* memory was provides (in case of forward hashing) */
      ASSERT(!BCK_IS_INVERTED(mem));
      new = mem;
    }
    else {
      new                = bck_create_node();
      BCK_PTR(new)->type = BCK_AND;
      ++bck_total_gates;
    }
    BCK_PTR(new)->left  = p1;
    BCK_PTR(new)->right = p2;
    BCK_PTR(new)->level = MAX(bck_get_level(p1), bck_get_level(p2)) + 1;
    bck_connect(p1, new);
    bck_connect(p2, new);
    bck_insert_hash(new);

    for(i = 0; i < num_and_service; i++) {
      if(bck_and_service[i] != NULL) bck_and_service[i](new);
    }
  }
  else {
    if(mem != BCK_NULL) {
      BCK_PTR(mem)->left  = p1;
      BCK_PTR(mem)->right = p2;
      BCK_PTR(mem)->level = MAX(bck_get_level(p1), bck_get_level(p2)) + 1;
      bck_connect(p1, mem);
      bck_connect(p2, mem);
      bck_insert_hash(mem);

      for(i = 0; i < num_and_service; i++) {
        if(bck_and_service[i] != NULL) bck_and_service[i](new);
      }
    }
  }
  new = bck_get_canonical(new);    /* get the "real" guy */

  return new;
}

/* ------------------------------------------------------------------------------- */
static void
bck_initialize_nodes(int begin, int end)
{
  int i;
  BCK_NODE *new;
  for(i=begin;i<end;i++){
    new            = BCK_PTR(i);
    new->type      = 0;
    new->id        = 0;
    new->level     = 0;
    new->out_count = 0;
    new->flags[0]     = 0;
    new->flags[1]     = 0;
    new->out       = BCK_NULL;
    new->left      = BCK_NULL;
    new->right     = BCK_NULL;
    new->next      = BCK_NULL;
  }
}
    

/* ------------------------------------------------------------------------------- */
static BCK
bck_create_node(void)
{
  unsigned  i, realloc_aux_fields = 0;
  BCK       n = bck_num_nodes++;
  BCK_NODE *new;

  if(bck_num_nodes >= bck_num_allocated) {
    realloc_aux_fields = 1;
    bck_num_allocated += bck_num_allocated >> 2;  /* increase by 25% */
    bck_nodes          = bck_realloc_nodes(bck_nodes, bck_num_allocated);
    bck_initialize_nodes(bck_num_nodes,bck_num_allocated);
  }
  
  for(i = 0; i < num_aux; i++) {
    if(realloc_aux_fields == 1) {
      aux_fields[i] =
        bck_realloc_convs(aux_fields[i], bck_num_allocated);
    }
    aux_fields[i][n].ptr = NULL;
    aux_fields[i][n].bck = BCK_NULL;
  }

  bck_connect(n, n);
  if(bck_aux_alloc_service != NULL) {
    for(i = 0; i < num_aux; i++) {
      if(bck_aux_alloc_service[i] != NULL) {
	bck_aux_alloc_service[i](&(aux_fields[i][n]).ptr);
      }
    }
  }
  return n;
}

/* ------------------------------------------------------------------------------- */
BCK
bck_create_not(BCK in)
{
  ASSERT(BCK_VALID(in));
  return BCK_INVERT(in);
}

/* ------------------------------------------------------------------------------- */
BCK
bck_create_var(void)
{
  BCK      var = bck_create_node();
  unsigned i;

  ++bck_normal_vars;
  BCK_PTR(var)->type = BCK_VARIABLE;
  BCK_PTR(var)->id   = bck_nr_variables++;

  if(bck_var_service != NULL) {
    for(i = 0; i < num_var_service; i++) {
      if(bck_var_service[i] != NULL) bck_var_service[i](var);
    }
  }
  return var;
}

/* ------------------------------------------------------------------------------- */
static BCK
bck_make_zero(void)
{
  BCK zero = bck_create_node();

  ASSERT(BCK_IDX(zero) == 1);
  BCK_PTR(zero)->type = BCK_CONSTANT_0;

  return zero;
}

/* ------------------------------------------------------------------------------- */
BCK
bck_create_zero(void)
{
  return bck_zero;
}

/* ------------------------------------------------------------------------------- */
BCK
bck_create_one(void)
{
  return BCK_INVERT(bck_zero);
}

/* ------------------------------------------------------------------------------- */
BCK
bck_create_null(void)
{
  return BCK_NULL;
}
/* ------------------------------------------------------------------------------- */
BCK
bck_get_constraint(void)
{
  return bck_get_canonical(bck_constraint);
}

/* ------------------------------------------------------------------------------- */
int
bck_set_input_constraint(BCK p)
{
  BCK f = bck_get_canonical(p);

  ASSERT(BCK_VALID(p));

  bck_constraint = bck_create_and(f, bck_constraint);

  if(!BCK_IS_AND(f)) {
    if(BCK_IS_ZERO(f)) return 0;          /* illegal constraint */
    bck_merge(bck_create_one(), f);       /* f is variable      */
  }

  /*
    In the following we have a special trick for constraints of the form:
       a == b
    We can just merge the two nodes instead of computing the function and merging 
    it with one.
   */
  else if(BCK_IS_INVERTED(f)) {
    if((BCK_IS_INVERTED(BCK_PTR(f)->left)) && (BCK_IS_INVERTED(BCK_PTR(f)->right))) {
      BCK_NODE *l = BCK_PTR(BCK_PTR(f)->left);
      BCK_NODE *r = BCK_PTR(BCK_PTR(f)->right);

      if((l->left == BCK_INVERT(r->left)) && (l->right == BCK_INVERT(r->right))) {
	bck_merge(l->left, l->right);
	return 1;
      }
    }
    /* bck_merge(bck_create_one(), f); */
  }
  else {
    if((BCK_IS_INVERTED(BCK_PTR(f)->left)) && (BCK_IS_INVERTED(BCK_PTR(f)->right))) {
      BCK_NODE *l = BCK_PTR(BCK_PTR(f)->left);
      BCK_NODE *r = BCK_PTR(BCK_PTR(f)->right);

      if((l->left == BCK_INVERT(r->left)) && (l->right == BCK_INVERT(r->right))) {
	bck_merge(l->left, BCK_INVERT(l->right));
	return 1;
      }
    }
    /*
      Recursive application for constraints of the form  
           assert((A&B) == 1) => assert(A == 1) & assert(B == 1))
     */
    if(!bck_set_input_constraint(BCK_PTR(f)->left))  return 0;
    if(!bck_set_input_constraint(BCK_PTR(f)->right)) return 0;
  }

  return 1;
}

/* ------------------------------------------------------------------------------- */
int 
bck_do_for_all_merged_nodes(void *parameter, BCK node, 
			    int (*service)(BCK node, BCK merged_node,
					   void *parameter))
{
  BCK tmp = node;

  while(BCK_PTR(tmp)->out[0] != node) {
    ASSERT(BCK_PTR(tmp)->out[0] != BCK_INVERT(node));
    ASSERT(BCK_IS_PASS_THROUGH(BCK_PTR(tmp)->out[0]));
    if(BCK_IS_INVERTED(tmp)) tmp = BCK_INVERT(BCK_PTR(tmp)->out[0]);
    else                     tmp = BCK_PTR(tmp)->out[0];
    if(!service(node, tmp, parameter)) return 0;
  }
  return 1;
}

/* ------------------------------------------------------------------------------- */
BCK
bck_get_through(BCK p)
{
  return bck_get_canonical(p);
}

/* ------------------------------------------------------------------------------- */
BCK
bck_get_canonical(BCK p)
{
  ASSERT(BCK_VALID(p));
  while(BCK_IS_PASS_THROUGH(p)) {
    BCK pp;

    ASSERT(BCK_VALID(p));
    ASSERT(BCK_PTR(p)->out_count > 0);
    pp = BCK_PTR(p)->out[0];
    if(BCK_IS_INVERTED(p)) pp = BCK_INVERT(pp);
    p  = pp;
  }
  return p;
}

/* ------------------------------------------------------------------------------- */
static void
bck_connect(BCK from, BCK to)
{
  BCK_NODE *f = BCK_PTR(from);

  ASSERT(BCK_VALID(from));
  ASSERT(BCK_VALID(to));

  f->out = bck_realloc_bcks(f->out, f->out_count+1);
  f->out[f->out_count++] = to;
}

/* ------------------------------------------------------------------------------- */
static void
bck_unconnect(BCK from, BCK to)
{
  unsigned  i;
  BCK_NODE *f = BCK_PTR(from);
  BCK       t = BCK_IDX(to);

  ASSERT(BCK_VALID(from));
  ASSERT(BCK_VALID(to));
  for(i = 1; i < f->out_count; i++) {
    if(f->out[i] == t) {
      unsigned count = f->out_count - i - 1;
      if(count > 0) memmove(&(f->out[i]), &(f->out[i+1]), count * sizeof(BCK));
      f->out_count--;
      return;
    }
    ASSERT(BCK_IDX(f->out[i]) != t);
  }
  ASSERT(0); /* should never happen */
}

/* ------------------------------------------------------------------------------- */
BCK
bck_merge(BCK pp1, BCK pp2)
{
  BCK      p1 = bck_get_canonical(pp1);
  BCK      p2 = bck_get_canonical(pp2);
  BCK      tmp;
  unsigned i;

  ASSERT(BCK_VALID(pp1));
  ASSERT(BCK_VALID(pp2));
  ASSERT(p1 != BCK_INVERT(p2));

  if(p1 == p2) return p1;

  if(BCK_IDX(p1) > BCK_IDX(p2)) {
    tmp = p2; p2 = p1; p1 = tmp;
  }
  if(BCK_IS_INVERTED(p2)) {
    p1 = BCK_INVERT(p1);
    p2 = BCK_INVERT(p2);
  }

  BCK_PTR(p2)->type = BCK_PASS_THROUGH;
  tmp = p1; 
  while(BCK_IDX(BCK_PTR(tmp)->out[0]) != BCK_IDX(p1)) {
    ASSERT(BCK_PTR(tmp)->out[0] != BCK_INVERT(p1));
    ASSERT(BCK_IS_PASS_THROUGH(BCK_PTR(tmp)->out[0]));
    if(BCK_IS_INVERTED(tmp)) tmp = BCK_INVERT(BCK_PTR(tmp)->out[0]);
    else                     tmp = BCK_PTR(tmp)->out[0];
  }
  if(BCK_IS_INVERTED(tmp)) BCK_PTR(tmp)->out[0] = BCK_INVERT(p2);
  else                     BCK_PTR(tmp)->out[0] = p2;
  BCK_PTR(p2)->out[0] = p1;

  while(bck_get_num_outputs(p2) > 0) {
    BCK new;
    BCK out = bck_get_outputs(p2)[0];

    ASSERT(!BCK_IS_INVERTED(out));

    bck_unconnect(BCK_PTR(out)->left, out);
    bck_unconnect(BCK_PTR(out)->right, out);
    bck_remove_hash(out);

    ASSERT((BCK_IDX(BCK_PTR(out)->left) == p2) || 
	   (BCK_IDX(BCK_PTR(out)->right) == p2));
    new = bck_create_learn_and1(out, BCK_PTR(out)->left, BCK_PTR(out)->right);
    bck_merge(new, out);
  }
  for(i = 0; i < num_merge_service; i++) {
    if(bck_merge_service[i] != NULL) bck_merge_service[i](p1, p2);
  }
  ASSERT(bck_get_num_outputs(p2) == 0);
  ASSERT(bck_get_canonical(p2) == bck_get_canonical(p1));

  return p1;
}



/* ------------------------------------------------------------------------------- */
static void
bck_update_levels(BCK node)
{
  BCK       out, left, right;
  unsigned  i, new_level;

  left  = BCK_PTR(node)->left;
  right = BCK_PTR(node)->right;
  if((new_level = MAX(BCK_PTR(left)->level, BCK_PTR(right)->level) + 1) != BCK_PTR(node)->level) {
    BCK_PTR(node)->level = new_level;

    if(bck_is_canonical(node)) return;

    for(i = 0; i < bck_get_num_outputs(node); i++) {
      out = bck_get_outputs(node)[i];
      bck_update_levels(out);
    }
  }
}

/* ------------------------------------------------------------------------------- */
static void
bck_init_hash_tables(void)
{
  unsigned j;

  for(j = 0; j < BCK_NODE_HASH_SIZE; ++j) {
    bck_hash_table[j] = BCK_NULL;
  }
}

/* ------------------------------------------------------------------------------- */
static unsigned
bck_hash(const BCK left, const BCK right)
{
  ASSERT(BCK_VALID(left));
  ASSERT(BCK_VALID(right));
  return ((((unsigned long) left) << 5) ^ (((unsigned long) right) << 3)) % BCK_NODE_HASH_SIZE;
}

/* ------------------------------------------------------------------------------- */

BCK
bck_get_and_node(BCK p1, BCK p2)
{
  unsigned ix;
  BCK      entry;

  if(BCK_IS_ZERO(p1))      return p1;
  if(BCK_IS_ZERO(p2))      return p2;
  if(BCK_IS_ONE(p1))       return p2;
  if(BCK_IS_ONE(p2))       return p1;
  if(p1 == p2)             return p1;
  if(p1 == BCK_INVERT(p2)) return bck_zero;

  if(BCK_IDX(p1) > BCK_IDX(p2)) {
    BCK tmp = p1; p1 = p2; p2 = tmp;
  }

  return bck_get_and_node1(p1, p2);
}

/*
-------------------------------------------------------------------------------
*/
static BCK
bck_get_and_node1(BCK left, BCK right)
{
  unsigned ix;
  BCK      entry;

  bck_hash_lookups++; 

  ASSERT(BCK_VALID(left));
  ASSERT(BCK_VALID(right));
  ASSERT(BCK_IDX(left) < BCK_IDX(right));

  ix = bck_hash(left, right);

  for(entry = bck_hash_table[ix]; entry > BCK_IDX(right);
      entry = BCK_PTR(entry)->next) {
    if(BCK_PTR(entry)->left == left && BCK_PTR(entry)->right == right) {
      bck_hash_hits++;
      return entry;
    }
  }
  return BCK_NULL;
}

/* ------------------------------------------------------------------------------- */
static void
bck_remove_hash(BCK pp)
{
  BCK      p  = BCK_IDX(pp);
  unsigned ix = bck_hash(BCK_PTR(p)->left, BCK_PTR(p)->right);

  bck_hash_lookups++;

  ASSERT(BCK_VALID(p));

  if(bck_hash_table[ix] == p) {
    bck_hash_table[ix] = BCK_PTR(p)->next;
  }
  else {
    BCK entry = bck_hash_table[ix];

    while(BCK_PTR(entry)->next > p) {
      entry = BCK_PTR(entry)->next;
    }
    ASSERT(BCK_PTR(entry)->next == p);   /* must be found otherwise error */
    BCK_PTR(entry)->next = BCK_PTR(BCK_PTR(entry)->next)->next;
  }
}

/* ------------------------------------------------------------------------------- */
static void
bck_insert_hash(BCK new)
{
  BCK      p  = BCK_IDX(new);
  unsigned ix = bck_hash(BCK_PTR(p)->left, BCK_PTR(p)->right);

  ASSERT(BCK_VALID(new));

  if(bck_hash_table[ix] < p) {
    BCK_PTR(p)->next   = bck_hash_table[ix];
    bck_hash_table[ix] = p;
  }
  else {
    BCK entry = bck_hash_table[ix];

    ASSERT(entry != p);
    while(BCK_PTR(entry)->next > p) {
      entry = BCK_PTR(entry)->next;
    }
    ASSERT(BCK_PTR(entry)->next < p);
    ASSERT(entry > p);

    BCK_PTR(p)->next     = BCK_PTR(entry)->next;
    BCK_PTR(entry)->next = p;
  }
}


/* ------------------------------------------------------------------------------- */
unsigned        
bck_get_level(BCK in)
{
  ASSERT(BCK_VALID(in));
  return BCK_PTR(in)->level;
}

/* ------------------------------------------------------------------------------- */
int             
bck_is_inverted(BCK in)
{
  ASSERT(BCK_VALID(in));
  return BCK_IS_INVERTED(in);
}

/* ------------------------------------------------------------------------------- */
int             
bck_is_constant(BCK in)
{
  ASSERT(BCK_VALID(in));
  return BCK_IS_CONSTANT(in);
}

/* ------------------------------------------------------------------------------- */
int             
bck_is_zero(BCK in)
{
  ASSERT(BCK_VALID(in));
  return BCK_IS_ZERO(in);
}

/* ------------------------------------------------------------------------------- */
int             
bck_is_one(BCK in)
{
  ASSERT(BCK_VALID(in));
  return BCK_IS_ONE(in);
}

/* ------------------------------------------------------------------------------- */
int             
bck_is_var(BCK in)
{
  ASSERT(BCK_VALID(in));
  return BCK_IS_VAR(in);
}

/* ------------------------------------------------------------------------------- */
int             
bck_is_and(BCK in)
{
  ASSERT(BCK_VALID(in));
  return BCK_IS_AND(in);
}

/* ------------------------------------------------------------------------------- */
int             
bck_is_null(BCK in)
{
   return (BCK_IDX(in) == BCK_NULL);
}

/* ------------------------------------------------------------------------------- */
int             
bck_is_canonical(BCK in)
{
  ASSERT(BCK_VALID(in));
  return !BCK_IS_PASS_THROUGH(in);
}


/* ------------------------------------------------------------------------------- */
int
bck_compare_ids(BCK p1, BCK p2)
{
  return BCK_IDX(p1) - BCK_IDX(p2);
}

/* ------------------------------------------------------------------------------- */
int
bck_compare_levels(BCK p1, BCK p2)
{
  return BCK_PTR(p1)->level - BCK_PTR(p2)->level;
}

/* ------------------------------------------------------------------------------- */
int
bck_do_in_topological_order(void *parameter,
			    int (*service)(BCK node, void *parameter))
{
  unsigned i;

  for(i = BCK_FIRST+1; i < bck_num_nodes; i++) {
    if(!service(i, parameter)) return 0;
  }
  return 1;
}

/* ------------------------------------------------------------------------------- */
unsigned
bck_alloc_flag(void)
{
  unsigned i;
  unsigned mask;

  assert(num_flag < BCK_FLAGS_BITS);
  num_flag++;

  for(i = 0; i < BCK_FLAGS_BITS; i++) {
    mask = (1 << (i&31));
    if((flags_allocated[i>>5] & mask) == 0) {
      flags_allocated[i>>5] |= mask;
      return i;
    }
  }
  ASSERT(0); /* should not happen */
}

/* ------------------------------------------------------------------------------- */
unsigned
bck_alloc_flag_reset(void)
{
  unsigned mask = bck_alloc_flag();

  bck_reset_all_flag(mask);
  return mask;
}

/* ------------------------------------------------------------------------------- */
void
bck_release_flag(unsigned id)
{
  unsigned mask = 1;
  mask <<= (id & 31);
  ASSERT(flags_allocated[id>>5] & mask);
  flags_allocated[id>>5] &= ~mask;
  num_flag--;
  
}

/* ------------------------------------------------------------------------------- */
void
bck_set_flag(BCK in, unsigned id)
{
  unsigned mask = 1;
  mask <<= (id & 31);  
  ASSERT(BCK_VALID(in));
  BCK_PTR(in)->flags[id>>5] |= mask;
}

/* ------------------------------------------------------------------------------- */
void
bck_set_all_flag(unsigned id)
{
  unsigned i;
  unsigned mask = 1;
  mask <<= (id & 31);

  for(i = BCK_FIRST; i < bck_num_nodes; i++) {
    BCK_PTR(i)->flags[id>>5] |= mask;
  }
}

/* ------------------------------------------------------------------------------- */
void
bck_reset_all_flag(unsigned id)
{
  unsigned i;
  unsigned mask = 1;
  mask <<= (id & 31);
  for(i = BCK_FIRST; i < bck_num_nodes; i++) {
    BCK_PTR(i)->flags[id>>5] &= ~mask;
  }
}

/* ------------------------------------------------------------------------------- */
void
bck_reset_flag(BCK in, unsigned id)
{
  unsigned mask = 1;
  mask <<= (id & 31);
  ASSERT(BCK_VALID(in));
  BCK_PTR(in)->flags[id>>5] &= ~mask;
}


/* ------------------------------------------------------------------------------- */
void
bck_toggle_flag(BCK in, unsigned id)
{
  unsigned mask = 1;
  mask <<= (id & 31);  
  ASSERT(BCK_VALID(in));
  BCK_PTR(in)->flags[id>>5] ^= mask;
}

/* ------------------------------------------------------------------------------- */
int
bck_is_flag_set(BCK in, unsigned id)
{
  unsigned mask = 1;
  mask <<= (id & 31);  
  ASSERT(BCK_VALID(in));
  return ((BCK_PTR(in)->flags[id>>5] & mask) != 0);
}

/* ------------------------------------------------------------------------------- */
unsigned
bck_alloc_aux(void (*service)(void **aux_field))
{
  unsigned i, j;

  for(i = 0; i < num_aux; i++) {
    if(aux_allocated[i] == 0) {
      aux_allocated[i]             = 1;
      bck_aux_alloc_service[i]     = service;

      ASSERT(aux_fields[i] != NULL);
      ASSERT(bck_aux_alloc_service != NULL);

      if(bck_aux_alloc_service[i] != NULL) {
	for(j = 0; j < bck_num_nodes; j++) {
          aux_fields[i][j].ptr = NULL;
          bck_aux_alloc_service[i](&(aux_fields[i][j].ptr));
        }
      }
      else {
        for(j = 0; j < bck_num_allocated; j++) {
          aux_fields[i][j].ptr = NULL;
          aux_fields[i][j].bck = BCK_NULL;
        }
      }
      return i;
    }
  }
  num_aux++;
  aux_allocated             = bck_realloc_auxs(aux_allocated, num_aux);
  aux_fields                = bck_realloc_convptrs(aux_fields, num_aux);
  bck_aux_alloc_service     = bck_realloc_aux_srv(bck_aux_alloc_service, num_aux);
  aux_allocated[num_aux-1]  = 1;
  aux_fields[num_aux-1]     = bck_alloc_convs(bck_num_allocated);

  bck_aux_alloc_service[num_aux-1]     = service;
  if (bck_aux_alloc_service[num_aux-1] != NULL ) {
    for(j = 0; j < bck_num_nodes; j++) {
      aux_fields[num_aux-1][j].ptr = NULL;      
      bck_aux_alloc_service[num_aux-1](&(aux_fields[num_aux-1][j].ptr));
    }
  }
  else {
    for(j = 0; j < bck_num_allocated; j++) {
      aux_fields[num_aux-1][j].ptr = NULL;
      aux_fields[num_aux-1][j].bck = BCK_NULL;
    }
  }
  return num_aux-1;
}

/* ------------------------------------------------------------------------------- */
void
bck_release_aux(unsigned id, void (*service)(void *aux_field))
{
  unsigned i;
  ASSERT(id < num_aux);
  ASSERT(aux_allocated[id]);

  if(service != NULL) {
    for(i = 0; i < bck_num_nodes; i++) service(&(aux_fields[id][i].ptr));
  }
  aux_allocated[id] = 0;
}

/* ------------------------------------------------------------------------------- */
void
bck_set_aux(BCK in, unsigned id, void *aux)
{
  ASSERT(id < num_aux);
  ASSERT(aux_fields[id] != NULL);
  ASSERT(BCK_VALID(in));

  aux_fields[id][BCK_IDX(in)].ptr = aux;
}

/* ------------------------------------------------------------------------------- */
void*
bck_get_aux(BCK in, unsigned id)
{
  ASSERT(id < num_aux);
  ASSERT(aux_fields[id] != NULL);
  ASSERT(BCK_VALID(in));

  return aux_fields[id][BCK_IDX(in)].ptr;
}

/* ------------------------------------------------------------------------------- */
void*
bck_get_aux_address(BCK in, unsigned id)
{
  ASSERT(id < num_aux);
  ASSERT(aux_fields[id] != NULL);
  ASSERT(BCK_VALID(in));

  return &(aux_fields[id][BCK_IDX(in)].ptr);
}

/* ------------------------------------------------------------------------------- */
void
bck_set_aux_bck(BCK in, unsigned id, BCK aux)
{
  ASSERT(id < num_aux);
  ASSERT(aux_fields[id] != NULL);
  ASSERT(BCK_VALID(in));

  aux_fields[id][BCK_IDX(in)].bck = aux;
}

/* ------------------------------------------------------------------------------- */
BCK
bck_get_aux_bck(BCK in, unsigned id)
{
  ASSERT(id < num_aux);
  ASSERT(aux_fields[id] != NULL);
  ASSERT(BCK_VALID(in));

  return aux_fields[id][BCK_IDX(in)].bck; 
}
/* ------------------------------------------------------------------------------- */
void
bck_set_fanin_cone_flag(BCK f, unsigned flag)
{
  BCK p = BCK_IDX(f);

  if(BCK_IS_AND(p)) {
    if(!bck_is_flag_set(p, flag)) {
      bck_set_fanin_cone_flag(BCK_PTR(p)->left, flag);
      bck_set_fanin_cone_flag(BCK_PTR(p)->right, flag);
    }
  }
  bck_set_flag(p, flag);
}

/* ------------------------------------------------------------------------------- */
void
bck_reset_fanin_cone_flag(BCK f, unsigned flag)
{
  BCK p = BCK_IDX(f);

  if(BCK_IS_AND(p)) {
    if(bck_is_flag_set(p, flag)) {
      bck_reset_fanin_cone_flag(BCK_PTR(p)->left, flag);
      bck_reset_fanin_cone_flag(BCK_PTR(p)->right, flag);
    }
  }
  bck_reset_flag(p, flag);
}

/* ------------------------------------------------------------------------------- */
void
bck_set_fanin_cone_aux(BCK f, unsigned aux, void *init_value)
{
  BCK p = BCK_IDX(f);

  if(BCK_IS_AND(p)) {
    if(bck_get_aux(p, aux) != init_value) {
      bck_set_fanin_cone_aux(BCK_PTR(p)->left, aux, init_value);
      bck_set_fanin_cone_aux(BCK_PTR(p)->right, aux, init_value);
    }
  }
  bck_set_aux(p, aux, init_value);
}

/* ------------------------------------------------------------------------------- */
static void
bck_pre(BCK f, unsigned comp_aux)
{
  bck_set_fanin_cone_aux(f,comp_aux,(void *)BCK_INVALID);
}

/* ------------------------------------------------------------------------------- */
static void
bck_post(BCK f, unsigned comp_aux)
{
  bck_set_fanin_cone_aux(f,comp_aux,(void *)BCK_NULL);
}

/* ------------------------------------------------------------------------------- */
static BCK
bck_comp(BCK f, unsigned comp_aux, void (*service)(BCK orig, BCK new))
{
  BCK b, p = BCK_IDX(f);

  if(BCK_IS_CONSTANT(f))                                                return f;
  if(BCK_IS_VAR(f) && (bck_get_aux(p, comp_aux) == (void*)BCK_INVALID)) return f;

  if(bck_get_aux_bck(p, comp_aux) == BCK_INVALID) {
    b = bck_create_and(bck_comp(BCK_PTR(p)->left, comp_aux, service), 
		       bck_comp(BCK_PTR(p)->right, comp_aux, service));
    bck_set_aux_bck(p, comp_aux, b);
    if (service != NULL) {
      service(p, b);
    }
  }

  b = bck_get_aux_bck(p, comp_aux);
  if(BCK_IS_INVERTED(f)) return BCK_INVERT(b);
  else                   return b;
}

/* ------------------------------------------------------------------------------- */
BCK*
bck_compose_vec(BCK *dest, BCK *out_f, unsigned length_out,
		BCK *in_v, BCK *in_h, unsigned length_in,
		void (*service)(BCK orig, BCK new))
{
  BCK      b;
  unsigned i, comp_aux;

  comp_aux = bck_alloc_aux(NULL);

  for(i = 0; i < length_out; i++) {
    bck_pre(bck_get_canonical(out_f[i]), comp_aux);
  }
  for(i = 0; i < length_in; i++) {
    BCK v = bck_get_canonical(in_v[i]);

    ASSERT(!BCK_IS_INVERTED(v));
    ASSERT(BCK_IS_VAR(v));

    b = bck_get_canonical(in_h[i]);
    bck_set_aux_bck(v, comp_aux, b);
    if (service != NULL) {
      service(v, b);
    }
  }

  for(i = 0; i < length_out; i++) {
    dest[i] = bck_comp(bck_get_canonical(out_f[i]), comp_aux, service);
  }
  for(i = 0; i < length_out; i++) {
    bck_post(bck_get_canonical(out_f[i]), comp_aux);
  }

  bck_release_aux(comp_aux, NULL);
  return dest;
}

/* ------------------------------------------------------------------------------- */
BCK
bck_get_index(BCK in)
{
  ASSERT(BCK_VALID(in));
  return BCK_IDX(in);
}

/* ------------------------------------------------------------------------------- */
BCK
bck_get_regular(BCK in)
{
  ASSERT(BCK_VALID(in));
  return BCK_IDX(in);
}

/* ------------------------------------------------------------------------------- */
BCK
bck_get_left_child(BCK in)
{
  return BCK_PTR(in)->left;
}

/* ------------------------------------------------------------------------------- */
BCK
bck_get_right_child(BCK in)
{
  return BCK_PTR(in)->right;
}

/* ------------------------------------------------------------------------------- */
unsigned
bck_get_num_outputs(BCK in)
{
  return BCK_PTR(in)->out_count-1;
}

/* ------------------------------------------------------------------------------- */
void
bck_traverse_pre(BCK bck, void *parameter, void (*service)(BCK node, void *parameter))
{
  unsigned visited_flag;

  ASSERT(!bck_is_canonical(bck));

  visited_flag = bck_alloc_flag();
  bck_reset_all_flag(visited_flag);
  bck_traverse_pre_aux(bck, visited_flag, parameter, service);
  bck_release_flag(visited_flag);
  return;
}

/* ------------------------------------------------------------------------------- */
static void
bck_traverse_pre_aux(BCK bck, unsigned visited_flag, void *parameter, void (*service)(BCK node, void *parameter))
{
  if(bck_is_flag_set(bck, visited_flag)) return;
  bck_set_flag(bck, visited_flag);

  if (service)
    service(bck, parameter);
  if(!BCK_IS_AND(bck)) return;

  bck_traverse_pre_aux(BCK_PTR(bck)->left, visited_flag, parameter, service);
  bck_traverse_pre_aux(BCK_PTR(bck)->right, visited_flag, parameter, service);

  return;
}

/* ------------------------------------------------------------------------------- */
void
bck_traverse_post(BCK bck, void *parameter, void (*service)(BCK node, void *parameter))
{
  unsigned visited_flag;

  ASSERT(!bck_is_canonical(bck));

  visited_flag = bck_alloc_flag();
  bck_reset_all_flag(visited_flag);
  bck_traverse_post_aux(bck, visited_flag, parameter, service);
  bck_release_flag(visited_flag);
  return;
}

/* ------------------------------------------------------------------------------- */
static void
bck_traverse_post_aux(BCK bck, unsigned visited_flag, void *parameter, void (*service)(BCK node, void *parameter))
{
  ASSERT(!bck_is_canonical(bck));

  if(bck_is_flag_set(bck, visited_flag)) return;
  bck_set_flag(bck, visited_flag);
  
  if(!BCK_IS_AND(bck)) {
    if (service)
      service(bck, parameter);
    return;
  }

  bck_traverse_pre_aux(BCK_PTR(bck)->left, visited_flag, parameter, service);
  bck_traverse_pre_aux(BCK_PTR(bck)->right, visited_flag, parameter, service);

  if (service)
    service(bck, parameter);

  return;
}
 
/* ------------------------------------------------------------------------------- */
BCK*
bck_get_outputs(BCK in)
{
  return &(BCK_PTR(in)->out[1]);
}
/* ------------------------------------------------------------------------------- */
void
bck_stat(FILE *out_file)
{
  int           i;
  unsigned long num_fanout_node = 0;

  fprintf(out_file,"bck: %lu hash lookups\n",     bck_hash_lookups);
  fprintf(out_file,"bck: %lu hash hits\n",        bck_hash_hits);
  
  fprintf(out_file,"bck: %lu normal variables\n", bck_normal_vars);
  fprintf(out_file,"bck: %lu and nodes\n",        bck_total_gates);
  fprintf(out_file,"bck: %lu bck nodes\n",        bck_num_nodes);


}

/* ------------------------------------------------------------------------------- */
BCK
bck_get_unique_phase(BCK in)
{
  return BCK_IDX(in);
}

/* ------------------------------------------------------------------------------- */
void
bck_print_fanin_cone(
BCK node, unsigned print_flag)
{
  if ( bck_is_flag_set(node,print_flag)){
    return;
  }
  bck_set_flag(node,print_flag);
  if (bck_is_var(node)){
    printf("VAR %d\n",bck_get_index(node));
    return;
  }else {
    printf("(%d) = ",bck_get_index(node));
    if ( bck_is_inverted(bck_get_left_child(node))){
      printf("!");
    }
    printf("(%d) * ",bck_get_index(bck_get_left_child(node)));

    if ( bck_is_inverted(bck_get_right_child(node))){
      printf("!");
    }
    printf("(%d)\n",bck_get_index(bck_get_right_child(node)));
  }

  bck_print_fanin_cone(bck_get_left_child(node),print_flag);
  bck_print_fanin_cone(bck_get_right_child(node),print_flag);
}

/* ------------------------------------------------------------------------------- */
unsigned
bck_get_num_nodes(void)
{
  return bck_num_nodes;
}  

/* ------------------------------------------------------------------------------- */
BCK
bck_get_last_node(void)
{
  return bck_num_nodes-1;
}  


/*--------------------------------------------------------------------------*/
BCK
bck_next_node(BCK p)
{
  if (BCK_IDX(p) >= bck_num_nodes) {
    return BCK_NULL;
  }
  return BCK_IDX(p)+1;
}

/*--------------------------------------------------------------------------*/
void
bck_release_node(BCK pp)
{
  int i;
  int ix;
  BCK p = BCK_IDX(pp);

  /* remove the bck nodes with id >= p */
  for(ix = 0; ix < BCK_NODE_HASH_SIZE; ++ix) {
    BCK entry = bck_hash_table[ix];
    while(entry && entry >= p) {
      entry = BCK_PTR(entry)->next;
    }
    bck_hash_table[ix] = entry;
  }
  
  for(i=p ; i < bck_num_nodes; i++) {
    BCK_NODE *node = BCK_PTR(i);
    BCK *out = node->out;
    if (BCK_IS_VAR(i)) {
      bck_normal_vars--;
      bck_nr_variables--;
    }else {
      if (BCK_IDX(node->left) < p) { 
        bck_unconnect(node->left, i);
      }
      if (BCK_IDX(node->right) < p) {
        bck_unconnect(node->right, i);
      }
    }
    bck_initialize_nodes(i,i+1);
    //Re-use the out space
    node->out = out;
  }
  bck_num_nodes = BCK_IDX(p);
}
    

/*--------------------------------------------------------------------------*/
BCK
bck_cofactor(BCK f, BCK var, unsigned cofactor_aux)
{
  BCK cofactor;
  ASSERT(BCK_IS_VAR(var));
  bck_set_fanin_cone_aux(f,cofactor_aux,(void *)BCK_INVALID);
  cofactor = bck_cofactor1(f,var,cofactor_aux);
  bck_set_fanin_cone_aux(f,cofactor_aux,(void *)BCK_NULL);
  return cofactor;
}

/*--------------------------------------------------------------------------*/
static BCK
bck_cofactor1(BCK f, BCK var, unsigned cofactor_aux)
{
  bck_conv conv;

  if (var == f) {
    return bck_create_one();
    
  } else if (var == BCK_INVERT(f)) {
    return bck_create_zero();
    
  }else if (BCK_IS_AND(f)){    
    if((conv.ptr = bck_get_aux(f, cofactor_aux)) == (void*)BCK_INVALID) {
      conv.bck = bck_create_and(bck_cofactor1(BCK_PTR(f)->left,var,
                                              cofactor_aux), 
                                bck_cofactor1(BCK_PTR(f)->right,var,
                                              cofactor_aux));
      bck_set_aux(f, cofactor_aux, conv.ptr);
    }
    if(BCK_IS_INVERTED(f)) return BCK_INVERT(conv.bck);
    else                   return conv.bck;
  
  }
  return f;
}



