/*
Andreas Kuehlmann
bck.h        : create circuit structure for AND/INVERTER graph
created      : 13/04/98
last changes : 
 09/25/98 Andreas Kuehlmann: update for work with Adnan & Malay
 11/22/98 Andreas Kuehlmann: major overhaul
 11/23/98 Andreas Kuehlmann: changed BCK to unsigned
 11/23/98 Andreas Kuehlmann: added access to children and outputs
 11/29/98 Andreas Kuehlmann: added service routines for bck_alloc_aux and 
                             bck_release_aux
 01/25/99 Malay K. Ganai   : added PIs bck_get_index, bck_get_first_index, bck_get_last_index
                             bck_create_null, bck_is_null
                             changed unsigned short to unsigned
                             fix a bug in bck_alloc_aux
                             bck_is_flag_set and bck_is_inverted is made to return 1/0
 02/13/99 Malay K. Ganai   : added bck_set_all_flag & bck_reset_all_flag routine
 02/17/99 Malay K. Ganai   : arg of bck_aux_alloc_service made (void **) from (void *) and 
                             also in bck_alloc_aux, bck_release_aux.
 04/28/99 Andreas Kuehlmann: added service routine for compose function
 08/26/99 Andreas Kuehlmann: quite some clean-up
 08/27/99 Andreas Kuehlmann: get_through -> get_canonical
 08/27/99 Andreas Kuehlmann: is_through -> is_canonical
 08/30/99 Andreas Kuehlmann: bck_do_for_all_merged_nodes added
 09/02/99 Andreas Kuehlmann: added service routine for merge
 09/02/99 Andreas Kuehlmann: added service routine for var
 09/06/99 Andreas Kuehlmann: added bck_alloc_flag_reset
 09/09/99 Andreas Kuehlmann: added bck_get_num_nodes and bck_get_num_vars
 11/01/00 Malay K. Ganai   : # flag field increased from 32->64. changed all related routines
 12/01/00 Malay K. Ganai   : added bck_cofactor
 12/05/00 Malay K. Ganai   : added bck_release_node for garbage collection
 
*/

#ifndef BCK_H
#define BCK_H

#include <malloc.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef NO_ASSERT
#define ASSERT(x) 
#else
#define ASSERT(x) assert(x)
#endif
#define BCK_NULL 0
#define bck_alloc(x) malloc(x)
#define bck_realloc(x, y) realloc(x, y)
#define bck_free(x) free(x)
#define bck_realloc_bcks(x,y) realloc(x,sizeof(BCK)*(y))
#define bck_free_bcks(x) free((void*)(x))
#define bck_realloc_nodes(x,y) realloc(x,sizeof(BCK_NODE)*(y))
#define bck_free_nodes(x) free((void*)(x))
#define bck_realloc_convptrs(x,y) realloc(x,sizeof(bck_conv *)*(y))
#define bck_free_convptrs(x) free((void*)(x))
#define bck_realloc_convs(x,y) realloc(x,sizeof(bck_conv)*(y))
#define bck_alloc_convs(y) malloc(sizeof(bck_conv)*(y))
#define bck_free_convs(x) free((void*)(x))
#define bck_realloc_auxs(x,y) realloc(x,sizeof(int)*(y))
#define bck_free_auxs(x) free((void*)(x))
#define bck_realloc_aux_srv(x,y) realloc(x,sizeof(alloc_srv)*(y))
#define bck_free_aux_srv(x) free((void*)(x))
#define bck_realloc_and_srv(x,y) realloc(x,sizeof(alloc_and_srv)*(y))
#define bck_free_and_srv(x) free((void*)(x))
#define bck_realloc_var_srv(x,y) realloc(x,sizeof(alloc_var_srv)*(y))
#define bck_free_var_srv(x) free((void*)(x))
#define bck_realloc_merge_srv(x,y) realloc(x,sizeof(alloc_merge_srv)*(y))
#define bck_free_merge_srv(x) free((void*)(x))


#define MAX(x,y) ((x)>(y)? (x) : (y))


#define BCK_NR_TYPE_BITS        2                   /* for type recognition        */
#define BCK_OUT_COUNTER_BITS    30                  /* counter for fanout array    */
#define BCK_FLAGS_BITS          64
#define BCK_NR_ID_BITS          32                  /* id counter for inputs       */
#define BCK_NR_LEVEL_BITS       32                  /* node level                  */

typedef unsigned BCK;

typedef struct bck_node {
  unsigned        type:         BCK_NR_TYPE_BITS;     /* encodes the node type     */
  unsigned        out_count:    BCK_OUT_COUNTER_BITS; /* # fanouts                 */  
  unsigned        flags[2];                           /* # of flags                */
  unsigned        id:           BCK_NR_ID_BITS;       /* id used for inputs, ...   */
  unsigned        level:        BCK_NR_LEVEL_BITS;    /* node level                */
  BCK             left;                               /* left operand              */
  BCK             right;                              /* right operand             */
  BCK             next;                               /* for hash collision chain  */
  BCK            *out;                                /* fanout structure          */
} BCK_NODE;

typedef void          (*alloc_srv)(void **t);
typedef void          (*alloc_and_srv)(BCK node);
typedef void          (*alloc_var_srv)(BCK node);
typedef void          (*alloc_merge_srv)(BCK canonical_node, BCK merged_node);
typedef void          (*BCK_PRINT)(const char *, ...);

                                            /* Prototypes for interface functions. */
extern void            bck_init(void);
extern void            bck_quit(void);

                                            /* construction routines               */
extern BCK             bck_create_and(BCK left, BCK right);
extern BCK             bck_create_not(BCK in);
extern BCK             bck_create_var(void);
extern BCK             bck_create_zero(void);
extern BCK             bck_create_one(void);
extern BCK             bck_create_null(void);
                                            /* set two nodes functional equal      */
extern BCK             bck_merge(BCK p1, BCK p2);

                                            /* vector compose operation            */
extern BCK            *bck_compose_vec (BCK *dest, BCK *out_f, unsigned length_out, 
					BCK *in_v, BCK *in_h, unsigned length_in,
					void (*service)(BCK orig, BCK new));

                                            /* set input constraints               */
                                            /* returns 0 if f == zero              */
extern int             bck_set_input_constraint(BCK f);

                                            /* get certain node infos              */
extern BCK             bck_get_regular(BCK in);

extern BCK             bck_get_index(BCK in);
extern BCK             bck_get_level(BCK in);
extern BCK             bck_get_left_child(BCK in);
extern BCK             bck_get_right_child(BCK in);
extern unsigned        bck_get_num_outputs(BCK in);
extern BCK            *bck_get_outputs(BCK in);
extern BCK             bck_get_and_node(BCK left, BCK right);
extern BCK             bck_get_unique_phase(BCK in);
extern BCK             bck_get_canonical(BCK p);

extern int             bck_is_inverted(BCK in);
extern int             bck_is_constant(BCK in);
extern int             bck_is_zero(BCK in);
extern int             bck_is_one(BCK in);
extern int             bck_is_var(BCK in);
extern int             bck_is_and(BCK in);
extern int             bck_is_null(BCK in);
extern int             bck_is_canonical(BCK in);
extern unsigned        bck_get_num_nodes(void);
extern unsigned        bck_get_num_vars(void);
extern void           bck_set_fanin_cone_flag(BCK p,unsigned flag);
extern void           bck_reset_fanin_cone_flag(BCK p,unsigned flag);
extern void           bck_set_fanin_cone_aux(BCK p,unsigned aux, void *val);
extern int             bck_compare_levels(BCK p1, BCK p2);
extern BCK             bck_get_through(BCK p);
/* 
   exec routine "service" for each node in topological order 
   parameter "parameter" gets passed to that routine
   return value for service routine: 1 continue the service for all nodes
                                     0 stop and return
   will be returned from main routine
*/
extern int             bck_do_in_topological_order(void *parameter,
						   int (*service)(BCK node,
								  void *parameter));
extern int             bck_do_for_all_merged_nodes(void *parameter, BCK node,
						   int (*service)(BCK node, 
								  BCK merged_node,
								  void *parameter));


/* flag operations on a node                                                       */
extern unsigned        bck_alloc_flag(void);                  /* register a flag   */
extern unsigned        bck_alloc_flag_reset(void);            /* resets flags      */
extern void            bck_release_flag(unsigned id);         /* release it        */
extern void            bck_set_flag(BCK in, unsigned id);
extern void            bck_reset_flag(BCK in, unsigned id);
extern void            bck_toggle_flag(BCK in, unsigned id);
extern int             bck_is_flag_set(BCK in, unsigned id);
extern void            bck_set_all_flag(unsigned id);
extern void            bck_reset_all_flag(unsigned id);
extern unsigned        bck_register_and_service(void (*and_service)(BCK node));
extern unsigned        bck_register_var_service(void (*var_service)(BCK node));
extern unsigned        bck_register_merge_service(void (*merge_service)(BCK canonical_node, BCK merged_node));

/* Aux field handling, the alloc and release routine take a service
   routine which is called for each field. This can be used for
   initialization. */
extern unsigned        bck_alloc_aux(void (*service)(void **aux_field));
extern void            bck_release_aux(unsigned id, void (*service)(void *aux_field));
extern void            bck_set_aux(BCK in, unsigned id, void *aux);
extern void           *bck_get_aux(BCK in, unsigned id);
extern void           *bck_get_aux_address(BCK in, unsigned id);
extern void            bck_set_aux_bck(BCK in, unsigned id, BCK aux);
extern BCK             bck_get_aux_bck(BCK in, unsigned id);
extern BCK            bck_cofactor(BCK f, BCK var, unsigned cofactor_aux);
extern void           bck_unregister_and_service(unsigned id);
extern void           bck_unregister_var_service(unsigned id);
extern void           bck_unregister_merge_service(unsigned id);
extern void           (*bck_error)(const char *, ...);
extern void           (*bck_debug)(const char *, ...);
extern void           (*bck_bug)(const char *, ...);
extern void           (*bck_exit)(unsigned);
extern void            bck_stat(FILE *out_file);
extern void            bck_print_fanin_cone(BCK node, unsigned print_flag);
extern void            bck_print(BCK bck);
extern BCK             bck_get_last_node(void);
extern BCK             bck_next_node(BCK p);
extern void            bck_release_node(BCK p);
#endif /* BCK_H */



