/**CFile**********************************************************************

  FileName                   [siva_atpg_solver.c]

  PackageName                [siva_core]

  Synopsis                   [Used to translate problem to SAT problem]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modifed: 01/11/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]
******************************************************************************/


#include "siva.h"

#define TEST 0
#define TEST_CNF "test.cnf"
#define SIVA_SET_BACK_TRACK_LIMIT(x) (siva_configure_grasp(_GRP_BACKTRACK_LIMIT_,(x)))
#define SIVA_SET_TIME_LIMIT(x) (siva_configure_grasp(_GRP_TIME_LIMIT_,(x)))

#define SIVA_SET_PREPROC_RECORDING(x) (siva_configure_grasp(_GRP_PREPROC_RECORDING_,(x)))

#define SIVA_SET_CONFLICT_LIMIT(x) (siva_configure_grasp(_GRP_CONFLICT_LIMIT_,(x)))
#define SIVA_SET_DEDUCTION_LEVEL(x) (siva_configure_grasp(_GRP_DEDUCTION_LEVEL_,(x)))
#define SIVA_SET_LEARNING_LEVEL(x) (siva_configure_grasp(_GRP_PREPROC_LEVEL_,(x)))
#define SIVA_SET_BACKTRACKING_STRATEGY(x) (siva_configure_grasp(_GRP_BACKTRACKING_STRATEGY_,(x)))

static var_set_t * siva_support_latch_vector;
static unsigned siva_var_id_aux;
static array_t *siva_clause_array;
static unsigned siva_num_clause;
static unsigned siva_num_var_id;
static unsigned siva_num_goal_remained;
static unsigned siva_grasp_conflict_limit;
static unsigned siva_grasp_back_track_limit;
static unsigned siva_grasp_time_limit;
static int siva_grasp_options[2*GRP_MAX_OPTION_SIZE];
static int **siva_pi_var_ids;
static int **siva_latch_var_ids;
static var_set_t **siva_witness_input_vectors;
static var_set_t **siva_witness_state_vectors;
static char *siva_witness_values;
static array_t *siva_atpg_target_array;


static void siva_fill_clause(array_t *node_array, int num_node, int reset_id);
static void siva_del_last_num_clause(int num_clause);
static void siva_fill_clause_aux(SIVA_NODE bck);
static void siva_dump_clause(char *filename, int *clauses, int num_clause, int num_var_id);
static void siva_clause_stat(void);
static void siva_set_interface_clause(var_set_t *support_latch_vector,
                                      int unfold_depth);
static void siva_set_state_vector_clause(var_set_t* state_vector);
static void siva_unset_state_vector_clause(void);
static void siva_configure_grasp(int grasp_mode, int value );
static void siva_configure_grasp_default(void);
static void siva_atpg_solver(int siva_num_clause, int num_var_id,
                             int *clauses, char *witness);

static void siva_set_objective_clause1(int var_id0, int var_id1, int var_id2);
static void siva_build_witness_input_sequence(int **pi_var_ids,
                                              int unfold_depth);
static void siva_build_witness_state_sequence(int **latch_var_ids,
                                              int unfold_depth);
static void siva_atpg_verify_witness(SIVA_NODE *targets,
                                     int *target_values,
                                     int num_target);
static void siva_sym_build_witness_sequence(array_t *input_nodes_array,
                                            int sym_depth,
                                            SIVA_STATE_WRAPPER *state_wrapper);
static void siva_set_et(SIVA_ET *et, int unfold_depth);
static void siva_atpg_grasp_solver(int num_clause, int num_var_id,
                                   int *clauses, char *witness);
static void siva_atpg_direct_search(int *latch_var_ids);

/*---------------------------------------------------------------------------*/
void
siva_init_atpg_solver(void)
{
  int i;
  siva_var_id_aux = bck_alloc_aux(siva_init_aux);
  siva_clause_array = array_alloc(int,64);
  siva_num_clause = 0;
  siva_atpg_target_array = array_alloc(int, 64);  
  siva_support_latch_vector = var_set_new(siva_num_latch);
  siva_witness_values = ALLOC(char,64);  
  siva_witness_input_vectors = ALLOC(var_set_t *,
                                     SIVA_ATPG_UNFOLD_DEPTH_MAX);  
  siva_pi_var_ids = ALLOC(int *,SIVA_ATPG_UNFOLD_DEPTH_MAX);
  for(i=0;i<SIVA_ATPG_UNFOLD_DEPTH_MAX;i++){
    siva_witness_input_vectors[i] = var_set_new(siva_num_primary_input);
    siva_pi_var_ids[i] = ALLOC(int,siva_num_primary_input);
  }
  siva_witness_state_vectors = ALLOC(var_set_t *,
                                     SIVA_ATPG_UNFOLD_DEPTH_MAX);
  siva_latch_var_ids = ALLOC(int *,SIVA_ATPG_UNFOLD_DEPTH_MAX);
  for(i=0;i<SIVA_ATPG_UNFOLD_DEPTH_MAX;i++){
    siva_witness_state_vectors[i] = var_set_new(siva_num_latch);
    siva_latch_var_ids[i] = ALLOC(int,siva_num_latch);
  }
  siva_configure_grasp_default();
}

/*---------------------------------------------------------------------------*/
void
siva_quit_atpg_solver(void)
{
  int i;

  for(i=0;i<SIVA_ATPG_UNFOLD_DEPTH_MAX;i++){
    var_set_free(siva_witness_state_vectors[i]);
    FREE(siva_latch_var_ids[i]);
  }
  FREE(siva_witness_state_vectors);
  FREE(siva_latch_var_ids);
  for(i=0;i<SIVA_ATPG_UNFOLD_DEPTH_MAX;i++){
    var_set_free(siva_witness_input_vectors[i]);
    FREE(siva_pi_var_ids[i]);
  }
  FREE(siva_pi_var_ids);
  FREE(siva_witness_input_vectors);
  FREE(siva_witness_values);
  var_set_free(siva_support_latch_vector);
  array_free(siva_atpg_target_array);
  array_free(siva_clause_array);
  bck_release_aux(siva_var_id_aux,NULL);

}

/*
  ---------------------------------------------------------------------------
  This routine is invoked to check whether the node is 
  tautology or not. This is invoked by siva_symbolic_sim1.
  ---------------------------------------------------------------------------
*/
int
siva_atpg_sym_solve(SIVA_NODE *goals,
                    SIVA_NODE *targets,
                    int *target_values,
                    int num_goal,
                    array_t *input_nodes_array,
                    int sym_depth,
                    SIVA_STATE_WRAPPER *state_wrapper,
                    void *para,
                    STORE_WITNESS_SRV store_witness_srv,
                    PRINT_WITNESS_SRV print_witness_srv)
{
  register int i,k;
  int *clauses;
  SIVA_CONV conv;
  int goal_values[num_goal];
  int num_target = 0;

  SIVA_DBG1("siva_atpg_sym_solve..\n");
  siva_configure_grasp_default();
  siva_num_clause = 0;
  siva_num_var_id = 0;


  var_set_clear(siva_support_latch_vector);
  for(i=0;i<num_goal;i++){
    goal_values[i] = 1 ^ bck_is_inverted(goals[i]) ;
  }  
  for(i=0;i<num_goal;i++){
    if (!bck_is_constant(goals[i])) {      
      array_insert(SIVA_NODE,siva_atpg_target_array,num_target++,goals[i]);
    }else if (bck_is_one(goals[i])){
      SIVA_VERB1_ARG1("Goal is trivially reachable");
    }else if (bck_is_zero(goals[i])){ 
      SIVA_VERB1_ARG1("Goal is trivially unreachable");
    }  
  }

#if 0  
  for(i=0;i<num_goal;i++){
    goal_values[i] = target_values[i] ^ bck_is_inverted(goals[i]) ;
  }
  
  for(i=0;i<num_goal;i++){
    if (!bck_is_constant(goals[i])) {      
      array_insert(SIVA_NODE,siva_atpg_target_array,num_target++,goals[i]);
    }else {
      if (bck_is_one(goals[i]) && !target_values[i]) {
        SIVA_ASSERT(0,"Goal is trivially unreachable");
      }
      if (bck_is_zero(goals[i]) && target_values[i]) {
        SIVA_ASSERT(0,"Goal is trivially unreachable");
      }
    }
  }
#endif

  if (num_target == 0) {
    SIVA_VERB2_ARG1("Trivially Satisified\n");
    return 1;
  }

  
  /* fill the clauses in siva_clause_array */
  siva_fill_clause(siva_atpg_target_array,num_target,1);


  /* set the objective clauses*/
  for(i=0;i<num_goal;i++){
    if (bck_is_constant(goals[i])) {
      continue;
    }
    conv.ptr = bck_get_aux(goals[i],siva_var_id_aux);
    if (goal_values[i] == 0){
      conv.int_value = -conv.int_value;
    }
    siva_set_objective_clause1(conv.int_value,0,0);
  }
  
  siva_witness_values = REALLOC(char,siva_witness_values,
                                siva_num_var_id+1);
  clauses = array_space(int,siva_clause_array);
  SIVA_VERB2(siva_clause_stat());
#if TEST
  {
    char buf[32];
    sprintf(buf,"new%d.cnf",sym_depth);
    siva_dump_clause(buf,clauses,siva_num_clause,siva_num_var_id);
  }
#endif
  siva_atpg_solver(siva_num_clause,
                   siva_num_var_id,clauses,siva_witness_values);
  
  if (siva_witness_values[0] == -1){
    SIVA_VERB2_ARG1("atpg_solver failed to solve, change options\n");      
  }else if (siva_witness_values[0] == 1){
    if (state_wrapper) {
      siva_sym_build_witness_sequence(input_nodes_array,
                                      sym_depth,
                                      state_wrapper);

      if (siva_error_check_level >= 1) {
        int value =  siva_evaluate_sym_node(goals[0],
                                            siva_witness_input_vectors,
                                            array_space(SIVA_NODE*,
                                                        input_nodes_array),
                                            sym_depth);
        SIVA_ASSERT(value == 1,
                    "sym valuation failed\n");
      }
      
      SIVA_ERR_CHK1(siva_atpg_verify_witness(targets,target_values,num_goal));

      SIVA_VERB2_ARG1("Printing witness for Target\n");
      if (print_witness_srv) {
        print_witness_srv(state_wrapper->state,sym_depth);
      }
      if (store_witness_srv) {
        /* from witness state to seed state */
        for(i=0;i<sym_depth;i++){               
          store_witness_srv(state_wrapper->state,
                            siva_witness_state_vectors[i],
                            sym_depth - i,
                            sym_depth,para);                                  
        }    
      }
    }
  }  

  SIVA_DBG1("siva_atpg_sym_solve..done\n");
  return siva_witness_values[0];

}


/*
  ---------------------------------------------------------------------------
  This is wrapper for symbolic simulation.
  This invoked by siva_solve.
  ---------------------------------------------------------------------------
*/
int
siva_atpg_solve(SIVA_NODE *goals, int *goal_values, int num_goal,
                array_t *state_wrapper_array, int num_state,
                void *para,
                STORE_WITNESS_SRV store_witness_srv,
                PRINT_WITNESS_SRV print_witness_srv)                         
{
  int i,t,j;
  SIVA_NODE targets[num_goal];
  int target_values[num_goal];
  int num_remained = num_goal;
  int num_target = num_remained;
  
  memcpy(targets,goals,num_goal*sizeof(SIVA_NODE));
  memcpy(target_values,goal_values,num_goal*sizeof(int));
  
  siva_symbolic_sim_depth = siva_atpg_unfold_depth_upper;
  
  for(j=0;j<num_state;j++) {
    SIVA_STATE_WRAPPER *state_wrapper =
      array_fetch_p(SIVA_STATE_WRAPPER,state_wrapper_array,j);
    num_remained = siva_symbolic_sim1(targets,
                                      target_values,num_target,
                                      state_wrapper,0,
                                      para,
                                      store_witness_srv,
                                      print_witness_srv);
    siva_gc_symbolic_sim();
    for(i=0;i<num_target;i++){
      SIVA_VERB2_ARG3("target %d value %d\n",targets[i],target_values[i]);
    }
    if (num_target == num_remained) {
      continue;
    }
    if (num_remained == 0) {
      break;
    }
    t = 0;    
    while(t < num_target) {
      if (target_values[t] == 2) {
        num_target--;
        memmove(targets+t,targets+t+1,
              (num_target-t) * sizeof(SIVA_NODE));
        memmove(target_values+t,target_values+t+1,
              (num_target-t) * sizeof(int));
      }else {
        t++;
      }
    }
    SIVA_VERB2_ARG2("Number target remained = %d\n",num_target);    
    SIVA_ASSERT(num_target == num_remained,"");
        
  }
  
  i = 0;
  for(j=0;j<num_remained;j++) {
    while(i<num_goal) {
      if (goals[i] == targets[j]) {
        i++;
        break;
      }else {
        goal_values[i] = 2;
        i++;
      }
    }
  }
  for(;i<num_goal;i++) {
    goal_values[i] = 2;
  }    
  

  return num_remained;
}



/*---------------------------------------------------------------------------*/
void
siva_store_target_witness_state(SIVA_STATE *seed_state,
                                var_set_t *wit_state_vector,
                                int rel_depth,
                                int unfold_depth,
                                void *para)

{
  unsigned state_value;
  SIVA_STATE *wit_state;
  int priority_value;  
  int is_created;
  
  var_set_compress(&state_value,wit_state_vector,
                   siva_state_init_hash_value);
  wit_state = siva_get_or_create_state(state_value,
                                       seed_state->state_value,
                                       rel_depth,
                                       wit_state_vector->data,
                                       &is_created);  
  priority_value = rel_depth;
  siva_put_wit_state(priority_value,wit_state);
  SIVA_VERB2_ARG3("SYMBOLIC-SIM: TARGET: Witness State %x, state_value = %x\n",
                  wit_state,state_value);
}

/*
  ---------------------------------------------------------------------------
  This is invoked to check if there exist any state which can reach
  the goal.
  ---------------------------------------------------------------------------
*/
int
siva_atpg_is_satisfiable(SIVA_NODE *goals, int *goal_values,
                        int num_goal)
{
  register int i,k;
  int *clauses;
  SIVA_CONV conv;
  int num_target;
  int reachable;
  
  SIVA_DBG1("siva_atpg_solve..");

  var_set_clear(siva_support_latch_vector);
  
  siva_num_clause = 0;
  siva_num_var_id = 0;
  
  num_target = 0;
  for(i=0;i<num_goal;i++){
    if (!bck_is_constant(goals[i])){
      array_insert(SIVA_NODE,siva_atpg_target_array,
                   num_target,goals[i]);
      num_target++;
    }else {
      if (bck_is_one(goals[i]) && !goal_values[i]) {        
        SIVA_VERB1_ARG2("Goal is trivially unreachable goal # %d\n",i);        
        return 0;
        SIVA_ASSERT(0,"Goal is trivially unreachable");
      }
      if (bck_is_zero(goals[i]) && goal_values[i]) {
        SIVA_VERB1_ARG2("Goal is trivially unreachable goal # %d\n",i);      
        return 0;
        SIVA_ASSERT(0,"Goal is trivially unreachable");
      }
    }      
  }
  var_set_clear(siva_support_latch_vector);
  /* fill the clauses in siva_clause_array */
  siva_fill_clause(siva_atpg_target_array,num_target,1);
  /* set the objective clauses */
  for(i=0;i<num_goal;i++){
    if (bck_is_constant(goals[i])) {
      continue;
    }        
    conv.ptr = bck_get_aux(goals[i],siva_var_id_aux);
    if (goal_values[i] == 0){
      conv.int_value = -conv.int_value;
    }
    siva_set_objective_clause1(conv.int_value,0,0);
  }
  siva_witness_values = REALLOC(char,siva_witness_values,
                                siva_num_var_id+1);
  clauses = array_space(int,siva_clause_array);
#if TEST
  siva_dump_clause(TEST_CNF,
                   clauses,siva_num_clause,siva_num_var_id);
#endif
  SIVA_VERB2(siva_clause_stat());  
  siva_atpg_solver(siva_num_clause,siva_num_var_id,clauses,
                   siva_witness_values);      
  if (siva_witness_values[0] == -1){
    SIVA_VERB1_ARG1("atpg failed to solve, change options\n");
    reachable = -1;
  }else if (siva_witness_values[0] == 1){
    reachable = 1;
  }else {
    reachable = 0;
  }
  return reachable;  
}


/*---------------------------------------------------------------------------*/
static void
siva_sym_build_witness_sequence(array_t *input_nodes_array,
                                int sym_depth,
                                SIVA_STATE_WRAPPER *state_wrapper)
{
  var_set_t *input_vector;
  var_set_t *state_vector;
  int i,k;
  SIVA_CONV conv;
  
  state_vector = siva_witness_state_vectors[sym_depth];
  var_set_assign(state_vector,state_wrapper->state_vector);
  for(i=0;i<=sym_depth;i++){
    SIVA_NODE *input_nodes = array_fetch(SIVA_NODE *,input_nodes_array,i);
    input_vector = siva_witness_input_vectors[sym_depth-i];
    var_set_clear(input_vector);
    for(k=0;k<siva_num_primary_input;k++){
      conv.ptr = bck_get_aux(input_nodes[k],siva_var_id_aux);
      if (conv.int_value == 0) {
        continue;
      }
      if (siva_witness_values[conv.int_value] == 1){
        var_set_set_elt(input_vector,k);
      }
    }
    if (i != sym_depth) {
      siva_sim_next_state_vector(siva_witness_state_vectors[sym_depth-i-1],
                                 input_vector,
                                 siva_witness_state_vectors[sym_depth-i]);
    }
  }

}

/*---------------------------------------------------------------------------*/
static void
siva_atpg_verify_witness(SIVA_NODE *targets,
                         int *target_values,
                         int num_target)
{
  int i;
  int values[num_target];
  siva_evaluate_node(targets,values,num_target,
                     siva_witness_input_vectors[0],
                     siva_witness_state_vectors[0]);
  for(i=0;i<num_target;i++){      
    SIVA_ASSERT(values[i] == target_values[i],
                "evaluated value does not match  target value");
    
  }
}
  
/*
  ---------------------------------------------------------------------------
  This prints the witness trace from the initial state to the target
  ---------------------------------------------------------------------------
*/
void
siva_atpg_print_witness_srv(SIVA_STATE *seed_state,
                            int unfold_depth)
{
  int num_prev_state;
  if (!array_n(siva_debug_node_array)){
    SIVA_VERB2_ARG1("Debug nodes do not exist\n");    
    return;
  }  
  num_prev_state = siva_print_state_trace(seed_state);
  siva_print_witness_trace(siva_witness_input_vectors,
                           siva_witness_state_vectors,
                           num_prev_state,
                           unfold_depth);
}

/*
  ---------------------------------------------------------------------------
  This prints the witness trace from the seed trace to the target
  ---------------------------------------------------------------------------
*/
void
siva_atpg_print_witness_srv1(SIVA_STATE *seed_state,
                             int unfold_depth)
{  
  if (!array_n(siva_debug_node_array)){
    SIVA_VERB2_ARG1("Debug nodes do not exist\n");    
    return;
  }  
  siva_print_witness_trace(siva_witness_input_vectors,
                           siva_witness_state_vectors,
                           0,unfold_depth);
}

/*---------------------------------------------------------------------------*/
static void
siva_clause_stat(void)
{
  SIVA_VERB1_ARG2("# clause %d\n",siva_num_clause);
  SIVA_VERB1_ARG2("# variables %d\n",siva_num_var_id);
  SIVA_VERB1_ARG2("# support latch %d\n",
                  var_set_n_elts(siva_support_latch_vector));
}


/* Links the latch id's with its next state ids */
/*---------------------------------------------------------------------------*/
static void
siva_set_interface_clause(var_set_t *support_latch_vector,
                          int current_unfold_depth)
{
  register int i;
  SIVA_CONV conv;
  int current_id;
  unsigned *var_ids;
  SIVA_NODE next_state_node;

  var_ids = siva_latch_var_ids[current_unfold_depth-1]; 
  for(i=0;i<siva_num_latch;i++){
    if (!var_set_get_elt(support_latch_vector,i)){
      continue;
    }
    next_state_node = siva_latch_nodes[i].input;
    current_id = var_ids[i];
    SIVA_ASSERT(current_id,"current_id should be in support");
    if (bck_is_constant(next_state_node)){
      if (bck_is_zero(next_state_node)){
        current_id = -current_id;
      }
      siva_set_objective_clause1(current_id,0,0);
      continue;
    }
    conv.ptr = bck_get_aux(next_state_node,siva_var_id_aux);
    if (bck_is_inverted(next_state_node)){
      conv.int_value = -conv.int_value;
    }
    SIVA_ASSERT((conv.int_value != 0),"id can't be zero");
    /* (!conv.int_value + current_id) */
    siva_set_objective_clause1(-conv.int_value,current_id,0);
    /* (conv.int_value + !current_id) */
    siva_set_objective_clause1(conv.int_value,-current_id,0);    
  }
}


/*---------------------------------------------------------------------------*/
static void
siva_set_objective_clause1(int var_id0,int var_id1, int var_id2)
{
  
  SIVA_ASSERT((var_id0 != 0),"id can't be zero");
  array_insert(int,siva_clause_array,3*siva_num_clause,var_id0);  
  array_insert(int,siva_clause_array,3*siva_num_clause+1,var_id1);
  array_insert(int,siva_clause_array,3*siva_num_clause+2,var_id2);
  siva_num_clause++;
  
}


/*---------------------------------------------------------------------------*/
static void
siva_set_state_vector_clause(var_set_t *state_vector)
{
  SIVA_CONV conv;
  int i;
  for(i=0;i<siva_num_latch;i++){
    if (!var_set_get_elt(siva_support_latch_vector,i)){
      continue;
    }
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_var_id_aux);
    if  (!var_set_get_elt(state_vector,i)){
      conv.int_value = -conv.int_value;
    }
    siva_set_objective_clause1(conv.int_value,0,0);
  }      
}

/*---------------------------------------------------------------------------*/
static void
siva_unset_state_vector_clause(void)
{  
  siva_del_last_num_clause(var_set_n_elts(siva_support_latch_vector));
}

/*---------------------------------------------------------------------------*/
static void
siva_del_last_num_clause(int num_clause)
{
  siva_num_clause -= num_clause;
}

/*---------------------------------------------------------------------------*/
static void
siva_fill_clause(array_t *node_array, int num_node, int reset_id)
{
  register int i;
  SIVA_CONV conv;
  
  /* clear the var_id aux */
  if (reset_id){
    conv.value = siva_var_id_aux;
    bck_do_in_topological_order(conv.ptr,siva_clear_aux_srv);
  }
  for(i=0;i<num_node;i++){
    SIVA_NODE node = array_fetch(SIVA_NODE,node_array,i);
    siva_fill_clause_aux(node);
  }
  for(i=0;i<num_node;i++){
    SIVA_NODE node = array_fetch(SIVA_NODE,node_array,i);
    bck_reset_fanin_cone_flag(node,siva_bck_visited_flag);
  }


}

/*---------------------------------------------------------------------------*/
static void
siva_fill_clause_aux(SIVA_NODE bck)
{
  if (bck_is_flag_set(bck,siva_bck_visited_flag)){
    return;
  }  
  if (bck_is_var(bck)){
    SIVA_CONV conv;    
    conv.int_value = ++siva_num_var_id;
    bck_set_aux(bck,siva_var_id_aux,conv.ptr);    
    if (bck_is_flag_set(bck,siva_latch_flag)){
      conv.ptr = bck_get_aux(bck,siva_latch_id_aux);
      var_set_set_elt(siva_support_latch_vector,conv.int_value);
    }    
  }else if (bck_is_and(bck)){
    SIVA_NODE left_child = bck_get_left_child(bck);
    SIVA_NODE right_child = bck_get_right_child(bck);

    /* recursive calls on left and right child */
    siva_fill_clause_aux(left_child);
    siva_fill_clause_aux(right_child);

    {
      SIVA_CONV conv;
      int node_id;
      int left_id;
      int right_id;
      
      node_id = conv.int_value = ++siva_num_var_id;
      bck_set_aux(bck,siva_var_id_aux,conv.ptr);
      /* get left child id */
      conv.ptr = bck_get_aux(left_child,siva_var_id_aux);
      SIVA_ASSERT(conv.int_value != 0,"not assigned var id");
      if (bck_is_inverted(left_child)){
        left_id = -conv.int_value;
      }else {
        left_id = conv.int_value;
      }
      /* get right child id */
      conv.ptr = bck_get_aux(right_child,siva_var_id_aux);
      SIVA_ASSERT(conv.int_value != 0,"not assigned var id");
      if (bck_is_inverted(right_child)){
        right_id = -conv.int_value;
      }else {
        right_id = conv.int_value;
      }
    
      /* (node_id + !left_id + !right_id) */
      array_insert(int,siva_clause_array,3*siva_num_clause,node_id);
      array_insert(int,siva_clause_array,3*siva_num_clause+1,-left_id);
      array_insert(int,siva_clause_array,3*siva_num_clause+2,-right_id);
      siva_num_clause++;
    
      /* (!node_id + left_id) */
      array_insert(int,siva_clause_array,3*siva_num_clause,-node_id);
      array_insert(int,siva_clause_array,3*siva_num_clause+1,left_id);
      array_insert(int,siva_clause_array,3*siva_num_clause+2,0);
      siva_num_clause++;
      
      /* (!node_id + right_id) */
      array_insert(int,siva_clause_array,3*siva_num_clause,-node_id);
      array_insert(int,siva_clause_array,3*siva_num_clause+1,right_id);
      array_insert(int,siva_clause_array,3*siva_num_clause+2,0);
      siva_num_clause++;
    }
  }
 end:
  bck_set_flag(bck,siva_bck_visited_flag);  
}

/*---------------------------------------------------------------------------*/
static void
siva_dump_clause(char *filename,
                  int *clauses, int num_clause, int num_var_id)
{
  FILE *file = fopen(filename,"w");
  register int i,j;

  fprintf(file,"c test file\n");
  fprintf(file,"p cnf %d %d\n",num_var_id, num_clause);
  for(i=0;i<num_clause;i++){
    for(j=0;j<3;j++){
      if (clauses[3*i+j] == 0){
        break;
      }
      fprintf(file,"%d ",clauses[3*i+j]);
    }
    fprintf(file,"0\n");
  }
  fflush(file);
  fclose(file);
}

/*---------------------------------------------------------------------------*/
static void
siva_configure_grasp_default(void)
{
  int i;
  for(i=0;i<GRP_MAX_OPTION_SIZE;i++){
    siva_grasp_options[(i<<1)] = 0;
  }
  siva_grasp_back_track_limit = siva_atpg_back_track_limit;
  SIVA_SET_BACK_TRACK_LIMIT(siva_grasp_back_track_limit);
  //siva_grasp_time_limit = 100;
  //siva_grasp_time_limit = 300;
  siva_grasp_time_limit = siva_max_time_limit;
  SIVA_SET_TIME_LIMIT(siva_grasp_time_limit);
  SIVA_SET_PREPROC_RECORDING(_SIMPLE_RECORDING_);
  siva_grasp_conflict_limit = siva_atpg_conflict_limit;
  SIVA_SET_CONFLICT_LIMIT(siva_grasp_conflict_limit);
  SIVA_SET_DEDUCTION_LEVEL(siva_atpg_deduction_level);
  SIVA_SET_LEARNING_LEVEL(siva_atpg_learning_level);
  //SIVA_SET_BACKTRACKING_STRATEGY(_CHRONOLOGICAL_B_);
}

/*---------------------------------------------------------------------------*/
static void
siva_configure_grasp(int grasp_mode, int value )
{
  siva_grasp_options[2*grasp_mode] = 1;
  siva_grasp_options[2*grasp_mode+1] = value;
}

/*---------------------------------------------------------------------------*/
static void
siva_atpg_grasp_solver(int num_clause, int num_var_id,
                       int *clauses, char *witness)
{
  int i;

  SIVA_DBG1("atpg-solver ..\n");

  siva_grasp_solver(num_clause, num_var_id,                    
                    clauses,witness,siva_grasp_options);


  siva_atpg_num_trial = 1;
  for(i=1;i<siva_atpg_num_trial;i++){
    if (witness[0] != -1){
      break;
    }
    fprintf(siva_output_file,"Trying harder with ...\n");
    siva_grasp_back_track_limit <<= 1;
    SIVA_SET_BACK_TRACK_LIMIT(siva_grasp_back_track_limit);
    fprintf(siva_output_file,"Backtrack limit = %d\n",
            siva_grasp_back_track_limit);
    siva_grasp_time_limit <<= 1;
    SIVA_SET_TIME_LIMIT(siva_grasp_time_limit);
    fprintf(siva_output_file,"Time limit = %d\n",
            siva_grasp_time_limit);

    siva_grasp_conflict_limit <<= 1;
    SIVA_SET_CONFLICT_LIMIT(siva_grasp_conflict_limit);
    fprintf(siva_output_file,"Conflict limit = %d\n",
            siva_grasp_conflict_limit);

    siva_grasp_solver(num_clause, num_var_id,                    
                      clauses,witness,siva_grasp_options);
  }
  SIVA_DBG1("atpg_solver ..done\n");
}

/*---------------------------------------------------------------------------*/
static void
siva_atpg_solver(int num_clause, int num_var_id,
                 int *clauses, char *witness)
{
  siva_atpg_grasp_solver(num_clause, num_var_id,
                         clauses, witness);
}

/*---------------------------------------------------------------------------*/
static void
siva_build_witness_input_sequence(int **pi_var_ids, int unfold_depth)
{
  unsigned *var_ids;
  int i,j;
  for(j=0;j<=unfold_depth;j++){    
    var_ids = pi_var_ids[j];
    for(i=0;i<siva_num_primary_input;i++){      
      (var_ids[i] == 0 || siva_witness_values[var_ids[i]] == 0 ||
       siva_witness_values[var_ids[i]] == 2) ?
        var_set_clear_elt(siva_witness_input_vectors[j],i):
        var_set_set_elt(siva_witness_input_vectors[j],i);
    }
  }
}

/*---------------------------------------------------------------------------*/
static void
siva_build_witness_state_sequence(int **latch_var_ids, int unfold_depth)
{
  int i,j;

  for(j=unfold_depth;j>0;j--){
    siva_sim_next_state_vector(siva_witness_state_vectors[j-1],
                               siva_witness_input_vectors[j],
                               siva_witness_state_vectors[j]);
    /* this is for sanitary check */
    if (siva_error_check_level >= 1) {
     unsigned * var_ids = latch_var_ids[j-1];
     int k;
      for(k=0;k<siva_num_latch;k++){
        if (var_ids[k] == 0 ||
            siva_witness_values[var_ids[k]] == 2) {
          continue;
        }
        SIVA_ASSERT1(siva_witness_values[var_ids[k]] ==
                    var_set_get_elt(siva_witness_state_vectors[j-1],k),
                    "Simulation does not match atpg at depth %d\n",j-1);
      }
    }                                       
  }
}

/*---------------------------------------------------------------------------*/
/* it constructs et from the witness vectors. it should be called after
   successful call of the atpg solver
*/
static void
siva_set_et(SIVA_ET *et, int unfold_depth)
{
  int k;
  int mask_size;
  var_set_t state_vector;
  var_set_t mask_vector;
  unsigned * var_ids = siva_latch_var_ids[unfold_depth];

  siva_set_state_vector(&mask_vector,et->mask);
  siva_set_state_vector(&state_vector,et->data);

  var_set_clear(&mask_vector);
  var_set_clear(&state_vector);
  
  for(k=0;k<siva_num_latch;k++){
    if (var_ids[k] == 0 ||
        siva_witness_values[var_ids[k]] == 2) {
      continue;
    }    
    var_set_set_elt(&mask_vector,k);
    if (siva_witness_values[var_ids[k]]) {
      var_set_set_elt(&state_vector,k);
    }
    SIVA_ASSERT(!bck_is_zero(siva_latch_nodes[k].input)||
                !var_set_get_elt(&state_vector,k),
                "why sv is not 0");
    SIVA_ASSERT(!bck_is_one(siva_latch_nodes[k].input)||
                var_set_get_elt(&state_vector,k),
                "why sv is not 1");
    
  }
  SIVA_VERB2(var_set_printx(siva_output_file,&mask_vector,
                           &state_vector));
  mask_size = var_set_n_elts(&mask_vector);
  SIVA_VERB2_ARG2("Mask_vector size = %d\n",mask_size);
}


/*---------------------------------------------------------------------------*/
/* using grasp to solve it. could have used chaff */
int
siva_atpg_find_input_sequence(var_set_t *target_state_vector,
                              var_set_t *seed_state_vector,
                              int rel_distance,
                              var_set_t **input_vectors,
                              var_set_t **state_vectors)
{

  register int i,j;
  int *clauses;
  int unfold_depth;
  SIVA_CONV conv;
  int num_target;
  SIVA_NODE goals[siva_num_latch];
  int goal_values[siva_num_latch];
  unsigned data[siva_latch_word_size];
  var_set_t support_latch_vector;
  int *var_ids;
  int found;
  int num_goal;
  
  SIVA_DBG1("siva_atpg_find_input_sequence..");
  SIVA_ASSERT(rel_distance <
              SIVA_ATPG_UNFOLD_DEPTH_MAX,"Increase unfold depth");

  siva_set_state_vector(&support_latch_vector,data);
  var_set_clear(siva_support_latch_vector);
  
  /* start with default setting */
  siva_configure_grasp_default();

  siva_num_clause = 0;
  siva_num_var_id = 0;
  num_goal = 0;
  for(i=0;i<siva_num_latch;i++){
    if (bck_is_constant(siva_latch_nodes[i].input)){
      continue;
    }
    goals[num_goal] = bck_get_index(siva_latch_nodes[i].input);
    goal_values[num_goal] = var_set_get_elt(target_state_vector,i) ^
      bck_is_inverted(siva_latch_nodes[i].input);
    num_goal++;
  }
  for(i=0;i<num_goal;i++){    
    array_insert(SIVA_NODE,siva_atpg_target_array,i,goals[i]);
  }
  num_target = num_goal;
  /* fill the clauses in siva_clause_array */
  var_set_clear(siva_support_latch_vector);
  siva_fill_clause(siva_atpg_target_array,num_target,1);

  /* set the objective clauses */
  for(i=0;i<num_goal;i++){    
    conv.ptr = bck_get_aux(goals[i],siva_var_id_aux);
    if (goal_values[i] == 0){
      conv.int_value = -conv.int_value;
    }
    siva_set_objective_clause1(conv.int_value,0,0);
  }
  var_ids = siva_pi_var_ids[0];
  for(i=0;i<siva_num_primary_input;i++){
    conv.ptr = bck_get_aux(siva_primary_inputs[i],siva_var_id_aux);
    var_ids[i] = conv.int_value;
  }
  var_ids = siva_latch_var_ids[0];
  for(i=0;i<siva_num_latch;i++){
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_var_id_aux);
    var_ids[i] = conv.int_value;
  }
  
  for(unfold_depth=1;unfold_depth<rel_distance;unfold_depth++){
    var_set_assign(&support_latch_vector,siva_support_latch_vector);
    var_set_clear(siva_support_latch_vector);
    num_target = 0;
    for(i=0;i<siva_num_latch;i++){
      if (!var_set_get_elt(&support_latch_vector,i)){
        continue;
      }
      array_insert(SIVA_NODE,siva_atpg_target_array,num_target,
                   siva_latch_nodes[i].input);
      num_target++;
    }
    /* Fill Clauses */
    siva_fill_clause(siva_atpg_target_array,num_target,1);
    /* Add interface clauses */
    siva_set_interface_clause(&support_latch_vector,unfold_depth);
    /* Store the var ids */
    var_ids = siva_pi_var_ids[unfold_depth];
    for(i=0;i<siva_num_primary_input;i++){
      conv.ptr = bck_get_aux(siva_primary_inputs[i],siva_var_id_aux);
      var_ids[i] = conv.int_value;
    }
    var_ids = siva_latch_var_ids[unfold_depth];
    for(i=0;i<siva_num_latch;i++){
      conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_var_id_aux);
      var_ids[i] = conv.int_value;
    }      
  }    

  siva_set_state_vector_clause(seed_state_vector);
  siva_witness_values = REALLOC(char,siva_witness_values,siva_num_var_id+1);
  clauses = array_space(int,siva_clause_array);

  SIVA_VERB2_ARG2("Current Depth %d\n",rel_distance);
  SIVA_VERB2(siva_clause_stat());
  
  siva_atpg_solver(siva_num_clause,siva_num_var_id,clauses,
                   siva_witness_values);          

  SIVA_ASSERT(siva_witness_values[0] !=0,
              "Witness generation failed");
  
  if (siva_witness_values[0] == -1){
    SIVA_VERB1_ARG1("grasp failed to solve, change options\n");    
    found = 0;
  }else {
    var_set_assign(siva_witness_state_vectors[rel_distance-1],
                   seed_state_vector);
    siva_build_witness_input_sequence(siva_pi_var_ids,
                                      rel_distance-1);
    siva_build_witness_state_sequence(siva_latch_var_ids,
                                      rel_distance-1);        
    for(i=0;i<rel_distance;i++){
      var_set_assign(input_vectors[i],
                     siva_witness_input_vectors[i]);
    }
    for(i=0;i<rel_distance;i++){
      var_set_assign(state_vectors[i],
                     siva_witness_state_vectors[i]);
    }
    found = 1;
  }
  return found;
}



/*---------------------------------------------------------------------------*/
/* take the goals and finds successive (e_factor) cubes                      */
int
siva_atpg_enlarge(SIVA_NODE *goals, int *goal_values,
                  int num_goal,
                  SIVA_ET *ets,
                  int e_factor)
{
  register int i;
  int unfold_depth;
  SIVA_CONV conv;
  int num_target;
  int *var_ids;
  int *clauses;
  unsigned data[siva_latch_word_size];
  var_set_t support_latch_vector;

  SIVA_DBG1("siva_atpg_solve_find_partial_et ..\n");
  SIVA_ASSERT(e_factor < SIVA_ATPG_UNFOLD_DEPTH_MAX,
              "increase unfold depth");
  SIVA_ASSERT(e_factor > 0,"");
  
  support_latch_vector.n_elts = siva_num_latch;
  support_latch_vector.n_words = siva_latch_word_size;
  support_latch_vector.data = data;
  siva_configure_grasp_default();
  siva_num_clause = 0;
  siva_num_var_id = 0;

  for(i=0;i<num_goal;i++){    
    array_insert(SIVA_NODE,siva_atpg_target_array,i,goals[i]);
  }
  num_target = num_goal;
  
  var_set_clear(siva_support_latch_vector);
  
  /* fill the clauses in siva_clause_array */
  siva_fill_clause(siva_atpg_target_array,num_target,1);
  for(i=0;i<num_goal;i++){    
    conv.ptr = bck_get_aux(goals[i],siva_var_id_aux);
    if (goal_values[i] == 0){
      conv.int_value = -conv.int_value;
    }
    siva_set_objective_clause1(conv.int_value,0,0);
  }

  var_ids = siva_pi_var_ids[0];
  for(i=0;i<siva_num_primary_input;i++){
    conv.ptr = bck_get_aux(siva_primary_inputs[i],siva_var_id_aux);
    var_ids[i] = conv.int_value;
  }
  var_ids = siva_latch_var_ids[0];
  for(i=0;i<siva_num_latch;i++){
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_var_id_aux);
    var_ids[i] = conv.int_value;
  }
  unfold_depth = 0;
  do{
    if (unfold_depth) {
      var_set_assign(&support_latch_vector,siva_support_latch_vector);
      var_set_clear(siva_support_latch_vector);
      num_target = 0;
      for(i=0;i<siva_num_latch;i++){
        if (!var_set_get_elt(&support_latch_vector,i)){
          continue;
        }
        array_insert(SIVA_NODE,siva_atpg_target_array,num_target,
                     siva_latch_nodes[i].input);
        num_target++;
      }
      /* Fill Clauses */
      siva_fill_clause(siva_atpg_target_array,num_target,1);
      /* Add interface clauses */
      siva_set_interface_clause(&support_latch_vector,unfold_depth);
      /* Store the var ids */
      var_ids = siva_pi_var_ids[unfold_depth];
      for(i=0;i<siva_num_primary_input;i++){
        conv.ptr = bck_get_aux(siva_primary_inputs[i],siva_var_id_aux);
        var_ids[i] = conv.int_value;
      }
      var_ids = siva_latch_var_ids[unfold_depth];
      for(i=0;i<siva_num_latch;i++){
        conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_var_id_aux);
        var_ids[i] = conv.int_value;
      }      
    }
    siva_witness_values = REALLOC(char,siva_witness_values,
                                  siva_num_var_id+1);
    clauses = array_space(int,siva_clause_array);
#if TEST
    siva_dump_clause(TEST_CNF,
                     clauses,siva_num_clause,siva_num_var_id);
#endif
    SIVA_VERB2_ARG2("Current Depth %d\n",unfold_depth);
    SIVA_VERB2(siva_clause_stat());
    siva_atpg_direct_search(siva_latch_var_ids[unfold_depth]);
    if (siva_witness_values[0] == -1){
      SIVA_VERB1_ARG1("grasp failed to solve, change options\n");
      return unfold_depth;
    }else if (siva_witness_values[0] == 1){
      siva_set_et(&ets[unfold_depth],unfold_depth);
      {
        var_set_t state_vector;
        int value=1;
        siva_set_state_vector(&state_vector,ets[unfold_depth].data);
        var_set_assign(siva_witness_state_vectors[unfold_depth],
                       &state_vector);
        siva_build_witness_input_sequence(siva_pi_var_ids,
                                          unfold_depth);
        siva_build_witness_state_sequence(siva_latch_var_ids,
                                          unfold_depth);
        SIVA_ERR_CHK1(siva_atpg_verify_witness(goals,goal_values,num_goal));

      }
    }else {
      int k;
      for(k=0;k<unfold_depth;k++){
        ets[k].is_unreachable  = 1;
      }
      SIVA_VERB2_ARG1("ET is unreachable\n");
      return 0;
    }
    unfold_depth++;
  }while (unfold_depth < e_factor);
  return e_factor;
  
}

/*---------------------------------------------------------------------------*/
/* directs towards reachable space                               
   Tries first with putting the clauses of the intertial latches
   and then without if it fails
*/
static void
siva_atpg_direct_search(int *latch_var_ids)
{
  int i;
  int num_inertial_clause = 0;
  int inertia;
  int num_constant_clause = 0;
  SIVA_NODE next_state_node;
  int current_id;
  
  
  for(i=0;i<siva_num_latch;i++){
    if (!var_set_get_elt(siva_support_latch_vector,i)){
      continue;
    }
    next_state_node = siva_latch_nodes[i].input;
    current_id = latch_var_ids[i];
    SIVA_ASSERT(current_id,"current_id should be in support");
    if (bck_is_constant(next_state_node)){
      if (bck_is_zero(next_state_node)){
        current_id = -current_id;
      }
      siva_set_objective_clause1(current_id,0,0);
      num_constant_clause++;
      continue;
    }
  }
  
  for(i=0;i<siva_num_latch;i++){
    if (latch_var_ids[i] == 0) {
      continue;
    }
    if (bck_is_constant(siva_latch_nodes[i].input)) {
      continue;
    }
    inertia = siva_get_latch_inertia(i,1);
    if (inertia == 100) {
      current_id = -latch_var_ids[i];
    }else if (inertia == 0) {
      current_id = latch_var_ids[i];
    }else {
      continue;
    }
    siva_set_objective_clause1(current_id,0,0);
    num_inertial_clause++;
  }
  siva_atpg_grasp_solver(siva_num_clause,
                         siva_num_var_id,array_space(int,siva_clause_array),
                         siva_witness_values);
  siva_del_last_num_clause(num_inertial_clause);
  if (siva_witness_values[0] == -1){
    SIVA_VERB1_ARG1("grasp failed to solve, change options\n");
  }else if (siva_witness_values[0] == 1){
    SIVA_VERB2_ARG1("ET found with inertial values\n");
  }else {
    SIVA_VERB2_ARG1("ET not found with inertial values\n");
  }
  
  if (siva_witness_values[0] != 0) {
    return;
  }
  siva_atpg_grasp_solver(siva_num_clause,
                         siva_num_var_id,array_space(int,siva_clause_array),
                         siva_witness_values);
  siva_del_last_num_clause(num_constant_clause);
}


