/**CFile**********************************************************************

  FileName                   [siva_auto_light_house.c]

  PackageName                [siva_core]

  Synopsis                   [Automatic lighthouse generation and use]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modifed: 01/01/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]
******************************************************************************/

#include "siva.h"

#define SIVA_DEFAULT_ALH_PICK_SIZE 5
#define SIVA_ALH_INERTIAL_THRESHOLD 10 /* in %*/

static array_t *siva_free_alh_array;
static unsigned siva_alh_state_heap_id;
static unsigned siva_alh_heap_id;
static unsigned siva_num_free_alh;

static void siva_put_alh(unsigned priority_value, SIVA_LIGHT_HOUSE *alh);
static int siva_get_alh(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr);
static int siva_get_best_alh(SIVA_LIGHT_HOUSE *alhs[],
                             int num_alh_req);
static void siva_store_alh_witness_state(SIVA_STATE *seed_state,
                                         var_set_t *wit_state_vector,
                                         int rel_depth,
                                         int unfold_depth,
                                         void *para);
static void siva_release_alh(SIVA_LIGHT_HOUSE *alh);
static SIVA_LIGHT_HOUSE* siva_alloc_alh(void);
static int siva_set_alh(SIVA_NODE node, int value);
static void siva_print_alh_heap(void);
static void siva_alh_heap_srv(int pos, SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr);
static void siva_check_alh_heap(int pos, SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr);


/*---------------------------------------------------------------------------*/
void
siva_init_alh(void)
{
  siva_free_alh_array = array_alloc(SIVA_LIGHT_HOUSE*,16);
  siva_alh_state_heap_id = siva_alloc_heap(siva_heap_unsigned_min_cmp);   
  siva_alh_heap_id = siva_alloc_heap(siva_heap_unsigned_min_cmp);
  siva_alh_state_priority = ALLOC(unsigned, SIVA_SIM_N_WORDS << 5);
  siva_num_free_alh = 0;
}

/*---------------------------------------------------------------------------*/
void
siva_quit_alh(void)
{
  int i;
  SIVA_CONV conv;
  SIVA_KEY key;

  FREE(siva_alh_state_priority);  
  siva_release_heap(siva_alh_state_heap_id);
  while(siva_get_alh(&key,&conv)){
    FREE(conv.lh);
  }  
  siva_release_heap(siva_alh_heap_id);
  for(i=0;i<siva_num_free_alh;i++){
    SIVA_LIGHT_HOUSE *alh = array_fetch(SIVA_LIGHT_HOUSE*,
                                        siva_free_alh_array,i);
    FREE(alh);
  }
  array_free(siva_free_alh_array);
}

/*
 --------------------------------------------------------------------------- 
 This is called every simulation step
 to clear all the priorities.
 ---------------------------------------------------------------------------
*/
void
siva_update_alh(void)
{

  memset(siva_alh_state_priority,0,SIVA_SIM_N_WORDS << 7);
}

/*---------------------------------------------------------------------------*/
static SIVA_LIGHT_HOUSE*
siva_alloc_alh(void)
{
  SIVA_LIGHT_HOUSE *alh;
  if (siva_num_free_alh) {
    alh = array_fetch(SIVA_LIGHT_HOUSE *,siva_free_alh_array,
                      siva_num_free_alh-1);
    siva_num_free_alh--;
  }else {
    alh = ALLOC(SIVA_LIGHT_HOUSE,1);
  }
  return alh;
}

/*---------------------------------------------------------------------------*/
static void
siva_release_alh(SIVA_LIGHT_HOUSE *alh) 
{
  array_insert(SIVA_LIGHT_HOUSE *,siva_free_alh_array,
               siva_num_free_alh,alh);;
  siva_num_free_alh++;
}

/*---------------------------------------------------------------------------*/
void
siva_find_alh_state(array_t *state_wrapper_array, int num_state)
{
  int i;
  int value;
  int num_alh;
  SIVA_CONV conv;
  int num_remained;
  SIVA_NODE alh_node;
  
  SIVA_LIGHT_HOUSE *alh; 
  SIVA_LIGHT_HOUSE *alhs[SIVA_DEFAULT_ALH_PICK_SIZE];

  /* less aggressive */
  int depth_upper =   siva_atpg_unfold_depth_upper;


  siva_atpg_unfold_depth_upper =
    (  siva_atpg_unfold_depth_upper >> 1);
  
  SIVA_VERB1_ARG1("Finding alh states ...\n");

  num_alh = siva_get_best_alh(alhs,SIVA_DEFAULT_ALH_PICK_SIZE);
  for(i=0;i<num_alh; i++){
    alh = alhs[i];
    value = bck_is_flag_set(alh->node,siva_alh_one_flag);
    conv.lh = alh;
    alh_node = bck_get_regular(alh->node);
    num_remained = siva_atpg_solve(&alh_node,&value,1,
                                   state_wrapper_array,num_state,
                                   conv.ptr,siva_store_alh_witness_state,NULL);
    
    if (num_remained==0) {
      bck_reset_flag(alh->node,siva_alh_one_flag);
      bck_reset_flag(alh->node,siva_alh_zero_flag);
    }    
  }
  siva_atpg_unfold_depth_upper =  depth_upper;
  
  /* Put back the unsolved ones; maybe with lower priority? */
  for(i=0;i<num_alh; i++){
    alh = alhs[i];
    if (bck_is_flag_set(alh->node,siva_alh_one_flag) ||
        bck_is_flag_set(alh->node,siva_alh_zero_flag)) {
      alh->priority <<= 1;
      siva_put_alh(alh->priority,alh);
    }else {
      siva_release_alh(alh);
    }
  }    
  SIVA_VERB1_ARG1("Finding alh states ...done\n");
}

/*---------------------------------------------------------------------------*/
static void
siva_store_alh_witness_state(SIVA_STATE *seed_state,
                             var_set_t *wit_state_vector,
                             int rel_depth,
                             int unfold_depth,
                             void *para)

{
  unsigned state_value;
  SIVA_STATE *wit_state;
  int priority_value;  
  int is_created;

  SIVA_CONV conv;
  
  /* only pick the witness state. ignore the rest of the sequence */
  if (rel_depth < unfold_depth) {
    return;
  }
  var_set_compress(&state_value,wit_state_vector,
                   siva_state_init_hash_value);
  wit_state = siva_get_or_create_state(state_value,
                                       seed_state->state_value,
                                       rel_depth,
                                       wit_state_vector->data,
                                       &is_created);  

  if (is_created == 0) {
    return;
  }
  conv.ptr = para;
  priority_value = conv.lh->priority; 
  siva_put_alh_state(priority_value,wit_state);
  SIVA_VERB1_ARG4("ALH: Witness State %x, state_value = %x priority = %d\n",
                  wit_state,state_value,priority_value);
}

/*--------------------------------------------------------------------------*/
int
siva_get_seed_alh_state(SIVA_STATE **state_p,
                        var_set_t *seed_state_vector)
{
  SIVA_KEY key;
  SIVA_CONV conv;
  while(siva_get_alh_state(&key,&conv)){
    /* regenerate state and check if valid state*/
    if (!siva_regenerate_state_vector(conv.state,
                                      seed_state_vector)){
      continue;
    }
    *state_p = conv.state;
    SIVA_VERB1_ARG3("ALH: Seed state value = %x, key = %u\n",
                    conv.state->state_value,key.unsigned_key);
    return 1;
  }  
  return 0;
}


/*--------------------------------------------------------------------------*/
void
siva_put_alh_state(unsigned priority_value, SIVA_STATE *state)
{
  SIVA_KEY key;
  SIVA_CONV conv;
  
  key.unsigned_key = priority_value;
  conv.state = state;
  siva_heap_insert(key,conv,siva_alh_state_heap_id);
}

/*--------------------------------------------------------------------------*/
int
siva_get_alh_state(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr)
{
  int found;
  found = siva_heap_extract_top(key_ptr,conv_ptr,siva_alh_state_heap_id);
  return found;
}

/*--------------------------------------------------------------------------*/
int
siva_is_alh_redundant(SIVA_LIGHT_HOUSE *alh)
{
  SIVA_NODE target;
  int is_redundant=0;
  if (alh->is_target_et) {
    if (alh->target.et->is_unreachable) {
      return 1;
    }
    target = siva_get_et_target_node(alh->target.et);
  }else {
    target = alh->target.node;
  }
  if (!bck_is_flag_set(target,siva_target_one_flag) &&
      !bck_is_flag_set(target,siva_target_zero_flag)){
    is_redundant = 1;
  }    
  return is_redundant;
}  

/*
  --------------------------------------------------------------------------
  Gets num_lh_req lighthouse from the heap; each lighthouse picked
  corresponds to atleast one unreached target
  --------------------------------------------------------------------------  
*/
static int
siva_get_best_alh(SIVA_LIGHT_HOUSE *alhs[],
                  int num_alh_req)
{
  SIVA_KEY key;
  SIVA_CONV conv;
  int num_alh = 0;
  SIVA_LIGHT_HOUSE *alh;
  SIVA_NODE target;
  int is_redundant;
  while(siva_get_alh(&key,&conv) && (num_alh < num_alh_req)){
    alh = conv.lh;
    is_redundant = siva_is_alh_redundant(alh);
    if (!is_redundant) {
      /* check if reset by simulation */
      if (!bck_is_flag_set(alh->node,siva_alh_one_flag) &&
          !bck_is_flag_set(alh->node,siva_alh_zero_flag)){
        is_redundant = 1;
      }    
    }
    if (is_redundant) {
      continue;
    }
    alhs[num_alh++] = alh;
    /* re-heapify later with lower priority ? */
    SIVA_VERB1_ARG3("ALH: %d priority value = %u\n",
                    bck_get_index(alh->node),
                    key.unsigned_key);
  }  
  return num_alh;
}


/*--------------------------------------------------------------------------*/
static void
siva_put_alh(unsigned priority_value, SIVA_LIGHT_HOUSE *alh)
{
  SIVA_KEY key;
  SIVA_CONV conv;
  int is_redundant = 0;

  key.unsigned_key = priority_value;
  conv.lh = alh;
  siva_heap_insert(key,conv,siva_alh_heap_id);
  SIVA_ERR_CHK1(siva_do_heap(siva_alh_heap_id,siva_check_alh_heap));
}

/*--------------------------------------------------------------------------*/
static int
siva_get_alh(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr)
{
  int found;    
  found = siva_heap_extract_top(key_ptr,conv_ptr,siva_alh_heap_id);
  SIVA_ERR_CHK1(siva_do_heap(siva_alh_heap_id,siva_check_alh_heap));
  return found;
}


/*
/*
  --------------------------------------------------------------------------  
  Those state variable whose valuation
  in the enlarge target has low probability of occurance are set as alh
  The priority of alh is also affected by the number of such low prob.   
  --------------------------------------------------------------------------
*/

void
siva_generate_alh(SIVA_ET *et)
{
  int i;
  int alh_needed;
  SIVA_NODE alh_node;
  unsigned priority;
  int value;
  SIVA_CONV conv;
  var_set_t mask_vector;
  var_set_t state_vector;
  SIVA_LIGHT_HOUSE *alh;
  int num_var_below_inertial_threshold = 0;
  int base_priority;
  
  if (!siva_alh_on) {
    return;
  }
  siva_set_state_vector(&mask_vector,et->mask);
  siva_set_state_vector(&state_vector,et->data);

  
  for(i=0;i<siva_num_latch;i++) {
    if (!var_set_get_elt(&mask_vector,i)) {
      continue;
    }
    value = var_set_get_elt(&state_vector,i);
    SIVA_ASSERT(!bck_is_zero(siva_latch_nodes[i].input)||
                !value,"This will never be reached");
    SIVA_ASSERT(!bck_is_one(siva_latch_nodes[i].input)||
                value,"This will never be reached");
    if (siva_get_latch_inertia(i,value) > SIVA_ALH_INERTIAL_THRESHOLD) {
      continue;
    }
    num_var_below_inertial_threshold++;
  }
  
  base_priority = (unsigned) (((float)num_var_below_inertial_threshold * 1000)/
                              siva_num_latch);
  
  for(i=0;i<siva_num_latch;i++) {
    if (!var_set_get_elt(&mask_vector,i)) {
      continue;
    }
    value = var_set_get_elt(&state_vector,i);
    SIVA_ASSERT(!bck_is_zero(siva_latch_nodes[i].input)||
                !value,"This will never be reached");
    SIVA_ASSERT(!bck_is_one(siva_latch_nodes[i].input)||
                value,"This will never be reached");
    priority = siva_get_latch_inertia(i,value) + 1;
    if (priority > SIVA_ALH_INERTIAL_THRESHOLD) {
      continue;
    }
    SIVA_ASSERT(!bck_is_constant(siva_latch_nodes[i].input),
                "input nodes should not be constant");
    value = value ^ bck_is_inverted(siva_latch_nodes[i].input);

    if (!siva_set_alh(siva_latch_nodes[i].input,value)) {
      continue;
    }
    alh = siva_alloc_alh();
    alh->target.et = et;
    alh_node = bck_get_regular(siva_latch_nodes[i].input);
    if (!value) {
      alh_node = bck_create_not(alh_node);
    }
    alh->node = alh_node;
    alh->priority = priority + base_priority;
    alh->is_target_et = 1;
    conv.lh = alh;
    bck_set_aux(alh_node,siva_lh_aux,conv.ptr);
    SIVA_VERB2_ARG6("setting alh: %d %d value %d priority %d base %d\n",
                    i,bck_get_index(alh_node),value,alh->priority,
                    base_priority);
    siva_put_alh(alh->priority,alh);
  }
}


/*--------------------------------------------------------------------------*/
static int
siva_set_alh(SIVA_NODE node, int value)
{
  int alh_needed = 0;
  if (value) {
    SIVA_ASSERT(!bck_is_flag_set(node,siva_alh_zero_flag),
                "This input has inertial 0");      
    if (bck_is_flag_set(node,siva_alh_one_flag)) {
      alh_needed = 0;
    }else {
      bck_set_flag(node,siva_alh_one_flag);
      alh_needed = 1;
    }
  }else {
    SIVA_ASSERT(!bck_is_flag_set(node,siva_alh_one_flag),
                "This input has inertial 1");      
    if (bck_is_flag_set(node,siva_alh_zero_flag)) {
      alh_needed = 0;
    }else {
      bck_set_flag(node,siva_alh_zero_flag);
      alh_needed = 1;
    }
  }      
  return alh_needed;
}


/*--------------------------------------------------------------------------*/
static void
siva_print_alh_heap(void)
{
  SIVA_VERB1_ARG1("Dumping ALH heap\n");
  siva_do_heap(siva_alh_heap_id,siva_alh_heap_srv);
}

/*--------------------------------------------------------------------------*/
static void
siva_alh_heap_srv(int pos, SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr)
{
  SIVA_VERB1_ARG5("%d: node = %d Priority = %d Key = %d\n",
                  pos,bck_get_index(conv_ptr->lh->node),
                  conv_ptr->lh->priority,key_ptr->unsigned_key);
}
  
/*--------------------------------------------------------------------------*/
static void
siva_check_alh_heap(int pos, SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr)
{
  if (conv_ptr->lh->priority != key_ptr->unsigned_key) {    
    SIVA_VERB1_ARG4("%d: Priority = %d Key = %d\n",
                    pos,conv_ptr->lh->priority,key_ptr->unsigned_key);
    assert(0);
  }
}








