/**CFile**********************************************************************

  FileName                   [siva_cmd.c]

  PackageName                [siva_core]

  Synopsis                   [Handle commands]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]

******************************************************************************/

#include "siva.h"

static int siva_read_blif_cmd(int argc, char ** argv);
static int siva_pre_process_cmd(int argc, char ** argv);
static int siva_solve_cmd(int argc, char ** argv);
static int siva_build_pi_free_target_cmd(int argc, char ** argv);
static int siva_solve_init_cmd(int argc, char ** argv);
static int siva_solve_quit_cmd(int argc, char ** argv);
static int siva_write_unresolved_target_cmd(int argc, char ** argv);
static int siva_write_blif_cmd(int argc, char ** argv);
static int siva_mv2blif_cmd(int argc, char **argv);
static int siva_symbolic_sim_cmd(int argc, char **argv);

/*---------------------------------------------------------------------------*/
/* Initializes the test package.                                             */
void 
siva_init(void)
{
  /*
   * Add a command to the global command table.  By using the leading
   * underscore, the command will be listed under "help -a" but not "help".
   */

  siva_init_ext_interface_done = 0;
  siva_solve_mode = 0;
  siva_init_latch_activity_done = 0;
  siva_build_pi_free_bck_network_done = 0;  

  /*
    Following parameters are set here although can be made as a part
    of user interface
  */
  siva_atpg_back_track_limit = SIVA_DEFAULT_ATPG_BACK_TRACK_LIMIT;
  siva_atpg_num_trial = SIVA_DEFAULT_ATPG_NUM_TRIAL;
  siva_atpg_unfold_depth_upper = SIVA_DEFAULT_ATPG_UNFOLD_DEPTH_UPPER;
  siva_atpg_unfold_depth_lower = SIVA_DEFAULT_ATPG_UNFOLD_DEPTH_LOWER;
  siva_atpg_conflict_limit = SIVA_DEFAULT_ATPG_CONFLCIT_LIMIT;
  siva_atpg_deduction_level = SIVA_DEFAULT_ATPG_DEDUCTION_LEVEL;
  siva_atpg_learning_level = SIVA_DEFAULT_ATPG_LEARNING_LEVEL;

  /* 
     making the program deterministic. same run at diff times with
     same sequence of commands should produce identical results 
  */
  srand(0);
  siva_rand_seed = SIVA_SEED(rand());

  Cmd_CommandAdd("read_blif", siva_read_blif_cmd,0);
  Cmd_CommandAdd("pre_process", siva_pre_process_cmd,0);
  Cmd_CommandAdd("solve_init", siva_solve_init_cmd,0);
  Cmd_CommandAdd("solve", siva_solve_cmd,0);
  Cmd_CommandAdd("write_unresolved_target", siva_write_unresolved_target_cmd,0);
  Cmd_CommandAdd("solve_quit", siva_solve_quit_cmd,0);
  Cmd_CommandAdd("build_pi_free_target",
                 siva_build_pi_free_target_cmd,0);
  Cmd_CommandAdd("mv2blif",siva_mv2blif_cmd,0);
  Cmd_CommandAdd("symbolic_sim",siva_symbolic_sim_cmd,0);
  Cmd_CommandAdd("write_blif",siva_write_blif_cmd,0);

  /* check for endian-ness */
  {
    int x = 1;
    if ((*(char *)&x) == 0) {
      siva_little_endian = 1;      
    }
    if (siva_little_endian) {
      //printf("It is little endian machine\n");
    }else {
      //printf("It is big endian machine\n");
    }
  } 

}


/*---------------------------------------------------------------------------*/
/* Ends the test package.*/
void
siva_quit(void)   
{
}


/*
  -----------------------------------------------------------------------------
  Handler for read_blif command.
  Invokes siva_exec_read_blif which reads input blif file.
  Consecutive reads overrides the previous read_blif command.
  However, in solve mode, read_blif is not allowed.
  -----------------------------------------------------------------------------
*/
static int
siva_read_blif_cmd(int argc, char **argv)
{
  int c;

  if (siva_solve_mode == 1){    
    fprintf(stdout,"Quit from solve_mode first. Command <read_blif> not performed\n");
    goto usage;
  }

  siva_input_file = NULL;
  siva_verbosity_level= SIVA_DEFAULT_VERBOSITY_LEVEL;
  siva_error_check_level = SIVA_DEFAULT_ERROR_CHECK_LEVEL;
  siva_output_file = stdout;

  util_getopt_reset(); 
  while((c = util_getopt(argc,argv,"o:e:v:h")) != EOF){
    switch(c){
    case 'o':
      if (util_optarg){
        siva_output_file = fopen(util_optarg,"w");
      }
      break;
    case 'e':
      siva_error_check_level = atoi(util_optarg);
      break;
    case 'v':
      siva_verbosity_level = atoi(util_optarg);
      break;
    default:
        goto usage; 	
    }
  }
  if (argc-1 != util_optind){
    goto usage;
  }

  siva_input_file = fopen(argv[util_optind],"r");  
  if ( siva_input_file == NULL) {
    goto usage;
  }
  if (siva_exec_read_blif() == 0){
    fprintf(stderr,"Error in reading file %s\n",argv[util_optind]);
    fclose(siva_input_file);
    return 0;
  }
  fclose(siva_input_file);
  return 1;    

  
 usage:

  fprintf(stdout, "read_blif [-o <file>] [-v <verbosity>] [-e <error check>] <blif-file>\n\n");
  fprintf(stdout, "   -o <file>\t\tOutput file\n");
  fprintf(stdout, "   -e <error check>\tError check level. Default %d.\n",
          SIVA_DEFAULT_ERROR_CHECK_LEVEL);  
  fprintf(stdout, "   -v <verbosity>\tVerbosity level. Default %d.\n",
          SIVA_DEFAULT_VERBOSITY_LEVEL);
  fprintf(stdout, "   -h      \t\tPrint the command usage.\n");
   return 0;             /* error exit */

}

/*
  ------------------------------------------------------------------------
  Handler for pre_process command.
  Invokes siva_exec_pre_process to eliminate easy targets.
  Consective invokation is allowed.
  However, in solve mode, this command is not allowed.
  -------------------------------------------------------------------------
*/
static int
siva_pre_process_cmd(int argc, char **argv)
{
  int c;

  if ( siva_solve_mode == 1){
    fprintf(stdout,"Quit from solve_mode first. Command <pre_process> not performed\n");
    goto usage;
  }
  
  if ( siva_init_ext_interface_done == 0){
    fprintf(stdout,"Error: Blif file not read\n");
    goto usage;
  }

  siva_sim_depth = SIVA_DEFAULT_SIM_DEPTH;
  siva_pre_process_num_trial = SIVA_DEFAULT_PRE_PROCESS_NUM_TRIAL;  
  siva_target_file = NULL;
  siva_max_time_limit = SIVA_DEFAULT_MAX_TIME_LIMIT;  
  siva_pruned_target_file = NULL;
  siva_verbosity_level = SIVA_DEFAULT_VERBOSITY_LEVEL;
  siva_error_check_level = SIVA_DEFAULT_ERROR_CHECK_LEVEL;


  
  util_getopt_reset(); 
  while((c = util_getopt(argc,argv,"n:r:t:l:p:v:e:h")) != EOF){
    switch(c){
    case 'n':
      siva_sim_depth  = atoi(util_optarg);
      break;
    case 'r':
      siva_pre_process_num_trial = atoi(util_optarg);
      break;
    case 't': 
      if (util_optarg){
        siva_target_file = fopen(util_optarg,"r");
      }
      break;
    case 'l':
      siva_max_time_limit = 1000 * atol(util_optarg);
      break;      
    case 'p':
      if (util_optarg){
        siva_pruned_target_file = fopen(util_optarg,"w");
      }
      break;
    case 'e':
      siva_error_check_level = atoi(util_optarg);
      break;
    case 'v':
      siva_verbosity_level = atoi(util_optarg);
      break;	
    default: 
        goto usage;
    }
  }

  siva_exec_pre_process();

  if (siva_target_file){
    fclose(siva_target_file);
  }
  if (siva_pruned_target_file){
    fclose(siva_pruned_target_file);
  }

  return 1;

 usage:

  fprintf(stdout, "pre_process [-n <depth>] [-r <repeat>] [-t <target>]\n");
  fprintf(stdout, "            [-l <time>] [-p <unresolved>]\n");
  fprintf(stdout, "            [-e <error_check>] [-v <verbosity>] [-h]\n\n");
  fprintf(stdout, "   -n <depth>\t\tDepth of random simulation. Default %d.\n",
          SIVA_DEFAULT_SIM_DEPTH);
  fprintf(stdout, "   -l <time>\t\tTime limit (sec) on execution. Default %d.\n",
          SIVA_DEFAULT_MAX_TIME_LIMIT/1000);
  fprintf(stdout, "   -r <repeat>\t\tNumber of restarts. Default  %d.\n",
          SIVA_DEFAULT_PRE_PROCESS_NUM_TRIAL);
  fprintf(stdout, "   -t <target>\t\tInput target file.\n");
  fprintf(stdout, "   -p <unresolved>\tFile to write unresolved target.\n");
  fprintf(stdout, "   -e <error check>\tError check level. Default %d.\n",
          SIVA_DEFAULT_ERROR_CHECK_LEVEL);  
  fprintf(stdout, "   -v <verbosity>\tVerbosity level. Default %d.\n",
          SIVA_DEFAULT_VERBOSITY_LEVEL);
  fprintf(stdout, "   -h      \t\tPrint the command usage.\n");
  return 0;             /* error exit */
   
}

/*
  ------------------------------------------------------------------------
  Handler for solve command.
  Invokes siva_exec_solve to solve the hard targets using different engines.
  User should call solve_init command before this command.
  -------------------------------------------------------------------------
*/
static int
siva_solve_cmd(int argc, char ** argv)
{
  int c;

  if ( siva_solve_mode == 0){
    fprintf(stdout,"Error: Allowed only in solve mode\n");
    goto usage;
  }

  siva_verbosity_level = SIVA_DEFAULT_VERBOSITY_LEVEL;
  siva_sim_depth = SIVA_DEFAULT_SIM_DEPTH;
  siva_symbolic_sim_depth = SIVA_DEFAULT_SYMBOLIC_SIM_DEPTH;  
  siva_alh_on = 0;
  siva_control_group = 0;
  siva_et_on = 0;
  siva_bfs_state_on = 0;
  siva_update_state_depth = SIVA_DEFAULT_SIM_DEPTH;
  siva_solve_resume = 0;
  siva_max_time_limit = SIVA_DEFAULT_MAX_TIME_LIMIT;
  siva_error_check_level = SIVA_DEFAULT_ERROR_CHECK_LEVEL;
  siva_atpg_num_trial = SIVA_DEFAULT_ATPG_NUM_TRIAL; 

  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "n:u:E:Lc:Bd:r:l:e:v:h")) != EOF) {
    switch(c) {
      /* Duration of the run in sec*/
    case 'n':
      siva_sim_depth  = atoi(util_optarg);
      break;
    case 'u':
      siva_symbolic_sim_depth  = atoi(util_optarg);
      break;
    case 'E':
      siva_et_on = 1;
      siva_target_enlarge_factor = atoi(util_optarg);
      break;
    case 'L':
      siva_alh_on = 1;
      break;	
    case 'c':
      siva_control_group = atoi(util_optarg);
      break;        
    case 'B':
      siva_bfs_state_on = 1;
      break;
    case 'd':
      siva_update_state_depth = atoi(util_optarg);
      break;
    case 'r':
      siva_solve_resume = 1;
      break;      
    case 'l':
      siva_max_time_limit = 1000 * atol(util_optarg);
      break;
    case 'e':
      siva_error_check_level = atoi(util_optarg);
      break;
    case 'v':
      siva_verbosity_level = atoi(util_optarg);
      break;	
    default:
      goto usage;
    }
  }

  siva_exec_solve();

  return 1;

  usage:

  fprintf(stdout, "solve [-n <rand depth>] [-u <fwd depth>] [-E <bwd depth>]\n");
  fprintf(stdout, "      [-L] [-c <group #>] [-B] [-d <storage depth>]\n");
  fprintf(stdout, "      [-r] [-l <time>] [-e <error check>] [-v <verbosity>] [-h]\n\n");
  
  fprintf(stdout, "   -n <rand depth>\tDepth of random simulation. Default %d.\n",
          SIVA_DEFAULT_SIM_DEPTH);
  fprintf(stdout, "   -u <fwd depth>\tDepth of symbolic simulation. Default %d\n",
          SIVA_DEFAULT_SYMBOLIC_SIM_DEPTH);
  fprintf(stdout, "   -E <bwd depth>\tDepth for trget enlargement. Default disable.\n");
  fprintf(stdout, "   -L      \t\tEnable automatic generation of lighthouses. Default disable.\n");
  fprintf(stdout, "   -c <group #>\t\tSpecify # of control group. Default disable.\n");
  fprintf(stdout, "   -B      \t\tEnable states visited during bfs traversal. Default disable.\n");
  fprintf(stdout, "   -d <storage depth>\tRestricts state storage depth. Default all.\n");
  fprintf(stdout, "   -r      \t\tResume. Default continue from initial state.\n");
  fprintf(stdout, "   -l <time>\t\tTime limit (sec) on execution. Default %d\n",
          SIVA_DEFAULT_MAX_TIME_LIMIT/1000);
  fprintf(stdout, "   -e <error check>\tError check level. Default %d.\n",
          SIVA_DEFAULT_ERROR_CHECK_LEVEL);
  fprintf(stdout, "   -v <verbosity>\tVerbosity level. Default %d.\n",
          SIVA_DEFAULT_VERBOSITY_LEVEL);
  fprintf(stdout, "   -h      \t\tPrint the command usage.\n");

  return 0;             /* error exit */


}

/*
  ------------------------------------------------------------------------
  Handler for symbolic_solve command.
  Invokes siva_symbolic_sim to solve the hard targets using symbolic simulation.
  User should call solve_init command before this command.
  -------------------------------------------------------------------------
*/
static int
siva_symbolic_sim_cmd(int argc, char ** argv)
{
  int c;

  if ( siva_solve_mode == 0){
    fprintf(stdout,"Error: Allowed only in solve mode\n");
    goto usage;
  }


  siva_max_time_limit = SIVA_DEFAULT_MAX_TIME_LIMIT;  
  siva_verbosity_level = SIVA_DEFAULT_VERBOSITY_LEVEL;
  siva_error_check_level = SIVA_DEFAULT_ERROR_CHECK_LEVEL;
  siva_symbolic_sim_depth = SIVA_DEFAULT_SYMBOLIC_SIM_DEPTH;
  siva_symbolic_sim_lower_depth = 0;
  siva_sym_check_unreachable = 0;  
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "u:s:Ul:e:v:h")) != EOF) {
    switch(c) {
    case 'u':
      siva_symbolic_sim_depth = atoi(util_optarg);
      break;
    case 's':
      siva_symbolic_sim_lower_depth = atoi(util_optarg);
      break;
    case 'U':
      siva_sym_check_unreachable = 1;
      break;      
    case 'l':
      siva_max_time_limit = 1000 * atol(util_optarg);
      break;
    case 'e':
      siva_error_check_level = atoi(util_optarg);
      break;
    case 'v':
      siva_verbosity_level = atoi(util_optarg);
      break;	
    default:
      goto usage;
    }
  }

  siva_symbolic_sim();

  return 1;

  usage:

  fprintf(stdout, "symbolic_sim [-u <fwd depth>] [-s <start depth>] [-U]\n");
  fprintf(stdout, "             [-l <time>] [-e <error check>] [-v <verbosity>] [-h]\n\n");
  fprintf(stdout, "   -u <fwd depth>\tDepth of symbolic simulation. Default %d.\n",
          SIVA_DEFAULT_SYMBOLIC_SIM_DEPTH);
  fprintf(stdout, "   -s <start depth>\tStart depth of symbolic simulation. Default 0.\n");
  fprintf(stdout, "   -U      \t\tEnable check for unreachability. Default disable.\n");
  fprintf(stdout, "   -l <time>\t\tTime limit (sec) on execution. Default %d.\n",
          SIVA_DEFAULT_MAX_TIME_LIMIT/1000);
  fprintf(stdout, "   -e <error check>\tError check level. Default %d.\n",
          SIVA_DEFAULT_ERROR_CHECK_LEVEL);
  fprintf(stdout, "   -v <verbosity>\tVerbosity level. Default %d.\n",
          SIVA_DEFAULT_VERBOSITY_LEVEL);
  fprintf(stdout, "   -h      \t\tPrint the command usage.\n");

  return 0;             /* error exit */
}

/*
  ------------------------------------------------------------------------
  Handler for build_pi_free_target command.
  Invokes siva_build_pi_free_bck_network to enlarge and quantify targets.
  User should call solve_init command before this command.
  Reinvokation not allowed.
  -------------------------------------------------------------------------
*/
static int
siva_build_pi_free_target_cmd(int argc, char ** argv)
{
  int c;

  if ( siva_solve_mode == 0){
    fprintf(stdout,"Error: Allowed only in solve mode\n");
    goto usage;
  }

  if (siva_build_pi_free_bck_network_done){
    fprintf(stdout,"Reinvokation not allowed.\n");
    goto usage;
  }

  siva_verbosity_level = SIVA_DEFAULT_VERBOSITY_LEVEL;  
  siva_unfold_depth=0;
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "u:v:h")) != EOF) {
    switch(c) {
    case 'u':
      siva_unfold_depth = atoi(util_optarg); 
      if (siva_unfold_depth < 0){
        goto usage;
      }
      break;
    case 'v':
      siva_verbosity_level = atoi(util_optarg);
      break;	
    default:
      goto usage;
    }
  }

  if (siva_build_pi_free_bck_network()){
    siva_build_pi_free_bck_network_done = 1;
  }


  return 1;

  usage:

  fprintf(stdout, "build_pi_free_target [-u <unfold depth>] [-v <verbosity>] [-h]\n");
  fprintf(stdout, "   -u <unfold depth>\t\tUnfold depth. Default 0.\n");
  fprintf(stdout, "   -h      \t\tPrint the command usage\n");
  fprintf(stdout, "   -v <verbosity>\tVerbosity level. Default %d.\n",
          SIVA_DEFAULT_VERBOSITY_LEVEL);
  fprintf(stdout, "   -h      \t\tPrint the command usage.\n");

  return 0;             /* error exit */


}

/*
  ------------------------------------------------------------------------
  Handler for solve_init command.
  Invokes siva_exec_solve_init to prepare for solving using
  several solve engines.
  Reinvokation is not allowed.
  -------------------------------------------------------------------------
*/
static int
siva_solve_init_cmd(int argc, char ** argv)
{
  int c;
  int init_stat=0;

  
  if ( siva_solve_mode == 1){
    fprintf(stdout,"Warn: Initialization is done. Use <solve_quit> to quit.\n");
    goto usage;
  }

  if ( siva_init_ext_interface_done == 0){
    fprintf(stdout,"Error: Blif file not read\n");
    goto usage;
  }

  siva_build_pi_free_bck_network_done = 0;    
  siva_control_state_check_on = 0;

  
  siva_target_file = NULL;
  siva_max_num_control_group = 0;
  siva_companding_method = NORMAL;
  siva_control_sub_group_on = 0;
  siva_partition_overlap_threshold = SIVA_DEFAULT_PARTITION_OVERLAP_THRESHOLD;
  siva_sim_depth = SIVA_DEFAULT_SIM_DEPTH;
  siva_pre_process_num_trial = SIVA_DEFAULT_PRE_PROCESS_NUM_TRIAL;    
  siva_debug_node_file = NULL;
  siva_error_check_level = SIVA_DEFAULT_ERROR_CHECK_LEVEL;
  siva_verbosity_level = SIVA_DEFAULT_VERBOSITY_LEVEL;

  
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "t:c:A:ST:n:r:D:e:v:h")) != EOF) {
    switch(c) {
    case 't': /* target file name */
      if (util_optarg){
        siva_target_file = fopen(util_optarg,"r");
      }
      break;
    case 'c': 
      siva_max_num_control_group = atoi(util_optarg);
      break;
    case 'A': 
      siva_companding_method = atoi(util_optarg);
      break;
    case 'S':
      siva_control_sub_group_on = 1;
      break;
    case 'T':
      siva_partition_overlap_threshold = atof(util_optarg);
      break;
    case 'n':
      siva_sim_depth  = atoi(util_optarg);
      break;
    case 'r':
      siva_pre_process_num_trial = atoi(util_optarg);
      break;
    case 'D':
      if (util_optarg){
        siva_debug_node_file = fopen(util_optarg,"r");
      }
      break;
    case 'e':
      siva_error_check_level = atoi(util_optarg);
      break;
    case 'v':
      siva_verbosity_level = atoi(util_optarg);
      break;	
    default:
      goto usage;
    }
  }

  init_stat = siva_exec_solve_init();

  if (siva_target_file){
    fclose(siva_target_file);
  }

  if (siva_debug_node_file){
    fclose(siva_debug_node_file);
  }

  if (init_stat){       
    siva_solve_mode = 1;    
  }

  return 1;

  usage:
  fprintf(stdout, "solve_init [-t <target>] [-c <max group>] [-A <companding>]\n");
  fprintf(stdout, "           [-S] [-T <threshold>] [-n <sim depth>] [-r <restarts>]\n");
  fprintf(stdout, "           [-D <debug node>] [-e <error_check>] [-v <verbosity>] [-h]\n\n");

  fprintf(stdout, "   -t <target>\t\tInput target file.\n");
  fprintf(stdout, "   -c <max group>\tMax number of active latch groups. Default 0.\n");
  fprintf(stdout, "   -A <companding>\t2 = ALOG, 1 = LOG, 0 = NORMAL (Default)\n");
  fprintf(stdout, "   -S      \t\tEnable partitioning of latch groups. Default Disable.\n");
  fprintf(stdout, "   -T <threshold>\tPartition overlap threshold. Default %f\n",          
          SIVA_DEFAULT_PARTITION_OVERLAP_THRESHOLD);
  fprintf(stdout, "   -n <depth>\t\tDepth of random simulation for latch activity. Default %d.\n",
          SIVA_DEFAULT_SIM_DEPTH);
  fprintf(stdout, "   -r <repeat>\t\tNumber of simulation restarts for latch activity. Default  %d.\n",
          SIVA_DEFAULT_PRE_PROCESS_NUM_TRIAL);
  
  fprintf(stdout, "   -D <debug node>\tDebug node file.\n");  
  fprintf(stdout, "   -e <error check>\tError check level. Default %d.\n",
          SIVA_DEFAULT_ERROR_CHECK_LEVEL);
  fprintf(stdout, "   -v <verbosity>\tVerbosity level. Default %d.\n",
          SIVA_DEFAULT_VERBOSITY_LEVEL);
  fprintf(stdout, "   -h      \t\tPrint the command usage.\n");

  return 0;             /* error exit */


}

/*
  ------------------------------------------------------------------------
  Handler for solve_quit command.
  Invokes siva_exec_solve_quit to release all resources used by 
  the solve engines.
  Reinvokation is not allowed.
  -------------------------------------------------------------------------
*/
static int
siva_solve_quit_cmd(int argc, char ** argv)
{
  int c;

  if ( siva_solve_mode == 0){
    fprintf(stdout,"Error: Allowed only in solve mode\n");
    goto usage;
  }

  siva_verbosity_level= SIVA_DEFAULT_VERBOSITY_LEVEL;
  siva_error_check_level= SIVA_DEFAULT_ERROR_CHECK_LEVEL;

  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "e:v:h")) != EOF) {
    switch(c) {
    case 'e':
      siva_error_check_level = atoi(util_optarg);
      break;        
    case 'v':
      siva_verbosity_level = atoi(util_optarg);
      break;
    default:
      goto usage;
    }
  }
  siva_exec_solve_quit();
  siva_solve_mode = 0;
  return 1;

  usage:
  fprintf(stdout, "solve_quit [-e <error_check>] [-v <verbosity>] [-h]\n\n");
  fprintf(stdout, "   -e <error check>\tError check level. Default %d.\n",
          SIVA_DEFAULT_ERROR_CHECK_LEVEL);
  fprintf(stdout, "   -v <verbosity>\tVerbosity level. Default %d.\n",
          SIVA_DEFAULT_VERBOSITY_LEVEL);
  fprintf(stdout, "   -h      \t\tPrint the command usage.\n");
  return 0;             /* error exit */


}

/*
  ------------------------------------------------------------------------
  Handler for write_unresolved_target command.
  Invokes siva_write_target_file to store the unresolved targets.
  Can invoke only in solve_mode.
  Reinvokation is allowed.
  -------------------------------------------------------------------------
*/
static int
siva_write_unresolved_target_cmd(int argc, char ** argv)
{
  int c;

  if ( siva_solve_mode == 0){
    fprintf(stdout,"Error: Allowed only in solve mode\n");
    goto usage;
  }
  
  siva_verbosity_level= SIVA_DEFAULT_VERBOSITY_LEVEL;
  siva_error_check_level= SIVA_DEFAULT_ERROR_CHECK_LEVEL;  
  siva_pruned_target_file = stdout;
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "p:e:v:h")) != EOF) {
    switch(c) {
    case 'p': /* pruned target file name */
      if (util_optarg){
        siva_pruned_target_file = fopen(util_optarg,"w");
      }
      break;
    case 'e':
      siva_error_check_level = atoi(util_optarg);
      break;        
    case 'v':
      siva_verbosity_level = atoi(util_optarg);
      break;
    default:
        goto usage;
    }
  }

  siva_write_target_file();
  
  if (siva_pruned_target_file != stdout){
    fclose(siva_pruned_target_file);
  }

  return 1;

  usage:

  fprintf(stdout, "write_unresolved_target [-p <file>] [-e <error_check>]\n");
  fprintf(stdout, "                        [-v <verbosity>]] [-h]\n\n");
  fprintf(stdout, "   -p <file>\t\tFile to write unresolved target. Default CONSOLE\n");
  fprintf(stdout, "   -e <error check>\tError check level. Default %d.\n",
          SIVA_DEFAULT_ERROR_CHECK_LEVEL);
  fprintf(stdout, "   -v <verbosity>\tVerbosity level. Default %d.\n",
          SIVA_DEFAULT_VERBOSITY_LEVEL);
  fprintf(stdout, "   -h      \t\tPrint the command usage.\n");
  return 0;             /* error exit */


}

/*
  ------------------------------------------------------------------------
  Handler for write_blif command.
  Invokes siva_exec_write_bck to write the internal representation in blif.
  Reinvokation is allowed.
  -------------------------------------------------------------------------
*/
static int
siva_write_blif_cmd(int argc, char ** argv)
{
  int c;
  if ( siva_solve_mode == 1){
    fprintf(stdout,"Quit from solve_mode first. Command <write_blif> not performed\n");
    goto usage;
  }
  
  if ( siva_init_ext_interface_done == 0){
    fprintf(stdout,"Error: Blif file not read\n");
    goto usage;
  }

  siva_verbosity_level= SIVA_DEFAULT_VERBOSITY_LEVEL;
  siva_error_check_level= SIVA_DEFAULT_ERROR_CHECK_LEVEL;  
  siva_bck_out_file = stdout;
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "o:e:v:h")) != EOF) {
    switch(c) {
    case 'o': 
      if (util_optarg){
        siva_bck_out_file = fopen(util_optarg,"w");
      }
      break;
    case 'e':
      siva_error_check_level = atoi(util_optarg);
      break;        
    case 'v':
      siva_verbosity_level = atoi(util_optarg);
      break;        
      default:
        goto usage;
    }
  }
  siva_exec_write_bck();  
  if (siva_bck_out_file != stdout){
    fclose(siva_bck_out_file);
  }
  return 1;

  usage:

  fprintf(stdout, "write_blif [-o <file>] [-e <error_check>]\n");
  fprintf(stdout, "           [-v <verbosity>]] [-h]\n\n");
  fprintf(stdout, "   -o <file>\t\tFile to write internal representation in blif.\n");
  fprintf(stdout, "            \t\tDefault CONSOLE.\n");
  fprintf(stdout, "   -e <error check>\tError check level. Default %d.\n",
          SIVA_DEFAULT_ERROR_CHECK_LEVEL);
  fprintf(stdout, "   -v <verbosity>\tVerbosity level. Default %d.\n",
          SIVA_DEFAULT_VERBOSITY_LEVEL);
  fprintf(stdout, "   -h      \t\tPrint the command usage.\n");
  return 0;             /* error exit */


}


/*
  ------------------------------------------------------------------------
  Handler for mv2blif command.
  Invokes siva_exec_mv2blif to convert mv to blif.
  -------------------------------------------------------------------------
*/

/*---------------------------------------------------------------------------*/
static int
siva_mv2blif_cmd(int argc, char ** argv)
{
  int c;
  siva_verbosity_level= SIVA_DEFAULT_VERBOSITY_LEVEL;
  siva_blif_file = stdout;
  siva_mv_file = NULL;  
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "w:v:h")) != EOF) {
    switch(c) {
    case 'w': /* blif file name */
      if (util_optarg){
        siva_blif_file = fopen(util_optarg,"w");
      }
      break;
    case 'v':
        siva_verbosity_level = atoi(util_optarg);
        break;	
    default:
      goto usage;
    }
  }
  if (argc-1 != util_optind){
    goto usage;
  }  
  siva_mv_file = fopen(argv[util_optind],"r");  
  if ( siva_mv_file == NULL) {
    goto usage;
  }
  siva_exec_mv2blif();
  fclose(siva_mv_file);
  if (siva_blif_file != stdout){
    fclose(siva_blif_file);
  }
  return 1;
  usage:

  fprintf(stdout, "mv2blif [-w <blif file>] [-v <verbosity>] <mv_file>\n");
  fprintf(stdout, "   -w <arg>\t\tSpecify blif file. Default CONSOLE.\n");
  fprintf(stdout, "   -v <verbosity>\tVerbosity level. Default %d.\n",
          SIVA_DEFAULT_VERBOSITY_LEVEL);
  fprintf(stdout, "   -h      \t\tPrint the command usage.\n");
  return 0;             /* error exit */

}  
    
