/*CFile**********************************************************************

  FileName                   [siva_complex_node.c]

  PackageName                [siva_core]

  Synopsis                   [Complex Node]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]
******************************************************************************/

#include "siva.h"

#define SIVA_DEFAULT_COMPLEX_NODE_TABLE_SIZE 1024
#define SIVA_COMPLEX_NODE_GROWTH_FACTOR 1

#define SIVA_CLEAR_VISIT(x) ((x)->visited = 0)
#define SIVA_GET_NODE(x) ((x)->bck)
#define SIVA_SET_NODE(x,y) ((x)->bck = y)
#define SIVA_GET_TARGET_ZERO(x) ((x)->tbl.target_zero)
#define SIVA_SET_TARGET_ZERO(x) ((x)->tbl.target_zero=1)
#define SIVA_CLEAR_TARGET_ZERO(x) ((x)->tbl.target_zero=0)
#define SIVA_GET_TARGET_ONE(x) ((x)->tbl.target_one)
#define SIVA_SET_TARGET_ONE(x) ((x)->tbl.target_one=1)
#define SIVA_CLEAR_TARGET_ONE(x) ((x)->tbl.target_one=0)
#define SIVA_IS_DEBUG_NODE(x) (x->tbl.debug_node == 1)
#define SIVA_SET_DEBUG_NODE(x) (x->tbl.debug_node = 1)
#define SIVA_CLEAR_DEBUG_NODE(x) (x->tbl.debug_node = 0)
#define SIVA_IS_LATCH_INPUT(x) (x->tbl.latch_input == 1)

static unsigned siva_total_num_complex_node;
static unsigned siva_num_complex_node_allocated;
static unsigned siva_num_complex_node;
static SIVA_COMPLEX_NODE *siva_complex_nodes;
static array_t *siva_complex_node_ptr_array;

static unsigned siva_complex_node_table_size;
static unsigned siva_num_hash_complex_node;
static SIVA_COMPLEX_NODE **siva_complex_node_table;

static void siva_resize_complex_node_table(void);
static unsigned siva_complex_node_hash(char *string);
static unsigned siva_string_hash(char *string);

/*---------------------------------------------------------------------------*/
void 
siva_init_complex_node(void)
{ 
  siva_complex_node_ptr_array = array_alloc(SIVA_COMPLEX_NODE *,16);
  siva_num_complex_node_allocated = 8192;
  siva_complex_nodes = ALLOC(SIVA_COMPLEX_NODE,
			     siva_num_complex_node_allocated);
  array_insert_last(SIVA_COMPLEX_NODE *,siva_complex_node_ptr_array,
                    siva_complex_nodes);

  siva_num_complex_node = 0;
  siva_total_num_complex_node = 0;
  siva_input_design_name = NIL(char);
  siva_init_complex_node_table();  
}

/*---------------------------------------------------------------------------*/
void 
siva_quit_complex_node(void)
{ 
  register int i,j;
  unsigned ptr_size = 8192;

  siva_quit_complex_node_table();

  if (siva_input_design_name){
    siva_free_string(siva_input_design_name);
  }
  for(j=0;j<array_n(siva_complex_node_ptr_array);j++){
    SIVA_COMPLEX_NODE *ptr = array_fetch(SIVA_COMPLEX_NODE *,
                                         siva_complex_node_ptr_array,j);

    if ( j == (array_n(siva_complex_node_ptr_array)-1)) {
       ptr_size = siva_num_complex_node;
    }
       
    for(i=0;i<ptr_size;i++){
      siva_free_string(ptr[i].name);
      siva_free_complex_node_input(&ptr[i]);
      siva_free_complex_node_entry(&ptr[i]);
    }
    FREE(ptr);
    ptr_size *= SIVA_COMPLEX_NODE_GROWTH_FACTOR;
  }
  array_free(siva_complex_node_ptr_array);
  siva_num_complex_node_allocated = 0;
  siva_num_complex_node = 0;
  siva_total_num_complex_node = 0;
}

/*-----------------------------------------`----------------------------------*/
void
siva_do_for_all_complex_node(void (*service)(SIVA_COMPLEX_NODE *node))
{

  register int i,j;
  unsigned ptr_size = 8192;
  for(j=0;j<array_n(siva_complex_node_ptr_array);j++){
    SIVA_COMPLEX_NODE *ptr = array_fetch(SIVA_COMPLEX_NODE *,
                                         siva_complex_node_ptr_array,j);
  
    if ( j == (array_n(siva_complex_node_ptr_array)-1)) {
       ptr_size = siva_num_complex_node;
    }
              
    for(i=0;i<ptr_size;i++){
      SIVA_VERB3_ARG2("Invoking %s..\n",ptr[i].name);
      service(&ptr[i]);
    }
    ptr_size *= SIVA_COMPLEX_NODE_GROWTH_FACTOR;
  }
    
}

/*---------------------------------------------------------------------------*/
/* Compare-function to topologically order nodes                             */
int
siva_node_cmp(const void *node_p1,const void *node_p2)
{
   return bck_compare_ids(*(SIVA_NODE *)node_p1,*(SIVA_NODE *)node_p2);
}
/*---------------------------------------------------------------------------*/
int
siva_complex_node_level_cmp(const void *node_p1,const void *node_p2)
{
   return (*(SIVA_COMPLEX_NODE **)node_p1)->level - 
          (*(SIVA_COMPLEX_NODE **)node_p2)->level;
}
/*---------------------------------------------------------------------------*/
int
siva_complex_node_bck_cmp(void *node_p1,void *node_p2)
{
   return bck_compare_ids(SIVA_GET_NODE(*(SIVA_COMPLEX_NODE **)node_p1),
                             SIVA_GET_NODE(*(SIVA_COMPLEX_NODE **)node_p2));
}


/*---------------------------------------------------------------------------*/
void
siva_set_target_srv(SIVA_COMPLEX_NODE *node)
{
  SIVA_SET_TARGET_ZERO(node);
  SIVA_SET_TARGET_ONE(node);    
  return;
}


/*---------------------------------------------------------------------------*/
void
siva_clear_visit_srv(SIVA_COMPLEX_NODE *node)
{
  SIVA_CLEAR_VISIT(node);
}

/*---------------------------------------------------------------------------*/
void
siva_clear_target_srv(SIVA_COMPLEX_NODE *node)
{
  SIVA_CLEAR_TARGET_ZERO(node);
  SIVA_CLEAR_TARGET_ONE(node);
  return;
}



/*---------------------------------------------------------------------------*/
void
siva_levelize_complex_node_srv(SIVA_COMPLEX_NODE *complex_node)
{  
  int max_level = 0;
  register int i;


  if (complex_node->level)                               return;
  if (complex_node->tbl.type == CONSTANT)                return;

  SIVA_ASSERT1(!complex_node->visited,
              "Cycle exists, check the node %s\n",
              complex_node->name);
  complex_node->visited = 1;
  

  SIVA_VERB3_ARG2("Levelizing %s\n",complex_node->name);
  
  if (complex_node->tbl.type == INPUT || complex_node->tbl.type == LATCH){
     complex_node->level = 1;
     goto end;
  }                   
  for(i=0;i<complex_node->tbl.num_input;i++){
    siva_levelize_complex_node_srv(complex_node->tbl.input[i]);
    if (max_level < complex_node->tbl.input[i]->level){
       max_level = complex_node->tbl.input[i]->level;
    }  
  } 
  complex_node->level = max_level+1;
 end:
  
  SIVA_VERB3_ARG3("Levelizing %s ..done %d\n",
                  complex_node->name,complex_node->level);
  return;
}

/*---------------------------------------------------------------------------*/
unsigned
siva_get_num_complex_node(void)
{
  return siva_total_num_complex_node;      
}

/*---------------------------------------------------------------------------*/
SIVA_COMPLEX_NODE *
siva_alloc_complex_node(unsigned type)
{
  SIVA_COMPLEX_NODE *node;
  if (siva_num_complex_node >= siva_num_complex_node_allocated){

    siva_num_complex_node_allocated *= SIVA_COMPLEX_NODE_GROWTH_FACTOR; 

    siva_complex_nodes         = ALLOC(SIVA_COMPLEX_NODE,
				       siva_num_complex_node_allocated);
    array_insert_last(SIVA_COMPLEX_NODE*, siva_complex_node_ptr_array,
                                          siva_complex_nodes);
    siva_num_complex_node = 0;
  }
  node = &siva_complex_nodes[siva_num_complex_node];
  node->tbl.type = type;
  node->tbl.input = NIL(SIVA_COMPLEX_NODE *);
  node->tbl.entry = NIL(unsigned *);
  node->tbl.out_phase = 0;
  node->tbl.num_minterm = 0;
  node->tbl.num_input = 0;
  node->tbl.latch_input = 0;
  node->tbl.debug_node = 0;
  node->level = 0;
  node->visited = 0;
  node->light_house_array = NIL(array_t);
  SIVA_SET_NODE(node,bck_create_null());
  SIVA_CLEAR_TARGET_ONE(node);
  SIVA_CLEAR_TARGET_ZERO(node);
  siva_num_complex_node ++;
  siva_total_num_complex_node ++;
  return node;
}


/*---------------------------------------------------------------------------*/
void
siva_alloc_complex_node_input(SIVA_COMPLEX_NODE *node,int size)
{
  node->tbl.num_input = size;
  node->tbl.input = ALLOC(SIVA_COMPLEX_NODE *,size);
}

/*---------------------------------------------------------------------------*/
void
siva_alloc_complex_node_entry(SIVA_COMPLEX_NODE *node,int num_minterm)
{
  register int i;
  node->tbl.num_minterm = num_minterm;
  node->tbl.entry = ALLOC(unsigned *,num_minterm);
  for(i=0;i<num_minterm;i++){
    node->tbl.entry[i] = ALLOC(unsigned,((node->tbl.num_input-1)/16)+1);
  }
}

/*---------------------------------------------------------------------------*/
void
siva_free_complex_node_input(SIVA_COMPLEX_NODE *node)
{
  if (  node->tbl.num_input ){
    FREE(node->tbl.input);
  }
}

/*---------------------------------------------------------------------------*/
void
siva_free_complex_node_entry(SIVA_COMPLEX_NODE *node)
{
  if (node->tbl.num_minterm){
    register int i;
    for(i=0;i<node->tbl.num_minterm;i++){
      FREE(node->tbl.entry[i]);
    }
    FREE(node->tbl.entry);
  }
}

/*---------------------------------------------------------------------------*/
int
siva_is_complex_node_primary_input(SIVA_COMPLEX_NODE *node)
{
  return (node->tbl.type == INPUT);
}


/*---------------------------------------------------------------------------*/
SIVA_NODE
siva_get_node(SIVA_COMPLEX_NODE *x)
{
  SIVA_NODE p = SIVA_GET_NODE(x);
  if (bck_is_null(p)) return p;
  return bck_get_through(p);
}

/*---------------------------------------------------------------------------*/
void
siva_set_node(SIVA_COMPLEX_NODE *x,SIVA_NODE y)
{
  SIVA_SET_NODE(x,y);
}

/*---------------------------------------------------------------------------*/
int
siva_is_target_zero_set(SIVA_COMPLEX_NODE *x)
{
  return (SIVA_GET_TARGET_ZERO(x) == 1);
}

/*---------------------------------------------------------------------------*/
int
siva_is_target_one_set(SIVA_COMPLEX_NODE *x)
{
  return (SIVA_GET_TARGET_ONE(x) == 1);
}

/*---------------------------------------------------------------------------*/
void
siva_set_target_one(SIVA_COMPLEX_NODE *x)
{
  SIVA_SET_TARGET_ONE(x);
}

/*---------------------------------------------------------------------------*/
void
siva_clear_target_one(SIVA_COMPLEX_NODE *x)
{
  SIVA_CLEAR_TARGET_ONE(x);
}

/*---------------------------------------------------------------------------*/
void
siva_set_target_zero(SIVA_COMPLEX_NODE *x)
{
  SIVA_SET_TARGET_ZERO(x);
}
/*---------------------------------------------------------------------------*/
void
siva_clear_target_zero(SIVA_COMPLEX_NODE *x)
{
  SIVA_CLEAR_TARGET_ZERO(x);
}


/*---------------------------------------------------------------------------*/
void
siva_clear_debug_node(SIVA_COMPLEX_NODE *x)
{
  SIVA_CLEAR_DEBUG_NODE(x);
}

/*---------------------------------------------------------------------------*/
int
siva_is_debug_node(SIVA_COMPLEX_NODE *x)
{
  return SIVA_IS_DEBUG_NODE(x);
}
/*---------------------------------------------------------------------------*/
void
siva_set_debug_node(SIVA_COMPLEX_NODE *x)
{
  SIVA_SET_DEBUG_NODE(x);
}


/*---------------------------------------------------------------------------*/
int
siva_is_latch_input(SIVA_COMPLEX_NODE *x)
{
  return SIVA_IS_LATCH_INPUT(x);
}

/*
  ---------------------------------------------------------------------------
  Complex node table and functions
  ---------------------------------------------------------------------------
*/
/*---------------------------------------------------------------------------*/
void
siva_init_complex_node_table(void)
{
  register unsigned j;
  siva_complex_node_table_size = SIVA_DEFAULT_COMPLEX_NODE_TABLE_SIZE;
  siva_complex_node_table = ALLOC(SIVA_COMPLEX_NODE*,
                                  siva_complex_node_table_size);
  for(j=0;j < siva_complex_node_table_size; j++){
    siva_complex_node_table[j] = NIL(SIVA_COMPLEX_NODE);
  }
}

/*---------------------------------------------------------------------------*/
void
siva_quit_complex_node_table(void)
{
  FREE(siva_complex_node_table);
}


/*---------------------------------------------------------------------------*/
SIVA_COMPLEX_NODE *
siva_find_complex_node_hash(char *string)
{
  SIVA_COMPLEX_NODE *entry;

  unsigned hash_value = siva_complex_node_hash(string);
  for(entry = siva_complex_node_table[hash_value]; 
      entry && (strcmp(entry->name,string) > 0); entry = entry->next);
  if ((entry)&&(strcmp(entry->name,string)==0)){
    return entry;
  }else {
    return NIL(SIVA_COMPLEX_NODE);
  }
}

/*---------------------------------------------------------------------------*/
void
siva_insert_complex_node_hash(SIVA_COMPLEX_NODE *node)
{
  unsigned hash_value = siva_complex_node_hash(node->name);
  if (!siva_complex_node_table[hash_value] ||
       (strcmp(siva_complex_node_table[hash_value]->name,node->name) < 0)){
    node->next = siva_complex_node_table[hash_value];
    siva_complex_node_table[hash_value] = node;
  }else {
    SIVA_COMPLEX_NODE *entry = siva_complex_node_table[hash_value];
    while(entry->next && 
          (strcmp(entry->next->name,node->name) > 0)){
      entry = entry->next;
    }
    node->next = entry->next;
    entry->next = node;
  }
  siva_num_hash_complex_node++;
  /* resize when density is 8 per bin */
  if ((siva_num_hash_complex_node >> 3) == siva_complex_node_table_size){
     siva_resize_complex_node_table();
  }

}
/*---------------------------------------------------------------------------*/
void
siva_remove_complex_node_hash(SIVA_COMPLEX_NODE *node)
{
  unsigned hash_value = siva_complex_node_hash(node->name);

  SIVA_ASSERT((siva_find_complex_node_hash(node->name) != 
               NIL(SIVA_COMPLEX_NODE)),"node not found");

  if (siva_complex_node_table[hash_value] == node){
    siva_complex_node_table[hash_value] = node->next;
  }else {
    SIVA_COMPLEX_NODE *entry = siva_complex_node_table[hash_value];
    while(strcmp(entry->next->name,node->name) > 0){
      entry = entry->next;
    }
    SIVA_ASSERT((entry->next == node),"");
    entry->next = entry->next->next;
  }

}

/*---------------------------------------------------------------------------*/
static unsigned 
siva_complex_node_hash(char *string)
{
  return (siva_string_hash(string) & (siva_complex_node_table_size-1));
}

/*---------------------------------------------------------------------------*/
static void
siva_resize_complex_node_table(void)
{
  register int i;

  siva_complex_node_table = REALLOC(SIVA_COMPLEX_NODE*,siva_complex_node_table,
                                    (siva_complex_node_table_size << 1));  

  for(i=0;i<siva_complex_node_table_size;i++){
    SIVA_COMPLEX_NODE lower_header;
    SIVA_COMPLEX_NODE *lower_ptr=&lower_header;

    SIVA_COMPLEX_NODE upper_header;
    SIVA_COMPLEX_NODE *upper_ptr=&upper_header;

    SIVA_COMPLEX_NODE *entry = siva_complex_node_table[i];

    lower_header.next = NIL(SIVA_COMPLEX_NODE);
    upper_header.next = NIL(SIVA_COMPLEX_NODE);
    while(entry){
      if (siva_string_hash(entry->name) & (siva_complex_node_table_size)){
        upper_ptr->next = entry;        
        upper_ptr = entry;
      }else{
        lower_ptr->next = entry;
        lower_ptr = entry;
      }
      entry = entry->next;
    }
    upper_ptr->next = NIL(SIVA_COMPLEX_NODE);
    lower_ptr->next = NIL(SIVA_COMPLEX_NODE);

    siva_complex_node_table[i] = lower_header.next;
    siva_complex_node_table[i+siva_complex_node_table_size] = 
                                 upper_header.next;
  }
  siva_complex_node_table_size = (siva_complex_node_table_size << 1);
}
