/**CFile**********************************************************************

  FileName                   [siva_control_state.c]

  PackageName                [siva_Core]

  Synopsis                   [control latch operation]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]

******************************************************************************/

#include "siva.h"

#define SIVA_DEFAULT_CONTROL_STATE_TABLE_SIZE 1024
#define SIVA_CONTROL_STATE_GROWTH_FACTOR 1

#define CONTROL_STATE_ON_LATCH_ACTIVITY 1


#define SIVA_MAX_GROUP_SIZE 100
#define SIVA_MIN_GROUP_SIZE 10
static void siva_collect_latch_srv(SIVA_COMPLEX_NODE *node);
static void siva_set_group_id_for_target_srv(SIVA_COMPLEX_NODE *node);

static st_table *siva_control_group_table;

static unsigned siva_num_control_sub_group;
static array_t *siva_control_sub_state_init_hash_value_array;
static array_t *siva_control_sub_group_mask_array;

static array_t *siva_latch_array;

static unsigned siva_control_state_table_size;
static unsigned siva_num_hash_control_state;
static SIVA_CONTROL_STATE **siva_control_state_table;

static unsigned siva_total_num_control_state;
static unsigned siva_num_control_state;
static unsigned siva_num_control_state_allocated;
static SIVA_CONTROL_STATE *siva_control_states;
static array_t *siva_control_state_ptr_array;
static var_set_t *siva_control_state_vector;

static unsigned siva_control_state_heap_id;
static array_t *siva_latch_group_array;

static void siva_init_control_state_table(void);
static void siva_quit_control_state_table(void);
static unsigned siva_control_state_hash(unsigned control_state_value);
static void siva_resize_control_state_table(void);
static void siva_control_state_table_stat(void);
static void siva_find_sub_groups(array_t *control_sub_group_mask_array);

/*---------------------------------------------------------------------------*/
void
siva_init_control_state(void)
{
  int i;
  siva_num_control_group=0;
  siva_num_control_sub_group = 0;
  siva_control_group_mask_array = array_alloc(var_set_t *,0);
  siva_control_sub_group_mask_array = array_alloc(var_set_t *,0);
  
  siva_num_control_state_allocated = 8192;
  siva_control_states = ALLOC(SIVA_CONTROL_STATE, 
                              siva_num_control_state_allocated);
  siva_num_control_state=0;
  siva_control_state_ptr_array = array_alloc(SIVA_CONTROL_STATE *,16);
  array_insert_last(SIVA_CONTROL_STATE *,siva_control_state_ptr_array,
                    siva_control_states);
  siva_total_num_control_state=0;
  siva_init_control_state_table();
  siva_control_state_vector = var_set_new(siva_num_latch);
  
  siva_control_state_heap_id = siva_alloc_heap(siva_heap_float_min_cmp);

  siva_active_latch_order = ALLOC(int, siva_num_latch);
  for(i=0;i<siva_num_latch;i++){
    siva_active_latch_order[i] = i;
  }
}

/*---------------------------------------------------------------------------*/
void
siva_quit_control_state(void)
{
  register int i;
  var_set_t *mask_vector;

  FREE(siva_active_latch_order);
  siva_release_heap(siva_control_state_heap_id);  
  var_set_free(siva_control_state_vector);

  for(i=0;i<array_n(siva_control_state_ptr_array);i++){
    FREE(array_fetch(SIVA_CONTROL_STATE *,siva_control_state_ptr_array,i));
  }

  array_free(siva_control_state_ptr_array);
  siva_quit_control_state_table();
  
  for(i=0;i<siva_num_control_group;i++){
    var_set_t *mask_vector = array_fetch(var_set_t *,
                                         siva_control_group_mask_array,i);
    var_set_free(mask_vector);
  }
  array_free(siva_control_group_mask_array);
  if (siva_num_control_group){
    array_free(siva_control_state_init_hash_value_array);
  }

  for(i=0;i<siva_num_control_sub_group;i++){
    var_set_t *mask_vector = array_fetch(var_set_t *,
                                         siva_control_sub_group_mask_array,i);
    var_set_free(mask_vector);
  }
  array_free(siva_control_sub_group_mask_array);
  if (siva_num_control_sub_group) {
    array_free(siva_control_sub_state_init_hash_value_array);
  }
}

/*---------------------------------------------------------------------------*/
void
siva_control_state(void)
{
  register int i;

  if (array_n(siva_control_group_mask_array)==0){
    if (siva_max_num_control_group){
      fprintf(siva_output_file,"Finding control groups...\n");
      siva_fill_active_latch_group(siva_control_group_mask_array,
                                   siva_active_latch_order);
      fprintf(siva_output_file,"Finding control groups...done\n");    
    }
  }

  if (siva_control_sub_group_on) {
    fprintf(siva_output_file,"Finding control sub groups...\n");
    siva_find_sub_groups(siva_control_sub_group_mask_array);
    fprintf(siva_output_file,"Finding control sub groups...done\n");    
  }

  srand(0);
  
  siva_num_control_sub_group = array_n(siva_control_sub_group_mask_array);
  if (siva_num_control_sub_group){
    siva_control_sub_state_init_hash_value_array =
      array_alloc(unsigned, siva_num_control_sub_group);
    
    for(i=0;i<siva_num_control_sub_group;i++){
      unsigned init_hash_value = rand();
      array_insert(unsigned,  siva_control_sub_state_init_hash_value_array,
                   i,init_hash_value);
      srand(init_hash_value);    
    }
    siva_control_state_check_on = 1;    
  }
  
  siva_num_control_group = array_n(siva_control_group_mask_array);
  if (siva_num_control_group){
    siva_control_state_init_hash_value_array =
      array_alloc(unsigned, siva_num_control_group);
    
    for(i=0;i<siva_num_control_group;i++){
      unsigned init_hash_value = rand();
      array_insert(unsigned,  siva_control_state_init_hash_value_array,
                   i,init_hash_value);
      srand(init_hash_value);    
    }
    siva_control_state_check_on = 1;    
  }
  
  siva_control_state_stat();

}



/*---------------------------------------------------------------------------*/
void
siva_control_state_stat()
{
  register int i;
  var_set_t *mask_vector;

  for(i=0;i<siva_num_control_group;i++){
    mask_vector = array_fetch(var_set_t *,siva_control_group_mask_array,i);
    fprintf(siva_output_file,"Size of control group %d: %d\n",
            i,var_set_n_elts(mask_vector));
    if (siva_verbosity_level > 1){
      var_set_print(siva_output_file,mask_vector);
    }
  }
  if (siva_verbosity_level >= 1){
    fprintf(siva_output_file,"# control group %d\n",siva_num_control_group);
  }

  for(i=0;i<siva_num_control_sub_group;i++){
    mask_vector = array_fetch(var_set_t *,siva_control_sub_group_mask_array,i);
    fprintf(siva_output_file,"Size of control sub group %d: %d\n",
            i,var_set_n_elts(mask_vector));
    if (siva_verbosity_level > 1){
      var_set_print(siva_output_file,mask_vector);
    }
  }
  if (siva_verbosity_level >= 1){
    fprintf(siva_output_file,"# control sub group %d\n",
            siva_num_control_sub_group);
  }
    
}

/*---------------------------------------------------------------------------*/
static void
siva_init_control_state_table(void)
{
  register unsigned j;
  siva_control_state_table_size = SIVA_DEFAULT_CONTROL_STATE_TABLE_SIZE;
  siva_control_state_table =
    ALLOC(SIVA_CONTROL_STATE*,siva_control_state_table_size);
  for(j=0;j < siva_control_state_table_size; j++){
    siva_control_state_table[j] = NIL(SIVA_CONTROL_STATE);
  }
  siva_num_hash_control_state = 0;
}
/*---------------------------------------------------------------------------*/
static void
siva_quit_control_state_table(void)
{

  FREE(siva_control_state_table);
}


/*---------------------------------------------------------------------------*/
static unsigned 
siva_control_state_hash(unsigned control_state_value)
{
  return (control_state_value & (siva_control_state_table_size-1));
}

/*---------------------------------------------------------------------------*/
SIVA_CONTROL_STATE *
siva_find_control_state_hash(unsigned control_state_value)
{
  SIVA_CONTROL_STATE *entry;

  unsigned hash_value = siva_control_state_hash(control_state_value);
  for(entry = siva_control_state_table[hash_value]; 
      entry && (entry->control_state_value > control_state_value);
      entry = entry->next);
  if ((entry)&&(entry->control_state_value == control_state_value)){
    return entry;
  }else {
    return NIL(SIVA_CONTROL_STATE);
  }
}

/*---------------------------------------------------------------------------*/
void
siva_insert_control_state_hash(SIVA_CONTROL_STATE *control_state)
{
  unsigned hash_value =
    siva_control_state_hash(control_state->control_state_value);
  if (!siva_control_state_table[hash_value] ||
       siva_control_state_table[hash_value]->control_state_value <
      control_state->control_state_value){
    control_state->next = siva_control_state_table[hash_value];
    siva_control_state_table[hash_value] = control_state;
  }else {
    SIVA_CONTROL_STATE *entry = siva_control_state_table[hash_value];
    while(entry->next && 
          (entry->next->control_state_value >
           control_state->control_state_value)){
      entry = entry->next;
    }
    control_state->next = entry->next;
    entry->next = control_state;
  }
  siva_num_hash_control_state++;
  /* resize when density is 8 per bin */
  if ((siva_num_hash_control_state >> 3) == siva_control_state_table_size){
    SIVA_VERB1(siva_control_state_table_stat());
    siva_resize_control_state_table();
  }
}
/*---------------------------------------------------------------------------*/
void
siva_remove_control_state_hash(SIVA_CONTROL_STATE *control_state)
{
  unsigned hash_value =
    siva_control_state_hash(control_state->control_state_value);

  SIVA_ASSERT((siva_find_control_state_hash(control_state->control_state_value)
               != NIL(SIVA_CONTROL_STATE)),
              "control state not found");

  if (siva_control_state_table[hash_value] == control_state){
    siva_control_state_table[hash_value] = control_state->next;
  }else {
    SIVA_CONTROL_STATE *entry = siva_control_state_table[hash_value];
    while(entry->next->control_state_value >
          control_state->control_state_value){
      entry = entry->next;
    }
    SIVA_ASSERT((entry->next == control_state),"");
    entry->next = entry->next->next;
  }
  siva_num_hash_control_state--;
  /* need to reduce table size ?*/
  /*
   resize when the density goes down to 2 per bin
  if ((siva_num_hash_control_state << 1) == siva_control_state_table_size){
     siva_resize_control_state_table();
  }
  */
}
/*---------------------------------------------------------------------------*/
static void
siva_resize_control_state_table(void)
{
  register int i;

  siva_control_state_table = REALLOC(SIVA_CONTROL_STATE*,
                                     siva_control_state_table,
                                     (siva_control_state_table_size << 1));  

  for(i=0;i<siva_control_state_table_size;i++){
    SIVA_CONTROL_STATE lower_header;
    SIVA_CONTROL_STATE *lower_ptr=&lower_header;

    SIVA_CONTROL_STATE upper_header;
    SIVA_CONTROL_STATE *upper_ptr=&upper_header;

    SIVA_CONTROL_STATE *entry = siva_control_state_table[i];

    lower_header.next = NIL(SIVA_CONTROL_STATE);
    upper_header.next = NIL(SIVA_CONTROL_STATE);
    while(entry){
      if (entry->control_state_value & (siva_control_state_table_size)){
        upper_ptr->next = entry;        
        upper_ptr = entry;
      }else{
        lower_ptr->next = entry;
        lower_ptr = entry;
      }
      entry = entry->next;
    }
    upper_ptr->next = NIL(SIVA_CONTROL_STATE);
    lower_ptr->next = NIL(SIVA_CONTROL_STATE);

    siva_control_state_table[i] = lower_header.next;
    siva_control_state_table[i+siva_control_state_table_size] =
      upper_header.next;
  }
  siva_control_state_table_size = (siva_control_state_table_size << 1);
}

/*---------------------------------------------------------------------------*/
static void
siva_control_state_table_stat(void)
{
  if (siva_verbosity_level > 1){
    register int i;
    float avg_bin_load =
      (float)siva_num_hash_control_state/siva_control_state_table_size;
    float bin_load_var;
    float std=0;
    for(i=0;i<siva_control_state_table_size;i++){
      SIVA_CONTROL_STATE *entry = siva_control_state_table[i];
      int num_entry=0;
      for(entry=siva_control_state_table[i];entry;entry=entry->next,num_entry++);
      bin_load_var = (num_entry-avg_bin_load);
      bin_load_var *= bin_load_var;
      std += bin_load_var;
    }
    fprintf(siva_output_file,"avg load = %f\n",avg_bin_load);
    fprintf(siva_output_file,"std load = %f\n",
            sqrt(std/siva_control_state_table_size));
    fprintf(siva_output_file,"table_size = %d\n",
            siva_control_state_table_size);
  }
  fprintf(siva_output_file,"# control state = %d\n",
          siva_num_hash_control_state);
}
/*---------------------------------------------------------------------------*/
unsigned 
siva_get_num_control_state(void)
{
  return siva_total_num_control_state;      
}

/*---------------------------------------------------------------------------*/
SIVA_CONTROL_STATE *
siva_alloc_control_state(void)
{
  SIVA_CONTROL_STATE  *control_state;

  if (siva_num_control_state >= siva_num_control_state_allocated){
    siva_num_control_state_allocated *= SIVA_CONTROL_STATE_GROWTH_FACTOR;
  
    siva_control_states = ALLOC(SIVA_CONTROL_STATE,
                                siva_num_control_state_allocated);
    array_insert_last(SIVA_CONTROL_STATE *,
                      siva_control_state_ptr_array,siva_control_states);
    siva_num_control_state = 0;
  }

  control_state = &siva_control_states[siva_num_control_state];
  control_state->control_state_value = 0;
  control_state->next = NIL(SIVA_CONTROL_STATE);


  siva_num_control_state++;
  siva_total_num_control_state++;
  return control_state;
}

/*--------------------------------------------------------------------------*/
int
siva_get_seed_control_state(SIVA_STATE **state_p,
                            var_set_t *seed_state_vector)
{
  SIVA_KEY key;
  SIVA_CONV conv;
  while(siva_get_control_state(&key,&conv)){    
    /* regenerate state and check if valid state*/
    if (!siva_regenerate_state_vector(conv.state,
                                      seed_state_vector)){
      continue;
    }
    *state_p = conv.state;
    SIVA_VERB1_ARG3("CS: Seed state value = %x, key = %f\n",
                    conv.state->state_value,key.float_key);
    return 1;
  }  
  return 0;
}

/*--------------------------------------------------------------------------*/
void
siva_put_control_state(float priority_value, SIVA_STATE *state)
{
  SIVA_KEY key;
  SIVA_CONV conv;

  key.float_key = priority_value;
  conv.state = state;
  siva_heap_insert(key,conv,siva_control_state_heap_id);

}

/*--------------------------------------------------------------------------*/
int
siva_get_control_state(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr)
{
  int found;

  found = siva_heap_extract_top(key_ptr,conv_ptr,siva_control_state_heap_id);
  return found;
}


/*--------------------------------------------------------------------------*/
int
siva_is_latch_active(SIVA_NODE latch, int group)
{
  SIVA_CONV conv;
  var_set_t *mask_vector;
  SIVA_ASSERT1(bck_is_flag_set(latch,siva_latch_flag),
               "%d is not latch\n",bck_get_regular(latch));

  if (group < 0 ) {
    return 0;
  }
  if (group >= array_n(siva_control_group_mask_array)){
    return 0;
  }
  mask_vector = array_fetch(var_set_t *,siva_control_group_mask_array,
                            group);
  conv.ptr = bck_get_aux(latch,siva_latch_id_aux);
  return (var_set_get_elt(mask_vector,conv.int_value) == 0);
}

/*---------------------------------------------------------------------------*/
static void
siva_find_sub_groups(array_t *control_sub_group_mask_array)
{

  int overlap;
  int support_size;
  var_set_t *latch_support_set = var_set_new(siva_num_latch);
  var_set_t *intersect_set = var_set_new(siva_num_latch);
  float max_overlap_frac;
  float overlap_frac;
  int i,k,j;
  int group_id;
  int num_group = 0;
  array_t *latch_group_array;
  array_t *latch_group_support_array = array_alloc(var_set_t*,16);
  var_set_t *latch_group_support_set;
  var_set_t *latch_group_set;

  siva_latch_group_array = array_alloc(array_t *,16);  
  bck_reset_all_flag(siva_bck_visited_flag);
  
  for(i=0;i<siva_num_latch;i++){
    if (bck_is_constant(siva_latch_nodes[i].input)){
      continue;
    }
    bck_set_fanin_cone_flag(siva_latch_nodes[i].input,
                            siva_bck_visited_flag);
    for(k=0;k<siva_num_latch;k++){
      if (bck_is_flag_set(siva_latch_nodes[k].latch,
                          siva_bck_visited_flag)){
        var_set_set_elt(latch_support_set,k);
      }else {
        var_set_clear_elt(latch_support_set,k);
      }
    }
    bck_reset_fanin_cone_flag(siva_latch_nodes[i].input,
                              siva_bck_visited_flag);
    
    max_overlap_frac = siva_partition_overlap_threshold;
    group_id = -1;
    for(j=0;j<num_group;j++){
      latch_group_array = array_fetch(array_t *,siva_latch_group_array,j);
      latch_group_support_set =
        array_fetch(var_set_t *,latch_group_support_array,j);
      support_size = var_set_n_elts(latch_group_support_set);
      var_set_and(intersect_set,latch_group_support_set,latch_support_set);
      overlap = var_set_n_elts(intersect_set);
      overlap_frac = (float)(overlap)/(support_size);
      if (overlap_frac > max_overlap_frac){
        group_id = j;
        max_overlap_frac = overlap_frac;
      }
    }
    if (group_id == -1){
      latch_group_support_set = var_set_copy(latch_support_set);
      latch_group_array = array_alloc(int,4);
      num_group++;      
      array_insert_last(var_set_t *,latch_group_support_array,
                        latch_group_support_set);      
      array_insert_last(SIVA_NODE,latch_group_array,i);
      array_insert_last(array_t *,siva_latch_group_array,
                        latch_group_array);
    }else {
      latch_group_support_set =
        array_fetch(var_set_t *, latch_group_support_array,
                    group_id);
      var_set_or(latch_group_support_set,latch_group_support_set,
                 latch_support_set);
      latch_group_array =
        array_fetch(array_t *,siva_latch_group_array,group_id);
      array_insert_last(int,latch_group_array,i);
    }
  }


  if (siva_verbosity_level >= 2){
    fprintf(siva_output_file,"# latch groups %d\n",
            array_n(siva_latch_group_array));
    for(i=0;i<array_n(siva_latch_group_array);i++){
      latch_group_array = array_fetch(array_t *,siva_latch_group_array,i);
      fprintf(siva_output_file,"Group %d: # Latch %d\n",i,
            array_n(latch_group_array));
      for(j=0;j<array_n(latch_group_array);j++){
        fprintf(siva_output_file,"Latch id %d\n",
                array_fetch(int,latch_group_array,j));
      }
    }
  }


  
  for(i=0;i<array_n(latch_group_support_array);i++){
    latch_group_support_set =
      array_fetch(var_set_t*,latch_group_support_array,i);
    var_set_free(latch_group_support_set);
  }

  array_free(latch_group_support_array);
  var_set_free(latch_support_set);  
  var_set_free(intersect_set);

  for(i=0;i<array_n(siva_latch_group_array);i++){
    latch_group_array = array_fetch(array_t *,
                                    siva_latch_group_array,i);

    if (array_n(latch_group_array) < 10) {
      continue;
    }

    latch_group_set = var_set_new(siva_num_latch);
    for(j=0;j<array_n(latch_group_array);j++){
      var_set_set_elt(latch_group_set,
                      array_fetch(int,latch_group_array,j));
    }
    array_insert_last(var_set_t *,
                      control_sub_group_mask_array,latch_group_set);
    SIVA_VERB1_ARG3("Control group %d: %d\n",
                   array_n(control_sub_group_mask_array),
                   var_set_n_elts(latch_group_set));
  }
}


/*---------------------------------------------------------------------------*/
/* returns  if 0 => rejected else admitted. state is in state_ptr            */
int
siva_control_sub_state_filter_in(var_set_t *state_vector)
{
  register int i;
  var_set_t *mask_vector;
  unsigned control_sub_state_value;
  unsigned num_found = 0;  
  SIVA_CONTROL_STATE *new_control_state;
  unsigned init_hash_value;
  var_set_t local_state_vector;
  unsigned data[siva_latch_word_size];

  siva_set_state_vector(&local_state_vector,data);
  var_set_clear(&local_state_vector);
  
  for(i=0;i<siva_num_control_sub_group;i++){
    mask_vector = array_fetch(var_set_t *,siva_control_sub_group_mask_array,i);
    init_hash_value =
      array_fetch(unsigned,siva_control_sub_state_init_hash_value_array,i);
    var_set_and(&local_state_vector,state_vector,mask_vector);
    var_set_compress(&control_sub_state_value,&local_state_vector,
                     init_hash_value);
    if (siva_verbosity_level > 2){
      fprintf(siva_output_file,"Mask %d, control value %x\n",i,
              control_sub_state_value);
      if (siva_verbosity_level > 3){
        fprintf(siva_output_file,"State vector\n");
        var_set_print(siva_output_file,state_vector);
        fprintf(siva_output_file,"Mask vector\n");
        var_set_print(siva_output_file,mask_vector);
        fprintf(siva_output_file,"Projected Vector\n");
        var_set_print(siva_output_file,&local_state_vector);
      }    
    }

    
    if (siva_find_control_state_hash(control_sub_state_value)
        ==NIL(SIVA_CONTROL_STATE)){
      new_control_state = siva_alloc_control_state();
      new_control_state->control_state_value = control_sub_state_value;
      siva_insert_control_state_hash(new_control_state);
    }else {
      num_found++;
    }
  }
  return num_found;
  
}
  


/*---------------------------------------------------------------------------*/
void
siva_control_project(var_set_t *state_vector, int control_group)
{
  var_set_t *mask_vector;
  /* array_fetch are not very thread safe function. why? */
  SIVA_ASSERT((control_group < array_n(siva_control_group_mask_array)),
              "check the control group");

  mask_vector = 
    array_fetch(var_set_t *,siva_control_group_mask_array,control_group);  
  assert(mask_vector->n_elts == siva_num_latch);
  var_set_and(state_vector,state_vector,mask_vector);
}


/*
  ---------------------------------------------------------------------------
  returns 0 if rejected else return the priority. lower the better.
  ---------------------------------------------------------------------------
*/
int
siva_control_state_filter_in(var_set_t *state_vector)
{
  int i,j;
  int num_found;
  var_set_t *mask_vector;
  unsigned control_state_value;
  unsigned priority = 0;
  SIVA_CONTROL_STATE *new_control_state;
  unsigned init_hash_value;
  int num_control_group;
  var_set_t local_state_vector;
  unsigned data[siva_latch_word_size];



  siva_set_state_vector(&local_state_vector,data);
  var_set_clear(&local_state_vector);
  
  num_control_group = (siva_control_group <siva_num_control_group)?
    (siva_control_group+1) : siva_num_control_group;

  if (num_control_group == 0) {
    num_found =
      siva_control_sub_state_filter_in(state_vector);
    priority = num_found + 1;
    if (priority > siva_num_control_sub_group){
      priority = 0;
    }
    return priority;
  }
      
  for(j=0;j<num_control_group;j++){
    var_set_t *mask_vector =
      array_fetch(var_set_t *,siva_control_group_mask_array,j);
    var_set_and(&local_state_vector,state_vector,mask_vector);
    init_hash_value =
      array_fetch(unsigned,siva_control_state_init_hash_value_array,j);
    var_set_and(&local_state_vector,state_vector,mask_vector);
    var_set_compress(&control_state_value,&local_state_vector,
                     init_hash_value);
    if (siva_find_control_state_hash(control_state_value)
        != NIL(SIVA_CONTROL_STATE)){
      continue;
    }
    num_found =
      siva_control_sub_state_filter_in(&local_state_vector);
    /* insert the  control state */
    if (num_found == siva_num_control_sub_group){
      new_control_state = siva_alloc_control_state();
      new_control_state->control_state_value = control_state_value;
      siva_insert_control_state_hash(new_control_state);
    }
     return  (1 + j*(siva_num_control_sub_group + 1) + num_found);
  }
  return 0;
}

