/*CFile**********************************************************************

  FileName                   [siva_et.c]

  PackageName                [siva_core]

  Synopsis                   [Enlarge targets]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modifed: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]
******************************************************************************/


#include "siva.h"
#define SIVA_DEFAULT_ET_PICK_SIZE 10
#define SIVA_DEFAULT_ET_TABLE_SIZE 1024
#define SIVA_ET_GROWTH_FACTOR 1
#define SIVA_ET_BASE_ARRAY_SIZE 8196

static unsigned siva_total_num_et;
static unsigned siva_num_et_allocated;
static unsigned siva_num_et;
static SIVA_ET *siva_ets;
static array_t *siva_et_ptr_array;
static unsigned siva_et_table_size;
static unsigned siva_num_hash_et;
static SIVA_ET **siva_et_table;
static unsigned siva_et_heap_id;

static int siva_built_et_frontiers_done;
static int siva_num_et_frontier;
static SIVA_ET **siva_et_frontiers;

static void siva_build_et_frontiers(void);
static void siva_destroy_et_frontiers(void);
static void siva_resize_et_table(void);
static unsigned siva_et_hash(unsigned state_value);
static void siva_et_table_stat(void);
static unsigned siva_et_init_mask_hash_value;
static unsigned siva_et_init_state_hash_value;
static void siva_init_et_table(void);
static void siva_quit_et_table(void);
static unsigned siva_et_hash(unsigned state_value);
static SIVA_ET *siva_find_et_hash(unsigned state_value);
static void siva_insert_et_hash(SIVA_ET *et);

static void siva_remove_et_hash(SIVA_ET *et);
static void siva_resize_et_table(void);


static void siva_put_et(unsigned priority_value, SIVA_ET *et);
static int siva_get_et(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr);
static unsigned siva_get_et_priority(SIVA_ET *et);
static int siva_et_solve1(SIVA_ET *et,
                          array_t *state_wrapper_array, int num_state);
static int siva_get_best_et(SIVA_ET *ets[],
                            int num_et_req);

static int siva_check_et_reached1(SIVA_ET *et,
                                  var_set_t *visited_state_vector);

static SIVA_ET* siva_store_ets(SIVA_ET *ets, void *target, int is_target_et,
                               int num_et);
static SIVA_ET* siva_get_or_create_et(var_set_t *mask_vector,
                                       var_set_t *state_vector,
                                       void *target,
                                       int is_target_et,
                                       int distance,
                                       int *is_created);
static int siva_get_et_target_distance(SIVA_ET *et);

static void siva_verify_et(SIVA_ET *et, var_set_t *sample_vector);
static void siva_et_print_witness(SIVA_ET *et);


/*---------------------------------------------------------------------------*/
void 
siva_init_et(void)
{ 
  siva_et_ptr_array = array_alloc(SIVA_ET *,16);
  siva_num_et_allocated = SIVA_ET_BASE_ARRAY_SIZE;
  siva_ets = ALLOC(SIVA_ET,siva_num_et_allocated);
  array_insert_last(SIVA_ET *,siva_et_ptr_array,
                    siva_ets);
  siva_num_et = 0;
  siva_total_num_et = 0;
  siva_init_et_table();
  srand(0);
  siva_et_init_mask_hash_value = rand();
  siva_et_init_state_hash_value = rand();  
  siva_et_heap_id = siva_alloc_heap(siva_heap_unsigned_max_cmp);
  siva_built_et_frontiers_done = 0;
}

/*---------------------------------------------------------------------------*/
void 
siva_quit_et(void)
{ 
  register int i,j;
  unsigned ptr_size = SIVA_ET_BASE_ARRAY_SIZE;

  siva_destroy_et_frontiers();
  siva_release_heap(siva_et_heap_id);  
  siva_quit_et_table();
  for(j=0;j<array_n(siva_et_ptr_array);j++){
    SIVA_ET *ptr = array_fetch(SIVA_ET *,
                               siva_et_ptr_array,j);
    if ( j == (array_n(siva_et_ptr_array)-1)) {
       ptr_size = siva_num_et;
    }       
    for(i=0;i<ptr_size;i++){
      FREE(ptr[i].data);
      FREE(ptr[i].mask);
    }
    FREE(ptr);
    ptr_size *= SIVA_ET_GROWTH_FACTOR;
  }
  array_free(siva_et_ptr_array);
  siva_num_et_allocated = 0;
  siva_num_et = 0;
  siva_total_num_et = 0;
}

/*--------------------------------------------------------------------------*/
void
siva_do_for_all_et(void (*service)(SIVA_ET *et))
{
  register int i,j;
  unsigned ptr_size = SIVA_ET_BASE_ARRAY_SIZE;
  for(j=0;j<array_n(siva_et_ptr_array);j++){
    SIVA_ET *ptr = array_fetch(SIVA_ET *,
                                         siva_et_ptr_array,j); 
    if ( j == (array_n(siva_et_ptr_array)-1)) {
       ptr_size = siva_num_et;
    }              
    for(i=0;i<ptr_size;i++){
      service(&ptr[i]);
    }
    ptr_size *= SIVA_ET_GROWTH_FACTOR;
  }    
}

/*--------------------------------------------------------------------------*/
SIVA_NODE
siva_get_et_target_node(SIVA_ET *et)
{
  while(et->is_target_et) {
    et = et->target.et;
  }
  return et->target.node;
}

/*--------------------------------------------------------------------------*/
static int
siva_get_et_target_distance(SIVA_ET *et)
{
  int distance = 0;
  while(et->is_target_et) {
    distance += et->distance;    
    et = et->target.et;
  }
  distance += (et->distance-1);
  return distance;
}
/*--------------------------------------------------------------------------*/
void
siva_et_stat(void)
{
  SIVA_ITER iter;
  SIVA_ET *et = siva_get_first_et(&iter);
  int reached=0;
  int total=0;
  int unreachable=0;
  SIVA_NODE tgt;
  while(et) {
    if (et->is_unreachable) {
      unreachable++;
    }else {
      tgt = siva_get_et_target_node(et);
      reached += (!bck_is_flag_set(tgt,siva_target_one_flag) &&
                  !bck_is_flag_set(tgt,siva_target_zero_flag));
    }
    total++;
    et = siva_get_next_et(&iter);
  }
  SIVA_VERB1_ARG2("ET: Total = %d, ",total);
  SIVA_VERB1_ARG2("Target reached = %d, ",reached);
  SIVA_VERB1_ARG2("ET unreachable = %d, ",unreachable);
  SIVA_VERB1_ARG2("Unresolved = %d\n",
                  (total - reached - unreachable));
}



/*--------------------------------------------------------------------------*/
SIVA_ET*
siva_get_first_et(SIVA_ITER *iter)
{
  SIVA_CONV conv;
  SIVA_ET *et;
  iter->array = (void *)siva_ets;
  if (siva_num_et == 0){
    et =  NIL(SIVA_ET);
  }else {
    iter->array_index = 0;    
    iter->index = 0;
    conv.ptr = iter->array;
    et =  &conv.et[iter->array_index];
  }
  return et;
  
}

/*--------------------------------------------------------------------------*/
SIVA_ET*
siva_get_next_et(SIVA_ITER *iter)
{
  SIVA_ET *et;  
  SIVA_CONV conv;
  iter->index++;
  iter->array_index ++;
  if (iter->array_index == siva_num_et) {
    et =  NIL(SIVA_ET);
  }else {
    conv.ptr = iter->array;
    et = &conv.et[iter->array_index];
  }
  return et;
}


/*---------------------------------------------------------------------------*/
unsigned
siva_get_num_et(void)
{
  return siva_total_num_et;      
}


/*
  ---------------------------------------------------------------------------
  mask vector and state vector represents a cube. This cube represents
  the enlarged target.
  ---------------------------------------------------------------------------
*/
static SIVA_ET *
siva_get_or_create_et(var_set_t *mask_vector,
                      var_set_t *state_vector,
                      void *target,
                      int is_target_et,
                      int distance,
                      int *is_created)
{
  SIVA_ET *et;
  unsigned state_value;
  unsigned state_value1;
  SIVA_CONV conv;
  
  var_set_compress(&state_value,mask_vector,siva_et_init_mask_hash_value);
  var_set_compress(&state_value1,state_vector,siva_et_init_state_hash_value);
  state_value = (state_value ^ state_value1);

  *is_created = 0;
  et = siva_find_et_hash(state_value);
  if (et) {
    SIVA_VERB2_ARG1("ET found in hash\n");
    return et;
  }
  et = siva_alloc_et();
  memcpy(et->data,state_vector->data,sizeof(int)*siva_latch_word_size);
  memcpy(et->mask,mask_vector->data,sizeof(int)*siva_latch_word_size);
  siva_insert_et_hash(et);  
  *is_created = 1;
  et->is_target_et = is_target_et;
  conv.ptr = target;
  et->target.et = conv.et;
  et->distance = distance;
  et->priority = siva_get_et_priority(et);

  SIVA_VERB2_ARG1("ET created\n");
  return et;
}


/*---------------------------------------------------------------------------*/
SIVA_ET *
siva_alloc_et(void)
{
  SIVA_ET *et;
  if (siva_num_et >= siva_num_et_allocated){
    /* TODO: allow more et */
    SIVA_ASSERT(0,"Too many et");
    siva_num_et_allocated *= SIVA_ET_GROWTH_FACTOR; 

    siva_ets         = ALLOC(SIVA_ET,
				       siva_num_et_allocated);
    array_insert_last(SIVA_ET*, siva_et_ptr_array,
                                          siva_ets);
    siva_num_et = 0;
  }
  et = &siva_ets[siva_num_et];
  et->mask = ALLOC(unsigned, siva_latch_word_size);
  et->data = ALLOC(unsigned, siva_latch_word_size);
  et->is_unreachable = 0;
  siva_num_et ++;
  siva_total_num_et ++;
  return et;
}

/*---------------------------------------------------------------------------*/
static void
siva_init_et_table(void)
{
  register unsigned j;
  siva_et_table_size = SIVA_DEFAULT_ET_TABLE_SIZE;
  siva_et_table = ALLOC(SIVA_ET*,
                                  siva_et_table_size);
  for(j=0;j < siva_et_table_size; j++){
    siva_et_table[j] = NIL(SIVA_ET);
  }
}

/*---------------------------------------------------------------------------*/
static void
siva_quit_et_table(void)
{
  FREE(siva_et_table);
}

/*---------------------------------------------------------------------------*/
static unsigned 
siva_et_hash(unsigned state_value)
{
  return (state_value & (siva_et_table_size-1));
}

/*---------------------------------------------------------------------------*/
static SIVA_ET *
siva_find_et_hash(unsigned state_value)
{
  SIVA_ET *entry;

  unsigned hash_value = siva_et_hash(state_value);
  for(entry = siva_et_table[hash_value]; 
      entry && (entry->state_value > state_value); entry = entry->next);
  if ((entry)&&(entry->state_value == state_value)){
    return entry;
  }else {
    return NIL(SIVA_ET);
  }
}

/*---------------------------------------------------------------------------*/
static void
siva_insert_et_hash(SIVA_ET *et)
{
  unsigned hash_value =
    siva_et_hash(et->state_value);
  if (!siva_et_table[hash_value] ||
       siva_et_table[hash_value]->state_value
      < et->state_value ){
    et->next = siva_et_table[hash_value];
    siva_et_table[hash_value] = et;
  }else {
    SIVA_ET *entry = siva_et_table[hash_value];
    while(entry->next && 
          (entry->next->state_value > et->state_value)){
      entry = entry->next;
    }
    et->next = entry->next;
    entry->next = et;
  }
  siva_num_hash_et++;
  /* resize when density is 8 per bin */
  if ((siva_num_hash_et >> 3) == siva_et_table_size){
     siva_resize_et_table();
  }
}

/*---------------------------------------------------------------------------*/
static void
siva_remove_et_hash(SIVA_ET *et)
{
  unsigned hash_value =
    siva_et_hash(et->state_value);

  SIVA_ASSERT((siva_find_et_hash(et->state_value) != 
               NIL(SIVA_ET)),"et not found");

  if (siva_et_table[hash_value] == et){
    siva_et_table[hash_value] = et->next;
  }else {
    SIVA_ET *entry = siva_et_table[hash_value];
    while(entry->next->state_value > et->state_value){
      entry = entry->next;
    }
    SIVA_ASSERT((entry->next == et),"");
    entry->next = entry->next->next;
  }

}

/*---------------------------------------------------------------------------*/
static void
siva_resize_et_table(void)
{
  register int i;

  siva_et_table = REALLOC(SIVA_ET*,
                          siva_et_table,
                          (siva_et_table_size << 1));  
  
  for(i=0;i<siva_et_table_size;i++){
    SIVA_ET lower_header;
    SIVA_ET *lower_ptr=&lower_header;

    SIVA_ET upper_header;
    SIVA_ET *upper_ptr=&upper_header;

    SIVA_ET *entry = siva_et_table[i];

    lower_header.next = NIL(SIVA_ET);
    upper_header.next = NIL(SIVA_ET);
    while(entry){
      if (entry->state_value & (siva_et_table_size)){
        upper_ptr->next = entry;        
        upper_ptr = entry;
      }else{
        lower_ptr->next = entry;
        lower_ptr = entry;
      }
      entry = entry->next;
    }
    upper_ptr->next = NIL(SIVA_ET);
    lower_ptr->next = NIL(SIVA_ET);

    siva_et_table[i] = lower_header.next;
    siva_et_table[i+siva_et_table_size] = 
                                 upper_header.next;
  }
  siva_et_table_size = (siva_et_table_size << 1);
}

/*---------------------------------------------------------------------------*/
static void
siva_et_table_stat(void)
{
  if (siva_verbosity_level > 1) {
    register int i;
    float avg_bin_load =
      (float)siva_num_hash_et/siva_et_table_size;
    float bin_load_var;
    float std=0;
    for(i=0;i<siva_et_table_size;i++){
      SIVA_ET *entry = siva_et_table[i];
      int num_entry=0;
      for(entry=siva_et_table[i];
          entry;entry=entry->next,num_entry++);
      bin_load_var = (num_entry-avg_bin_load);
      bin_load_var *= bin_load_var;
      std += bin_load_var;
    }
    fprintf(siva_output_file,"avg load = %f\n",avg_bin_load);
    fprintf(siva_output_file,"std load = %f\n",
            sqrt(std/siva_et_table_size));
    fprintf(siva_output_file,"table_size = %d\n",
            siva_et_table_size);
    fprintf(siva_output_file,"load = %d\n",
            siva_num_hash_et);
  }
  fprintf(siva_output_file,"# et = %d\n",
          siva_total_num_et);
}




/*--------------------------------------------------------------------------*/
static void
siva_put_et(unsigned priority_value, SIVA_ET *et)
{
  SIVA_KEY key;
  SIVA_CONV conv;
  key.unsigned_key = priority_value;
  conv.et = et;
  SIVA_ASSERT(et->priority == key.unsigned_key,"");
  siva_heap_insert(key,conv,siva_et_heap_id);
}

/*--------------------------------------------------------------------------*/
static int
siva_get_et(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr)
{
  int found;

  found = siva_heap_extract_top(key_ptr,conv_ptr,
                                siva_et_heap_id);
  return found;
}


/*--------------------------------------------------------------------------*/
static unsigned
siva_get_et_priority(SIVA_ET *et)
{
  int i;
  unsigned value = 0;

  var_set_t mask_vector;
  var_set_t state_vector;

  siva_set_state_vector(&mask_vector,et->mask);
  siva_set_state_vector(&state_vector,et->data);
  
  for(i=0;i<siva_num_latch;i++) {
    if (!var_set_get_elt(&mask_vector,i)) {
      value += 100;
    }else {
      value += siva_get_latch_inertia(i,var_set_get_elt(&state_vector,i));
    }
  }
  return value;
}

/*---------------------------------------------------------------------------*/
void
siva_et_solve(array_t *state_wrapper_array, int num_state)
{
  int i;
  int value;
  int num_et;
  int is_resolved;
  SIVA_NODE target;  
  SIVA_ET *ets[SIVA_DEFAULT_ET_PICK_SIZE];

  num_et = siva_get_best_et(ets,SIVA_DEFAULT_ET_PICK_SIZE);  
  SIVA_VERB1_ARG1("Solving best ets ...\n");
  for(i=0;i<num_et; i++){
    target = siva_get_et_target_node(ets[i]);
    is_resolved = 0;
    if (!bck_is_flag_set(target,siva_target_one_flag) &&
        !bck_is_flag_set(target,siva_target_zero_flag)){
      is_resolved = 1;
    }
    if (is_resolved) {
      continue;
    }
    is_resolved = siva_et_solve1(ets[i],state_wrapper_array,num_state);    
    if (is_resolved) {
      SIVA_VERB1_ARG3("SYMBOLIC-SIM ET: Target (%d=%d) reachable\n",
                      bck_get_index(target),
                      bck_is_flag_set(target,siva_target_one_flag));
      if (bck_is_flag_set(target,siva_target_one_flag)){
        bck_reset_flag(target,siva_target_one_flag);    
        siva_num_one_target--;
      }else if (bck_is_flag_set(target,siva_target_zero_flag)) {
        bck_reset_flag(target,siva_target_zero_flag);    
        siva_num_zero_target--;
      }
    }
  }
  
  /* Put back the unsolved ones; maybe with lower priority? */
  for(i=0;i<num_et; i++){
    SIVA_NODE target = siva_get_et_target_node(ets[i]);
    if (bck_is_flag_set(target,siva_target_one_flag) ||
        bck_is_flag_set(target,siva_target_zero_flag)) {
      siva_put_et(ets[i]->priority,ets[i]);
    }
  }    
  SIVA_VERB1_ARG1("Solving et ...done\n");
}

/*--------------------------------------------------------------------------*/
/* Gets num_lh_req lighthouse from the heap; each lighthouse picked
   corresponds to atleast one unreached target
*/
static int
siva_get_best_et(SIVA_ET *ets[],
                 int num_et_req)
{
  SIVA_KEY key;
  SIVA_CONV conv;
  int num_et = 0;
  SIVA_ET *et;
  SIVA_NODE target;
  int is_redundant;

  
  while(siva_get_et(&key,&conv) && (num_et < num_et_req)){
    et = conv.et;
    if (et->is_unreachable) {
      continue;
    }
    if (et->is_target_et) {
      target = siva_get_et_target_node(et);
    }else {
      target = et->target.node;
    }
    is_redundant = 0;
    if (!bck_is_flag_set(target,siva_target_one_flag) &&
        !bck_is_flag_set(target,siva_target_zero_flag)){
      is_redundant = 1;
    }
    if (is_redundant) {
      continue;
    }
    ets[num_et++] = et;
    /* re-heapify later with lower priority ? */
    SIVA_ASSERT(et->priority == key.unsigned_key,"");
    SIVA_VERB1_ARG2("ET: priority value = %u\n",
                    key.unsigned_key);
  }  
  return num_et;
  
}


/*---------------------------------------------------------------------------*/
static int
siva_et_solve1(SIVA_ET *et,
              array_t *state_wrapper_array, int num_state)
{
  int i,j;
  var_set_t state_vector;
  var_set_t mask_vector;
  int reachable;
  int num_goal;
  int goal_values[siva_num_latch];
  SIVA_NODE goals[siva_num_latch];
  SIVA_CONV conv;

  SIVA_ASSERT(!et->is_unreachable,
              "ET is proved unreachable before");
  siva_set_state_vector(&mask_vector,et->mask);
  siva_set_state_vector(&state_vector,et->data);

  num_goal = 0;
  for(i=0;i<siva_num_latch;i++){
    if (!var_set_get_elt(&mask_vector,i)){
      continue;
    }
    if (bck_is_constant(siva_latch_nodes[i].input)){
      SIVA_ASSERT(var_set_get_elt(&state_vector,i) ==
                  bck_is_one(siva_latch_nodes[i].input),
                  "mismatch in constant node");
      continue;
    }
    goal_values[num_goal] = var_set_get_elt(&state_vector,i) ^
        bck_is_inverted(siva_latch_nodes[i].input);
    goals[num_goal] = bck_get_regular(siva_latch_nodes[i].input);
    num_goal++;
  }

  /* be less aggressive ?*/
  //siva_symbolic_sim_depth =>> 1;
  
  for(j=0;j<num_state;j++) {
    SIVA_STATE_WRAPPER *state_wrapper =
      array_fetch_p(SIVA_STATE_WRAPPER,state_wrapper_array,j);
    //conv.et = et;
    reachable = siva_symbolic_sim1(goals,
                                   goal_values,num_goal,
                                   state_wrapper,1,
                                   NULL,
                                   siva_store_target_witness_state,
                                   siva_atpg_print_witness_srv);

    if (reachable == 1) {
      siva_et_print_witness(et);
      return 1;
    }
    siva_gc_symbolic_sim();    
  }

  return 0;
}

/*---------------------------------------------------------------------------*/
/* take the goals and returns the enlarged target                            */
SIVA_ET*
siva_enlarge_et(SIVA_ET *et)
{
  SIVA_ET ets[siva_target_enlarge_factor];
  float priority;
  int i;
  SIVA_NODE goals[siva_num_latch];
  int goal_values[siva_num_latch];
  var_set_t state_vector;
  var_set_t mask_vector;
  int num_goal = 0;
  int num_found;
  unsigned data[siva_latch_word_size * siva_target_enlarge_factor * 2];

  SIVA_CONV conv;
  conv.value_ptr = data;
  for(i=0;i<siva_target_enlarge_factor;i++){
    ets[i].mask = conv.value_ptr;
    conv.value_ptr += siva_latch_word_size;
    ets[i].data = conv.value_ptr;
    conv.value_ptr += siva_latch_word_size;    
  }
  
  siva_set_state_vector(&mask_vector,et->mask);
  siva_set_state_vector(&state_vector,et->data);
    
  for(i=0;i<siva_num_latch;i++){
    if (!var_set_get_elt(&mask_vector,i)){
      continue;
    }
    if (bck_is_constant(siva_latch_nodes[i].input)){
      SIVA_ASSERT(var_set_get_elt(&state_vector,i) ==
                  bck_is_one(siva_latch_nodes[i].input),
                  "mismatch in constant node");
      continue;
    }
    goal_values[num_goal] = var_set_get_elt(&state_vector,i) ^
      bck_is_inverted(siva_latch_nodes[i].input);
    goals[num_goal] = bck_get_regular(siva_latch_nodes[i].input);
    num_goal++;
  }
 
  num_found = siva_atpg_enlarge(goals,goal_values,num_goal,
                    ets,siva_target_enlarge_factor);
  if (!num_found) {
    return NIL(SIVA_ET);
  }
  return siva_store_ets(ets,(void *)et,1,num_found);
}

/*---------------------------------------------------------------------------*/
/* take the goals and returns the enlarged target                            */
static SIVA_ET*
siva_store_ets(SIVA_ET *ets, void *target, int is_target_et,
               int num_et)
{
  int i;
  var_set_t state_vector;
  var_set_t mask_vector;
  SIVA_ET *new_et;
  int is_created;

  SIVA_ASSERT(num_et>0,"");
  for(i=num_et-1 ; i >=0 ;i--) {
    siva_set_state_vector(&mask_vector,ets[i].mask);
    siva_set_state_vector(&state_vector,ets[i].data);
    new_et = siva_get_or_create_et(&mask_vector,
                                   &state_vector,
                                   target,
                                   is_target_et,
                                   i+1,
                                   &is_created);

    SIVA_ERR_CHK1(siva_verify_et(new_et,&state_vector));
    
    if (!is_created) {
      continue;
    }
    siva_put_et(new_et->priority,new_et);
    siva_generate_alh(new_et);      
  }
  return new_et;
}


/*---------------------------------------------------------------------------*/
/* take an et and returns the enlarged target                                */
SIVA_ET*
siva_enlarge_target(SIVA_NODE target, int target_value)
{
  SIVA_ET ets[siva_target_enlarge_factor];
  unsigned priority;
  int i;
  int num_found;
  unsigned data[siva_latch_word_size * siva_target_enlarge_factor * 2];

  SIVA_CONV conv;
  conv.value_ptr = data;
  for(i=0;i<siva_target_enlarge_factor;i++){
    ets[i].mask = conv.value_ptr;
    conv.value_ptr += siva_latch_word_size;
    ets[i].data = conv.value_ptr;
    conv.value_ptr += siva_latch_word_size;    
  }
  SIVA_ASSERT(!bck_is_inverted(target),"target should not be inverted");
  num_found =
    siva_atpg_enlarge(&target,&target_value,1,ets,siva_target_enlarge_factor);

  if (!num_found) {
    return NIL(SIVA_ET);
  }
  return siva_store_ets(ets,(void *)target,0,num_found);
}

/*--------------------------------------------------------------------------*/
void
siva_backward_sim(void)
{
  register int i,k;
  int reachable;
  SIVA_NODE target;
  
  if (!siva_et_on) {
    return;
  }
  if (!siva_built_et_frontiers_done) {
    siva_build_et_frontiers();
    siva_built_et_frontiers_done = 1;
    return;
  }
  SIVA_VERB1_ARG1("Backward Simulation: ..\n");  
  for(i=0;i<siva_num_et_frontier;i++){
    if (!siva_et_frontiers[i]) {
      continue;
    }
    target = siva_get_et_target_node(siva_et_frontiers[i]);
    
    if (!bck_is_flag_set(target,
                        siva_target_one_flag) &&
        !bck_is_flag_set(target,
                         siva_target_zero_flag)){
      continue;
    }
    siva_et_frontiers[i] = siva_enlarge_et(siva_et_frontiers[i]);
  }
  SIVA_VERB1_ARG1("Backward Simulation: ..done\n");    

} 


/*---------------------------------------------------------------------------*/
static void
siva_build_et_frontiers(void)
{
  int num_target;
  int i;
  SIVA_CONV conv;
  SIVA_NODE *targets;
  if (siva_num_zero_target + siva_num_one_target ==0) {
    fprintf(siva_output_file,"No targets to resolve\n");
    return;
  }
  SIVA_VERB1_ARG1("Building et frontiers: ..\n");    
  siva_target_stat();  
  /* initialize for targets */
  num_target = siva_num_zero_target+siva_num_one_target;
  targets = ALLOC(SIVA_NODE,num_target);  
  siva_fill_targets(targets);

  siva_num_et_frontier = num_target;
  siva_et_frontiers = ALLOC(SIVA_ET*,
                            siva_num_et_frontier);  
  for(i=0;i<num_target;i++){
    siva_et_frontiers[i] =
      siva_enlarge_target(targets[i],
                          bck_is_flag_set(targets[i],
                                          siva_target_one_flag));
    SIVA_ASSERT(siva_et_frontiers[i],
                "Target is unreachable?");
               
  }
  FREE(targets);
  SIVA_VERB1_ARG1("Building et frontiers: ..done\n");
}

/*---------------------------------------------------------------------------*/
static void
siva_destroy_et_frontiers(void)
{
  if (siva_num_et_frontier) {
     FREE(siva_et_frontiers);
  }
}

/*
  ---------------------------------------------------------------------------
  Given a state vector, it returns 1 if any of ets are reached and
  the corresponding target is reset.
  ---------------------------------------------------------------------------
*/

int
siva_check_et_reached(SIVA_STATE *state,
                      var_set_t *state_vector)
{
  SIVA_ITER iter;
  SIVA_ET *et;
  int witness_state = 0;

  et = siva_get_first_et(&iter);
  if (!et) {
    return 0;
  }
  do {
    int reachable;
    SIVA_NODE target;
    if (et->is_unreachable ) {
      continue;
    }
    target = siva_get_et_target_node(et);
    if (!bck_is_flag_set(target,siva_target_one_flag) &&
        !bck_is_flag_set(target,siva_target_zero_flag)) {
      continue;
    }
    reachable = siva_check_et_reached1(et,state_vector);
    if (reachable) {
      witness_state = 1;
      {
        int distance;
        SIVA_VERB1_ARG4("CYCLE-SIM ET: (Depth=%d): Reached Target (%d=%d)\n",
                        siva_current_sim_depth,bck_get_index(target),
                        bck_is_flag_set(target,siva_target_one_flag));
        distance = siva_get_et_target_distance(et);
        SIVA_VERB2_ARG2("# enlargements required = %d\n",distance);
      }
      siva_print_state_trace(state);
      siva_et_print_witness(et);
      if (bck_is_flag_set(target,siva_target_one_flag)){
        bck_reset_flag(target,siva_target_one_flag);    
        siva_num_one_target--;
      }else if (bck_is_flag_set(target,siva_target_zero_flag)) {
        bck_reset_flag(target,siva_target_zero_flag);    
        siva_num_zero_target--;
      }
    }
  }while (et = siva_get_next_et(&iter)) ;

  return witness_state;

}

/*---------------------------------------------------------------------------*/
static int
siva_check_et_reached1(SIVA_ET *et,
                       var_set_t *visited_state_vector)
{
  unsigned data[siva_latch_word_size];
  var_set_t project_vector;
  var_set_t mask_vector;
  var_set_t state_vector;
  int reachable;
  SIVA_ASSERT(!et->is_unreachable,
              "This et is marked unreachable");

  siva_set_state_vector(&mask_vector,et->mask);
  siva_set_state_vector(&state_vector,et->data);
  siva_set_state_vector(&project_vector,data);
  var_set_clear(&project_vector);
  
  var_set_and(&project_vector,&mask_vector,visited_state_vector);
  reachable = var_set_equal(&project_vector,&state_vector);
  if (reachable) {
    SIVA_ERR_CHK1(siva_verify_et(et,visited_state_vector));
  }
    
  return reachable;
}

/*---------------------------------------------------------------------------*/
static void
siva_verify_et(SIVA_ET *et, var_set_t *sample_vector)
{
  SIVA_STATE_WRAPPER wrapper;
  SIVA_NODE target = siva_get_et_target_node(et);
  int target_value = bck_is_flag_set(target,siva_target_one_flag);
  int prev_depth =   siva_symbolic_sim_depth;
  int reachable;
  wrapper.state_vector = sample_vector;
  wrapper.state = NIL(SIVA_STATE);
  SIVA_ASSERT(!bck_is_inverted(target),"");
  siva_symbolic_sim_depth = siva_get_et_target_distance(et);
  reachable = siva_symbolic_sim1(&target,
                                 &target_value,1,
                                 &wrapper,1,
                                 NIL(void),NULL,NULL);
  SIVA_ASSERT(reachable,              
              "ET: is not reachable?");
  siva_symbolic_sim_depth = prev_depth;  
  SIVA_VERB1_ARG1("Verified ET\n");
}

/*---------------------------------------------------------------------------*/
static void
siva_et_print_witness(SIVA_ET *et)
{
  var_set_t state_vector;
  SIVA_STATE_WRAPPER wrapper;
  SIVA_NODE target;
  int target_value;
  int prev_depth;
  int reachable;

  if (!array_n(siva_debug_node_array)) {
    return;
  }
  target = siva_get_et_target_node(et);
  target_value = bck_is_flag_set(target,siva_target_one_flag);
  prev_depth =   siva_symbolic_sim_depth;

  siva_set_state_vector(&state_vector,et->data);  
  wrapper.state_vector = &state_vector;
  wrapper.state = NIL(SIVA_STATE);
  SIVA_ASSERT(!bck_is_inverted(target),"");
  siva_symbolic_sim_depth = siva_get_et_target_distance(et);
  SIVA_VERB1_ARG3("Printing trace from ET to the target (%d=%d)\n",
                  target,target_value);
  reachable = siva_symbolic_sim1(&target,
                                 &target_value,1,
                                 &wrapper,1,
                                 NIL(void),NULL,
                                 siva_atpg_print_witness_srv1);
  SIVA_ASSERT(reachable,              
              "ET: is not reachable?");
  siva_symbolic_sim_depth = prev_depth;  
}

