/**CFile**********************************************************************

  FileName                   [siva_exec.c]

  PackageName                [siva]

  Synopsis                   [External command execution routines] 

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/11/01]

  Copyright                  []
******************************************************************************/

#include "siva.h"

static array_t *siva_ess_complex_node_array;
static SIVA_NODE siva_latch_first_cycle_high;
static SIVA_NODE siva_latch_first_cycle_low;
static SIVA_NODE siva_latch_pos_clock;
static SIVA_NODE siva_latch_neg_clock;

static void siva_init_ext_interface(void);
static void siva_quit_ext_interface(void);
static void siva_build_ess_complex_node_array(void);
static void siva_build_bck_all(void);
static void siva_store_node_identification(void);
static void siva_remove_constant_latch_srv(SIVA_COMPLEX_NODE *node);
static void siva_store_node_identification_srv(SIVA_COMPLEX_NODE *node);
static void siva_store_target_identification_srv(SIVA_COMPLEX_NODE *node);
static void siva_release_node_identification(void);
static void siva_destroy_bck_srv(SIVA_COMPLEX_NODE *node);
static void siva_prune_target_srv(SIVA_COMPLEX_NODE *complex_node);
static void siva_build_ess_complex_node_array_srv(SIVA_COMPLEX_NODE *complex_node);
static void siva_build_light_house_node_array_srv(SIVA_COMPLEX_NODE *complex_node);
static void siva_print_complex_node_bck_srv(SIVA_COMPLEX_NODE *node);
static void siva_move_latch_target(void);
static void siva_target_identification(void);
static void siva_store_target_srv(SIVA_COMPLEX_NODE *complex_node);

/*
  ---------------------------------------------------------------------------
  Initializes external interface.
  If already done, closes it first.
  Reads the blif file.
  Marks all nodes as potential targets.
  ---------------------------------------------------------------------------
*/
int 
siva_exec_read_blif(void)
{
  /* consecutive read */
  if (siva_init_ext_interface_done){
    siva_quit_ext_interface();
  }
  siva_init_ext_interface_done = 1; 
  siva_init_ext_interface();

  /* check if problem in reading blif */
  if (siva_read_blif_file() == 0){
    siva_init_ext_interface_done = 0;
    siva_quit_ext_interface();
    return 0;
  }
  siva_do_for_all_complex_node(siva_clear_visit_srv);  
  /* TODO: need to check cycle */
  siva_do_for_all_complex_node(siva_levelize_complex_node_srv);
  SIVA_VERB2(siva_do_for_all_complex_node(siva_print_complex_node_srv));
  /* Mark all node as potential targets */
  siva_do_for_all_complex_node(siva_set_target_srv);
  return 1;

}


/*
  ---------------------------------------------------------------
  Reads the targets from the siva_target_file or targets unresolved
  if no such file exists. Build BCK network of the next state nodes
  and the target nodes. Invokes siva_pre_process_target to solve
  the targets. Stores the unresolved target in a specified file.
  Releases the BCK network after it is done.
  ---------------------------------------------------------------
*/
int 
siva_exec_pre_process(void)
{

  /* init steps */  
  siva_init_bck();
  siva_ess_complex_node_array = array_alloc(SIVA_COMPLEX_NODE *,32); 
  /* read the target file */
  siva_read_target_file();  
  /* based on the target list, build the essential array */
  siva_build_ess_complex_node_array();
  /* build the bck for only target + next state nodes*/
  siva_build_bck_all();
  array_free(siva_ess_complex_node_array);
  SIVA_VERB2(bck_stat(siva_output_file));  
  /* identify bck nodes as latches and pi */
  siva_store_node_identification();
  /*
    Identify bck nodes as targets with target value.
    Latch targets are moved to its next state nodes.
  */  
  siva_target_identification();
  /* write the network information in a local file */
  siva_map_file = fopen("id-name","w");
  siva_do_for_all_complex_node(siva_print_complex_node_srv);
  fprintf(siva_map_file,"\n");
  fclose(siva_map_file);
  SIVA_VERB1(siva_network_stat());
  SIVA_VERB1(siva_target_stat());
  /* build initial state vector */
  siva_build_initial_state_vector();
  /* pre process the targets */
  siva_pre_process_target();
  siva_free_initial_state_vector();
  SIVA_VERB1(siva_target_stat());
  /* Unmark the complex node set as target if it is reachable */
  siva_do_for_all_complex_node(siva_prune_target_srv);
  /*
    Mark the complex node set as target if it is not reachable.
    This is required because when latch targets are moved to its input,
    the corresponding complex node might not have been targetted.
  */
  siva_do_for_all_complex_node(siva_store_target_srv);
  /* write out the unreached targets */
  siva_write_target_file();
  /* release the node information */
  siva_release_node_identification();
  /* destroy the bck information from node field of complex node*/
  siva_do_for_all_complex_node(siva_destroy_bck_srv);
  siva_quit_bck();
  return 1;
}

/*
  ---------------------------------------------------------------------------
  Invokes solve target. Should be called in solve mode
  ---------------------------------------------------------------------------
*/

int
siva_exec_solve(void)
{
  siva_solve_target();
  SIVA_VERB1(siva_target_stat());
}


/*
  ---------------------------------------------------------------------------
  Writes the internal representation in Blif format.
  Can be invoked only in non-solve mode.
  ---------------------------------------------------------------------------
*/

int
siva_exec_write_bck(void)
{
  /* init steps */  
  siva_init_bck();
  siva_ess_complex_node_array = array_alloc(SIVA_COMPLEX_NODE *,0); 
  /* based on the target list, build the essential array */
  siva_build_ess_complex_node_array();
  /* build the bck for only target + next state */
  siva_build_bck_all();
  array_free(siva_ess_complex_node_array);
  SIVA_VERB2(bck_stat(siva_output_file));
  /* latches and pi */
  siva_store_node_identification();
  siva_write_bck_file();
  siva_release_node_identification();
  /* destroy the bck information from node field of complex node*/
  siva_do_for_all_complex_node(siva_destroy_bck_srv);
  siva_quit_bck();
  return 1;
}

/*
  ---------------------------------------------------------------------------
  Reads the targets from the siva_target_file or targets unresoved if
  no such file exists. Prepares network for solving the targets.
  Releases the external interface resources.
  Depdending on the options, it finds the active control groups,
  partitions using overlap threshold. Finds out the latch activity and
  affinity.
  ---------------------------------------------------------------------------
*/
int
siva_exec_solve_init(void)
{

  /* init steps */  
  siva_init_bck();
  siva_debug_node_array = array_alloc(SIVA_NODE,16);
  siva_debug_node_name_array = array_alloc(char*,16);
  
  siva_ess_complex_node_array = array_alloc(SIVA_COMPLEX_NODE *,0); 
  siva_read_target_file();
  siva_read_debug_node_file();
  /* based on the target list, build the essential array */
  siva_build_ess_complex_node_array();
  /* build the bck for only target + next state */
  siva_build_bck_all();
  array_free(siva_ess_complex_node_array);

  /* idetify bck nodes as latches and pi */
  siva_store_node_identification();
  /*
    Identify bck nodes as targets with target value.
    Latch targets are moved to its next state nodes.
  */    
  siva_target_identification();

  siva_init_witness();  
  siva_init_heap(); 
  siva_init_state(); 
  siva_init_control_state();
  siva_init_wit_state();
  siva_init_bfs_state();

  siva_init_symbolic_sim();  
  siva_init_alh();
  siva_init_et();
  siva_init_mlh();
  siva_init_latch_activity();
 
  
  siva_build_mlh();
  siva_build_initial_state_vector();
  siva_create_initial_state();

  /* put another command for this */
  siva_evaluate_latch_activity();
  siva_control_state();

  siva_network_stat();
  siva_target_stat();
  /* print out the information in a file */
  siva_map_file = fopen("id-name","w");  
  siva_do_for_all_complex_node(siva_print_complex_node_srv);
  fclose(siva_map_file);
  
  /* quit pre process */
  siva_quit_ext_interface();
  siva_init_ext_interface_done=0;
  return 1;
}

/*
  ---------------------------------------------------------------------------
  Release the resources used during solve_mode
  ---------------------------------------------------------------------------
*/

int
siva_exec_solve_quit(void)
{
  siva_network_stat();
  siva_target_stat();
  siva_state_table_stat();  
  siva_quit_latch_activity();
  siva_quit_mlh();
  siva_quit_et();
  siva_quit_alh();
  siva_quit_symbolic_sim();  
  siva_quit_wit_state();
  siva_quit_bfs_state();
  siva_quit_control_state();
  siva_quit_state();  
  siva_quit_heap();  
  siva_quit_witness();  

  /* release the node information */
  siva_release_node_identification();
  array_free(siva_debug_node_array);
  array_free(siva_debug_node_name_array);
  siva_quit_bck();
}

/*
  -------------------------------------------------------------------------
  Static Routines follows
  -------------------------------------------------------------------------
*/

                                                                            
/*
  ---------------------------------------------------------------------------
  Initializes external interface.
  ---------------------------------------------------------------------------
*/

static void
siva_init_ext_interface(void)
{
  siva_init_complex_node();
}

/*
  ---------------------------------------------------------------------------
  quits external interface. Releases the complex_node data structure.
  ---------------------------------------------------------------------------
*/

static void
siva_quit_ext_interface(void)
{
  siva_quit_complex_node();
}  

/*
  ---------------------------------------------------------------------------
  Builds the array of complex node for which BCK structure will be
  built.
  ---------------------------------------------------------------------------
*/

static void
siva_build_ess_complex_node_array(void)
{
  siva_do_for_all_complex_node(siva_build_ess_complex_node_array_srv);
  siva_do_for_all_complex_node(siva_build_light_house_node_array_srv);
}
  
/*
  ---------------------------------------------------------------------------
  Insert complex node that are latch inputs and
  marked as target and debug nodes in siva_ess_complex_node_array
  ---------------------------------------------------------------------------
*/

static void
siva_build_ess_complex_node_array_srv(SIVA_COMPLEX_NODE *complex_node)
{

  if (siva_is_target_zero_set(complex_node) ||
      siva_is_target_one_set(complex_node) ||
      siva_is_latch_input(complex_node) ||
      siva_is_debug_node(complex_node)){          

    array_insert_last(SIVA_COMPLEX_NODE *,
                      siva_ess_complex_node_array,complex_node);
  }

  return;

}

/*
  ---------------------------------------------------------------------------
  Inserts manual lighthouses into siva_ess_complex_node_array.
  ---------------------------------------------------------------------------
*/

static void
siva_build_light_house_node_array_srv(SIVA_COMPLEX_NODE *complex_node)
{
  register int i;
  
  /* Lighthouse might be already entered. Duplicate insertion of
     node is inefficient but ok.
  */
  if (siva_is_target_zero_set(complex_node) ^
      siva_is_target_one_set(complex_node)){
    array_t *light_house_array =  complex_node->light_house_array;
    if (light_house_array){
      SIVA_COMPLEX_NODE *node;
      for(i=0;i<array_n(light_house_array);i+=2){
        node = array_fetch(SIVA_COMPLEX_NODE *,
                           light_house_array,i);
        array_insert_last(SIVA_COMPLEX_NODE *,
                          siva_ess_complex_node_array,node);
      }
    }
  } 
  return;
}


/*
  --------------------------------------------------------------------------
  Builds BCK network for the complex nodes in
  siva_ess_complex_node_array.
  --------------------------------------------------------------------------
*/

static void
siva_build_bck_all(void)
{
  register int i;
  array_t *complex_node_array = array_alloc(SIVA_COMPLEX_NODE*,16);  
  /* clears the visit flag */
  siva_do_for_all_complex_node(siva_clear_visit_srv);  
  for(i=0;i<array_n(siva_ess_complex_node_array);i++){
      SIVA_COMPLEX_NODE *complex_node = 
        array_fetch(SIVA_COMPLEX_NODE *,siva_ess_complex_node_array,i);
      siva_fill_fanin_cone(complex_node_array,complex_node);
  }  
  /*
    array_sort(complex_node_array,siva_complex_node_level_cmp);
  */  
  for(i=0;i<array_n(complex_node_array);i++){
    SIVA_COMPLEX_NODE *complex_node = 
               array_fetch(SIVA_COMPLEX_NODE *,complex_node_array,i);
    siva_build_bck(complex_node);    
  }
  array_free(complex_node_array);  
  SIVA_VERB2(bck_reset_all_flag(siva_print_flag));
  SIVA_VERB2(siva_do_for_all_complex_node(siva_print_complex_node_bck_srv));
}


/*
  ---------------------------------------------------------------------------
  Print the fanin cone nodes
  ---------------------------------------------------------------------------
*/

static void
siva_print_complex_node_bck_srv(SIVA_COMPLEX_NODE *node)
{
  if (bck_is_null(siva_get_node(node))){
    return;
  }
  siva_print_fanin_cone(siva_get_node(node));
  bck_reset_fanin_cone_flag(siva_get_node(node),siva_print_flag);
}



 
/*
  ---------------------------------------------------------------------------
  Release the association of complex_node and the corresponding node.
  ---------------------------------------------------------------------------
*/

static void
siva_destroy_bck_srv(SIVA_COMPLEX_NODE *complex_node)
{
  siva_set_node(complex_node,bck_create_null());
}


/*
  ---------------------------------------------------------------------------
  Identify latches and inputs.
  Merges and propogates constant latches.
  Order latches on the basis of their dependency.
  ---------------------------------------------------------------------------
*/
static void
siva_store_node_identification(void)
{
  int num_primary_input=siva_num_primary_input;
  int num_latch=siva_num_latch;
  register int i;

  siva_last_network_node = bck_get_last_node();
  siva_primary_inputs = siva_alloc_node(siva_num_primary_input);
  siva_latch_nodes = siva_alloc_latch_node(siva_num_latch);

  siva_num_latch=0;
  siva_num_primary_input=0;

  /* Propogate the constant latches and also merge the first_cycle high/low latches */
  siva_latch_first_cycle_high = BCK_NULL;
  siva_latch_first_cycle_low = BCK_NULL;

  /* remove all the constant latches */
  siva_do_for_all_complex_node(siva_remove_constant_latch_srv);
  /* identify nodes as latch or input */
  siva_do_for_all_complex_node(siva_store_node_identification_srv);

  SIVA_ASSERT((num_latch>=siva_num_latch),"latches less?");
  SIVA_ASSERT((num_primary_input>=siva_num_primary_input),"input less?");

  /* sort the inputs in the ascending order of their node ids */
  qsort(siva_primary_inputs,siva_num_primary_input,sizeof(SIVA_NODE),
        siva_node_cmp);
  
  /* Order the latches in the order of the dependency.
     If L1 is input to L2, then L2 has index less than L1 in siva_latch_nodes.
  */
  siva_order_latch_nodes();

  /* set the latch and input ids  */
  for(i = 0; i <siva_num_latch; i++){
    SIVA_CONV conv;
    conv.int_value = i;
    bck_set_aux(siva_latch_nodes[i].latch,siva_latch_id_aux,conv.ptr);
  }
  for(i = 0; i <siva_num_primary_input; i++){
    SIVA_CONV conv;
    conv.int_value = i;
    bck_set_aux(siva_primary_inputs[i],siva_pi_id_aux,conv.ptr);
  }

}

/*
  ---------------------------------------------------------------------------
  Mark the target nodes.
  ---------------------------------------------------------------------------
*/

static void
siva_target_identification(void)
{
  siva_num_zero_target = 0;
  siva_num_one_target = 0;
  siva_do_for_all_complex_node(siva_store_target_identification_srv);
  siva_move_latch_target();
  bck_do_in_topological_order(NULL,siva_target_count_srv);
  SIVA_VERB2(siva_target_stat());
}


/*
  ---------------------------------------------------------------------------
  Move latch targets to latch inputs.
  Moving the target flag from latch to the latch input will always
  end up in non-latch node. order_latch_nodes ensures that 
  the latch input, if also a latch, will be positioned
  after the latch in siva_latch_nodes.
  
  Following pruning operations are done:
  1) Remove the latch targets that matching the latch init values
  2) Move latch targets to its input until the input is not a latch;
  if such input does not exist then remove the latch target as those latches are in loop.
  Those loop latches might be required to generate clocking operation, for example.
  ---------------------------------------------------------------------------
*/ 
static void
siva_move_latch_target(void)
{
  int i;
  for(i=0;i<siva_num_latch;i++){
    if (bck_is_flag_set(siva_latch_nodes[i].latch,siva_target_zero_flag)){
      bck_reset_flag(siva_latch_nodes[i].latch,siva_target_zero_flag);
      if (bck_is_inverted(siva_latch_nodes[i].latch)){
        fprintf(siva_output_file,"Latch %d target 0 is init reachable\n",
                bck_get_regular(siva_latch_nodes[i].latch));
      }else if (bck_is_zero(siva_latch_nodes[i].input)){
        fprintf(siva_output_file,
                "Latch Input %d is constant 0; reachable in init-next step\n",
                bck_get_regular(siva_latch_nodes[i].input));
      }else if (bck_is_one(siva_latch_nodes[i].input)){
        fprintf(siva_output_file,
                "Latch Input %d is constant 1; target un-reachable\n",
                bck_get_regular(siva_latch_nodes[i].input));
      }else  if (!bck_is_flag_set(siva_latch_nodes[i].input,
                                siva_primary_input_flag)){
        fprintf(siva_output_file,
                "Moving 0 tgt from latch %d to its input %d\n",
                bck_get_regular(siva_latch_nodes[i].latch),
                bck_get_regular(siva_latch_nodes[i].input));
        if (bck_is_inverted(siva_latch_nodes[i].input)){
          bck_set_flag(siva_latch_nodes[i].input,siva_target_one_flag);        
        }else {
          bck_set_flag(siva_latch_nodes[i].input,siva_target_zero_flag);
        }
      }else {
        fprintf(siva_output_file,
                "Data input of target latch %d is PI %d, hence not targettting\n",
                bck_get_regular(siva_latch_nodes[i].latch),
                bck_get_regular(siva_latch_nodes[i].input));
      }
    }
    if (bck_is_flag_set(siva_latch_nodes[i].latch,siva_target_one_flag)){
      bck_reset_flag(siva_latch_nodes[i].latch,siva_target_one_flag);
      if (!bck_is_inverted(siva_latch_nodes[i].latch)){
        fprintf(siva_output_file,"Latch %d target 1 is init reachable\n",
                bck_get_regular(siva_latch_nodes[i].latch));
      }else if (bck_is_one(siva_latch_nodes[i].input)){
        fprintf(siva_output_file,
                "Latch Input %d is constant 1; reachable in init-next step\n",
                bck_get_regular(siva_latch_nodes[i].input));
      }else if (bck_is_zero(siva_latch_nodes[i].input)){
        fprintf(siva_output_file,
                "Latch Input %d is constant 0; target un-reachable\n",
                bck_get_regular(siva_latch_nodes[i].input));
      }else if (!bck_is_flag_set(siva_latch_nodes[i].input,
                                siva_primary_input_flag)){
        fprintf(siva_output_file,
                "Moving 1 tgt from latch %d to its input %d\n",
                bck_get_regular(siva_latch_nodes[i].latch),
                bck_get_regular(siva_latch_nodes[i].input));
        if (bck_is_inverted(siva_latch_nodes[i].input)){
          bck_set_flag(siva_latch_nodes[i].input,siva_target_zero_flag);        
        }else {
          bck_set_flag(siva_latch_nodes[i].input,siva_target_one_flag);
        }
      }else {
        fprintf(siva_output_file,
                "Data input of target latch %d is PI %d, hence not targettting\n",
                bck_get_regular(siva_latch_nodes[i].latch),
                bck_get_regular(siva_latch_nodes[i].input));
      }
    }
  }
  for(i=0;i<siva_num_latch;i++){
    SIVA_ASSERT((bck_get_regular(siva_latch_nodes[i].latch) !=
                 siva_latch_nodes[i].input),
                "latch is constant, not removed?");  
    if (bck_is_flag_set(siva_latch_nodes[i].latch,siva_target_zero_flag)){
      fprintf(siva_output_file,
              "Latch %d is in latch-loop. Not targetted for 0.\n",
              bck_get_regular(siva_latch_nodes[i].latch));
      bck_reset_flag(siva_latch_nodes[i].latch,siva_target_zero_flag);
    }
    if (bck_is_flag_set(siva_latch_nodes[i].latch,siva_target_one_flag)){
      fprintf(siva_output_file,
              "Latch %d is in latch-loop. Not targetted for 1.\n",
              bck_get_regular(siva_latch_nodes[i].latch));
      bck_reset_flag(siva_latch_nodes[i].latch,siva_target_one_flag);
    }
  }

}

/*
  ---------------------------------------------------------------------------
  Identify latches, inputs and debug nodes.
  ---------------------------------------------------------------------------
*/

static void
siva_store_node_identification_srv(SIVA_COMPLEX_NODE *complex_node)
{
  SIVA_NODE bck = siva_get_node(complex_node);

  if (siva_is_debug_node(complex_node)){
    SIVA_CONV conv;
    SIVA_ASSERT1(!bck_is_null(bck),"Structure not built for debug node %s\n",
                complex_node->name);
    conv.str = siva_alloc_copy_string(complex_node->name);
    array_insert_last(char *,siva_debug_node_name_array,conv.str);
    array_insert_last(SIVA_NODE,siva_debug_node_array,bck);    
  }

  if (bck_is_null(bck)) return;
  if (bck_is_constant(bck)) return;

  
  if ( complex_node->tbl.type == INPUT) {
    /*
      if input is constrained to be dependent on another input,
      this might occur.
    */
    if (bck_is_flag_set(bck,siva_primary_input_flag)){
      return;
    }
    siva_primary_inputs[siva_num_primary_input] = bck;
    bck_set_flag(siva_primary_inputs[siva_num_primary_input],
                 siva_primary_input_flag);      
    SIVA_VERB2_ARG3("input %d: %s\n",siva_num_primary_input,
                    complex_node->name);
    if (siva_verbosity_level >= 1){
      SIVA_CONV conv;
      conv.str = siva_alloc_copy_string(complex_node->name);
      bck_set_aux(bck,siva_bck_name_aux,conv.ptr);
    }
    siva_num_primary_input++;
  }

  if ( complex_node->tbl.type == LATCH) {
    /* if latch is merge to another, this might occur */
    if (bck_is_flag_set(bck,siva_latch_flag)){
      return;
    }    
    if (complex_node->tbl.out_phase == 0) {
      siva_latch_nodes[siva_num_latch].latch = 	bck_create_not(bck);
    }else{
      siva_latch_nodes[siva_num_latch].latch = bck;
    }
    siva_latch_nodes[siva_num_latch].input =
      siva_get_node(complex_node->tbl.input[0]);
    bck_set_flag(siva_latch_nodes[siva_num_latch].latch,siva_latch_flag);
    bck_set_flag(siva_latch_nodes[siva_num_latch].input,
		 siva_latch_input_flag);
    if (siva_verbosity_level >= 1){
      SIVA_CONV conv;
      conv.str = siva_alloc_copy_string(complex_node->name);
      bck_set_aux(bck,siva_bck_name_aux,
                  conv.ptr);
    }
    siva_num_latch++;
  }
}

/*
  ---------------------------------------------------------------------------
  Identify and remove the constant latches. Propogate the
  constants in the network.
  ---------------------------------------------------------------------------
*/
static void
siva_remove_constant_latch_srv(SIVA_COMPLEX_NODE *complex_node)
{
  if ( complex_node->tbl.type == LATCH) {
    SIVA_NODE bck = siva_get_node(complex_node);
    SIVA_NODE latch_input;
    
    if (bck_is_null(bck)) return;
    if (bck_is_constant(bck)) return;

    latch_input = siva_get_node(complex_node->tbl.input[0]);

    SIVA_ASSERT(!bck_is_inverted(bck),"latch var can't be inverted");

    if (siva_verbosity_level >= 1){
      if (complex_node->tbl.out_phase == 0 && bck_is_zero(latch_input)){
        fprintf(siva_output_file,"Removing the constant 0 latch %s\n",complex_node->name);
      }else if (complex_node->tbl.out_phase == 1 && bck_is_one(latch_input)){
        fprintf(siva_output_file,"Removing the constant 1 latch %s\n",complex_node->name);
      }else if (complex_node->tbl.out_phase == 1 && bck_is_zero(latch_input)){
        fprintf(siva_output_file,"Merging the first cycle high latch %s\n",complex_node->name);
      }else if (complex_node->tbl.out_phase == 0 && bck_is_one(latch_input)){
        fprintf(siva_output_file,"Merging the first cycle low latch %s\n",complex_node->name);
      }else if (latch_input == bck) {
        fprintf(siva_output_file,"Merging the self-loop constant latch %s\n",complex_node->name);
      }else if (bck_get_index(latch_input) == bck) {
        if (complex_node->tbl.out_phase) {
          fprintf(siva_output_file,"Merging the pos clock latch %s\n",complex_node->name);
        }else {
          fprintf(siva_output_file,"Merging the neg clock latch %s\n",complex_node->name);
        }
      }              
    }
    if (complex_node->tbl.out_phase == 0 && bck_is_zero(latch_input)){
      /* propogate the constant 0 latch */
      siva_bck_merge_and_propogate(0,bck);
    }else if (complex_node->tbl.out_phase == 1 && bck_is_one(latch_input)){
      /* propogate the constant 1 latch */
      siva_bck_merge_and_propogate(1,bck);
    }else if (complex_node->tbl.out_phase == 1 && bck_is_zero(latch_input)){
      /* merge all the first_cycle_high latches */
      if (bck_is_null(siva_latch_first_cycle_high)){
        siva_latch_first_cycle_high = bck;
      }else {
        siva_latch_first_cycle_high = bck_merge(siva_latch_first_cycle_high,bck);
      }
    }else if (complex_node->tbl.out_phase == 0 && bck_is_one(latch_input)){
      /* merge all the first_cycle_low latches */
      if (bck_is_null(siva_latch_first_cycle_low)){
        siva_latch_first_cycle_low = bck;
      }else {
        siva_latch_first_cycle_low = bck_merge(siva_latch_first_cycle_low,bck);
      }
    }else if (latch_input == bck) {
      siva_bck_merge_and_propogate(complex_node->tbl.out_phase,bck);
    }else if (bck_get_index(latch_input) == bck) {
      /* merge all the pos clock latches */
      if (complex_node->tbl.out_phase) {
        if (bck_is_null(siva_latch_pos_clock)){
          siva_latch_pos_clock = bck;
        }else {
          siva_latch_pos_clock = bck_merge(siva_latch_pos_clock,bck);
        }
      }else {
        if (bck_is_null(siva_latch_neg_clock)){
          siva_latch_neg_clock = bck;
        }else {
          siva_latch_neg_clock = bck_merge(siva_latch_neg_clock,bck);
        }        
      }      
    }
  }
}

/*
  ---------------------------------------------------------------------------
  Set the flags siva_target_one_flag and siva_target_zero_flag of the
  target nodes.
  ---------------------------------------------------------------------------
*/

static void
siva_store_target_identification_srv(SIVA_COMPLEX_NODE *complex_node)
{

  if (bck_is_null(siva_get_node(complex_node)))       return;
  
  if (siva_verbosity_level > 1) {
    if (siva_is_target_zero_set(complex_node) ||
        siva_is_target_one_set(complex_node)){
      SIVA_ASSERT(!bck_is_null(siva_get_node(complex_node)),
                  "Target node can't be null");                             
      if (bck_is_constant(siva_get_node(complex_node))){
        fprintf(siva_output_file,"%s is Constant. Not targettting\n",
                complex_node->name);
      }
      if (bck_is_flag_set(siva_get_node(complex_node),
                          siva_primary_input_flag)){
        fprintf(siva_output_file,"%s is PI. Not targettting\n",
                complex_node->name);
      }
    }
  }

  if (bck_is_constant(siva_get_node(complex_node)))    return;
  if (bck_is_flag_set(siva_get_node(complex_node),
                      siva_primary_input_flag))    return;
  
  if (siva_is_target_zero_set(complex_node)){    
    SIVA_NODE node = siva_get_node(complex_node);
    if (bck_is_inverted(node)){
      if (siva_verbosity_level >= 2){
        if (bck_is_flag_set(node,siva_target_one_flag)){
          fprintf(siva_output_file,"%s Target %d already set to 1\n",
                  complex_node->name,bck_get_regular(node));
        }
      }
      bck_set_flag(node,siva_target_one_flag);
    }else {
      if (siva_verbosity_level >= 2){
        if (bck_is_flag_set(node,siva_target_zero_flag)){
          fprintf(siva_output_file,"%s Target %d already set to 0\n",
                  complex_node->name,bck_get_regular(node));
        }
      }
      bck_set_flag(node,siva_target_zero_flag);
    }
  }

  if (siva_is_target_one_set(complex_node)){
    SIVA_NODE node = siva_get_node(complex_node);
    if (bck_is_inverted(node)){
      if (siva_verbosity_level >= 2){
        if (bck_is_flag_set(node,siva_target_zero_flag)){
          fprintf(siva_output_file,"%s Target %d already set to 0\n",
                  complex_node->name,bck_get_regular(node));
        }
      }
      bck_set_flag(node,siva_target_zero_flag);
    }else {
      if (siva_verbosity_level >= 2){
        if (bck_is_flag_set(node,siva_target_one_flag)){
          fprintf(siva_output_file,"%s Target %d already set to 1\n",
                  complex_node->name,bck_get_regular(node));
        }
      }
      bck_set_flag(node,siva_target_one_flag);
    }
  }
}



/*
  ---------------------------------------------------------------------------
  Releases resources allocated for inputs and latches.
  ---------------------------------------------------------------------------
*/

static void
siva_release_node_identification(void)
{
  if (siva_verbosity_level >= 1){
    SIVA_CONV conv;
    int i;
    for(i=0;i<siva_num_primary_input;i++){
      conv.ptr = bck_get_aux(siva_primary_inputs[i],siva_bck_name_aux);
      siva_free_string(conv.str);
      bck_set_aux(siva_primary_inputs[i],siva_bck_name_aux,
                  (void *)0);
    }
    for(i=0;i<siva_num_latch;i++){
      conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_bck_name_aux);
      siva_free_string(conv.str);
      bck_set_aux(siva_latch_nodes[i].latch,siva_bck_name_aux,
                  (void *)0);
    }
  }

  siva_free_node(siva_primary_inputs);
  siva_free_latch_node(siva_latch_nodes);
}


/*
  ---------------------------------------------------------------------------
  This is service routine called for each complex_node.
  If the target node for the complex_node is resolved, 
  then unmark the complex_node if set as  target.
  ---------------------------------------------------------------------------
*/

static void
siva_prune_target_srv(SIVA_COMPLEX_NODE *complex_node)
{
  
  if (bck_is_null(siva_get_node(complex_node))) return;

  if (siva_is_target_zero_set(complex_node)){
    SIVA_NODE node = siva_get_node(complex_node);

    SIVA_VERB3_ARG4("Checking %s 0, Bck %d %d..\n",
                    complex_node->name,bck_get_index(node),
                    (bck_is_inverted(node)==0));
    if ((bck_is_inverted(node) &&
	 (!bck_is_flag_set(node,siva_target_one_flag))) ||
	(!bck_is_inverted(node) && 
	 (!bck_is_flag_set(node,siva_target_zero_flag)))){
      
      siva_clear_target_zero(complex_node);
    }else {
      SIVA_VERB2_ARG1("Unreached\n");
    }    
  }
  
  if (siva_is_target_one_set(complex_node)){
    SIVA_NODE node = siva_get_node(complex_node);
    SIVA_VERB3_ARG4("Checking %s 1, Bck %d %d..\n",
                    complex_node->name,bck_get_index(node),
                    (bck_is_inverted(node)==0));
    if ((bck_is_inverted(node) &&
	 (!bck_is_flag_set(node,siva_target_zero_flag))) ||
	(!bck_is_inverted(node) && 
	 (!bck_is_flag_set(node,siva_target_one_flag)))){
      siva_clear_target_one(complex_node);
    }else {
      SIVA_VERB2_ARG1("Unreached\n");      
    }
  }  
}

/*
  ---------------------------------------------------------------------------
  This is service routine called for each complex_node.
  If the target node for the complex_node is unresolved,
  mark the complex_node as target.
  ---------------------------------------------------------------------------
*/

static void
siva_store_target_srv(SIVA_COMPLEX_NODE *complex_node)
{

  SIVA_NODE node = siva_get_node(complex_node);

  if (bck_is_null(node)) return;

  if ((bck_is_inverted(node) &&
       bck_is_flag_set(node,siva_target_one_flag)) ||
      (!bck_is_inverted(node) &&
       bck_is_flag_set(node,siva_target_zero_flag))){
    siva_set_target_zero(complex_node);
  }
  if ((bck_is_inverted(node) &&
       bck_is_flag_set(node,siva_target_zero_flag)) ||
      (!bck_is_inverted(node) &&
       bck_is_flag_set(node,siva_target_one_flag))){
    siva_set_target_one(complex_node);
  }  
}

