/*CFile**********************************************************************

  FileName                   [siva_heap.c]

  PackageName                [siva_core]

  Synopsis                   [heap functions for priority queues]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modifed: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]
******************************************************************************/

#include "siva.h"

#define SIVA_MAX_NUM_HEAP 32

#define SIVA_LEFT_HEAP(i)  ((i<<1)|1)
#define SIVA_RIGHT_HEAP(i)  ((i+1)<<1)
#define SIVA_PARENT_HEAP(i) (((i+1)>>1)-1)

static array_t *siva_heaps[SIVA_MAX_NUM_HEAP];
static unsigned siva_heap_sizes[SIVA_MAX_NUM_HEAP];
typedef int (*heap_cmps)(SIVA_HEAP heap1, SIVA_HEAP heap2);
static heap_cmps siva_heap_cmps[SIVA_MAX_NUM_HEAP];
static unsigned siva_num_heap_allocated;
static unsigned siva_heap_allocated_flag;

static void siva_heapify(int i,unsigned heap_id);
static int siva_heap_find_pos(SIVA_HEAP node_heap, unsigned id);
static SIVA_HEAP siva_heap_find_min(unsigned id);



/*---------------------------------------------------------------------------*/
void 
siva_init_heap(void)
{
  register int i;
  for(i=0;i<SIVA_MAX_NUM_HEAP;i++){
    siva_heap_cmps[i] = 0;
    siva_heaps[i] = 0;
    siva_heap_sizes[i] = 0;
  }
  siva_num_heap_allocated = 0;
  siva_heap_allocated_flag = 0; 
}

/*---------------------------------------------------------------------------*/
void 
siva_quit_heap(void)
{
  register int i;
  for(i=0;i<SIVA_MAX_NUM_HEAP;i++){
    siva_release_heap(i);
  }
}

/*---------------------------------------------------------------------------*/
int
siva_heap_int_max_cmp(SIVA_HEAP heap1, SIVA_HEAP heap2)
{
  return (heap1.key.int_key-heap2.key.int_key);
}

/*---------------------------------------------------------------------------*/
int
siva_heap_int_min_cmp(SIVA_HEAP heap2, SIVA_HEAP heap1)
{
  return (heap1.key.int_key-heap2.key.int_key);
}

/*---------------------------------------------------------------------------*/
int
siva_heap_unsigned_max_cmp(SIVA_HEAP heap1, SIVA_HEAP heap2)
{
  return (heap1.key.unsigned_key < heap2.key.unsigned_key)? -1 :
    (heap1.key.unsigned_key == heap2.key.unsigned_key)?  0 : 1;
}

/*---------------------------------------------------------------------------*/
int
siva_heap_unsigned_min_cmp(SIVA_HEAP heap2, SIVA_HEAP heap1)
{
  return (heap1.key.unsigned_key < heap2.key.unsigned_key)? -1 :
    (heap1.key.unsigned_key == heap2.key.unsigned_key)?  0 : 1;
}

/*---------------------------------------------------------------------------*/
int
siva_heap_float_max_cmp(SIVA_HEAP heap1, SIVA_HEAP heap2)
{
  return (heap1.key.float_key < heap2.key.float_key)? -1 :
    (heap1.key.float_key == heap2.key.float_key)?  0 : 1;
}

/*---------------------------------------------------------------------------*/
int
siva_heap_float_min_cmp(SIVA_HEAP heap2, SIVA_HEAP heap1)
{
  return (heap1.key.float_key < heap2.key.float_key)? -1 :
    (heap1.key.float_key == heap2.key.float_key)?  0 : 1;
}

/*---------------------------------------------------------------------------*/
int
siva_heap_str_min_cmp(SIVA_HEAP heap2, SIVA_HEAP heap1)
{
  return strcmp(heap1.key.str_key,heap2.key.str_key);
}

/*---------------------------------------------------------------------------*/
int
siva_heap_str_max_cmp(SIVA_HEAP heap1, SIVA_HEAP heap2)
{
  return strcmp(heap1.key.str_key,heap2.key.str_key);
}


/*---------------------------------------------------------------------------*/
unsigned 
siva_alloc_heap(int (*cmp_srv)(SIVA_HEAP heap1, SIVA_HEAP heap2))
{
  register unsigned i;
  for(i=0;i<SIVA_MAX_NUM_HEAP;i++){
    if (siva_heap_allocated_flag & (1<<i)){
      continue;
    }
    siva_num_heap_allocated++;
    siva_heap_allocated_flag |= (1<<i);
    siva_heap_cmps[i] = cmp_srv;
    siva_heaps[i] = array_alloc(SIVA_HEAP,256);
    siva_heap_sizes[i] = 0;
    return i;
  }
  SIVA_ASSERT(0,"heap is full\n");
}

/*---------------------------------------------------------------------------*/
void
siva_release_heap(unsigned heap_id)
{
  if (siva_heap_allocated_flag & (1<<heap_id)){
    SIVA_ASSERT(siva_num_heap_allocated,"no heap to release");
    siva_num_heap_allocated--;
    siva_heap_allocated_flag &= ~(1<<heap_id);
    siva_heap_cmps[heap_id] = 0;
    array_free(siva_heaps[heap_id]);
    siva_heap_sizes[heap_id] = 0;
  }
}
/*---------------------------------------------------------------------------*/
int
siva_heap_size(unsigned id)
{
  SIVA_ASSERT((siva_heap_allocated_flag & (1<<id)),"heap not valid");
  return siva_heap_sizes[id];
}

/*---------------------------------------------------------------------------*/
void
siva_do_heap(unsigned id, void (*service)(int pos,
                                           SIVA_KEY *key_p,
                                           SIVA_CONV *value_p))
{
  register int i;
  for(i=0;i<siva_heap_sizes[id];i++){
    SIVA_HEAP *node_heap = array_fetch_p(SIVA_HEAP,siva_heaps[id],i);  
    if (service) {
      service(i,&node_heap->key,&node_heap->value);
    }
  }
}

/*---------------------------------------------------------------------------*/
void
siva_heap_print(unsigned id)
{
  register int i;  
  for(i=0;i<siva_heap_sizes[id];i++){
    SIVA_HEAP node_heap = array_fetch(SIVA_HEAP,siva_heaps[id],i);
    /* only valid for int key */
    fprintf(siva_output_file,"%d \n",node_heap.key.int_key);
  }
  fprintf(siva_output_file,"\n");
}


/*---------------------------------------------------------------------------*/
void 
siva_heap_insert(SIVA_KEY key, SIVA_CONV value, unsigned id)
{
  int i;
  SIVA_HEAP node_heap;
  SIVA_HEAP parent_heap;
  node_heap.key = key;
  node_heap.value = value;
  array_insert(SIVA_HEAP,siva_heaps[id],siva_heap_sizes[id],node_heap);
  i = siva_heap_sizes[id]++;
  if (!i) {
    return;
  }
  parent_heap = array_fetch(SIVA_HEAP,siva_heaps[id],SIVA_PARENT_HEAP(i));
  while(i && (siva_heap_cmps[id](parent_heap,node_heap) < 0)) {
    array_insert(SIVA_HEAP,siva_heaps[id],i,parent_heap);    
    i = SIVA_PARENT_HEAP(i);
    if (!i) {
      break;
    }
    parent_heap = array_fetch(SIVA_HEAP,siva_heaps[id],SIVA_PARENT_HEAP(i));
  }
  array_insert(SIVA_HEAP,siva_heaps[id],i,node_heap);
}  

/*---------------------------------------------------------------------------*/
int
siva_heap_peek_top(SIVA_KEY *key_ptr, SIVA_CONV *value_ptr, unsigned id)
{
  SIVA_HEAP node_heap;

  if (siva_heap_sizes[id]==0){
    return 0;
  }
  node_heap = array_fetch(SIVA_HEAP,siva_heaps[id],0);  
  *key_ptr = node_heap.key;
  *value_ptr = node_heap.value;
  return 1;
}

/*---------------------------------------------------------------------------*/
int 
siva_heap_extract_top(SIVA_KEY *key_ptr, SIVA_CONV *value_ptr,unsigned id)
{
  SIVA_HEAP node_heap;

  if (siva_heap_sizes[id]==0){
    return 0;
  }
  node_heap = array_fetch(SIVA_HEAP,siva_heaps[id],0);  
  *key_ptr = node_heap.key;
  *value_ptr = node_heap.value;
  array_insert(SIVA_HEAP,siva_heaps[id],0,
	       array_fetch(SIVA_HEAP,siva_heaps[id],--siva_heap_sizes[id]));
  siva_heapify(0,id);
  return 1;
} 

/*---------------------------------------------------------------------------*/
void
siva_heap_replace(int i, SIVA_KEY key, SIVA_CONV value, unsigned id)
{

  SIVA_HEAP node_heap;
  SIVA_HEAP parent_heap;
  node_heap.key = key;
  node_heap.value = value;
  array_insert(SIVA_HEAP,siva_heaps[id],i,node_heap);
  siva_heapify(i,id);
  if (!i) {
    return;
  }
  node_heap = array_fetch(SIVA_HEAP,siva_heaps[id],i);
  parent_heap = array_fetch(SIVA_HEAP,siva_heaps[id],SIVA_PARENT_HEAP(i));
  while(i && (siva_heap_cmps[id](parent_heap,node_heap) < 0)) {
    array_insert(SIVA_HEAP,siva_heaps[id],i,parent_heap);    
    i = SIVA_PARENT_HEAP(i);
    if (!i) {
      break;
    }      
    parent_heap = array_fetch(SIVA_HEAP,siva_heaps[id],SIVA_PARENT_HEAP(i));
  }
  array_insert(SIVA_HEAP,siva_heaps[id],i,node_heap);
}  

/*---------------------------------------------------------------------------*/
static int
siva_heap_find_pos(SIVA_HEAP node_heap, unsigned id)
{
  register int i;
  
  for(i=siva_heap_sizes[id]-1;i>=0;i--){
    SIVA_HEAP node_heap_i = 
      array_fetch(SIVA_HEAP,siva_heaps[id],i);
    if (siva_heap_cmps[id](node_heap,node_heap_i) == 0){
      return i;
    }
  }
  SIVA_ASSERT(0,"key not found");
}

/*---------------------------------------------------------------------------*/
static SIVA_HEAP
siva_heap_find_min(unsigned id)
{
  int i;
  SIVA_HEAP min_heap = array_fetch(SIVA_HEAP,siva_heaps[id],
                                   siva_heap_sizes[id]-1);
  for(i = SIVA_PARENT_HEAP(siva_heap_sizes[id]-1)+1;
      i < siva_heap_sizes[id]-1; i++){
    SIVA_HEAP node_heap_i = 
      array_fetch(SIVA_HEAP,siva_heaps[id],i);
    if (siva_heap_cmps[id](min_heap,node_heap_i) > 0){
      min_heap = node_heap_i;
    }
  }
  return min_heap;
}
                   
/*---------------------------------------------------------------------------*/
static void
siva_heapify(int i, unsigned id)
{
  int l = SIVA_LEFT_HEAP(i);
  int r = SIVA_RIGHT_HEAP(i);
  SIVA_HEAP node_heap = array_fetch(SIVA_HEAP,siva_heaps[id],i);
  SIVA_HEAP max_heap = node_heap;
  int max_index = i;

  if (l<siva_heap_sizes[id]){  
    SIVA_HEAP left_heap = array_fetch(SIVA_HEAP,siva_heaps[id],l);
    if (siva_heap_cmps[id](left_heap,node_heap) > 0){
      max_index = l;
      max_heap = left_heap;
    }
  }   
  if (r<siva_heap_sizes[id]){ 
    SIVA_HEAP right_heap = array_fetch(SIVA_HEAP,siva_heaps[id],r);
    if (siva_heap_cmps[id](right_heap,max_heap) > 0){
      max_index = r;
      max_heap = right_heap;
    }
  } 
  if ( max_index != i){
    array_insert(SIVA_HEAP,siva_heaps[id],max_index,node_heap);
    array_insert(SIVA_HEAP,siva_heaps[id],i,max_heap);      
    siva_heapify(max_index,id);
  }
}


