/**CFile**********************************************************************

  FileName                   [siva_io.c]

  PackageName                [siva_core]

  Synopsis                   [io operation]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modifed: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]
******************************************************************************/

#include "siva.h"


static unsigned siva_written_flag;
static void siva_write_bck_target_file(void);
static int siva_write_bck_file_srv(SIVA_NODE node, void *para);

/*
  ---------------------------------------------------------------------------
  Opens the "id-name" file and calls siva_write_bck_target_file
  to write the targets.
  ---------------------------------------------------------------------------
*/
void
siva_write_target_file(void)
{

  if (!siva_pruned_target_file){
    return;
  }
  siva_map_file = fopen("id-name","r");
  if (!siva_map_file){
    fprintf(siva_output_file,
            "Error opening <id-name>. No targets written\n");
    return;
  }
  siva_written_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_written_flag);  
  siva_write_bck_target_file();
  fclose(siva_map_file);    
  bck_release_flag(siva_written_flag);
}


/*
  ---------------------------------------------------------------------------
  Read the targets and lighthouses from the file.
  ---------------------------------------------------------------------------
*/
  
int 
siva_read_target_file(void)
{
  int i;
  SIVA_COMPLEX_NODE *complex_node;
  char node_name[256];
  int target_value;
  int ignore;
  int num_light_house;
  SIVA_NODE light_house;
  array_t *light_house_array;
  SIVA_CONV conv;
  int clear_target=0;
  
  if (!siva_target_file ) {
    return 0; 
  }
  SIVA_VERB1_ARG1("Reading targets from file..\n");

  while(!feof(siva_target_file)) {
    complex_node=NIL(SIVA_COMPLEX_NODE);
    ignore=0;
    fscanf(siva_target_file,"%s %d\n",node_name,&target_value);
    if ((complex_node = siva_find_complex_node_hash(node_name)) ==
	NIL(SIVA_COMPLEX_NODE)){
        printf("Warning: Target Node %s does not exist\n",node_name);
        ignore = 1;
    }else {
      if (!clear_target){
        clear_target=1;
	siva_do_for_all_complex_node(siva_clear_target_srv);
      }

      SIVA_VERB2_ARG3("Target: Node %s, Value %d\n",node_name,target_value); 
            
      if (target_value ){
	siva_set_target_one(complex_node);
      }else {
	siva_set_target_zero(complex_node);
      }        
    }
    fscanf(siva_target_file,"%d\n",&num_light_house);
    if (!ignore) {
      if (num_light_house) {
        complex_node->light_house_array = 
        light_house_array = array_alloc(SIVA_COMPLEX_NODE*,2*num_light_house);

      }    
      for(i=0;i<num_light_house;i++) {
        fscanf(siva_target_file,"%s %d\n",node_name,&conv.value);
        if ((complex_node = siva_find_complex_node_hash(node_name)) ==
	    NIL(SIVA_COMPLEX_NODE)){
            printf("Warning: Light House %s does not exist\n",node_name);
            continue;
        }        
        array_insert(SIVA_COMPLEX_NODE*,light_house_array,2*i,complex_node);
        array_insert(SIVA_COMPLEX_NODE*,light_house_array,2*i+1,
                     conv.complex_node);      
      }    
    }else {
      continue;
    }          
  }
  if (clear_target==0){
    SIVA_VERB1_ARG1("No target found. Retaining earlier targets\n");
  }
  SIVA_VERB1_ARG1("Reading targets from file..done\n");

  return 1;
  
} 

/*
  ---------------------------------------------------------------------------
  Read the debug nodes  from the file.
  Witness will printed using these debug nodes.
  ---------------------------------------------------------------------------
*/   
int 
siva_read_debug_node_file(void)
{
  int i;
  SIVA_COMPLEX_NODE *complex_node;
  char node_name[256];
  SIVA_CONV conv;
  int clear_debug_node=0;
  if (!siva_debug_node_file ) {
    return 0; 
  }
  SIVA_VERB1_ARG1("Reading debug nodes from file..\n");
  while(!feof(siva_debug_node_file)) {
    complex_node=NIL(SIVA_COMPLEX_NODE);
    fscanf(siva_debug_node_file,"%s\n",node_name);
    if ((complex_node = siva_find_complex_node_hash(node_name)) ==
	NIL(SIVA_COMPLEX_NODE)){
        printf("Warning: Debug Node %s does not exist\n",node_name);
    }else {
      if (!clear_debug_node){
        clear_debug_node=1;
	siva_do_for_all_complex_node(siva_clear_debug_node);
      }
      
      SIVA_VERB2_ARG2("Debug: Node %s\n",node_name);                    
      siva_set_debug_node(complex_node);
    } 
  }
  if (clear_debug_node==0){
    SIVA_VERB1_ARG1("No debug nodes found. Retaining earlier debug nodes\n");
  }
  SIVA_VERB1_ARG1("Reading debug nodes from file..done\n");

  return 1;
  
} 


/*
  ---------------------------------------------------------------------------
  Reads the input blif file.
  Builds the complex node structure.
  ---------------------------------------------------------------------------
*/
int
siva_read_blif_file(void)
{
  char buf[5000];
  char temp_buf[5000];
  char *token;
  char *first_token;
  char minterms[20][1000];
  int num_minterm;
  char *list_of_tokens[500];
  int i,j;
  int num_token;
  char read_already=0;
  char reached_end=0;
  SIVA_COMPLEX_NODE *temp_node;
  SIVA_COMPLEX_NODE *node;
  SIVA_COMPLEX_NODE *fanin;
  int node_id;
  int input_parse_cont=0;
  int output_parse_cont=0;
  char first_time=0;
  char *string;
  char *node_string;
  long file_pos;
  char *pre_string;
  char *post_string;
  array_t * group_latch_array;

  SIVA_ASSERT(siva_input_file,"input file not fopen-ed");

  while ( !feof(siva_input_file) ){
    if (!read_already){
      fgets(buf,5000,siva_input_file);
    }
    SIVA_VERB2_ARG2("%s",buf);
    
    read_already=0;

    /* model name */

    if (( buf[0] == '.')&&( buf[1] == 'm')){
      token = strtok(buf," \n");
      token = strtok(NULL," \n") ;
      siva_input_design_name = siva_alloc_copy_string(token); 
      continue;
    }

    /* inputs */

    if ((input_parse_cont)||(( buf[0] == '.')&&( buf[1] == 'i'))){
      strcpy(temp_buf,buf);
      first_time=1;
      while(1){
        if ( first_time ) {
           token = strtok(temp_buf,"\\ \t\n");
           first_time=0;
           if (!input_parse_cont) {
             continue;
           }
        }else {
           token = strtok(NULL,"\\ \t\n") ;
           if (token==NULL) {
             break;
           }
        }
        if ((node = siva_find_complex_node_hash(token)) ==
            NIL(SIVA_COMPLEX_NODE)){
          node = siva_alloc_complex_node(INPUT);
          node->name = siva_alloc_copy_string(token); 
          siva_insert_complex_node_hash(node);         
        }else {
          if ( node->tbl.type == GATE ) { 
            node->tbl.type = INPUT ;
          }else {
            printf("siva_read_blif_file: conflicting types %s",node->name);
            exit(1);
          }
        }
      }
      if (strrchr(buf,'\\')){
        input_parse_cont=1;
      }else {
        input_parse_cont=0;
      }
      continue;
    }

    /* latches */
    /* assuming that the latch has only input, output and reset parameter*/

    if (( buf[0] == '.')&&( buf[1] == 'l')){
      strcpy(temp_buf,buf);
      first_token = strtok(temp_buf," \t\n");
      num_token=0;
      while((token = strtok(NULL," \t\n")) != NULL ){
        list_of_tokens[num_token] = token;
        num_token++;
      }
      /* next state is defined yet? */      
      if ((node = siva_find_complex_node_hash(list_of_tokens[0])) ==
	NIL(SIVA_COMPLEX_NODE)){
         /* assume gate, could be latch but not pi */
         node = siva_alloc_complex_node(GATE); 
         node->name = siva_alloc_copy_string(list_of_tokens[0]);
         siva_insert_complex_node_hash(node);
      } 
      /* inserting all input nodes of latches as psuedo outputs */
      node->tbl.latch_input = 1;
      temp_node = node ;   
      if ((node = siva_find_complex_node_hash(list_of_tokens[1])) ==
	NIL(SIVA_COMPLEX_NODE)){
         node = siva_alloc_complex_node(LATCH);
         node->name = siva_alloc_copy_string(list_of_tokens[1]); 
         siva_insert_complex_node_hash(node);         
      }else {
         if ( node->tbl.type == GATE ) { 
            node->tbl.type = LATCH ;
         }else {
            printf("siva_read_blif_file: conflicting types %s",node->name);
            exit(1);
         }
      }     
      siva_alloc_complex_node_input(node,1);
      node->tbl.input[0] = temp_node;
      /* bubble for latch indicates the intial state value 0/1 */
      if (atoi(list_of_tokens[2])) {
        node->tbl.out_phase = 1;
      }      
      continue;
    }


    /* the names */

   if (( buf[0] == '.')&&( buf[1] == 'n')){
      strcpy(temp_buf,buf);
      first_token = strtok(temp_buf," \t\n");
      num_token=0;
      while((token = strtok(NULL," \t\n")) != NULL ){
        list_of_tokens[num_token] = token;
        num_token++;
      }
      /* Look ahead number of minterms there */
      num_minterm = 0;
      file_pos = ftell(siva_input_file);
      while(!feof(siva_input_file)){
        fgets(buf,5000,siva_input_file);
        if (( buf[0] == '.')){
           break;
        }
        num_minterm++;
      }
      fseek(siva_input_file,file_pos,SEEK_SET);
      /* handle constant node separately */
      if ((num_minterm == 0) || (num_token == 1)){
        if (num_minterm) {
          fgets(buf,5000,siva_input_file);
        }
        if ((node = siva_find_complex_node_hash(list_of_tokens[0])) ==
	    NIL(SIVA_COMPLEX_NODE)){
           node = siva_alloc_complex_node(CONSTANT);
           node->name = siva_alloc_copy_string(list_of_tokens[0]);
           node->tbl.out_phase = ((num_minterm == 1)&&(buf[0] == '1')) ? 1 : 0;
           siva_insert_complex_node_hash(node);
        }else {
           node->tbl.type = CONSTANT;
           node->tbl.out_phase = ((num_minterm == 1)&&(buf[0] == '1')) ? 1 : 0;
        }
        continue;
      }

      if ((node = siva_find_complex_node_hash(list_of_tokens[num_token-1])) ==
	NIL(SIVA_COMPLEX_NODE)){  
        node = siva_alloc_complex_node(GATE);
        node->name = siva_alloc_copy_string(list_of_tokens[num_token-1]);
        siva_insert_complex_node_hash(node);
      }
      node->tbl.type = GATE;
      siva_alloc_complex_node_input(node,num_token-1);
      siva_alloc_complex_node_entry(node,num_minterm);
      for(i=0;i<num_token-1;i++){
        if ((fanin = siva_find_complex_node_hash(list_of_tokens[i])) 
              == NIL(SIVA_COMPLEX_NODE)){  
           fanin = siva_alloc_complex_node(GATE);
           fanin->name = siva_alloc_copy_string(list_of_tokens[i]); 
           siva_insert_complex_node_hash(fanin);
        }
        node->tbl.input[i] = fanin;
      }
 
      j=0;
      while (!feof(siva_input_file)) {
        fgets(buf,5000,siva_input_file);
        if (( buf[0] == '.')){
          read_already = 1;
          break;
        }
        node->tbl.out_phase = (buf[num_token] == '0') ? 0 : 1;
        for(i=0;i<num_token-1;i++){
          if ( buf[i] == '-' ) {
            node->tbl.entry[j][i/16] |= (1 << (( 2 * (i%16) )+1 ));
          } else {
            node->tbl.entry[j][i/16] &= ~(1 << (( 2 * (i%16) )+1 ));
            if ( buf[i] == '0') {
              node->tbl.entry[j][i/16] &= ~(1 << (2 * (i%16) ));
            } else {
              node->tbl.entry[j][i/16] |= (1 << (2 * (i%16) ));
            }
          }
        }
        j++;
      }
      continue;
    }
    if (( buf[0] == '.')&&( buf[1] == 'e')){
      break;
    }

  }
  return 1;
}

/*
  ---------------------------------------------------------------------------
  Writes out the BCK structure in blif format.
  ---------------------------------------------------------------------------
*/
int
siva_write_bck_file(void)
{
  int i;
  time_t t;
  SIVA_CONV conv;
  SIVA_COMPLEX_NODE *latch;
  SIVA_COMPLEX_NODE *latch_input;
  
  if (siva_input_design_name) {
    fprintf(siva_bck_out_file,".model %s\n",siva_input_design_name);
  }else {
    fprintf(siva_bck_out_file,".model anonymous\n");
  }  
  for(i=0;i<siva_num_primary_input;i++){
    conv.ptr = bck_get_aux(siva_primary_inputs[i],siva_bck_name_aux);
    fprintf(siva_bck_out_file,".inputs %s\n",conv.str);
  }
  for(i=0;i<siva_num_latch;i++){
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_bck_name_aux);
    latch = siva_find_complex_node_hash(conv.str);    
    fprintf(siva_bck_out_file,".latch %s %s %d\n",
            latch->tbl.input[0]->name,conv.str,
            1^bck_is_inverted(siva_latch_nodes[i].latch));
  }
  fprintf(siva_bck_out_file,".names BCK_1\n0\n");
  for(i=0;i<siva_num_primary_input;i++){
    conv.ptr = bck_get_aux(siva_primary_inputs[i],siva_bck_name_aux);
    fprintf(siva_bck_out_file,".names %s BCK_%d\n1 1\n",
            conv.str,bck_get_index(siva_primary_inputs[i]));
  }
  for(i=0;i<siva_num_latch;i++){
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_bck_name_aux);
    fprintf(siva_bck_out_file,".names %s BCK_%d\n1 1\n",
            conv.str,bck_get_index(siva_latch_nodes[i].latch));
  }

  for(i=0;i<siva_num_latch;i++) {
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_bck_name_aux);
    latch = siva_find_complex_node_hash(conv.str);
    latch_input = siva_find_complex_node_hash(latch->tbl.input[0]->name);
    if (latch_input->tbl.type == INPUT ||
        latch_input->tbl.type == LATCH) {
      continue;
    }
    fprintf(siva_bck_out_file,".names BCK_%d %s\n0 %d\n",
            bck_get_index(siva_latch_nodes[i].input),
            latch->tbl.input[0]->name,
            bck_is_inverted(siva_latch_nodes[i].input));
  }
  bck_do_in_topological_order(NULL,siva_write_bck_file_srv);
  fprintf(siva_bck_out_file,".end\n");
}



/*
  ---------------------------------------------------------------------------
  Service to print the mapping information between the external and
  internal representation in the map-file
  ---------------------------------------------------------------------------
*/
void 
siva_print_complex_node_srv(SIVA_COMPLEX_NODE *node)
{

  SIVA_NODE bck = siva_get_node(node);

  
  fprintf(siva_map_file,"NODE: %s\n",node->name);
  
  if (bck_is_null(bck)) {
    fprintf(siva_map_file,"id: 0\n");
    return;
  }

  fprintf(siva_map_file,"id: %d\n",
	  bck_get_index(bck));

  fprintf(siva_map_file,"inv: %d\n",
	  bck_is_inverted(bck));

  switch(node->tbl.type){
  case INPUT:
    fprintf(siva_map_file,"TYPE: INPUT\n");
    break;
  case LATCH:
    fprintf(siva_map_file,"TYPE: LATCH\n");
    break;
  case CONSTANT:
    fprintf(siva_map_file,"TYPE: CONSTANT\n");
    break;
  case GATE:
    fprintf(siva_map_file,"TYPE: GATE\n");
    break;
  }

  if ( bck_is_flag_set(bck,siva_latch_flag)) {
    fprintf(siva_map_file,"type: LATCH\n");
  }else if ( bck_is_var(bck)){
    fprintf(siva_map_file,"type: PI\n");
  }else if (bck_is_constant(bck)){
    fprintf(siva_map_file,"type: CONST\n");
  } else if (bck_is_null(bck)){
    fprintf(siva_map_file,"type: NULL\n");
  }else {
    fprintf(siva_map_file,"type: AND\n");
    fprintf(siva_map_file,"left_child:  %d\n",
	    bck_get_index(bck_get_left_child(bck)));
    fprintf(siva_map_file,"left_child_inverted:  %d\n",
	    bck_is_inverted(bck_get_left_child(bck)));
    fprintf(siva_map_file,"right_child:  %d\n",
	    bck_get_index(bck_get_right_child(bck)));
    fprintf(siva_map_file,"right_child_inverted:  %d\n",
	    bck_is_inverted(bck_get_right_child(bck)));
  }

  fprintf(siva_map_file,"latch_input: %d\n",
	  bck_is_flag_set(bck,siva_latch_input_flag));
  fprintf(siva_map_file,"target_one: %d\n",
	  bck_is_flag_set(bck,siva_target_one_flag));
  fprintf(siva_map_file,"target_zero: %d\n",
	  bck_is_flag_set(bck,siva_target_zero_flag));
  
}


 
/*
  ---------------------------------------------------------------------------
  Print the fanincone of the nodes
  ---------------------------------------------------------------------------
*/

void
siva_print_fanin_cone(SIVA_NODE node)
{
  SIVA_CONV conv;

  if ( bck_is_flag_set(node,siva_print_flag)){
    return;
  }
  bck_set_flag(node,siva_print_flag);

  if (bck_is_one(node)){
    fprintf(siva_output_file,"CONST 1\n");
    return;
  }else if (bck_is_zero(node)){
    fprintf(siva_output_file,"CONST 0\n");
    return;
  }else if (bck_is_var(node)){
    if ( bck_is_flag_set(node,siva_latch_flag) ){
      fprintf(siva_output_file,"LATCH %d\n",bck_get_index(node));
    }else {
      fprintf(siva_output_file,"INPUT %d\n",bck_get_index(node));
    }
    return;
  }else {
    fprintf(siva_output_file,"(%d) = ",bck_get_index(node));
    if ( bck_is_inverted(bck_get_left_child(node))){
      fprintf(siva_output_file,"!");
    }
    fprintf(siva_output_file,"(%d) * ",bck_get_index(bck_get_left_child(node)));
    
    if ( bck_is_inverted(bck_get_right_child(node))){
      fprintf(siva_output_file,"!");
    }
    fprintf(siva_output_file,"(%d)\n",bck_get_index(bck_get_right_child(node)));  
  }


  siva_print_fanin_cone(bck_get_left_child(node));
  siva_print_fanin_cone(bck_get_right_child(node));
}

/*
  ---------------------------------------------------------------------------
  Network status
  ---------------------------------------------------------------------------
*/

void
siva_network_stat(void)
{
  SIVA_VERB2(bck_stat(siva_output_file));
  fprintf(siva_output_file,"# pi = %d\n",siva_num_primary_input);
  fprintf(siva_output_file,"# latch = %d\n",siva_num_latch);
  fprintf(siva_output_file,"# 2-input and = %d\n",
          bck_get_index(bck_get_last_node())-siva_num_primary_input
          -siva_num_latch);


}

/*
  ---------------------------------------------------------------------------
  Print the target status
  ---------------------------------------------------------------------------
*/
void
siva_target_stat(void)
{
  siva_num_zero_target = 0;
  siva_num_one_target = 0;
  bck_do_in_topological_order(NULL,siva_target_count_srv);


  if (siva_verbosity_level >= 2){
    if (!(siva_num_zero_target+siva_num_one_target)){
      fprintf(siva_output_file,"Resolved all targets\n");
    }else{
      register int i;
      SIVA_CONV conv;
      SIVA_NODE targets[siva_num_zero_target+siva_num_one_target];
      SIVA_NODE *targets_ptr = targets;
      conv.bck_ptr_ptr      = &targets_ptr;
      bck_do_in_topological_order(conv.ptr,siva_fill_target_srv); 
      fprintf(siva_output_file,"List of unresolved targets\n");
      for(i=0;i<siva_num_zero_target+siva_num_one_target;i++){    
        if (bck_is_flag_set(targets[i],siva_target_one_flag) &&
            bck_is_flag_set(targets[i],siva_target_zero_flag)){
          fprintf(siva_output_file,"targets[%d]: %d= 0 & 1\n",i,targets[i]);
        }else{
          fprintf(siva_output_file,"targets[%d]: %d=%d\n",i,      
                  targets[i],bck_is_flag_set(targets[i],siva_target_one_flag));
        }      
      }
    }
  }

  SIVA_VERB1_ARG2("# zero target = %d\n",siva_num_zero_target);
  SIVA_VERB1_ARG2("# one  target = %d\n",siva_num_one_target);

}


/*
  ---------------------------------------------------------------------------
  Print the state vector using index
  ---------------------------------------------------------------------------
*/

void
siva_print_index_state_vector(var_set_t *state_vector)
{

  register int i;
  fprintf(siva_output_file,"State vector\n");
 
  for(i=0;i<siva_num_latch;i++){
    if(var_set_get_elt(state_vector,i)){
      fprintf(siva_output_file,"(%d=1) ",
              bck_get_index(siva_latch_nodes[i].latch));
    }else{
      fprintf(siva_output_file,"(%d=0) ",
              bck_get_index(siva_latch_nodes[i].latch));
    }
    /* eol after 8 or last*/
    if ((i && (!(i%8)))||(i == (siva_num_latch-1))){
      fprintf(siva_output_file,"\n");
    }
  }

}


/*
  ---------------------------------------------------------------------------
  Print the state vector using names
  ---------------------------------------------------------------------------
*/
void
siva_print_name_state_vector(var_set_t *state_vector)
{
  register int i;
  
  fprintf(siva_output_file,"State vector\n");
  for(i=0;i<siva_num_latch;i++){
    SIVA_CONV conv;
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,
                           siva_bck_name_aux);
    if(var_set_get_elt(state_vector,i)){
      fprintf(siva_output_file,"(%s=1)\n ",conv.str);
    }else{
      fprintf(siva_output_file,"(%s=0)\n ",conv.str);
    }
  }
}

/*
  ---------------------------------------------------------------------------
  Print the input vector using index
  ---------------------------------------------------------------------------
*/
void
siva_print_index_input_vector(var_set_t* input_vector)
{

  register int i;
  fprintf(siva_output_file,"Input vector\n");
  for(i=0;i<siva_num_primary_input;i++){
    if(var_set_get_elt(input_vector,i)){
      fprintf(siva_output_file,"(%d=1) ",
	      bck_get_index(siva_primary_inputs[i]));
    }else{
      fprintf(siva_output_file,"(%d=0) ",
	      bck_get_index(siva_primary_inputs[i]));
    }
    /* eol after 8 or last*/
    if ((i & (!(i%8)))||(i == (siva_num_primary_input-1))){
      fprintf(siva_output_file,"\n");
    }
  }
}

/*
  ---------------------------------------------------------------------------
  Print the input vector using names
  ---------------------------------------------------------------------------
*/
void
siva_print_name_input_vector(var_set_t* input_vector)
{
  register int i;
  fprintf(siva_output_file,"Input vector\n");
  for(i=0;i<siva_num_primary_input;i++){
    SIVA_CONV conv;
    conv.ptr = bck_get_aux(siva_primary_inputs[i],
                           siva_bck_name_aux);
    if(var_set_get_elt(input_vector,i)){
      fprintf(siva_output_file,"(%s=1)\n ",conv.str);
    }else{
      fprintf(siva_output_file,"(%s=0)\n ",conv.str);
    }
  }
}
/*
  ---------------------------------------------------------------------------
  Reads the file "id-name" in the local directory and then check the
  siva_target_one_flag and siva_target_zero_flag for the node
  corresponding to each complex node.
  Write it out in the file specified.
  ---------------------------------------------------------------------------
*/
static void
siva_write_bck_target_file(void)
{
  char node_name[200];
  char para[30];
  int id;
  int inverted;
  char *ptr;  
  int num_zero_target=0;
  int num_one_target=0;

  while(!feof(siva_map_file)){
    if (fscanf(siva_map_file,"%s %s\n",para,node_name) == EOF){
      return;
    }
    if (!(strcmp(para,"id:")==0 ||
          strcmp(para,"NODE:")==0 ||
          strcmp(para,"inv:")==0 ||
          strcmp(para,"TYPE:")==0 ||
          strcmp(para,"type:")==0 ||
          strcmp(para,"latch_input:")==0 ||
          strcmp(para,"target_one:")==0 ||
          strcmp(para,"target_zero:")==0 ||
          strcmp(para,"left_child:")==0 ||
          strcmp(para,"left_child_inverted:")==0 ||
          strcmp(para,"right_child:")==0 ||
          strcmp(para,"right_child_inverted:")==0)) {
      assert(0);
    }
    SIVA_VERB2_ARG3("%s %s\n",para,node_name);
    if (strcmp(para,"NODE:") == 0){
      fscanf(siva_map_file,"%s %d\n",para,&id);
      SIVA_ASSERT((strcmp(para,"id:")==0),
                  "para should be id:\n");
      if (id && !bck_is_constant(id = bck_get_canonical(id))){
        if (bck_is_flag_set((BCK)id,siva_written_flag)) {
          continue;
        }
        fscanf(siva_map_file,"%s %d\n",para,&inverted);
        SIVA_ASSERT((strcmp(para,"inv:")==0),
                    "para should be id:\n");

        inverted ^= bck_is_inverted(id);
        id = bck_get_regular(id);
        if(bck_is_flag_set(id,siva_target_one_flag)){
          fprintf(siva_pruned_target_file,"%s %d\n0\n",node_name,
                  1^inverted);
          num_one_target++;
          SIVA_VERB2_ARG4("%s(%d) %d\n0\n",node_name,id,
                          1^inverted);        
        }
        if(bck_is_flag_set(id,siva_target_zero_flag)){
          fprintf(siva_pruned_target_file,"%s %d\n0\n",node_name,
                  inverted);
          num_zero_target++;
          SIVA_VERB2_ARG4("%s(%d) %d\n0\n",node_name,id,
                          inverted);
        }
        bck_set_flag((BCK)id,siva_written_flag);
      }      
    }    
  }
  SIVA_ASSERT((num_one_target == siva_num_one_target),
              "# zero target don't match");
  SIVA_ASSERT((num_zero_target == siva_num_zero_target),
              "# one target don't match");
  SIVA_VERB1_ARG2("Written # zero target = %d\n",num_zero_target);
  SIVA_VERB1_ARG2("Written # one target = %d\n",num_one_target);
}

/*
  ---------------------------------------------------------------------------
  Service routine to write the blif name table for each bck node
  ---------------------------------------------------------------------------
*/
static int
siva_write_bck_file_srv(SIVA_NODE node, void *para)
{

  SIVA_NODE left;
  SIVA_NODE right;

  if (!bck_is_and(node)) {
    return 1;
  }
  left = bck_get_left_child(node);
  right = bck_get_right_child(node);
  fprintf(siva_bck_out_file,".names BCK_%d BCK_%d BCK_%d\n%d%d 1\n",
          bck_get_index(left),
          bck_get_index(right),
          bck_get_index(node),
          1^bck_is_inverted(left),
          1^bck_is_inverted(right));
  return 1;
}
