/**CFile**********************************************************************

  FileName                   [siva_mlh.c]

  PackageName                [siva_core]

  Synopsis                   [manual lighthouse]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]
******************************************************************************/
#include "siva.h"

static array_t *siva_mlh_array;
static unsigned siva_mlh_state_heap_id;
static void siva_store_mlh_witness_state(SIVA_STATE *seed_state,
                                         var_set_t *wit_state_vector,
                                         int rel_depth,
                                         int unfold_depth,
                                         void *para);

static void siva_build_mlh_srv(SIVA_COMPLEX_NODE *complex_node);
static void siva_mlh_stat(void);
static void siva_insert_mlh(SIVA_NODE lh_node,
                            SIVA_NODE target,
                            unsigned priority);


/*---------------------------------------------------------------------------*/
void
siva_init_mlh(void)
{
  siva_mlh_array = array_alloc(SIVA_LIGHT_HOUSE,16);
  siva_mlh_state_priority = ALLOC(unsigned, SIVA_SIM_N_WORDS << 5);
  siva_mlh_state_heap_id = siva_alloc_heap(siva_heap_unsigned_min_cmp);
}

/*---------------------------------------------------------------------------*/
void
siva_quit_mlh(void)
{
  siva_release_heap(siva_mlh_state_heap_id);
  FREE(siva_mlh_state_priority);
  array_free(siva_mlh_array);
}


/*---------------------------------------------------------------------------*/
static void
siva_mlh_stat(void)
{
  int i;
  for(i=0;i<array_n(siva_mlh_array);i++){
    SIVA_LIGHT_HOUSE *mlh =
      array_fetch_p(SIVA_LIGHT_HOUSE,siva_mlh_array,i);
    SIVA_VERB1_ARG5("MLH %d: %d value = %d Target = %d\n",
                    i,bck_get_index(mlh->node),
                    bck_is_flag_set(mlh->node,siva_mlh_one_flag),
                    bck_get_index(mlh->target.node));
  }
}

/*---------------------------------------------------------------------------*/
void
siva_build_mlh(void)
{
  int i;
  SIVA_CONV conv;
  siva_do_for_all_complex_node(siva_build_mlh_srv);
  siva_update_mlh();
  for(i=0;i<array_n(siva_mlh_array);i++){
    SIVA_LIGHT_HOUSE *mlh = array_fetch_p(SIVA_LIGHT_HOUSE,
                                          siva_mlh_array,i);
    conv.lh = mlh;
    bck_set_aux(mlh->target.node,siva_lh_aux,conv.ptr);
  }
  siva_mlh_stat();
}
  
/*---------------------------------------------------------------------------*/
static void
siva_build_mlh_srv(SIVA_COMPLEX_NODE *complex_node)
{
  register int i;
  SIVA_COMPLEX_NODE *lh_node;
  SIVA_NODE lh_bck;  
  SIVA_CONV conv;
  int reachable;
  array_t *light_house_array =  complex_node->light_house_array;   
  SIVA_NODE target = siva_get_node(complex_node);
  if (bck_is_null(target) || bck_is_constant(target)) {
    return;
  }
  if (!bck_is_flag_set(target,siva_target_zero_flag) &&
      !bck_is_flag_set(target,siva_target_one_flag)){
    return;
  }
  if (!light_house_array) {
    return;
  }
  for(i=0;i<array_n(light_house_array);i+=2){
    lh_node = array_fetch(SIVA_COMPLEX_NODE *,light_house_array,i);
    lh_bck = siva_get_node(lh_node);
    SIVA_ASSERT(!bck_is_null(lh_bck),"lh can't be null");
    if (bck_is_constant(lh_bck)){
      continue;
    }
    conv.complex_node = array_fetch(SIVA_COMPLEX_NODE *,
                                    light_house_array,i+1);
    if (conv.int_value == 0) {
      lh_bck = bck_create_not(lh_bck);
    }
    siva_insert_mlh(lh_bck,target,
                    (i>>1)+1);
    
  }
}

/*---------------------------------------------------------------------------*/
void
siva_update_mlh(void)
{
  int i;
  /* clear all the priorities */
  memset(siva_mlh_state_priority,0,SIVA_SIM_N_WORDS << 7);

  /* Retain only those mlh whose corresponding target is not
     yet resolved. Takes care of the cases mlh points to 2 or more targets
  */
  for(i=0;i<array_n(siva_mlh_array);i++){
    SIVA_LIGHT_HOUSE *mlh = array_fetch_p(SIVA_LIGHT_HOUSE,
                                          siva_mlh_array,i);
    bck_reset_flag(mlh->node,siva_mlh_one_flag);
    bck_reset_flag(mlh->node,siva_mlh_zero_flag);    
  }
  for(i=0;i<array_n(siva_mlh_array);i++){
    SIVA_LIGHT_HOUSE *mlh = array_fetch_p(SIVA_LIGHT_HOUSE,
                                          siva_mlh_array,i);  
    if (bck_is_flag_set(mlh->target.node,siva_target_one_flag) ||
        bck_is_flag_set(mlh->target.node,siva_target_zero_flag)){
      if (bck_is_inverted(mlh->node)){
        SIVA_ASSERT1(!bck_is_flag_set(mlh->node,siva_mlh_one_flag),
                    "MLH %d is set for both one and zero\n",
                    bck_get_index(mlh->node));
        bck_set_flag(mlh->node,siva_mlh_zero_flag);
      }else {
        SIVA_ASSERT1(!bck_is_flag_set(mlh->node,siva_mlh_zero_flag),
                    "MLH %d is set for both one and zero\n",
                    bck_get_index(mlh->node));
        bck_set_flag(mlh->node,siva_mlh_one_flag);
      }
    }
  }
}

/*---------------------------------------------------------------------------*/
static void
siva_insert_mlh(SIVA_NODE lh_node,
                SIVA_NODE target,
                unsigned priority)
{
  int i;
  SIVA_LIGHT_HOUSE mlh;
  SIVA_CONV conv;
  SIVA_ASSERT(!bck_is_flag_set(target,siva_target_one_flag) ||
              !bck_is_flag_set(target,siva_target_zero_flag),
              "target has both flag set?");
  if (!bck_is_inverted(lh_node)) {
    SIVA_ASSERT1(!bck_is_flag_set(lh_node,siva_mlh_zero_flag),
                 "MLH %d can't be set for both one and zero\n",
                 bck_get_index(lh_node));
    bck_set_flag(lh_node,siva_mlh_one_flag);
  }else {
    SIVA_ASSERT1(!bck_is_flag_set(lh_node,siva_mlh_one_flag),
                 "MLH %d can't be set for both one and zero\n",
                 bck_get_index(lh_node));
    bck_set_flag(lh_node,siva_mlh_zero_flag);
  }  
  mlh.target.node = target;
  SIVA_ASSERT(priority > 0,"MLH priority can't be zero");
  mlh.priority = priority;
  mlh.is_target_et = 0;
  mlh.node = lh_node;
  array_insert_last(SIVA_LIGHT_HOUSE,siva_mlh_array,mlh);
  conv.lh = array_fetch_p(SIVA_LIGHT_HOUSE,siva_mlh_array,
                          array_n(siva_mlh_array)-1);
  bck_set_aux(lh_node,siva_lh_aux,conv.ptr);
}



/*---------------------------------------------------------------------------*/
void
siva_find_mlh_state(array_t *state_wrapper_array, int num_state)
{
  int i;
  int value;
  int num_goal;
  SIVA_CONV conv;
  SIVA_NODE mlh_node;
  if (!array_n(siva_mlh_array)) {
    return;
  }
      
  SIVA_VERB1_ARG1("Finding mlh states ...\n");


  for(i=0;i<array_n(siva_mlh_array); i++){
    SIVA_LIGHT_HOUSE *mlh = array_fetch_p(SIVA_LIGHT_HOUSE,
                                          siva_mlh_array,i);
    /* Do nothing if mlh is not required */
    if (!bck_is_flag_set(mlh->node,siva_mlh_zero_flag) &&
        !bck_is_flag_set(mlh->node,siva_mlh_one_flag)){
      continue;
    }
    /* Do nothing if corresponding target is reached */
    if (!bck_is_flag_set(mlh->target.node,siva_target_zero_flag) &&
        !bck_is_flag_set(mlh->target.node,siva_target_one_flag)){
      continue;
    }
    value = bck_is_flag_set(mlh->node,siva_mlh_one_flag);
    conv.lh = mlh;
    mlh_node = bck_get_regular(mlh->node);
    siva_atpg_solve(&mlh_node,&value,1,
                    state_wrapper_array,num_state,
                    conv.ptr,siva_store_mlh_witness_state,NULL);
  }
  SIVA_VERB1_ARG1("Finding mlh states ...done\n");
}

/*---------------------------------------------------------------------------*/
static void
siva_store_mlh_witness_state(SIVA_STATE *seed_state,
                             var_set_t *wit_state_vector,
                             int rel_depth,
                             int unfold_depth,
                             void *para)

{
  unsigned state_value;
  SIVA_STATE *wit_state;
  int priority_value;  
  int is_created;

  SIVA_CONV conv;

  /* only pick the witness state. ignore the rest of the sequence */
  if (rel_depth < unfold_depth) {
    return;
  }
  var_set_compress(&state_value,wit_state_vector,
                   siva_state_init_hash_value);
  wit_state = siva_get_or_create_state(state_value,
                                       seed_state->state_value,
                                       rel_depth,
                                       wit_state_vector->data,
                                       &is_created);  

  if (is_created == 0) {
    return;
  }
  conv.ptr = para;
  priority_value = conv.lh->priority; 
  siva_put_mlh_state(priority_value,wit_state);
  SIVA_VERB1_ARG4("MLH: Witness State %x, state_value = %x priority = %d\n",
                  wit_state,state_value,priority_value);
}

/*--------------------------------------------------------------------------*/
int
siva_get_seed_mlh_state(SIVA_STATE **state_p,
                        var_set_t *seed_state_vector)
{
  SIVA_KEY key;
  SIVA_CONV conv;
  while(siva_get_mlh_state(&key,&conv)){
    /* regenerate state and check if valid state*/
    if (!siva_regenerate_state_vector(conv.state,
                                      seed_state_vector)){
      continue;
    }
    *state_p = conv.state;
    SIVA_VERB1_ARG3("MLH: Seed state value = %x, key = %u\n",
                    conv.state->state_value,key.unsigned_key);
    return 1;
  }  
  return 0;
}


/*--------------------------------------------------------------------------*/
void
siva_put_mlh_state(unsigned priority_value, SIVA_STATE *state)
{
  SIVA_KEY key;
  SIVA_CONV conv;

  key.unsigned_key = priority_value;
  conv.state = state;
  siva_heap_insert(key,conv,siva_mlh_state_heap_id);

}

/*--------------------------------------------------------------------------*/
int
siva_get_mlh_state(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr)
{
  int found;
  found = siva_heap_extract_top(key_ptr,conv_ptr,siva_mlh_state_heap_id);
  return found;
}
