/**CFile**********************************************************************

  FileName                   [siva_sim.c]

  PackageName                [siva_core]

  Synopsis                   [simulation functions]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]

******************************************************************************/

#include "siva.h"
#include "siva_state_lookup.h"



static unsigned siva_latch_activity_aux;
static unsigned siva_latch_one_aux;
static unsigned siva_latch_zero_aux;

static unsigned siva_num_sim_drag_space;
static unsigned *siva_sim_drag_space;



static int siva_sim_space_alloc_srv(SIVA_NODE node, void *para);
static int siva_sim_space_assign_srv(SIVA_NODE node, void *para);
static void siva_sim_alloc_drag_space(void);
static void siva_sim_release_drag_space(void);
static int siva_sim_space_flag_clear_srv(SIVA_NODE in, void *para);
static void siva_evaluate(SIVA_NODE node);

static void siva_check_target_reached(SIVA_NODE node);
static void siva_check_mlh_reached(SIVA_NODE node);
static void siva_check_alh_reached(SIVA_NODE node);
static void siva_build_next_state_vector(int word_index, int bit_index,
                                         var_set_t *state_vector);



static var_set_t *siva_activity_vector;
static void siva_store_sim_witness_state(int word_index, int bit_index);
static void siva_apply_reset_vector(void);
static int siva_set_coi_srv(SIVA_NODE in, void *para);
static int siva_count_coi_srv(SIVA_NODE in, void *para);
static void siva_check_enlarged_target_reached(SIVA_NODE node);
static int siva_latch_activity_cmp(const void *p1, const void *p2);


static int siva_evaluate_node1(SIVA_NODE nodes,
                               unsigned visited_flag,
                               unsigned value_flag);
static void
siva_sim_verify_witness(unsigned word_index,
                        unsigned bit_index,
                        SIVA_NODE target,
                        int goal);
static void siva_sim_print_witness(int wi, int bi, int offset);
static SIVA_STATE* siva_sim_build_witness_state(int wi, int bi);
static void siva_normalize_latch_activity(void);
static void siva_update_latch_activity(void);

/*
  ---------------------------------------------------------------------------
  Service routine for allocating space.
  ---------------------------------------------------------------------------
*/
static int
siva_sim_space_alloc_srv(SIVA_NODE node, void *para)
{
  SIVA_CONV conv;
  SIVA_NODE left_child;
  SIVA_NODE right_child;

  if (bck_compare_ids(node,siva_last_network_node) > 0) {
    return 0;
  }
  if (!bck_is_flag_set(node,siva_coi_flag)){
    return 1;
  }
  
  if (bck_is_var(node)){
    /* this space should be virgin */
    conv.value = siva_num_sim_drag_space++;
    bck_set_aux(node,siva_sim_space_aux,conv.ptr);    
    return 1;
  }

  /* check if sim is required at the node */
  if (bck_is_flag_set(node,siva_sim_space_flag)){
    if (!siva_stack_empty()) {
      conv.ptr = siva_stack_pop();    
    } else {
      conv.value = siva_num_sim_drag_space++  ;
    }
  }else {
    conv.value = 1;
  }
  bck_set_aux(node,siva_sim_space_aux,conv.ptr);    
  
  /* set the output reference count */
  conv.value = bck_get_num_outputs(node);
  bck_set_aux(node,siva_global_aux,conv.ptr);

  /* left child space can be resued only if it is not var 
       and not a next state node 
  */
  left_child = bck_get_left_child(node);

  if ( !bck_is_null(left_child) && !bck_is_var(left_child) && 
       !bck_is_flag_set(left_child,siva_latch_input_flag)){

    conv.ptr = bck_get_aux(left_child,siva_global_aux);
    conv.value--;
    bck_set_aux(left_child,siva_global_aux,conv.ptr);
    if ( conv.value == 0 ) {
      conv.ptr = bck_get_aux(left_child,siva_sim_space_aux);
      siva_stack_push(conv.ptr);
    }else {
      bck_set_aux(left_child,siva_global_aux,conv.ptr); 
    }
  }

  right_child = bck_get_right_child(node);
  
  if ( !bck_is_null(right_child) && !bck_is_var(right_child) &&
       !bck_is_flag_set(right_child,siva_latch_input_flag)){

    conv.ptr = bck_get_aux(right_child,siva_global_aux);
    conv.value--;
    bck_set_aux(right_child,siva_global_aux,conv.ptr);
    if ( conv.value == 0 ) {
      conv.ptr = bck_get_aux(right_child,siva_sim_space_aux);
      siva_stack_push(conv.ptr);
    }else {
      bck_set_aux(right_child,siva_global_aux,conv.ptr); 
    }
  }
  return 1;
}

/*
  ---------------------------------------------------------------------------
  Assign the space pointers to each node. Space can be resued
  as defined by "Space dragging".
  ---------------------------------------------------------------------------
*/
static int
siva_sim_space_assign_srv(SIVA_NODE node, void *para)
{
  SIVA_CONV conv;
  if (bck_compare_ids(node,siva_last_network_node) > 0) {
    return 0;
  }

  if (!bck_is_flag_set(node,siva_coi_flag)){
    return 1;
  }

  conv.ptr = bck_get_aux(node,siva_sim_space_aux);
  
  SIVA_VERB2_ARG3("bck = %d, space = %d\n",
                  bck_get_index(node),conv.value);
  
  conv.ptr = &(siva_sim_drag_space[conv.value * SIVA_SIM_N_WORDS]);
  bck_set_aux(node,siva_sim_space_aux,conv.ptr);
  

  if (bck_is_and(node)){

    conv.ptr = bck_get_aux(bck_get_left_child(node),siva_sim_space_aux);
    if ( bck_is_inverted(bck_get_left_child(node))){
      conv.value |= 01;
      bck_set_aux(node,siva_sim_lc_space_aux,conv.ptr); 
    } else {
      bck_set_aux(node,siva_sim_lc_space_aux,conv.ptr); 
    }

    conv.ptr = bck_get_aux(bck_get_right_child(node),siva_sim_space_aux);

    if ( bck_is_inverted(bck_get_right_child(node))){
      conv.value |= 01;
      bck_set_aux(node,siva_sim_rc_space_aux,conv.ptr); 
    } else {
      bck_set_aux(node,siva_sim_rc_space_aux,conv.ptr); 
    }
    
  }

  return 1;
}
/*---------------------------------------------------------------------------*/
void 
siva_init_sim(void)
{
  register int i;

  /* Identify nodes that should be simulated */  
  bck_do_in_topological_order(NULL,siva_set_coi_srv);
  
  siva_sim_space_aux = bck_alloc_aux(siva_init_aux);
  siva_sim_lc_space_aux = bck_alloc_aux(siva_init_aux);
  siva_sim_rc_space_aux = bck_alloc_aux(siva_init_aux);
  siva_num_sim_drag_space = 0;
  siva_sim_drag_space = NULL;
  bck_set_all_flag(siva_sim_space_flag);
  bck_do_in_topological_order(NULL,siva_sim_space_flag_clear_srv);
  siva_sim_alloc_drag_space();
  /* Note; siva_sim_value_flag for bck_zero is cleared
     at the time of allocation */
}
/*---------------------------------------------------------------------------*/
static int
siva_sim_space_flag_clear_srv(SIVA_NODE in, void *para)
{
  /* some var may not be any node's support; but to keep
     less complicated they are retained  
  */
  if (bck_compare_ids(in,siva_last_network_node) > 0) {
    return 0;
  }

  if (bck_is_var(in)) return 1;
  if (bck_is_canonical(in) &&
      bck_is_flag_set(in,siva_latch_input_flag)) return 1;

  if (!bck_is_flag_set(in,siva_coi_flag)){
    bck_reset_flag(in,siva_sim_space_flag);
  }
  
  if (bck_get_num_outputs(in) == 0){
    bck_reset_flag(in,siva_sim_space_flag);
  }
  return 1;
  
}
/*---------------------------------------------------------------------------*/
void 
siva_sim_stat(void)
{
  int num_node=0;
  fprintf(siva_output_file,"# drag space blocks = %d\n",
          siva_num_sim_drag_space);
  //fprintf(siva_output_file,"number of latches = %d\n",siva_num_latch);
  //fprintf(siva_output_file,"number of pi = %d\n",siva_num_primary_input);
  bck_do_in_topological_order((void *)&num_node,siva_count_coi_srv);
  fprintf(siva_output_file,"# fanin cone nodes = %d\n",num_node);
  //bck_stat(siva_output_file);  
}

/*---------------------------------------------------------------------------*/
void 
siva_quit_sim(void)
{  
  siva_sim_release_drag_space();
  siva_num_sim_drag_space = 0;
  siva_sim_drag_space = NULL;
  bck_release_aux(siva_sim_space_aux,NULL);
  bck_release_aux(siva_sim_lc_space_aux,NULL);
  bck_release_aux(siva_sim_rc_space_aux,NULL);
  bck_reset_all_flag(siva_coi_flag);
}


/*
  ---------------------------------------------------------------------------
  "Space Dragging" to allocate & assign space to bck node.
  ---------------------------------------------------------------------------
*/  
static void 
siva_sim_alloc_drag_space(void)
{
  
  SIVA_CONV conv;
  register int i;
  siva_global_aux = bck_alloc_aux(NULL); /* for reference count */
  /* for nodes which is not input to any node */
  siva_num_sim_drag_space = 2;
  
  siva_init_stack(); 
  bck_do_in_topological_order(NULL,siva_sim_space_alloc_srv);
  siva_quit_stack();
  siva_sim_drag_space = ALLOC(unsigned,sizeof(unsigned) * 
                              siva_num_sim_drag_space * SIVA_SIM_N_WORDS);
  conv.value_ptr = &(siva_sim_drag_space[0]); 
  bck_set_aux(bck_create_zero(),siva_sim_space_aux,conv.ptr);  
  for(i=0;i<SIVA_SIM_N_WORDS;i++){
   *conv.value_ptr++ = 0;
  }
  bck_do_in_topological_order(NULL,siva_sim_space_assign_srv);  
  bck_release_aux(siva_global_aux,NULL);
}

/*---------------------------------------------------------------------------*/
static void 
siva_sim_release_drag_space(void)
{
  SIVA_CONV conv;

  conv.value = siva_sim_space_aux;
  bck_do_in_topological_order(conv.ptr,siva_clear_aux_srv);
  conv.value = siva_sim_lc_space_aux;
  bck_do_in_topological_order(conv.ptr,siva_clear_aux_srv);
  conv.value = siva_sim_rc_space_aux;
  bck_do_in_topological_order(conv.ptr,siva_clear_aux_srv);
  FREE(siva_sim_drag_space);
  siva_num_sim_drag_space = 0;
}    

/*--------------------------------------------------------------------------*/
static void
siva_evaluate(SIVA_NODE node)
{
  SIVA_CONV left_child_cv;
  SIVA_CONV right_child_cv;
  SIVA_CONV node_cv;
  register int i;
  register unsigned index=0;
  /*
  if (!bck_is_and(node)){
    node_cv.ptr = bck_get_aux(node,siva_sim_space_aux);
    fprintf(siva_output_file,"%d = %x\n",
	    bck_get_index(node),*node_cv.value_ptr);
  }
  */
  if (!bck_is_and(node))                                      return;

  left_child_cv.ptr = bck_get_aux(node,siva_sim_lc_space_aux);
  index = (left_child_cv.value & 01)? 1 : 0;    
  left_child_cv.value &= ~01;

  right_child_cv.ptr = bck_get_aux(node,siva_sim_rc_space_aux);
  index |= (right_child_cv.value & 01)? 2 : 0;    
  right_child_cv.value &= ~01;

  node_cv.ptr = bck_get_aux(node,siva_sim_space_aux);
  
  switch(index){
  case 0:
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      (*node_cv.value_ptr++) = (*right_child_cv.value_ptr++) & 
	(*left_child_cv.value_ptr++);
    }
    break;

  case 1:
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      (*node_cv.value_ptr++) = (*right_child_cv.value_ptr++) & 
	~(*left_child_cv.value_ptr++);
    }
    break;
  case 2:
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      (*node_cv.value_ptr++) = ~(*right_child_cv.value_ptr++) & 
	(*left_child_cv.value_ptr++);
    }
    break;
  case 3:
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      (*node_cv.value_ptr++) = ~(*right_child_cv.value_ptr++) & 
	~(*left_child_cv.value_ptr++);
    }
    break;
  }
#if 0
  node_cv.ptr = bck_get_aux(node,siva_sim_space_aux);
  fprintf(siva_output_file,"N:%d = %x ",
	  bck_get_index(node),*node_cv.value_ptr);
  left_child_cv.ptr = bck_get_aux(node,siva_sim_lc_space_aux);
  left_child_cv.value &= ~01;
  if (index & 01){
    fprintf(siva_output_file,"L:%d = %x ",
	    bck_get_index(bck_get_left_child(node)),
	    ~(*left_child_cv.value_ptr));
  }else{
    fprintf(siva_output_file,"L:%d = %x ",
	    bck_get_index(bck_get_left_child(node)),
	    (*left_child_cv.value_ptr));
  }
  right_child_cv.ptr = bck_get_aux(node,siva_sim_rc_space_aux);
  right_child_cv.value &= ~01;
  if (index & 02){
    fprintf(siva_output_file,"R:%d = %x\n",
	    bck_get_index(bck_get_right_child(node)),
	    ~(*right_child_cv.value_ptr));
  }else{
    fprintf(siva_output_file,"R:%d = %x\n",
	    bck_get_index(bck_get_right_child(node)),
	    (*right_child_cv.value_ptr));
  }
#endif

}

/*---------------------------------------------------------------------------*/
static SIVA_STATE*
siva_sim_build_witness_state(int wi, int bi)
{

  var_set_t local_state_vector;
  var_set_t local_input_vector;
  unsigned state_data[siva_latch_word_size];
  unsigned pi_data[siva_primary_input_word_size];
  SIVA_STATE *wit_state;
  int is_created;
  unsigned state_value;
  
  siva_set_state_vector(&local_state_vector,state_data);
  siva_set_state_vector(&local_input_vector,pi_data);
  var_set_clear(&local_state_vector);
  var_set_clear(&local_input_vector);
  siva_build_witness_state(wi,bi,&local_state_vector);
  siva_build_witness_input(wi,bi,&local_input_vector);
  var_set_compress(&state_value,&local_state_vector,
                   siva_state_init_hash_value);  
  wit_state =
    siva_get_or_create_state(state_value,
                             siva_last_stored_states[(wi<<5)+bi]->state_value,
                             (siva_current_sim_depth - 
                              siva_last_stored_state_depths[(wi<<5)+bi]),
                             state_data,
                             &is_created);
  return wit_state;
}


/*---------------------------------------------------------------------------*/
static void
siva_sim_print_witness(int wi, int bi, int offset)
{
  var_set_t local_state_vector;
  var_set_t local_input_vector;
  unsigned state_data[siva_latch_word_size];
  unsigned pi_data[siva_primary_input_word_size];
  var_set_t *witness_state_vectors[1];
  var_set_t *witness_input_vectors[1];
    
  if (!array_n(siva_debug_node_array)){
    return;
  }
  siva_set_state_vector(&local_state_vector,state_data);
  siva_set_state_vector(&local_input_vector,pi_data);
  var_set_clear(&local_state_vector);
  var_set_clear(&local_input_vector);
  witness_state_vectors[0] = &local_state_vector;
  witness_input_vectors[0] = &local_input_vector;
  siva_build_witness_state(wi,bi,&local_state_vector);
  siva_build_witness_input(wi,bi,&local_input_vector);
  siva_print_witness_trace(witness_input_vectors,
                           witness_state_vectors,
                           offset,0);
}

/*---------------------------------------------------------------------------*/
static void
siva_sim_verify_witness(unsigned word_index,
                        unsigned bit_index,
                        SIVA_NODE target,
                        int goal)
{

  var_set_t local_state_vector;
  var_set_t local_input_vector;
  unsigned state_data[siva_latch_word_size];
  unsigned pi_data[siva_primary_input_word_size];
  int value;


  siva_set_state_vector(&local_state_vector,state_data);
  siva_set_state_vector(&local_input_vector,pi_data);

  var_set_clear(&local_state_vector);
  var_set_clear(&local_input_vector);
  
  SIVA_VERB1_ARG3("SIM: Verifying the witness for %d=%d ..\n",
                      target,goal);
  SIVA_ASSERT(!bck_is_inverted(target),
              "Target should not be inverted");
  
  siva_build_witness_state(word_index,bit_index,&local_state_vector);
  siva_build_witness_input(word_index,bit_index,&local_input_vector);

  siva_evaluate_node(&target,&value,1,&local_input_vector,
                     &local_state_vector);
  SIVA_ASSERT(goal == value,
              "Target verification failed");
  SIVA_VERB1_ARG3("Succeeded %d = %d\n",
                  target,goal);
  
}

/*--------------------------------------------------------------------------*/
static void
siva_check_pre_process_target_reached(SIVA_NODE node)
{
  if (bck_is_flag_set(node,siva_target_zero_flag)){
    SIVA_CONV node_conv;
    register int i;
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);    
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      if ( (~(*node_conv.value_ptr))){        
        SIVA_VERB1_ARG3("CYCLE-SIM (depth=%d): Reached Target (%d=0)\n",
                        siva_current_sim_depth,bck_get_index(node));
        bck_reset_flag(node,siva_target_zero_flag);
        siva_num_zero_target--;
        break;
      }
      node_conv.value_ptr++;
    }    
  }
  if (bck_is_flag_set(node,siva_target_one_flag)){
    SIVA_CONV node_conv;
    register int i;
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);    
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      if (*node_conv.value_ptr){          
        SIVA_VERB1_ARG3("CYCLE-SIM (depth=%d): Reached Target (%d=1)\n",
                        siva_current_sim_depth,bck_get_index(node));
        bck_reset_flag(node,siva_target_one_flag);
        siva_num_one_target--;        
        break;
      }
      node_conv.value_ptr++;
    }
  }

}

/*--------------------------------------------------------------------------*/
static void
siva_check_target_reached(SIVA_NODE node)
{
  int witness_word_index = -1;
  
  if (bck_is_flag_set(node,siva_target_zero_flag)){
    SIVA_CONV node_conv;
    register int i;
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);    
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      if ((~(*node_conv.value_ptr))){
        siva_wit_state_pos[i] |= (~(*node_conv.value_ptr));
        witness_word_index = i;
        break;
      }
      node_conv.value_ptr++;      
    }
  }

  if (bck_is_flag_set(node,siva_target_one_flag)){
    SIVA_CONV node_conv;
    register int i;
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);    
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      if (*node_conv.value_ptr){
        siva_wit_state_pos[i] |= ((*node_conv.value_ptr));
        witness_word_index = i;
        break;
      }
      node_conv.value_ptr++;      
    }
  }

  if (witness_word_index == -1) {
    return;
  }
    
  SIVA_VERB1_ARG4("CYCLE-SIM (Depth=%d): Reached Target (%d=%d)\n",
                  siva_current_sim_depth,bck_get_index(node),
                  bck_is_flag_set(node,siva_target_one_flag));
  if (siva_error_check_level >= 1){  
    int j;
    int obj = bck_is_flag_set(node,siva_target_one_flag);
    SIVA_CONV node_conv;
    unsigned witness_word;
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);
    node_conv.value_ptr += witness_word_index;
    witness_word =
      ((obj)? *node_conv.value_ptr : ~(*node_conv.value_ptr));

    for(j = 0; j < 32 ; j++){
      if (witness_word & (1<<j)){
        siva_sim_verify_witness(witness_word_index,j,
                                node,obj);
        break;
      }
    }
  }
  

  if (array_n(siva_debug_node_array)){  
    int j;
    int obj = bck_is_flag_set(node,siva_target_one_flag);
    SIVA_CONV node_conv;
    unsigned witness_word;
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);
    node_conv.value_ptr += witness_word_index;
    witness_word =
      ((obj)? *node_conv.value_ptr : ~(*node_conv.value_ptr));

    for(j = 0; j < 32 ; j++){
      if (witness_word & (1<<j)){
        SIVA_STATE* wit_state;
        int num_prev_state;
        wit_state = 
          siva_sim_build_witness_state(witness_word_index,j);
        num_prev_state = siva_print_state_trace(wit_state);
        siva_sim_print_witness(witness_word_index,j,
                               num_prev_state);
        break;
      }
    }
  }
  

  if (bck_is_flag_set(node,siva_target_one_flag)){
    bck_reset_flag(node,siva_target_one_flag);    
    siva_num_one_target--;
  }else if (bck_is_flag_set(node,siva_target_zero_flag)) {
    bck_reset_flag(node,siva_target_zero_flag);    
    siva_num_zero_target--;
  }

}


/*--------------------------------------------------------------------------*/
static void
siva_check_enlarged_target_reached(SIVA_NODE node)
{
  int witness_word_index = -1;
  
  if (bck_is_flag_set(node,siva_enlarged_target_zero_flag)){
    SIVA_CONV node_conv;
    register int i;
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);    
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      if ((~(*node_conv.value_ptr))){
        siva_wit_state_pos[i] |= (~(*node_conv.value_ptr));

        witness_word_index = i;
      }
      node_conv.value_ptr++;      
    }
  }

  if (bck_is_flag_set(node,siva_enlarged_target_one_flag)){
    SIVA_CONV node_conv;
    register int i;
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);    
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      if (*node_conv.value_ptr){
        siva_wit_state_pos[i] |= ((*node_conv.value_ptr));

        witness_word_index = i;
      }
      node_conv.value_ptr++;      
    }
  }

  if (witness_word_index == -1) {
    return;
  }
  

  SIVA_VERB1_ARG4("CYCLE-SIM (Depth=%d): Reached Enlarged Target (%d=%d)\n",
                  siva_current_sim_depth,bck_get_index(node),
                  bck_is_flag_set(node,siva_enlarged_target_one_flag));

  if (siva_error_check_level >= 1){
    int j;
    int obj = bck_is_flag_set(node,siva_enlarged_target_one_flag);
    SIVA_CONV node_conv;
    unsigned witness_word;
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);
    node_conv.value_ptr += witness_word_index;
    witness_word =
      ((obj)? *node_conv.value_ptr : ~(*node_conv.value_ptr));

    for(j = 0; j < 32 ; j++){
      if (witness_word & (1<<j)){
        siva_sim_verify_witness(witness_word_index,j,
                                node,obj);
        break;
      }
    }
  }
  
  if (array_n(siva_debug_node_array)){
    int j;
    int obj = bck_is_flag_set(node,siva_target_one_flag);
    SIVA_CONV node_conv;
    unsigned witness_word;
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);
    node_conv.value_ptr += witness_word_index;
    witness_word =
      ((obj)? *node_conv.value_ptr : ~(*node_conv.value_ptr));
    for(j = 0; j < 32 ; j++){
      if (witness_word & (1<<j)){
        SIVA_STATE* wit_state =
          siva_sim_build_witness_state(witness_word_index,j);
        int num_prev_state = siva_print_state_trace(wit_state);
        siva_sim_print_witness(witness_word_index,j,
                               num_prev_state);
        break;
      }
    }
  }
    


  bck_reset_flag(node,siva_enlarged_target_one_flag);
  bck_reset_flag(node,siva_enlarged_target_zero_flag);

  

}


/*--------------------------------------------------------------------------*/
int
siva_evaluate_pre_process_srv(SIVA_NODE node, void *para)
{
  if (bck_compare_ids(node,siva_last_network_node) > 0) {
    return 0;
  }  
  if (!bck_is_flag_set(node,siva_coi_flag)){
    return 1;
  }
  siva_evaluate(node);
  siva_check_pre_process_target_reached(node);
  
  return 1;
}

/*--------------------------------------------------------------------------*/
int
siva_evaluate_latch_activity_srv(SIVA_NODE node, void *para)
{
  if (bck_compare_ids(node,siva_last_network_node) > 0) {
    return 0;
  }
  
  if (!bck_is_flag_set(node,siva_coi_flag)){
    return 1;
  }
  siva_evaluate(node);
  return 1;
}

/*--------------------------------------------------------------------------*/
int
siva_evaluate_solve_srv(SIVA_NODE node, void *para)
{
  if (bck_compare_ids(node,siva_last_network_node) > 0) {
    return 0;
  }

  if (!bck_is_flag_set(node,siva_coi_flag)){
    return 1;
  }
  siva_evaluate(node);

  if (siva_build_pi_free_bck_network_done){
    siva_check_enlarged_target_reached(node);
  }else {
    siva_check_target_reached(node);
  }
  
  siva_check_alh_reached(node);
  siva_check_mlh_reached(node);

  return 1;
}

/*--------------------------------------------------------------------------*/
static void
siva_check_alh_reached(SIVA_NODE node)
{
  int found = 0;
  int k;

  if (bck_is_flag_set(node,siva_alh_zero_flag)){
    SIVA_CONV node_conv;
    register int i;
    SIVA_CONV conv;
    unsigned word;
    conv.ptr = bck_get_aux(node,siva_lh_aux);
    if (siva_is_alh_redundant(conv.lh)) {
      return;
    }
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);    
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      word =  (~(*node_conv.value_ptr));

      for(k=0;k<32;k++){
        if (word & (1<<k)) {
          siva_alh_state_priority[(i<<5)+k] = conv.lh->priority;
          found = 1;
          break;
        }
      }
    }
    node_conv.value_ptr++;   
  }
  if (bck_is_flag_set(node,siva_alh_one_flag)){
    SIVA_CONV node_conv;
    register int i;
    SIVA_CONV conv;
    unsigned word;
    conv.ptr = bck_get_aux(node,siva_lh_aux);
    if (siva_is_alh_redundant(conv.lh)) {
      return;
    }    
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);    
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      word =  ((*node_conv.value_ptr));

      for(k=0;k<32;k++){
        if (word & (1<<k)) {
          siva_alh_state_priority[(i<<5)+k] = conv.lh->priority;
          found = 1;
          break;
        }
      }
    }
    node_conv.value_ptr++;   
  }
  
  if (!found) {
    return;
  }

  bck_reset_flag(node,siva_alh_one_flag);
  bck_reset_flag(node,siva_alh_zero_flag);

  
}


/*--------------------------------------------------------------------------*/
static void
siva_check_mlh_reached(SIVA_NODE node)
{
  if (bck_is_flag_set(node,siva_mlh_zero_flag)){
    SIVA_CONV node_conv;
    register int i,k;
    SIVA_CONV conv;
    unsigned word;

    conv.ptr = bck_get_aux(node,siva_lh_aux);
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);    
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      word =  (~(*node_conv.value_ptr));

      for(k=0;k<32;k++){
        if (word & (1<<k)) {
          siva_mlh_state_priority[(i<<5)+k] = conv.lh->priority;          
        }
      }
    }
    node_conv.value_ptr++;   
  }
  if (bck_is_flag_set(node,siva_mlh_one_flag)){
    SIVA_CONV node_conv;
    register int i,k;
    unsigned word;
    SIVA_CONV conv;
    
    conv.ptr = bck_get_aux(node,siva_lh_aux);
    node_conv.ptr =  bck_get_aux(node,siva_sim_space_aux);    
    for(i = 0; i < SIVA_SIM_N_WORDS; i++){
      word =  ((*node_conv.value_ptr));

      for(k=0;k<32;k++){
        if (word & (1<<k)) {
          siva_mlh_state_priority[(i<<5)+k] = conv.lh->priority;
        }
      }
    }
    node_conv.value_ptr++;   
  }
}

/*--------------------------------------------------------------------------*/
void
siva_generate_input(void)
{
  register int i,j;
  SIVA_CONV conv[siva_num_primary_input];


  for(i=0;i<siva_num_primary_input;i++){
    conv[i].ptr = bck_get_aux(siva_primary_inputs[i],siva_sim_space_aux);
  }
  
  for(j=0;j<SIVA_SIM_N_WORDS;j++){
    siva_rand_seed = SIVA_SEED(siva_rand_seed++);
    srand(siva_rand_seed);      
    for(i=0;i<siva_num_primary_input;i++){
      *conv[i].value_ptr++ = (unsigned) rand();
    }
  }
  
  
}


/*--------------------------------------------------------------------------*/
void
siva_update_current_with_next_state(void)
{
  register int i,j;
  SIVA_CONV conv1;
  SIVA_CONV conv2;
  for(i=0;i<siva_num_latch;i++){
    conv1.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_sim_space_aux);
    conv2.ptr = bck_get_aux(siva_latch_nodes[i].input,siva_sim_space_aux);
    if (bck_is_inverted(siva_latch_nodes[i].input)){
      for (j=0;j<SIVA_SIM_N_WORDS;j++){
	*conv1.value_ptr++ = ~(*conv2.value_ptr++);
      }
    }else {
      for (j=0;j<SIVA_SIM_N_WORDS;j++){
	*conv1.value_ptr++ = *conv2.value_ptr++;
      }
    }
  }  
}

/*---------------------------------------------------------------------------*/
int 
siva_evaluate_sym_node(SIVA_NODE sym_node,
                       var_set_t **input_vectors,
                       SIVA_NODE **input_node_vectors,
                       int sym_depth)
{
  int i,j;
  int value;
  for(i=0;i<=sym_depth;i++){
    SIVA_NODE* input_nodes = input_node_vectors[i];
    for(j=0;j<siva_num_primary_input;j++){
      if (var_set_get_elt(input_vectors[sym_depth-i],j)){
        bck_set_flag(input_nodes[j],siva_sim_value_flag);
      }else {
        bck_reset_flag(input_nodes[j],siva_sim_value_flag);
      }
    }
  }
  value = siva_evaluate_node1(sym_node,
                              siva_bck_visited_flag,
                              siva_sim_value_flag);
  bck_reset_fanin_cone_flag(sym_node,siva_bck_visited_flag);
  return value;
}


/*---------------------------------------------------------------------------*/
void
siva_evaluate_node(SIVA_NODE *nodes,
                   int *node_values,
                   int num_node,
                   var_set_t *input_vector,
                   var_set_t *state_vector)
{
  int i;  
  if (input_vector) {
    siva_valuate_primary_input(input_vector);
  }
  if (state_vector) {
    siva_valuate_latch(state_vector);
  }  
  for(i=0;i<num_node;i++){
    node_values[i] = siva_evaluate_node1(nodes[i],
                                         siva_bck_visited_flag,
                                         siva_sim_value_flag);
  }
  for(i=0;i<num_node;i++){
    bck_reset_fanin_cone_flag(nodes[i],siva_bck_visited_flag);
  }  
}

/*---------------------------------------------------------------------------*/
static int
siva_evaluate_node1(SIVA_NODE node,
                    unsigned visited_flag,
                    unsigned value_flag)
{
  SIVA_CONV conv;
  int value;

  if (bck_is_zero(node)) return 0;
  if (bck_is_one(node))  return 1;
  if (bck_is_flag_set(node,visited_flag)){
    return bck_is_inverted(node) ^
      bck_is_flag_set(node,value_flag);
  }
  bck_set_flag(node,visited_flag);
  if (bck_is_and(node)){
    value =
      siva_evaluate_node1(bck_get_left_child(node),
                          visited_flag,value_flag)
      & siva_evaluate_node1(bck_get_right_child(node), 
                           visited_flag,value_flag);
    
    (value) ?
      bck_set_flag(node,value_flag):
      bck_reset_flag(node,value_flag);    
  }
  return bck_is_inverted(node) ^
    bck_is_flag_set(node,value_flag);
}

/*--------------------------------------------------------------------------*/
void
siva_update_current_with_seed_state(var_set_t *state_vector)
{
  int i;
  SIVA_CONV conv;
  int byte;
  for(i=0;i<siva_num_latch;i++){
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_sim_space_aux);
    byte = (var_set_get_elt(state_vector,i)) ? 0xff : 0;
    memset(conv.value_ptr,byte,(SIVA_SIM_N_WORDS)<<2);
  }
  
}

/*--------------------------------------------------------------------------*/
void
siva_update_current_with_seed_state_array(array_t *seed_state_wrapper_array,
                                          int num_seed_state)
{
  register int i,j,k;
  SIVA_CONV conv;
  int word_per_seed;
  int extra_word;
  var_set_t *vectors[num_seed_state];
  SIVA_STATE *states[num_seed_state];
  SIVA_STATE **last_stored_states;
  
  SIVA_ASSERT(num_seed_state,"Number of seed state can't be 0");
  
  word_per_seed = SIVA_SIM_N_WORDS/num_seed_state;
  extra_word = SIVA_SIM_N_WORDS % num_seed_state;


  for(k=0;k<num_seed_state;k++) {
    SIVA_STATE_WRAPPER *state_wrapper =
      array_fetch_p(SIVA_STATE_WRAPPER,seed_state_wrapper_array,k);
    vectors[k] = state_wrapper->state_vector;
    states[k] = state_wrapper->state;
  }

  /* Initialize the last stored states and depths */
  memset(siva_last_stored_state_depths,0,(SIVA_SIM_N_WORDS<<7));
  last_stored_states = siva_last_stored_states;
  for(k=0;k<num_seed_state;k++) {
    int num_word = word_per_seed;
    if (k==num_seed_state-1) {
      num_word += extra_word;
    }
    for(i=0;i<num_word<<5;i++){
      last_stored_states[i] = states[k];
    }
    last_stored_states += (num_word<<5);
  }
    
  /* set the current state of the latches */  
  for(i=0;i<siva_num_latch;i++){
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_sim_space_aux);
    for(k=0;k<num_seed_state;k++){
      int byte = (var_set_get_elt(vectors[k],i)) ? 0xff : 0;
      if (k==num_seed_state-1) {
        memset(conv.value_ptr,byte,(extra_word + word_per_seed)<<2);
      }else {
        memset(conv.value_ptr,byte,word_per_seed<<2);      
        conv.value_ptr += word_per_seed;
      }
    }
  }
}



/*--------------------------------------------------------------------------*/
int 
siva_evaluate_bit(SIVA_NODE node,void *para)
{
  if (!bck_is_and(node))                       return 1;
  
  if ((bck_is_inverted(bck_get_left_child(node)) ^
       bck_is_flag_set(bck_get_left_child(node),siva_sim_value_flag)) &
      (bck_is_inverted(bck_get_right_child(node)) ^
       bck_is_flag_set(bck_get_right_child(node),siva_sim_value_flag))){
    bck_set_flag(node,siva_sim_value_flag);
  }else{
    bck_reset_flag(node,siva_sim_value_flag);
  }
  return 1;
}

/*--------------------------------------------------------------------------*/
void
siva_valuate_primary_input(var_set_t *input_vector)
{
  register int i;

  for(i=0;i<siva_num_primary_input;i++){
    if (var_set_get_elt(input_vector,i)){
      bck_set_flag(siva_primary_inputs[i],siva_sim_value_flag);
    }else {
      bck_reset_flag(siva_primary_inputs[i],siva_sim_value_flag);
    }
  }
}

/*
  --------------------------------------------------------------------------
  state_vector to latch
  --------------------------------------------------------------------------
*/
  
void
siva_valuate_latch(var_set_t *state_vector)
{
  register int i;

  for(i=0;i<siva_num_latch;i++){
    if (var_set_get_elt(state_vector,i)){
      bck_set_flag(siva_latch_nodes[i].latch,siva_sim_value_flag);
    }else {
      bck_reset_flag(siva_latch_nodes[i].latch,siva_sim_value_flag);
    }
  }
}
/*
  --------------------------------------------------------------------------
  latch input to state vector
  --------------------------------------------------------------------------
*/
void
siva_fill_state_vector(var_set_t *state_vector)
{
  register int i;
  for(i=0;i<siva_num_latch;i++){
    SIVA_ASSERT(bck_is_canonical(siva_latch_nodes[i].input),
                "latch input is pass thru");
    if (bck_is_flag_set(siva_latch_nodes[i].input,siva_sim_value_flag) ^
	bck_is_inverted(siva_latch_nodes[i].input)){
      var_set_set_elt(state_vector,i);
    }else {
      var_set_clear_elt(state_vector,i);
    }
  }
}


/*
  --------------------------------------------------------------------------
  latch input to latch
  --------------------------------------------------------------------------
*/
   
void
siva_update_latch_with_input(void)
{
  register int i;

  for(i=0;i<siva_num_latch;i++){
    if (bck_is_flag_set(siva_latch_nodes[i].input,siva_sim_value_flag) ^
	bck_is_inverted(siva_latch_nodes[i].input)){
      bck_set_flag(siva_latch_nodes[i].latch,siva_sim_value_flag);
    }else {
      bck_reset_flag(siva_latch_nodes[i].latch,siva_sim_value_flag);
    }
  }
}


/*--------------------------------------------------------------------------*/
static void
siva_build_next_state_vector(int word_index, int bit_index,
                             var_set_t *state_vector)
{
  register int i;
  SIVA_CONV next_state_conv;
  unsigned value;
  unsigned *ptr;
  for(i=0;i<siva_num_latch;i++){
    next_state_conv.ptr = 
      bck_get_aux(siva_latch_nodes[i].input,siva_sim_space_aux);
    ptr = next_state_conv.value_ptr + word_index;
    value =  bck_is_inverted(siva_latch_nodes[i].input) ?
      (*ptr ^ ~0) : *ptr;
    //fprintf(siva_output_file,"%d: Next state %x\n",i,value);
    if (value & (1<<bit_index)){
      var_set_set_elt(state_vector,i);
    }else {
      var_set_clear_elt(state_vector,i);
    }
  }
}

/*--------------------------------------------------------------------------*/
void
siva_verify_simulation(void)
{
  SIVA_CONV next_state_conv[siva_num_latch];
  SIVA_CONV current_state_conv[siva_num_latch];
  SIVA_CONV pi_conv[siva_num_primary_input];
  register int i,j,k;
  var_set_t *next_state_vector = var_set_new(siva_num_latch);

  SIVA_VERB1_ARG1("Verifying simulation result..\n");

  for(i=0;i<siva_num_latch;i++){
    current_state_conv[i].ptr =  bck_get_aux(siva_latch_nodes[i].latch,
					  siva_sim_space_aux);
    next_state_conv[i].ptr =  bck_get_aux(siva_latch_nodes[i].input,
					  siva_sim_space_aux);
    /*
      fprintf(siva_output_file,"L(%d): %x, LI(%d): %x\n",
	      bck_get_index(siva_latch_nodes[i].latch),
	      *current_state_conv[i].value_ptr,
	      bck_get_index(siva_latch_nodes[i].input),
	      *next_state_conv[i].value_ptr);
	      */
  }
  for(i=0;i<siva_num_primary_input;i++){
    pi_conv[i].ptr =  bck_get_aux(siva_primary_inputs[i],
				  siva_sim_space_aux);
  }

  for(j=0;j<SIVA_SIM_N_WORDS;j++){
    for(k=0;k<32;k++){ 
      /* build the state vector */
      for(i=0;i<siva_num_latch;i++){
	if ((*(current_state_conv[i].value_ptr+j))&(1<<k)){
	  var_set_set_elt(siva_state_vector,i);
	}else {
	  var_set_clear_elt(siva_state_vector,i);	
	}
      }
      /* build the input vector */
      for(i=0;i<siva_num_primary_input;i++){
	if ((*(pi_conv[i].value_ptr+j))&(1<<k)){
	  var_set_set_elt(siva_input_vector,i);
	}else {
	  var_set_clear_elt(siva_input_vector,i);	
	}
      }
      siva_valuate_latch(siva_state_vector);
      siva_valuate_primary_input(siva_input_vector);
      bck_do_in_topological_order(NULL,siva_evaluate_bit);
      siva_fill_state_vector(siva_state_vector);

      /* build the next_state vector */
      for(i=0;i<siva_num_latch;i++){
	if (bck_is_inverted(siva_latch_nodes[i].input) ^
	     (((*(next_state_conv[i].value_ptr+j))&(1<<k))!=0)){
	  var_set_set_elt(next_state_vector,i);
	}else {
	  var_set_clear_elt(next_state_vector,i);	
	}
      }

      if (!var_set_equal(siva_state_vector,next_state_vector)){
	fprintf(siva_output_file,"simulation verification failed\n");
	fprintf(siva_output_file,"word_index = %d, bit_index =%d\n",j,k);
	fprintf(siva_output_file,"bit simulation:\n");
	siva_print_index_state_vector(siva_state_vector);
	fprintf(siva_output_file,"word simulation:\n");
	siva_print_index_state_vector(next_state_vector);

	SIVA_ASSERT(0,"");
      }
    }
  }
  var_set_free(next_state_vector);
  fprintf(siva_output_file,".............................OK\n");


}

/*--------------------------------------------------------------------------*/
void
siva_evaluate_latch_activity(void)
{
  register int i;
  int num_trial=0;

  SIVA_VERB1_ARG1("Evaluating latch activity...\n");
  siva_init_sim();
  siva_sim_stat();
  num_trial = 0;
  
  while(num_trial < siva_pre_process_num_trial){    
    siva_update_current_with_seed_state(siva_initial_state_vector);
    siva_current_rand_seed = siva_rand_seed;


    for(i=0; i < siva_sim_depth; i++){
      siva_current_sim_depth = i;
      siva_generate_input();
      bck_do_in_topological_order(NULL,siva_evaluate_latch_activity_srv);
      siva_update_latch_activity();            
      siva_update_current_with_next_state();
    }
    num_trial++;
  }
  

  siva_quit_sim();  
  siva_normalize_latch_activity();
  SIVA_VERB1_ARG1("Evaluating latch activity...done\n");
}

/*--------------------------------------------------------------------------*/
void
siva_init_latch_activity(void)
{
  int i;
  SIVA_ASSERT(!siva_init_latch_activity_done,
              "latch activity already initialized");

  siva_activity_vector = var_set_new(32 * SIVA_SIM_N_WORDS);
  siva_latch_activity_aux = bck_alloc_aux(siva_init_aux);
  siva_latch_one_aux = bck_alloc_aux(siva_init_aux);
  siva_latch_zero_aux = bck_alloc_aux(siva_init_aux);
  siva_init_latch_activity_done = 1;
}

/*--------------------------------------------------------------------------*/
void
siva_quit_latch_activity(void)
{
  SIVA_ASSERT(siva_init_latch_activity_done,
              "init latch activity?");  
  siva_init_latch_activity_done = 0;
  var_set_free(siva_activity_vector);
  bck_release_aux(siva_latch_one_aux,NULL);
  bck_release_aux(siva_latch_zero_aux,NULL);
  bck_release_aux(siva_latch_activity_aux,NULL);
}

/*--------------------------------------------------------------------------*/
static void
siva_normalize_latch_activity(void)
{
  int i;
  unsigned max_activity = 0;
  SIVA_CONV conv;
  SIVA_CONV conv1;
  SIVA_CONV conv0;
  SIVA_CONV conv2;  
  float f;



  SIVA_ASSERT(siva_init_latch_activity_done,
              "init latch activity?");
  for(i=0;i<siva_num_latch;i++){
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,
                           siva_latch_activity_aux);
    if (conv.value > max_activity){
      max_activity = conv.value;
    }
  }
  SIVA_ASSERT(max_activity,"No latch actvity?");
  if (siva_verbosity_level > 1){
    fprintf(siva_output_file,"Max latch activity = %u\n",max_activity);
  }

  for(i=0;i<siva_num_latch;i++){
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,
                           siva_latch_activity_aux);
            
    if (siva_companding_method == LOG) {      
      f = log(1+(float)conv.value)/log(1+(float)max_activity);
    }else  if (siva_companding_method == ALOG) {
      f = ((float)conv.value)/max_activity;
      f = (exp(f) - 1)/(exp(1.0) - 1);
    }else {
      f = ((float)conv.value)/max_activity;
    }
    if (siva_verbosity_level > 1){
      fprintf(siva_output_file,"Latch[%d]: %d, Activity %d, Norm %f\n",
              i,bck_get_regular(siva_latch_nodes[i].latch),
              conv.value,
              ((float)conv.value)/max_activity);
    }
    conv.float_value = f;
    bck_set_aux(siva_latch_nodes[i].latch,siva_latch_activity_aux,conv.ptr);

    /* set the interial value with large prob of occurance */
    conv1.ptr = bck_get_aux(siva_latch_nodes[i].latch,
                            siva_latch_one_aux);
    
    conv0.ptr = bck_get_aux(siva_latch_nodes[i].latch,
                            siva_latch_zero_aux);
    f = (conv1.value)/((float) (conv1.value + conv0.value));
    conv.value = (int) (f*100);
    bck_set_aux(siva_latch_nodes[i].latch,siva_latch_inertial_aux,conv.ptr);
    conv2.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_bck_name_aux);
    SIVA_VERB2_ARG7("Latch[%d]: %s %d num_one =%d num_zero =%d Inertia =%d\n",
                   i,conv2.str,bck_get_regular(siva_latch_nodes[i].latch),
                    conv1.value,conv0.value,conv.value);      
  }

}

/*--------------------------------------------------------------------------*/
/* returns the interia value for the specified value                        */
int
siva_get_latch_inertia(int latch_id, int value)
{
  SIVA_CONV conv;
  conv.ptr = bck_get_aux(siva_latch_nodes[latch_id].latch,
                         siva_latch_inertial_aux);
  if (value == 1) {
    return conv.value;
  }else {
    return 100 - conv.value;
  }  
}
     
/*--------------------------------------------------------------------------*/
static int
siva_latch_activity_cmp(const void *p1, const void *p2)
{
  SIVA_CONV conv1,conv2;

  conv1.ptr = bck_get_aux(siva_latch_nodes[*(int *)p1].latch,
                          siva_latch_activity_aux);
  conv2.ptr = bck_get_aux(siva_latch_nodes[*(int *)p2].latch,
                          siva_latch_activity_aux);

  /* if you remove the equality, there is a bug in qsort */
  return ((conv1.float_value <= conv2.float_value) ? -1 : 1);
}



/*--------------------------------------------------------------------------*/
void
siva_fill_active_latch_group(array_t *group_vector_array, int *latch_order)
{

  SIVA_CONV conv;
  int num_group = siva_max_num_control_group;
  int groups[num_group+1];

  int index;
  float f;
  int i,j;
  var_set_t *group_vectors[num_group];

  /* order the latch nodes based on activity */
  qsort(latch_order,siva_num_latch,sizeof(int),siva_latch_activity_cmp);
  if (siva_verbosity_level > 1){
    fprintf(siva_output_file,"Latch Activity order..\n");
    for(j=0;j<siva_num_latch;j++){
      i = latch_order[j];
      conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,
                             siva_latch_activity_aux);
      f = conv.float_value;
      fprintf(siva_output_file,"Latch[%d]: %d, Activity Norm %f\n",
              i,bck_get_regular(siva_latch_nodes[i].latch),f);
    }
  }
  
  for(i=0;i<num_group;i++){
    group_vectors[i] = var_set_new(siva_num_latch);
  }
  
  for(i=0;i<=num_group;i++){
    groups[i] = 0;
  }
  for(i=0;i<siva_num_latch;i++){
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_latch_activity_aux);
    f = conv.float_value;
    f *= num_group;
    index = (int)f;
    if (siva_verbosity_level > 1 ){
      fprintf(siva_output_file,"activity %f, index = %d\n",
              conv.float_value,index);
    }
    groups[index]++;
    /* do not store the last 2 groups */
    for(j=index;j<num_group-1;j++){
      var_set_set_elt(group_vectors[j],i);
    }      
  }

  
  for(i=0;i<num_group;i++){
    fprintf(siva_output_file,"%d-%d\%: %d\n",
            (100 * i)/num_group,(100 * (i+1))/num_group,
            groups[i]);
  }
  fprintf(siva_output_file,"100\%: %d\n",groups[num_group]);

  for(i=0;i<num_group-1;i++){
    if (groups[i] < 10) {
      var_set_free(group_vectors[i]);
      continue;
    }
    if (siva_verbosity_level > 1){
      fprintf(siva_output_file,"Group %d: Size = %d\n",i,
              var_set_n_elts(group_vectors[i]));
    }
    array_insert_last(var_set_t *,group_vector_array,group_vectors[i]);    
  }
}

/*--------------------------------------------------------------------------*/
static void
siva_update_latch_activity(void)
{

  SIVA_CONV next_state_conv;
  SIVA_CONV current_state_conv;
  int i,j;
  SIVA_CONV conv;
  unsigned inversion_word;
  int num_one;
  var_set_t one_set;

  one_set.n_elts = SIVA_SIM_N_WORDS << 5;
  one_set.n_words = SIVA_SIM_N_WORDS;

  
  SIVA_ASSERT(siva_init_latch_activity_done,
              "init latch activity?");

  /* ignoring the first clock */
  if (siva_current_sim_depth == 0){
    return;
  }
  for(i=0;i<siva_num_latch;i++){
    current_state_conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,
                                            siva_sim_space_aux);
    next_state_conv.ptr =  bck_get_aux(siva_latch_nodes[i].input,
					  siva_sim_space_aux);
    if(bck_is_inverted(siva_latch_nodes[i].input)){
      inversion_word = 0xffffffff;
    }else{
      inversion_word = 0;      
    }

    /* count the number of ones and zeros */
    one_set.data = current_state_conv.value_ptr;    
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,
                           siva_latch_one_aux);
    num_one = var_set_n_elts(&one_set);
    conv.value += num_one;    
    bck_set_aux(siva_latch_nodes[i].latch,siva_latch_one_aux,conv.ptr);
    
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,
                           siva_latch_zero_aux);
    conv.value += ((SIVA_SIM_N_WORDS << 5) - num_one);    
    bck_set_aux(siva_latch_nodes[i].latch,siva_latch_zero_aux,conv.ptr);
      
    
    for(j=0;j<SIVA_SIM_N_WORDS;j++){
      siva_activity_vector->data[j] = (*current_state_conv.value_ptr++) ^
        ((inversion_word) ^ (*next_state_conv.value_ptr++)) ;
    }

    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,
                           siva_latch_activity_aux);    

    conv.value += var_set_n_elts(siva_activity_vector);
    bck_set_aux(siva_latch_nodes[i].latch,siva_latch_activity_aux,conv.ptr);
  }
}


/*--------------------------------------------------------------------------*/
static int
siva_count_coi_srv(SIVA_NODE in, void *para)
{
  SIVA_CONV conv;
  conv.ptr = para;
  if (bck_is_flag_set(in,siva_coi_flag)){
    (*conv.value_ptr)++;
  }
  return 1;
}

/*--------------------------------------------------------------------------*/
static int
siva_set_coi_srv(SIVA_NODE in, void *para)
{

  if (bck_is_var(in) ||
      bck_is_flag_set(in,siva_latch_input_flag) ||
      bck_is_flag_set(in,siva_mlh_zero_flag) ||
      bck_is_flag_set(in,siva_mlh_one_flag)){      
    bck_set_fanin_cone_flag(in,siva_coi_flag);
    return 1;
  }
  if (siva_build_pi_free_bck_network_done){
    if (bck_is_flag_set(in,siva_enlarged_target_one_flag) ||
        bck_is_flag_set(in,siva_enlarged_target_zero_flag)){
      bck_set_fanin_cone_flag(in,siva_coi_flag);
    }
    return 1;    
  }
  if (bck_is_flag_set(in,siva_target_one_flag) ||
      bck_is_flag_set(in,siva_target_zero_flag)){
    bck_set_fanin_cone_flag(in,siva_coi_flag);    
    return 1;
  }

  return 1;
}

/*--------------------------------------------------------------------------*/
int
siva_verify_next_state_vector(var_set_t *next_state_vector,
                              var_set_t *input_vector,
                              var_set_t *current_state_vector)
{
  unsigned data[siva_latch_word_size];
  var_set_t state_vector;

  siva_set_state_vector(&state_vector,data);
  var_set_clear(&state_vector);
  
  siva_sim_next_state_vector(&state_vector,input_vector,current_state_vector);
  if (!var_set_equal(next_state_vector,&state_vector)) {
    fprintf(siva_output_file,"Expected Result..\n");    
    var_set_print(siva_output_file,next_state_vector);
    fprintf(siva_output_file,"Found Result..\n");    
    var_set_print(siva_output_file,&state_vector);
    fprintf(siva_output_file,"Found Result..\n");
    var_set_xor(&state_vector,next_state_vector,&state_vector);
    fprintf(siva_output_file,"Diff Result..\n");
    var_set_print(siva_output_file,&state_vector);
  }
  SIVA_ASSERT(var_set_equal(next_state_vector,&state_vector),
              "Verification of state vector failed");
  return var_set_equal(next_state_vector,&state_vector);
}

/*--------------------------------------------------------------------------*/
void
siva_sim_next_state_vector(var_set_t *next_state_vector,
                           var_set_t *input_vector,
                           var_set_t *current_state_vector)
{  

  if (input_vector) {
    siva_valuate_primary_input(input_vector);
  }
  if (current_state_vector) {
    siva_valuate_latch(current_state_vector);
  }
  bck_do_in_topological_order(NULL,siva_evaluate_bit);
  siva_fill_state_vector(next_state_vector);  

}


