/**CFile**********************************************************************

  FileName                   [siva_solve.c]

  PackageName                [siva_core]

  Synopsis                   [solve target list]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]

******************************************************************************/
#include "siva.h"

#define SIVA_NUM_SOLVE_METHOD 2
#define SIVA_ATPG 0
#define SIVA_SIM 1

static char *siva_state_pick_strategies[] = {
  "WT",
  "MLH",  
  "ALH",
  "CS",
  "BFS"
};

#define SIVA_OVERLAP_THRESHOLD (0.01)
#define SIVA_MAX_GROUP_SIZE 100
#define SIVA_MAX_NUM_STATE 10


static int siva_get_seed_state1(SIVA_STATE **seed_state_p,
                                int *strategy_p,
                                var_set_t *vector);

static int siva_seed_fills[SIVA_NUM_SOLVE_METHOD];
static array_t *siva_seed_state_wrapper_array;
static array_t *siva_atpg_seed_state_wrapper_array;
static array_t *siva_sim_seed_state_wrapper_array;
static int siva_num_seed_state_wrapper;
static int siva_num_sim_seed_state_wrapper;
static int siva_num_atpg_seed_state_wrapper;
static void siva_fill_seed_state(int method, array_t *state_wrapper_array,
                                int *num_seed_wrapper);
static int siva_fill_seed_state1(void);
static int siva_copy_state_wrapper_array(array_t *state_wrapper_array);
static void siva_fill_seed_state_init(void);
static void siva_fill_seed_state_quit(void);
static void siva_solve_target_sim(void);
static void siva_solve_target_atpg(void);
static void siva_atpg_et(int start);
static int siva_count_fanin_node_srv(SIVA_NODE in, void *para);
static int siva_update_target_reachability(void);
static void siva_atpg(void);
static void siva_group_targets(void);
static void siva_ungroup_targets(void);


static unsigned siva_fill_seed_status;
static int siva_atpg_terminate;
static int siva_sim_terminate;
static int siva_num_seed;
static long siva_atpg_end_time;
static long siva_sim_end_time;
static int siva_num_fanin_node;



/*--------------------------------------------------------------------------*/
void
siva_solve_target(void)
{
  register int i,j;


  SIVA_CONV conv;
  int status;

  if (siva_num_zero_target + siva_num_one_target ==0) {
    SIVA_VERB1_ARG1("No targets to resolve\n");
    return;
  }

  siva_init_atpg_solver();
  siva_fill_seed_state_init();
  siva_init_sim();
  siva_sim_stat();
  siva_group_targets();
  
  siva_atpg_end_time =siva_sim_end_time =util_cpu_time()+siva_max_time_limit;
  siva_atpg_terminate=0;
  siva_sim_terminate=0;
  
  while(1){
    siva_solve_target_sim();
    if (siva_sim_terminate){
      break;
    }
    siva_solve_target_atpg();
    if (siva_atpg_terminate){
      break;
    }
  }
  fprintf(siva_output_file,"Number of seed used = %d\n",siva_num_seed);
  fprintf(siva_output_file,"Number of states reached = %d\n",
	  siva_get_num_state());
  fprintf(siva_output_file,"Number of control states reached = %d\n",
	  siva_get_num_control_state());

  siva_ungroup_targets();
  siva_quit_sim();
  siva_fill_seed_state_quit();
  siva_quit_atpg_solver();
  SIVA_VERB1_ARG1("ALL stat:\n");
  util_print_cpu_stats(siva_output_file);
}

/*--------------------------------------------------------------------------*/
static void
siva_solve_target_sim(void)
{
  int i,j,k;
  int print_freq;
  time_t sim_start_time;
  time_t sim_end_time;

  SIVA_VERB1_ARG1("Getting Seed States...\n");      
  siva_fill_seed_state(SIVA_SIM,
                       siva_sim_seed_state_wrapper_array,
                       &siva_num_sim_seed_state_wrapper);
  if (siva_sim_terminate){
    goto end;
  }
  if (siva_fill_seed_status == 1 ||
      siva_fill_seed_status == 3) {
    for(i=0;i<siva_num_seed_state_wrapper;i++){
      SIVA_STATE_WRAPPER *state_wrapper =
        array_fetch_p(SIVA_STATE_WRAPPER,siva_seed_state_wrapper_array,i);
      SIVA_STATE_WRAPPER *state_wrapper1 =
        array_fetch_p(SIVA_STATE_WRAPPER,siva_sim_seed_state_wrapper_array,i);
      assert(var_set_equal(state_wrapper1->state_vector,
                           state_wrapper->state_vector));
      assert(state_wrapper1->state == state_wrapper->state);
      assert(state_wrapper1->strategy == state_wrapper->strategy);
      assert( siva_num_seed_state_wrapper = siva_num_sim_seed_state_wrapper);
    }
  }
  SIVA_VERB1_ARG1("Getting Seed States...done\n");
  
  SIVA_VERB1_ARG1("<<<<<<<<< Simulation Checker... >>>>>>>>\n");    
  SIVA_VERB1(siva_target_stat());
  
  
  siva_update_current_with_seed_state_array(siva_sim_seed_state_wrapper_array,
                                            siva_num_sim_seed_state_wrapper);

  for(k=0;k<siva_num_sim_seed_state_wrapper;k++) {
    SIVA_STATE_WRAPPER *state_wrapper =
      array_fetch_p(SIVA_STATE_WRAPPER,siva_sim_seed_state_wrapper_array,k);
    SIVA_VERB2_ARG4("%s Seed State %d: state_value %x\n",
                    siva_state_pick_strategies[state_wrapper->strategy],
                    siva_num_seed,
                    state_wrapper->state->state_value);
    SIVA_VERB3(siva_print_index_state_vector(state_wrapper->state_vector));
    siva_num_seed++;
  }
        
  siva_current_rand_seed = siva_rand_seed;
  print_freq = SIVA_PRINT_FREQ;
  for(i=0; i < siva_sim_depth; i++){        
    SIVA_DBG2("Simulation At depth %d\n",i);
    if (--print_freq == 0) {
      SIVA_VERB1_ARG3("Simulation At depth %d(%d)\n",i,
                      siva_sim_depth);
      print_freq = SIVA_PRINT_FREQ;
    }
    siva_current_sim_depth = i;
    siva_generate_input();              
    bck_do_in_topological_order(NULL,siva_evaluate_solve_srv);
    if (siva_num_one_target+siva_num_zero_target == 0){
      goto end;
    }
    SIVA_VERB3(siva_verify_simulation());
    siva_update_state_table();
    siva_update_current_with_next_state();
    siva_update_mlh();
    siva_update_alh();
    memset(siva_wit_state_pos,0,SIVA_SIM_N_WORDS * 4);
  }
  siva_update_target_reachability();
  if (siva_num_one_target+siva_num_zero_target == 0) {
    goto end;
  }
  if (util_cpu_time() > siva_sim_end_time ) { 
    fprintf(siva_output_file,"Simulation time out\n");
    goto end;          
  }
  SIVA_VERB1_ARG1("<<<<<<<<< Simualation Checker...Done >>>>>>>>\n");      
  return;  
 end:
  siva_sim_terminate = 1;
  SIVA_VERB1_ARG1("<<<<<<<<< Simualation Checker...Done >>>>>>>>\n");      
  return;
}

/*--------------------------------------------------------------------------*/
static void
siva_solve_target_atpg(void)
{
  SIVA_VERB1_ARG1("<<<<<<<<< Symbolic Solver... >>>>>>>>\n");
  SIVA_VERB1(siva_target_stat());

  SIVA_VERB1(siva_et_stat());

  siva_fill_seed_state(SIVA_ATPG,
                       siva_atpg_seed_state_wrapper_array,
                       &siva_num_atpg_seed_state_wrapper);
  if (siva_atpg_terminate){
    goto end;
  }  
  if (util_cpu_time() > siva_atpg_end_time ) {    
    fprintf(siva_output_file,"Symbolic Solver timed out\n");
    goto end;
  }    
  if (siva_num_one_target+siva_num_zero_target == 0) {
    goto end;
  }
  /* solve targets */
  siva_atpg();  
  /* enlarge targets */
  siva_backward_sim();
  siva_et_solve(siva_atpg_seed_state_wrapper_array,
                siva_num_atpg_seed_state_wrapper);
  siva_update_target_reachability();
  if (siva_num_one_target+siva_num_zero_target ==0) {
    goto end;
  }
  siva_find_alh_state(siva_atpg_seed_state_wrapper_array,
                      siva_num_atpg_seed_state_wrapper);
  siva_find_mlh_state(siva_atpg_seed_state_wrapper_array,
                      siva_num_atpg_seed_state_wrapper);  
  SIVA_VERB1_ARG1("<<<<<<<<< Symbolic Solver...Done >>>>>>>>\n");
  return;  
 end:
  SIVA_VERB1_ARG1("<<<<<<<<< Symbolic Solver...Done >>>>>>>>\n");  
  siva_atpg_terminate = 1;
  return;
}



/*--------------------------------------------------------------------------*/
static void
siva_fill_seed_state_init(void)
{
  int i;

  siva_num_seed = 0;  
  siva_seed_state_wrapper_array = array_alloc(SIVA_STATE_WRAPPER,
                                              SIVA_MAX_NUM_STATE);  
  for(i=0;i<SIVA_MAX_NUM_STATE;i++){
    SIVA_STATE_WRAPPER state_wrapper;
    state_wrapper.state_vector = var_set_new(siva_num_latch);
    array_insert(SIVA_STATE_WRAPPER,siva_seed_state_wrapper_array,i,
                 state_wrapper);
  }

  siva_atpg_seed_state_wrapper_array =
    array_alloc(SIVA_STATE_WRAPPER,SIVA_MAX_NUM_STATE);
  
  for(i=0;i<SIVA_MAX_NUM_STATE;i++){
    SIVA_STATE_WRAPPER state_wrapper;
    state_wrapper.state_vector = var_set_new(siva_num_latch);
    array_insert(SIVA_STATE_WRAPPER,siva_atpg_seed_state_wrapper_array,i,
                 state_wrapper);
  }

  siva_sim_seed_state_wrapper_array =
    array_alloc(SIVA_STATE_WRAPPER,SIVA_MAX_NUM_STATE);

  for(i=0;i<SIVA_MAX_NUM_STATE;i++){
    SIVA_STATE_WRAPPER state_wrapper;
    state_wrapper.state_vector = var_set_new(siva_num_latch);
    array_insert(SIVA_STATE_WRAPPER,siva_sim_seed_state_wrapper_array,i,
                 state_wrapper);
  }
  if (!siva_solve_resume){
    SIVA_STATE_WRAPPER *state_wrapper =
      array_fetch_p(SIVA_STATE_WRAPPER,siva_seed_state_wrapper_array,0);
    var_set_assign(state_wrapper->state_vector,siva_initial_state_vector);
    state_wrapper->state = siva_initial_state;
    state_wrapper->strategy = 0;
    siva_num_seed_state_wrapper = 1;  
    siva_fill_seed_status = 0;
  }  else {
    siva_fill_seed_status = 3;
  }

}

/*--------------------------------------------------------------------------*/
static void
siva_fill_seed_state_quit(void)
{
  int i;
  for(i=0;i<SIVA_MAX_NUM_STATE;i++){
    SIVA_STATE_WRAPPER state_wrapper =
      array_fetch(SIVA_STATE_WRAPPER, siva_seed_state_wrapper_array,i);
    var_set_free(state_wrapper.state_vector);
  }
  array_free(siva_seed_state_wrapper_array);
  
  for(i=0;i<SIVA_MAX_NUM_STATE;i++){
    SIVA_STATE_WRAPPER state_wrapper =
      array_fetch(SIVA_STATE_WRAPPER, siva_atpg_seed_state_wrapper_array,i);
    var_set_free(state_wrapper.state_vector);
  }
  array_free(siva_atpg_seed_state_wrapper_array);
  
  for(i=0;i<SIVA_MAX_NUM_STATE;i++){
    SIVA_STATE_WRAPPER state_wrapper =
      array_fetch(SIVA_STATE_WRAPPER, siva_sim_seed_state_wrapper_array,i);
    var_set_free(state_wrapper.state_vector);
  }
  array_free(siva_sim_seed_state_wrapper_array);  
}

/*--------------------------------------------------------------------------*/
static void
siva_fill_seed_state(int solve_method, array_t *state_wrapper_array,
                     int *num_state)
{
  int i;

  SIVA_ASSERT1((solve_method < SIVA_NUM_SOLVE_METHOD),
              "method value = %d\n",solve_method);
  if (siva_fill_seed_status == 3){
    siva_fill_seed_status = 0;
    if (!siva_fill_seed_state1()){
      if (solve_method == SIVA_ATPG){
        siva_atpg_terminate = 1;
      }else if (solve_method == SIVA_SIM){
        siva_sim_terminate = 1;
      }
      return;
    }
  }
  if (solve_method == SIVA_ATPG){
    siva_fill_seed_status |= 0x01;
  }else if (solve_method == SIVA_SIM){
    siva_fill_seed_status |= 0x02;
  }
  *num_state = siva_copy_state_wrapper_array(state_wrapper_array);  
}


/*--------------------------------------------------------------------------*/
static int
siva_copy_state_wrapper_array(array_t *state_wrapper_array)  
{
  int i;
  SIVA_ASSERT(siva_num_seed_state_wrapper,
              "There are no seed states\n");
  for(i=0;i<siva_num_seed_state_wrapper;i++){
    SIVA_STATE_WRAPPER *state_wrapper =
      array_fetch_p(SIVA_STATE_WRAPPER,siva_seed_state_wrapper_array,i);
    SIVA_STATE_WRAPPER *state_wrapper1 =
      array_fetch_p(SIVA_STATE_WRAPPER,state_wrapper_array,i);
    var_set_assign(state_wrapper1->state_vector,
                   state_wrapper->state_vector);
    state_wrapper1->state = state_wrapper->state;
    state_wrapper1->strategy = state_wrapper->strategy;
  }
  return siva_num_seed_state_wrapper;
}

/*--------------------------------------------------------------------------*/
static int
siva_fill_seed_state1(void)  
{
  int i;
  /* get new set of seed states */    
  for(i=0;i<SIVA_MAX_NUM_STATE;i++){
    SIVA_STATE_WRAPPER *state_wrapper =
      array_fetch_p(SIVA_STATE_WRAPPER,
                    siva_seed_state_wrapper_array,i);
    if (!siva_get_seed_state1(&(state_wrapper->state),
                              &(state_wrapper->strategy),
                              state_wrapper->state_vector)){
      break;
    }
  }
  siva_num_seed_state_wrapper = i;
  if (!siva_num_seed_state_wrapper){
    fprintf(siva_output_file,"no more states\n");
  }
  return siva_num_seed_state_wrapper;
}

/*--------------------------------------------------------------------------*/
static int
siva_get_seed_state1(SIVA_STATE **seed_state_p, int *strategy_p,
                     var_set_t *seed_state_vector)
{
  if (siva_get_seed_wit_state(seed_state_p,seed_state_vector)){
    *strategy_p = 0;
    return 1;
  }
  if (siva_get_seed_mlh_state(seed_state_p,seed_state_vector)){
    *strategy_p = 1;
    return 1;
  }
  if (siva_get_seed_alh_state(seed_state_p,seed_state_vector)){
    *strategy_p = 2;
    return 1;    
  }
  if (siva_get_seed_control_state(seed_state_p,seed_state_vector)){
    *strategy_p = 3;
    return 1;
  }
  if (siva_get_seed_bfs_state(seed_state_p,seed_state_vector)){
    *strategy_p = 4;
    return 1;
  }

  return 0;

}


/*--------------------------------------------------------------------------*/
static void
siva_atpg(void)
{
  register int i,j;

  int num_target;
  int num_solve;
  int num_goal;
  
  array_t *goal_array;
  array_t *goal_value_array;
  array_t *target_array;
  

  SIVA_NODE *goals;
  int *goal_values;

  


  goal_array = array_alloc(SIVA_NODE,16);
  goal_value_array = array_alloc(int,16);
  
  
  for(i=0;i<array_n(siva_target_group_array);i++){
    target_array = array_fetch(array_t *,siva_target_group_array,i);
    /* all targets of the group reached */
    if (!target_array) {
      continue;
    }
    num_goal=0;    
    for(j=0;j<array_n(target_array);j++){
      SIVA_NODE target =
        array_fetch(SIVA_NODE,target_array,j);
      if (bck_is_flag_set(target,siva_target_one_flag)){
        array_insert(SIVA_NODE,goal_array,num_goal,
                     bck_get_regular(target));
        array_insert(int,goal_value_array,num_goal,1);
        num_goal++;
      }else if (bck_is_flag_set(target,siva_target_zero_flag)){
        array_insert(SIVA_NODE,goal_array,num_goal,
                     bck_get_regular(target));
        array_insert(int,goal_value_array,num_goal,0);
        num_goal++;
      }
    }
    /* solved by other engine */
    if (num_goal == 0){
      array_free(target_array);
      array_insert(array_t *,siva_target_group_array,i,NIL(array_t));
      continue;
    }
    
    SIVA_ASSERT(num_goal,"num_target can't be 0");
    goals = array_space(SIVA_NODE,goal_array);
     goal_values = array_space(int,goal_value_array);

    siva_atpg_solve(goals,goal_values,num_goal,
                    siva_atpg_seed_state_wrapper_array,
                    siva_num_atpg_seed_state_wrapper,
                    NULL,siva_store_target_witness_state,
                    siva_atpg_print_witness_srv);

    num_solve=0;

    for(j=0;j<num_goal;j++){
      SIVA_VERB2_ARG3("goal %d value %d\n",goals[j],goal_values[j]);
      /* objective reached */
      if (goal_values[j] == 2){
        num_solve++;
        if (bck_is_flag_set(goals[j],siva_target_one_flag)){
          SIVA_VERB1_ARG2("SYMBOLIC-SIM: Target (%d=1) reachable\n",
                          bck_get_index(goals[j]));
          bck_reset_flag(goals[j],siva_target_one_flag);
          siva_num_one_target--;
        }else if (bck_is_flag_set(goals[j],siva_target_zero_flag)){
          SIVA_VERB1_ARG2("SYMBOLIC-SIM: Target (%d=0) reachable\n",
                          bck_get_index(goals[j]));
          bck_reset_flag(goals[j],siva_target_zero_flag);
          siva_num_zero_target--;
        }
      }
    }

    if (num_solve == num_goal){
      SIVA_VERB2_ARG2("Solved all goals in the group %d\n",i);      
      array_free(target_array);
      array_insert(array_t *,siva_target_group_array,i,NIL(array_t));
      
    }
  }
  array_free(goal_array);
  array_free(goal_value_array);
               
}

/*---------------------------------------------------------------------------*/
static void
siva_group_targets(void)
{

  int overlap;
  var_set_t *support_set;
  int support_size;
  var_set_t *target_support_set = var_set_new(siva_num_latch);
  var_set_t *intersect_set = var_set_new(siva_num_latch);
  float max_overlap_frac;
  float overlap_frac;

  SIVA_NODE *targets;
  int num_target;
  
  int i,k,j;
  int group_id;
  int num_group = 0;
  array_t *target_array;

  array_t *group_support_array = array_alloc(var_set_t*,16);
  unsigned bck_visited_flag = bck_alloc_flag();

  siva_target_group_array = array_alloc(array_t *,16);

  if (siva_num_zero_target + siva_num_one_target == 0) {
    SIVA_VERB1("No targets to resolve\n");
    return;
  }
  SIVA_VERB2(siva_target_stat());
  
  /* initialize for targets */
  num_target = siva_num_zero_target+siva_num_one_target;
  targets = ALLOC(SIVA_NODE,num_target);  
  siva_fill_targets(targets);
  
  bck_reset_all_flag(bck_visited_flag);
  
  for(i=0;i<num_target;i++){
    bck_set_fanin_cone_flag(targets[i],bck_visited_flag);
    for(k=0;k<siva_num_latch;k++){
      if (bck_is_flag_set(siva_latch_nodes[k].latch,bck_visited_flag)){
        var_set_set_elt(target_support_set,k);
      }else {
        var_set_clear_elt(target_support_set,k);
      }
    }
    bck_reset_fanin_cone_flag(targets[i],bck_visited_flag);
    
    max_overlap_frac = SIVA_OVERLAP_THRESHOLD;
    group_id = -1;
    for(j=0;j<num_group;j++){
      target_array = array_fetch(array_t *,siva_target_group_array,j);
      if (array_n(target_array) >= SIVA_MAX_GROUP_SIZE){
        continue;
      }
      support_set = array_fetch(var_set_t *,group_support_array,j);
      support_size = var_set_n_elts(support_set);
      var_set_and(intersect_set,support_set,target_support_set);
      overlap = var_set_n_elts(intersect_set);
      overlap_frac = (float)(overlap)/(support_size);
      if (overlap_frac > max_overlap_frac){
        group_id = j;
        max_overlap_frac = overlap_frac;
      }
    }
    if (group_id == -1){
      support_set = var_set_copy(target_support_set);
      array_insert_last(var_set_t *,group_support_array,support_set);
      target_array = array_alloc(SIVA_NODE,4);
      array_insert_last(SIVA_NODE,target_array,targets[i]);
      array_insert_last(array_t *,siva_target_group_array,target_array);
      num_group++;
    }else {
      support_set = array_fetch(var_set_t *, group_support_array,group_id);
      var_set_or(support_set,support_set,target_support_set);
      target_array = array_fetch(array_t *,siva_target_group_array,group_id);
      array_insert_last(SIVA_NODE,target_array,targets[i]);
    }
  }
  for(i=0;i<array_n(group_support_array);i++){
    support_set = array_fetch(var_set_t*,group_support_array,i);
    var_set_free(support_set);
  }
  array_free(group_support_array);
  var_set_free(target_support_set);
  var_set_free(intersect_set);
  bck_release_flag(bck_visited_flag);

#ifndef GPROF
  if (siva_verbosity_level >= 2){
    fprintf(siva_output_file,"# target groups %d\n",
            array_n(siva_target_group_array));
    for(i=0;i<array_n(siva_target_group_array);i++){
      target_array = array_fetch(array_t *,siva_target_group_array,i);
      fprintf(siva_output_file,"Group %d: # Targets %d\n",i,
            array_n(target_array));
      for(j=0;j<array_n(target_array);j++){
        fprintf(siva_output_file,"Bck %d\n",
                bck_get_regular(array_fetch(SIVA_NODE,target_array,j)));
      }
    }
  }
#endif

  FREE(targets);
}


/*---------------------------------------------------------------------------*/
static void
siva_ungroup_targets(void)
{
  int i;
  
  for(i=0;i<array_n(siva_target_group_array);i++){
    array_t *target_array = array_fetch(array_t *,siva_target_group_array,i);
    array_free(target_array);
  }  
  array_free(siva_target_group_array);
}


/*--------------------------------------------------------------------------*/
static int
siva_update_target_reachability(void)
{
  int i;

  if (!array_n(siva_enlarged_target_array)){
    goto end;
  }

  for(i=0;i<array_n(siva_enlarged_target_array);i+=2){
    SIVA_NODE target = array_fetch(SIVA_NODE,siva_enlarged_target_array,i);
    SIVA_NODE enlarged_target =
      array_fetch(SIVA_NODE,siva_enlarged_target_array,i+1);
    int target_obj = 
      bck_is_flag_set(target,siva_target_one_flag) ? 1 :
      bck_is_flag_set(target,siva_target_zero_flag) ? 0 : -1;
    int enlarged_target_obj =
      bck_is_flag_set(enlarged_target,siva_enlarged_target_one_flag) ? 1 :
      bck_is_flag_set(enlarged_target,siva_enlarged_target_zero_flag) ? 0 : -1;

    if (target_obj == -1 && enlarged_target_obj == -1) {
      continue;
    }
    if (target_obj != -1 && enlarged_target_obj != -1) {
      continue;
    }
    /* enlarged target is reached */
    if (target_obj == 0) {
      bck_reset_flag(target,siva_target_zero_flag);
      siva_num_zero_target--;
    }else if (target_obj == 1) {
      bck_reset_flag(target,siva_target_one_flag);
      siva_num_one_target--;
    }
    /* target is reached */
    if (enlarged_target_obj == 0) {      
      bck_reset_flag(enlarged_target,siva_enlarged_target_zero_flag);
    }else if (enlarged_target_obj == 1) {
      bck_reset_flag(enlarged_target,siva_enlarged_target_one_flag);      
    }
  }

 end:
  return;
  
}
