/**CFile**********************************************************************

  FileName                   [siva_state.c]

  PackageName                [siva_core]

  Synopsis                   [State storage and retrieval]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]
******************************************************************************/

#include "siva.h"
#include "siva_state_lookup.h"

#define SIVA_DEFAULT_STATE_TABLE_SIZE 1024
#define SIVA_DEFAULT_STATE_AUX_TABLE_SIZE 4096


#define SIVA_STATE_GROWTH_FACTOR 1
#define SIVA_TARGET_GROWTH_FACTOR 1


static unsigned siva_current_state_values[SIVA_SIM_N_WORDS << 5]; 
static unsigned **siva_current_state_data;

static unsigned siva_total_num_state;
static unsigned siva_num_state_allocated;
static unsigned siva_num_state;
static SIVA_STATE *siva_states;
static array_t *siva_state_ptr_array;
static unsigned siva_state_table_size;
static unsigned siva_num_hash_state;
static SIVA_STATE **siva_state_table;

static unsigned siva_state_aux_table_size;
static unsigned siva_num_hash_state_aux;
static SIVA_STATE_AUX **siva_state_aux_table;
static unsigned siva_num_state_aux;
static unsigned siva_num_state_aux_allocated;
static SIVA_STATE_AUX *siva_state_auxs;
static unsigned *siva_state_aux_data;
static unsigned siva_state_record_size;
static unsigned *siva_state_records;
static unsigned siva_current_state_aux_table;

static void siva_resize_state_table(void);
static unsigned siva_state_hash(unsigned state_value);
void siva_state_table_stat(void);
static unsigned siva_state_aux_hash(unsigned state_value);
static SIVA_STATE_AUX *siva_find_state_aux_hash(unsigned state_value);
static unsigned  siva_state_hash(unsigned state_value);
static void siva_flush_state_aux(int table);
static void siva_write_buffer(void);
static void siva_write_buffer0(void);
static void siva_write_buffer1(void);
static int siva_build_trace(SIVA_STATE *state,
                            array_t *state_vector_array,
                            array_t *input_vector_array);


static unsigned siva_bfs_state_heap_id;

/*---------------------------------------------------------------------------*/
void 
siva_init_state(void)
{ 
  register int i;

  siva_state_ptr_array = array_alloc(SIVA_STATE *,16);
  /* var_set: code hacking */
  siva_primary_input_word_size = ((siva_num_primary_input-1)/32) + 1;
  siva_num_state_allocated = 8192;
  siva_states = ALLOC(SIVA_STATE, 
                      siva_num_state_allocated);
  siva_num_state=0;
  array_insert_last(SIVA_STATE *,siva_state_ptr_array,siva_states);
  siva_total_num_state=0;
  siva_init_state_table();
  srand(0);
  siva_state_init_hash_value = rand();
  siva_init_update_state_table();
}

/*---------------------------------------------------------------------------*/
void
siva_quit_state(void)
{
  register int j;
  siva_quit_update_state_table();
  siva_quit_state_table();
  for(j=0;j<array_n(siva_state_ptr_array);j++){
    FREE(array_fetch(SIVA_STATE *,siva_state_ptr_array,j));
  }
  array_free(siva_state_ptr_array);
}

/*---------------------------------------------------------------------------*/
SIVA_STATE *
siva_get_or_create_state(unsigned state_value,
                         unsigned prev_state_value,
                         int rel_depth,
                         unsigned *state_data,
                         int *is_created)
{
  SIVA_STATE *state;
  SIVA_STATE_AUX *state_aux;  

  *is_created = 0;
  state = siva_find_state_hash(state_value);
  if (state) {
    SIVA_VERB3("State found in hash\n");
    return state;
  }
  *is_created = 1;
  state = siva_alloc_state();
  state->state_value = state_value;
  siva_insert_state_hash(state);

  SIVA_ASSERT(!prev_state_value || rel_depth,"Relative depth can not be zero");
  state_aux = siva_alloc_state_aux();
  state_aux->state_value = state_value;
  state_aux->prev_state_value = prev_state_value;
  state_aux->rel_depth = rel_depth;
  memcpy(state_aux->state_data,state_data,
         sizeof(unsigned)*siva_latch_word_size);  
  siva_insert_state_aux_hash(state_aux);

  SIVA_VERB3("State created\n");
  return state;
}
                  

/*---------------------------------------------------------------------------*/
SIVA_STATE *
siva_alloc_state(void)
{
  SIVA_STATE  *state;
  if (siva_num_state >= siva_num_state_allocated){
    siva_num_state_allocated *= SIVA_STATE_GROWTH_FACTOR;  
    siva_states = ALLOC(SIVA_STATE,
			siva_num_state_allocated);
    array_insert_last(SIVA_STATE *,siva_state_ptr_array,siva_states);
    siva_num_state = 0;
  }
  state = &siva_states[siva_num_state];
  state->state_value = 0;
  siva_num_state++;
  siva_total_num_state++;
  return state;
}

/*---------------------------------------------------------------------------*/
SIVA_STATE_AUX *
siva_alloc_state_aux(void)
{
  SIVA_STATE_AUX  *state_aux;
  int offset;
  
  if (siva_num_state_aux >= (siva_num_state_aux_allocated>>1)){      
    SIVA_VERB2_ARG2("Flushing states (table %d) requested ..\n",
                    siva_current_state_aux_table);  
    siva_write_buffer();
    siva_current_state_aux_table ^= 1;
    siva_num_state_aux = 0;
  }
  if (siva_current_state_aux_table == 0) {
    offset = 0;
  }else {
    offset = siva_num_state_aux_allocated>>1;
  }  
  state_aux = &siva_state_auxs[siva_num_state_aux+offset];
  siva_num_state_aux++;    
  return state_aux;
}

   
/*---------------------------------------------------------------------------*/
unsigned 
siva_get_num_state(void)
{
  return siva_total_num_state;      
}

/*---------------------------------------------------------------------------*/
unsigned
siva_get_state_value(SIVA_STATE *state)
{
  return state->state_value;
}

/*---------------------------------------------------------------------------*/
void
siva_set_state_value(SIVA_STATE *state, unsigned state_value)
{
  state->state_value = state_value;
}

/*---------------------------------------------------------------------------*/
void
siva_init_state_table(void)
{
  register unsigned j;
  siva_state_table_size = SIVA_DEFAULT_STATE_TABLE_SIZE;
  siva_state_table = ALLOC(SIVA_STATE*,siva_state_table_size);
  for(j=0;j < siva_state_table_size; j++){
    siva_state_table[j] = NIL(SIVA_STATE);
  }
  siva_num_hash_state = 0;
}
/*---------------------------------------------------------------------------*/
void
siva_quit_state_table(void)
{
  FREE(siva_state_table);
}


/*---------------------------------------------------------------------------*/
static void
siva_write_buffer(void)
{
  siva_flush_state_aux(siva_current_state_aux_table);
}


/*---------------------------------------------------------------------------*/
static void
siva_flush_state_aux(int table)
{
  FILE *file;
  register int j;
  char buf[20];  
  int num_record;
  SIVA_STATE_AUX *src;
  unsigned *dest;
  int total_num_record = 0;
  int offset;
  
  if (table == 1) {
    offset = siva_state_aux_table_size>>1;
  }else {
    offset = 0;
  }
  for(j=0;j<siva_state_aux_table_size>>1;j++){
    num_record=0;
    sprintf(buf,"cache/%03x",j);
    file = fopen(buf,"a");
    src = siva_state_aux_table[j+offset];
    dest = siva_state_records;
    while(src){
      *dest++ = src->state_value;
      *dest++ = src->prev_state_value;
      *dest++ = src->rel_depth;
      memcpy(dest,src->state_data,sizeof(unsigned) * siva_latch_word_size);
      dest += siva_latch_word_size;
      num_record++;
      src = src->next;
    }
    if (num_record) {
      fwrite(&num_record,sizeof(unsigned),1,file);
      fwrite(siva_state_records,sizeof(unsigned)*siva_state_record_size,
             num_record,file);
      siva_state_aux_table[j+offset] = NIL(SIVA_STATE_AUX);
    }
    total_num_record += num_record;
    fclose(file);
  }

  SIVA_VERB2_ARG2("# flushed = %d\n",total_num_record);  
}


/*---------------------------------------------------------------------*/
void
siva_build_witness_state(int word_index, int bit_index,
                         var_set_t *state_vector)
{
  register unsigned n,i,m,k1,k2;
  SIVA_CONV conv[siva_num_latch];
  unsigned *ptr = state_vector->data;
  /* modulo 8 operation */
  k1 = bit_index & 7;
  k2 = bit_index >> 3;

  if (siva_little_endian) {
    k2 = 3-k2;
  }
  
  for(i=0;i<siva_num_latch;i++){
    conv[i].ptr =  bck_get_aux(siva_latch_nodes[i].latch,
			       siva_sim_space_aux);
    conv[i].value_ptr += word_index;
  }
  
  for(i=0,n=0;i<siva_latch_word_size-1;i++){
    *ptr = 0;
    for(m=0;m<32;m++){
      *ptr |= bit_lookup[k1][conv[n++].byte_ptr->byte[k2]][m];
    }
    ptr++;
  }
  
  /* go for last 32 bit block of latch */
  for(m=0,*ptr=0;n<siva_num_latch;m++){
    *ptr |= bit_lookup[k1][conv[n++].byte_ptr->byte[k2]][m];
  }
	
  /* do some sanitary check */
  if (siva_error_check_level >= 1) {
    for(i=0;i<siva_num_latch;i++){
      SIVA_CONV conv1;
      conv1.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_sim_space_aux);
      SIVA_ASSERT((var_set_get_elt(state_vector,i) ==
                   (((*(conv1.value_ptr+word_index)) & (1<<bit_index)) != 0)),
                  "problem in state extracting");
    }
  }
}

/*--------------------------------------------------------------------------*/
void
siva_build_witness_input(int word_index, int bit_index,
                         var_set_t *input_vector)
{
  register int i;
  SIVA_CONV conv;
  
  for(i=0;i<siva_num_primary_input;i++){
    conv.ptr = bck_get_aux(siva_primary_inputs[i],siva_sim_space_aux);
    if ((*(conv.value_ptr+word_index)) & (1<<bit_index)){
      var_set_set_elt(input_vector,i);
    }else {
      var_set_clear_elt(input_vector,i);
    }
  }
}

/*--------------------------------------------------------------------------*/
int
siva_regenerate_state_vector(SIVA_STATE *state,
                             var_set_t *state_vector)
{
  siva_upload_state_vector(state,state_vector);
  return 1;
}

/*---------------------------------------------------------------------------*/
void
siva_upload_state_vector(SIVA_STATE *state,
                         var_set_t *state_vector)
{
  SIVA_STATE_AUX state_aux;

  state_aux.state_value = state->state_value;
  state_aux.state_data = state_vector->data;
  siva_upload_state_aux(&state_aux);
  
}

/*---------------------------------------------------------------------------*/
void
siva_upload_state_aux(SIVA_STATE_AUX *state_aux)
{
  unsigned hash_value;
  FILE *file;
  char buf[20];
  unsigned num_record;
  SIVA_STATE_AUX *entry;
  register int j;
  unsigned *ptr;


  /* do a local search in memory */
  entry = siva_find_state_aux_hash(state_aux->state_value);

  if (entry) {
    state_aux->prev_state_value = entry->prev_state_value;
    state_aux->rel_depth = entry->rel_depth;
    memcpy(state_aux->state_data,entry->state_data,
           sizeof(unsigned) * siva_latch_word_size);
    return;
  }
  /* search in file */
  hash_value = siva_state_aux_hash(state_aux->state_value);
  sprintf(buf,"cache/%03x",hash_value);
  file = fopen(buf,"r");
  SIVA_ASSERT1(file,"File <%s> open error\n",buf);
  while(!feof(file)){
    fread(&num_record,sizeof(unsigned),1,file);
    SIVA_ASSERT(num_record,"Num of record can't be zero");
    fread(siva_state_records,sizeof(unsigned)*siva_state_record_size,
           num_record,file);
    ptr = siva_state_records;
    for(j=0;j<num_record;j++){
      if (*ptr > state_aux->state_value) {
        ptr += siva_state_record_size;
        continue;
      }
      if (*ptr != state_aux->state_value){
        break;
      }
      ptr ++;
      state_aux->prev_state_value = *ptr;
      ptr ++;
      state_aux->rel_depth = *ptr;
      ptr ++;
      memcpy(state_aux->state_data,ptr,sizeof(unsigned) *
             siva_latch_word_size);
      fclose(file);

      return;
    }
  }  
  fclose(file);
  SIVA_ASSERT(0,"state not found in the file");
}


/*---------------------------------------------------------------------------*/
void
siva_init_state_aux(void)
{
  register unsigned j;
  unsigned *ptr;

  struct sched_param rr_param;
  
  siva_state_aux_table_size = SIVA_DEFAULT_STATE_AUX_TABLE_SIZE;
  siva_state_aux_table = ALLOC(SIVA_STATE_AUX*,siva_state_aux_table_size);
  for(j=0;j < siva_state_aux_table_size; j++){
    siva_state_aux_table[j] = NIL(SIVA_STATE_AUX);
  }
  siva_num_hash_state_aux = 0;
  siva_num_state_aux = 0;
  siva_num_state_aux_allocated = siva_state_aux_table_size << 5;
  siva_state_auxs = ALLOC(SIVA_STATE_AUX, siva_num_state_aux_allocated);
  ptr = siva_state_aux_data = ALLOC(unsigned, siva_latch_word_size *
                                    siva_num_state_aux_allocated);
  
  for(j=0;j<siva_num_state_aux_allocated;j++){
    siva_state_auxs[j].state_data = ptr;
    ptr += siva_latch_word_size;
  }

  siva_state_record_size = 3 + siva_latch_word_size;
  siva_state_records = ALLOC(unsigned,siva_state_record_size * 
                             siva_num_state_aux_allocated);

  siva_current_state_aux_table = 0;
  SIVA_VERB2_ARG1("Removing the directory cache and its contents...\n");
  system("\\rm -rf cache");
  SIVA_VERB2_ARG1("Removing the directory cache and its contents...done\n");
  system("mkdir cache");
}

/*---------------------------------------------------------------------------*/
void
siva_quit_state_aux(void)
{
  FREE(siva_state_records);
  FREE(siva_state_aux_table);
  FREE(siva_state_auxs);
  FREE(siva_state_aux_data);
  FREE(siva_state_aux_table);  
}

/*---------------------------------------------------------------------------*/
static unsigned 
siva_state_aux_hash(unsigned state_value)
{
  return (state_value & ((siva_state_aux_table_size>>1)-1));
}

/*---------------------------------------------------------------------------*/
void
siva_insert_state_aux_hash(SIVA_STATE_AUX *state_aux)
{
  unsigned hash_value = siva_state_aux_hash(state_aux->state_value);

  SIVA_VERB3_ARG1("insert state aux ..\n");
  if (siva_current_state_aux_table) {
    hash_value += (siva_state_aux_table_size>>1);
  }
  SIVA_ASSERT(hash_value < siva_state_aux_table_size,
              "hash value is more than table size");
  if (!siva_state_aux_table[hash_value] ||
       siva_state_aux_table[hash_value]->state_value <
      state_aux->state_value){
    state_aux->next = siva_state_aux_table[hash_value];
    siva_state_aux_table[hash_value] = state_aux;
  }else {
    SIVA_STATE_AUX *entry = siva_state_aux_table[hash_value];
    while(entry->next && 
          (entry->next->state_value > state_aux->state_value)){
      entry = entry->next;
    }    
    state_aux->next = entry->next;
    entry->next = state_aux;
  }
  SIVA_VERB3_ARG1("insert state aux ..done\n");
}

/*---------------------------------------------------------------------------*/
static SIVA_STATE_AUX *
siva_find_state_aux_hash(unsigned state_value)
{
  SIVA_STATE_AUX *entry;
  unsigned hash_value = siva_state_aux_hash(state_value);
  
  if (siva_current_state_aux_table) {
    hash_value += (siva_state_aux_table_size>>1);
  }
  for(entry = siva_state_aux_table[hash_value]; 
      entry && (entry->state_value > state_value); entry = entry->next);
  if ((entry)&&(entry->state_value == state_value)){
    return entry;
  }else {
    return NIL(SIVA_STATE_AUX);
  }
}

/*---------------------------------------------------------------------------*/
static unsigned 
siva_state_hash(unsigned state_value)
{
  return (state_value & (siva_state_table_size-1));
}

/*---------------------------------------------------------------------------*/
SIVA_STATE *
siva_find_state_hash(unsigned state_value)
{
  SIVA_STATE *entry;

  unsigned hash_value = siva_state_hash(state_value);
  for(entry = siva_state_table[hash_value]; 
      entry && (entry->state_value > state_value); entry = entry->next);
  if ((entry)&&(entry->state_value == state_value)){
    return entry;
  }else {
    return NIL(SIVA_STATE);
  }
}

/*---------------------------------------------------------------------------*/
void
siva_insert_state_hash(SIVA_STATE *state)
{
  unsigned hash_value = siva_state_hash(state->state_value);
  if (!siva_state_table[hash_value] ||
       siva_state_table[hash_value]->state_value < state->state_value){
    state->next = siva_state_table[hash_value];
    siva_state_table[hash_value] = state;
  }else {
    SIVA_STATE *entry = siva_state_table[hash_value];
    while(entry->next && 
          (entry->next->state_value > state->state_value)){
      entry = entry->next;
    }

    
    state->next = entry->next;
    entry->next = state;
  }
  siva_num_hash_state++;
  SIVA_ASSERT(siva_num_hash_state==siva_total_num_state,
              "allocated does not match hashed");
  /* resize when density is 8 per bin */
  if ((siva_num_hash_state >> 3) == siva_state_table_size){
    SIVA_VERB2(siva_state_table_stat());
    siva_resize_state_table();
  }
}

/*---------------------------------------------------------------------------*/
void
siva_remove_state_hash(SIVA_STATE *state)
{
  unsigned hash_value = siva_state_hash(state->state_value);

  SIVA_ASSERT((siva_find_state_hash(state->state_value) != NIL(SIVA_STATE)),
              "state not found");

  if (siva_state_table[hash_value] == state){
    siva_state_table[hash_value] = state->next;
  }else {
    SIVA_STATE *entry = siva_state_table[hash_value];
    while(entry->next->state_value > state->state_value){
      entry = entry->next;
    }
    SIVA_ASSERT((entry->next == state),"");
    entry->next = entry->next->next;
  }
  siva_num_hash_state--;
  /* need to reduce table size ?*/
  /*
    resize when the density goes down to 2 per bin
    if ((siva_num_hash_state << 1) == siva_state_table_size){
    siva_resize_state_table();
    }
  */
}

/*---------------------------------------------------------------------------*/
void
siva_state_table_stat(void)
{
  if (siva_verbosity_level > 1) {
    register int i;
    float avg_bin_load = (float)siva_num_hash_state/siva_state_table_size;
    float bin_load_var;
    float std=0;
    for(i=0;i<siva_state_table_size;i++){
      SIVA_STATE *entry = siva_state_table[i];
      int num_entry=0;
      for(entry=siva_state_table[i];entry;entry=entry->next,num_entry++);
      bin_load_var = (num_entry-avg_bin_load);
      bin_load_var *= bin_load_var;
      std += bin_load_var;
    }
    fprintf(siva_output_file,"avg load = %f\n",avg_bin_load);
    fprintf(siva_output_file,"std load = %f\n",
            sqrt(std/siva_state_table_size));
    fprintf(siva_output_file,"table_size = %d\n",siva_state_table_size);
    fprintf(siva_output_file,"load = %d\n",siva_num_hash_state);
  }
  fprintf(siva_output_file,"# state = %d\n",siva_total_num_state);
  
}

/*--------------------------------------------------------------------------*/
void
siva_init_update_state_table(void)
{
  int i;

  siva_last_stored_state_depths = ALLOC(unsigned,(SIVA_SIM_N_WORDS<<5));
  siva_last_stored_states = ALLOC(SIVA_STATE *,(SIVA_SIM_N_WORDS<<5));

  siva_current_state_data = ALLOC(unsigned *,SIVA_SIM_N_WORDS);
  for(i=0;i<SIVA_SIM_N_WORDS;i++){
    siva_current_state_data[i] = ALLOC(unsigned,(siva_latch_word_size << 5));
  }
  siva_init_state_aux();

}
  
/*--------------------------------------------------------------------------*/
void
siva_quit_update_state_table(void)
{
  int i;

  siva_quit_state_aux();
  for(i=0;i<SIVA_SIM_N_WORDS;i++){
    FREE(siva_current_state_data[i]);
  }
  FREE(siva_current_state_data);
  FREE(siva_last_stored_state_depths);
  FREE(siva_last_stored_states);
}
     
/*
  --------------------------------------------------------------------------
  the new approach of state storage instead of using var_set_word_assign  
  yields 20% faster execution though at the cost of 32kb mem for lookup
  --------------------------------------------------------------------------
*/
void
siva_update_state_table(void)
{
  SIVA_CONV next_state_conv[siva_num_latch];
  SIVA_CONV conv_value;
  register int i,j,k,d;
  SIVA_STATE *next_state;
  SIVA_STATE *current_state;
  unsigned *ptr;
  unsigned *next_state_data_ptr;
  unsigned *current_state_data_ptr;
  unsigned next_state_values[32];  

  int rel_depth;
  int priority_value;
  int control_state_priority_value=0;
  /* code hacking */
  var_set_t state_vector;
  int is_created;
  int witness_state;

  SIVA_DBG1("Updating state table ..\n");

  
  state_vector.n_elts = siva_num_latch;
  state_vector.n_words = siva_latch_word_size;


  for(i=0;i<siva_num_latch;i++){
    next_state_conv[i].ptr =  bck_get_aux(siva_latch_nodes[i].input,
					  siva_sim_space_aux);    
  }

  for(j=0;j<SIVA_SIM_N_WORDS;j++){

    register unsigned m; /* index to byte */
    register unsigned n; /* index to latch */
    register unsigned c; /* index to bit in byte */

    /* Next State Values */
    ptr = siva_state_word_vector->data;

    /* loop on each byte */
    for(m=0;m<4;m++){
      int byte_index = m;      
      if (siva_little_endian) {
        byte_index = 3-m;
      }
      /* loop on each bit of byte */
      for(c=0;c<8;c++){
	/* go for all each 32 bit block of latch */
	for(i=0,n=0;i<siva_latch_word_size-1;i++){
          *ptr = 0;
	  for(k=0;k<32;k++){
	    *ptr |= 
	      bit_lookup[c][siva_latch_inversion_bytes[n] ^ 
			   next_state_conv[n].byte_ptr->byte[byte_index]][k];
	    n++;
	  }
          ptr++;
	}
	/* go for last 32 bit block of latch */
        for(*ptr=0,k=0;n<siva_num_latch;k++){
	  *ptr |= 
	    bit_lookup[c][siva_latch_inversion_bytes[n] ^ 
			 next_state_conv[n].byte_ptr->byte[byte_index]][k];
	  n++;
	}
	ptr++;
      }
    }
    var_set_word_compress(next_state_values,siva_state_word_vector,
                          siva_state_init_hash_value);


    current_state_data_ptr = siva_current_state_data[j];
    next_state_data_ptr = siva_state_word_vector->data;
    for(k=0;k<32;k++,current_state_data_ptr += siva_latch_word_size,
          next_state_data_ptr += siva_latch_word_size){

      if (siva_current_sim_depth &&
          (siva_wit_state_pos[j] & (1<<k) ||
           siva_mlh_state_priority[(j<<5)+k] ||
           siva_alh_state_priority[(j<<5)+k])) {
        current_state =
          siva_get_or_create_state(siva_current_state_values[(j<<5)+k],
                                   siva_last_stored_states[(j<<5)+k]->state_value,
                                   (siva_current_sim_depth - 
                                    siva_last_stored_state_depths[(j<<5)+k]),
                                   current_state_data_ptr,
                                   &is_created);

        if (is_created){
          siva_last_stored_states[(j<<5)+k] = current_state;
          siva_last_stored_state_depths[(j<<5)+k] = siva_current_sim_depth;        
          if (siva_wit_state_pos[j] & (1<<k)) {
            priority_value = siva_current_sim_depth;
            siva_put_wit_state(priority_value,current_state);
            SIVA_VERB2_ARG3("CYCLE-SIM: TARGET: Witness State %x, state_value = %x\n",
                            current_state,
                            siva_current_state_values[(j<<5)+k]);          
          }else if (siva_mlh_state_priority[(j<<5)+k]) {
            siva_put_mlh_state(siva_mlh_state_priority[(j<<5)+k],current_state);
          }else if (siva_alh_state_priority[(j<<5)+k]) {
            siva_put_alh_state(siva_alh_state_priority[(j<<5)+k],current_state);
          }
        }
      }

      state_vector.data = next_state_data_ptr;

      next_state = siva_find_state_hash(next_state_values[k]);
      if (next_state) {
        continue;
      }
      if (siva_current_sim_depth >= siva_update_state_depth){
        return;
      }

      witness_state = siva_check_et_reached(next_state,
                                            &state_vector);
      
      control_state_priority_value = 0;
      if (siva_control_state_check_on){
        control_state_priority_value =
          siva_control_state_filter_in(&state_vector);
      }

      if (control_state_priority_value == 0 &&
          siva_bfs_state_on == 0 && witness_state == 0){
        continue;
      }      
      next_state =
          siva_get_or_create_state(next_state_values[k],
                                   siva_last_stored_states[(j<<5)+k]->state_value,
                                   (siva_current_sim_depth+1 - 
                                    siva_last_stored_state_depths[(j<<5)+k]),
                                   next_state_data_ptr,
                                   &is_created);
      if (is_created){
        SIVA_ERR_CHK3(siva_verify_state(next_state));
      }
      siva_last_stored_states[(j<<5)+k] = next_state;
      siva_last_stored_state_depths[(j<<5)+k] = siva_current_sim_depth+1;

      if (witness_state) {
        priority_value = siva_current_sim_depth;
        siva_put_wit_state(priority_value,next_state);
        SIVA_VERB2_ARG3("CYCLE-SIM: ET: Witness State %x, state_value = %x\n",
                        current_state,
                        siva_current_state_values[(j<<5)+k]);          
      }else if (control_state_priority_value){        
        float p;
        p = (float)control_state_priority_value;
        assert(p);          
        p += (1/((float)(siva_current_sim_depth+2)));
        assert(p);
        siva_put_control_state(p,next_state);      
      }else if (siva_bfs_state_on){
        siva_put_bfs_state(siva_current_sim_depth+1,next_state);
      }
    }

    memcpy(siva_current_state_data[j],siva_state_word_vector->data,
           sizeof(unsigned)*(siva_latch_word_size << 5));    
    memcpy(&siva_current_state_values[j<<5],
           next_state_values,sizeof(unsigned) * 32);

    for(i=0;i<siva_num_latch;i++){        
      next_state_conv[i].value_ptr++;
    }
  }

  SIVA_DBG1("Updating state table ..done\n");
}

/*--------------------------------------------------------------------------*/
void
siva_build_initial_state_vector(void)
{
  register int i;

  siva_initial_state_vector =  var_set_new(siva_num_latch);
  for(i=0;i<siva_num_latch;i++){
    if (!bck_is_inverted(siva_latch_nodes[i].latch)){
      var_set_set_elt(siva_initial_state_vector,i);
    }else{
      var_set_clear_elt(siva_initial_state_vector,i);
    }
  }
  if (siva_verbosity_level > 1){
    siva_print_index_state_vector(siva_initial_state_vector);
  }

}

/*--------------------------------------------------------------------------*/
void
siva_free_initial_state_vector(void)
{
  var_set_free(siva_initial_state_vector);
}

/*--------------------------------------------------------------------------*/
void
siva_create_initial_state(void)
{
  int j;
  int is_created;
  unsigned state_value;

  var_set_compress(&state_value,
		   siva_initial_state_vector,
                   siva_state_init_hash_value);
  siva_initial_state = 
    siva_get_or_create_state(state_value,0,0,siva_initial_state_vector->data,
                             &is_created);
  
  SIVA_ASSERT(is_created,"Initial state should be created");
  SIVA_VERB2_ARG3("Init state: %x state_value = %x\n\n",
                  siva_initial_state,siva_initial_state->state_value);
  
}


/*--------------------------------------------------------------------------*/
void
siva_init_bfs_state(void)
{
  siva_bfs_state_heap_id = siva_alloc_heap(siva_heap_unsigned_min_cmp);
}

/*--------------------------------------------------------------------------*/
void
siva_quit_bfs_state(void)
{
  siva_release_heap(siva_bfs_state_heap_id);
}

/*--------------------------------------------------------------------------*/
void
siva_put_bfs_state(unsigned priority_value, SIVA_STATE *state)
{
  SIVA_KEY key;
  SIVA_CONV conv;
  key.unsigned_key = priority_value;
  conv.state = state;
  siva_heap_insert(key,conv,siva_bfs_state_heap_id);
}

/*--------------------------------------------------------------------------*/
int
siva_get_bfs_state(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr)
{
  int found;
  found = siva_heap_extract_top(key_ptr,conv_ptr,siva_bfs_state_heap_id);
  return found;
}

/*--------------------------------------------------------------------------*/
int
siva_get_seed_bfs_state(SIVA_STATE **state_p,
                        var_set_t *seed_state_vector)
{
  SIVA_KEY key;
  SIVA_CONV conv;
  while(siva_get_bfs_state(&key,&conv)){
    /* regenerate state and check if valid state*/
    if (!siva_regenerate_state_vector(conv.state,
                                      seed_state_vector)){
      continue;
    }
    *state_p = conv.state;
    SIVA_VERB1_ARG3("BFS: Seed state value = %x, key = %u\n",
                    conv.state->state_value,key.unsigned_key);
    return 1;
  }  
  return 0;
}


/*---------------------------------------------------------------------------*/
static void
siva_resize_state_table(void)
{
  register int i;

  siva_state_table = REALLOC(SIVA_STATE*,siva_state_table,
                             (siva_state_table_size << 1));  

  for(i=0;i<siva_state_table_size;i++){
    SIVA_STATE lower_header;
    SIVA_STATE *lower_ptr=&lower_header;

    SIVA_STATE upper_header;
    SIVA_STATE *upper_ptr=&upper_header;

    SIVA_STATE *entry = siva_state_table[i];

    lower_header.next = NIL(SIVA_STATE);
    upper_header.next = NIL(SIVA_STATE);
    while(entry){
      if (entry->state_value & (siva_state_table_size)){
        upper_ptr->next = entry;        
        upper_ptr = entry;
      }else{
        lower_ptr->next = entry;
        lower_ptr = entry;
      }
      entry = entry->next;
    }
    upper_ptr->next = NIL(SIVA_STATE);
    lower_ptr->next = NIL(SIVA_STATE);

    siva_state_table[i] = lower_header.next;
    siva_state_table[i+siva_state_table_size] = upper_header.next;
  }
  siva_state_table_size = (siva_state_table_size << 1);
}

/*
  ---------------------------------------------------------------------------
  It prints the trace for the given state; returns
  number the length of the trace.
  ---------------------------------------------------------------------------
*/
int
siva_print_state_trace(SIVA_STATE *state)
{
  int num_prev_state;
  array_t *state_vector_array;
  array_t *input_vector_array;
  int i;
  
  
  state_vector_array =  array_alloc(var_set_t *,8);
  input_vector_array =  array_alloc(var_set_t *,8);
  
  if (!siva_build_trace(state,
                        state_vector_array,
                        input_vector_array)){
    SIVA_VERB1_ARG1("Failed to build the trace\n");
    goto end;
  }
  num_prev_state = array_n(state_vector_array);
  if (num_prev_state == 0){
    goto end;
  }
  
  for(i=num_prev_state-1;i>=0;i--){
    var_set_t *input_vector =
      array_fetch(var_set_t*, input_vector_array,i);
    var_set_t *state_vector =
      array_fetch(var_set_t*, state_vector_array,i);

    siva_print_witness_trace(&input_vector,
                             &state_vector,
                             (num_prev_state-1-i),
                             0);
  }
 end:
  for(i=0;i<array_n(state_vector_array);i++){
    var_set_t *state_vector =
      array_fetch(var_set_t *,state_vector_array,i);
  }
  array_free(state_vector_array);
  for(i=0;i<array_n(input_vector_array);i++){
    var_set_t *input_vector =
      array_fetch(var_set_t *,input_vector_array,i);
  }
  array_free(input_vector_array);

  return num_prev_state;
}

/*
  ---------------------------------------------------------------------------
  This prints the debug trace.
  Format:
  (invar (cntr = <clock>) -> (
  (<debug_node>=<debug_value) &
  ... &
  1)
  This trace can be plugged into other tool like SMV for validating.
  ---------------------------------------------------------------------------
*/
void
siva_print_witness_trace(var_set_t **input_vectors,
                         var_set_t **state_vectors,
                         int offset,
                         int unfold_depth)
{
  register int i,j;
  int num_debug_node = array_n(siva_debug_node_array);
  int values[num_debug_node];
  SIVA_CONV conv;
  SIVA_NODE* debug_nodes;
  char **debug_node_names;

  SIVA_ASSERT(num_debug_node,"No debug nodes");
  
  debug_nodes = array_space(SIVA_NODE,
                            siva_debug_node_array);
  debug_node_names =
    array_space(char *,siva_debug_node_name_array);
  
  for(j=unfold_depth;j>=0;j--){    
    siva_evaluate_node(debug_nodes,values,num_debug_node,
                       input_vectors[j],
                       state_vectors[j]);
    fprintf(siva_output_file,"------- Debug Nodes @(posedge clk) %d-----\n",
            (unfold_depth-j)+offset);
    fprintf(siva_output_file,"invar((cntr=%d) -> (\n",(unfold_depth-j)+offset);
    for(i=0;i<num_debug_node;i++){
      if (bck_is_constant(debug_nodes[i])) {
        continue;
      }
      fprintf(siva_output_file,"(%s =%d) &\n",debug_node_names[i],
              values[i]);
    }
    fprintf(siva_output_file,"1));\n");
  }    
}


/*--------------------------------------------------------------------------*/
int
siva_verify_state(SIVA_STATE *state)
{
  array_t *state_vector_array;
  array_t *input_vector_array;
  int i;
  state_vector_array =  array_alloc(var_set_t *,8);
  input_vector_array =  array_alloc(var_set_t *,8);

  if (!siva_build_trace(state,
                        state_vector_array,
                        input_vector_array)){
    SIVA_ASSERT(0,"Failed to build the state\n");
    return 0;
  }
  for(i=0;i<array_n(state_vector_array);i++){
    var_set_t *state_vector =
      array_fetch(var_set_t *,state_vector_array,i);
  }
  array_free(state_vector_array);
  for(i=0;i<array_n(input_vector_array);i++){
    var_set_t *input_vector =
      array_fetch(var_set_t *,input_vector_array,i);
  }
  array_free(input_vector_array);
  return 1;
}

/*--------------------------------------------------------------------------*/
static int
siva_build_trace(SIVA_STATE *state,
                 array_t *state_vector_array,
                 array_t *input_vector_array)
{
  int i,j;
  SIVA_STATE_AUX state_aux;
  unsigned data[siva_latch_word_size];
  var_set_t *seed_vector;
  var_set_t *target_vector;
  array_t *distance_array;
  array_t *stored_state_vector_array;
  var_set_t **input_vectors;
  var_set_t **state_vectors;
  int prev_size;
  int succeed;
  
  distance_array = array_alloc(int,16);
  stored_state_vector_array = array_alloc(var_set_t *,16);
  
  state_aux.state_data = data;
  state_aux.state_value = state->state_value;
  while(1){
    var_set_t *vector;
    siva_upload_state_aux(&state_aux);
    vector = var_set_new(siva_num_latch);
    memcpy(vector->data,state_aux.state_data,sizeof(unsigned)*
           siva_latch_word_size);
    array_insert_last(var_set_t *,stored_state_vector_array,vector);    
    if (state_aux.rel_depth == 0) {
      break;
    }
    array_insert_last(int,distance_array,state_aux.rel_depth);
    state_aux.state_value = state_aux.prev_state_value;
  }
  
  target_vector =
    array_fetch(var_set_t *,stored_state_vector_array,0);
  
  for(i=1;i<array_n(stored_state_vector_array);i++){
    var_set_t *seed_vector =
      array_fetch(var_set_t *,stored_state_vector_array,i);
    int rel_depth = array_fetch(int,distance_array,i-1);
    prev_size = array_n(input_vector_array);
    SIVA_ASSERT(array_n(input_vector_array) ==
                array_n(state_vector_array),
                "input/state array mismatch");
    for(j=0;j<rel_depth;j++){
      array_insert_last(var_set_t *,input_vector_array,
                        var_set_new(siva_num_primary_input));
      array_insert_last(var_set_t *,state_vector_array,
                        var_set_new(siva_num_latch));
    }
    /* array gets resize and hence space pointer will change */
    input_vectors = array_space(var_set_t*,input_vector_array);
    state_vectors = array_space(var_set_t*,state_vector_array);
    if (!siva_atpg_find_input_sequence(target_vector,seed_vector,rel_depth,
                                       input_vectors+prev_size,
                                       state_vectors+prev_size)){
      succeed = 0;
      goto end;
    }
    SIVA_ASSERT(var_set_equal((state_vectors+prev_size)[rel_depth-1],
                              seed_vector),
                "state vector has different value than seed vector");
    target_vector = seed_vector;
  }
  /* verifying the state sequence */
  target_vector =
    array_fetch(var_set_t *,stored_state_vector_array,0);
  for(i=0;i<array_n(state_vector_array);i++){
    var_set_t *seed_vector =
      array_fetch(var_set_t *,state_vector_array,i);
    var_set_t *input_vector =
      array_fetch(var_set_t *,input_vector_array,i);
    siva_verify_next_state_vector(target_vector,input_vector,seed_vector);
    target_vector = seed_vector;
  }
      
  succeed = 1;
  
 end:
  for(i=0;i<array_n(stored_state_vector_array);i++){
    var_set_free(array_fetch(var_set_t *,stored_state_vector_array,i));    
  }
  array_free(stored_state_vector_array);
  array_free(distance_array);
  return succeed;
}

