/**CFile**********************************************************************

  FileName                   [siva_symbolic_sim.c]

  PackageName                [siva]

  Synopsis                   [symbolic simulation]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]

******************************************************************************/

#include "siva.h"

static int siva_check_target_reached_disj(SIVA_NODE *symbolic_values,
                                          SIVA_NODE *nodes,
                                          int *goal_values,
                                          int num_symbolic_value,
                                          array_t *input_nodes_array,
                                          int sym_depth,
                                          SIVA_STATE_WRAPPER *state_wrapper,
                                          void *para,
                                          STORE_WITNESS_SRV store_witness_srv,
                                          PRINT_WITNESS_SRV print_witness_srv);

static int siva_check_target_reached(SIVA_NODE *symbolic_values,
                                     SIVA_NODE *nodes,
                                     int *goal_values,
                                     int num_symbolic_value,
                                     array_t *input_nodes_array,
                                     int sym_depth,
                                     SIVA_STATE_WRAPPER *state_wrapper,
                                     void *para,
                                     STORE_WITNESS_SRV store_witness_srv,
                                     PRINT_WITNESS_SRV print_witness_srv);
static void siva_sym_verify_tautology(SIVA_NODE target,
                                      int taut_value,
                                      var_set_t *seed_state_vector,
                                      int depth);
static int siva_symbolic_fixed_point(SIVA_NODE target, int goal_value);

static array_t* siva_input_nodes_array;

/*--------------------------------------------------------------------------*/
void
siva_init_symbolic_sim(void)
{
  siva_input_nodes_array = array_alloc(SIVA_NODE *,16);
  array_insert_last(SIVA_NODE*,siva_input_nodes_array,
                    siva_primary_inputs);
}

/*--------------------------------------------------------------------------*/
void
siva_quit_symbolic_sim(void)
{
  int i;
  for(i=1;i<array_n(siva_input_nodes_array);i++) {
    SIVA_NODE* input_nodes = array_fetch(SIVA_NODE*,siva_input_nodes_array,i);
    FREE(input_nodes);
  }
  array_free(siva_input_nodes_array);
}

/*
  --------------------------------------------------------------------------
  Garbage collect bck nodes
  --------------------------------------------------------------------------
*/

void
siva_gc_symbolic_sim(void)
{
  int i;
  SIVA_NODE p;
  for(i=1;i<array_n(siva_input_nodes_array);i++) {
    SIVA_NODE* input_nodes = array_fetch(SIVA_NODE*,siva_input_nodes_array,i);
    FREE(input_nodes);
  }
  array_free(siva_input_nodes_array);
  siva_input_nodes_array = array_alloc(SIVA_NODE *,16);
  array_insert_last(SIVA_NODE*,siva_input_nodes_array,
                    siva_primary_inputs);

  SIVA_VERB2(bck_stat(siva_output_file));
  if ((p = bck_next_node(siva_last_network_node)) != BCK_NULL) {
    bck_release_node(p);
  }
  SIVA_VERB2(bck_stat(siva_output_file));  
}

/*
  --------------------------------------------------------------------------
  Invoke symbolic simulation from the initial state
  --------------------------------------------------------------------------
*/

void
siva_symbolic_sim(void)
{
  register int i;
  int num_target;
  SIVA_NODE *targets;
  SIVA_STATE_WRAPPER wrapper;
  SIVA_STATE_WRAPPER *state_wrapper;
  SIVA_NODE goal;
  int goal_value;
  int reachable;
  int fixed_point;

  
  wrapper.state = siva_initial_state;
  wrapper.state_vector = siva_initial_state_vector;
  state_wrapper = &wrapper;
  if (siva_num_zero_target + siva_num_one_target == 0) {
    SIVA_VERB1_ARG1("No targets to resolve\n");
    return;
  }
  siva_init_atpg_solver();
  siva_target_stat();
  num_target = siva_num_zero_target+siva_num_one_target;
  targets = ALLOC(SIVA_NODE,num_target);  
  siva_fill_targets(targets);
  /* testing the unreachable solve mode */

  for(i=0;i<num_target;i++) {
    goal_value = bck_is_flag_set(targets[i],siva_target_one_flag);
    goal = bck_get_regular(targets[i]);


    /* reachable ?*/
    SIVA_VERB1_ARG3("Checking reachability from initial state of target (%d = %d)..\n",
                    bck_get_index(goal),goal_value);    
    reachable = siva_symbolic_sim1(&goal,&goal_value,1,
                                   state_wrapper,1,
                                   NULL,NULL,
                                   siva_atpg_print_witness_srv);
    if (reachable == 1) {
      SIVA_VERB1_ARG3("Target (%d=1) is reachable\n",
                      goal,goal_value );
    }             
    SIVA_VERB1_ARG3("Checking reachability from initial state of target (%d = %d)...done\n",
                    bck_get_index(goal),goal_value);  
    siva_gc_symbolic_sim();

    fixed_point = 0;
    if (reachable == 0) {
      /* if not reachble, then fix point? */
      if (siva_sym_check_unreachable) {

        SIVA_VERB1_ARG3("Checking unreachability target (%d = %d)..\n",
                        bck_get_index(goal),goal_value);    
        fixed_point = siva_symbolic_fixed_point(goal,goal_value);
        if (fixed_point == 1) {
          SIVA_VERB1_ARG3("Target (%d=1) is un-reachable\n",
                          goal,goal_value );
        }             
        siva_gc_symbolic_sim();
        SIVA_VERB1_ARG3("Checking unreachability target (%d = %d)..done\n",
                        bck_get_index(goal),goal_value);            
      }
    }
    if (reachable == 1 || fixed_point == 1) {
      if (bck_is_flag_set(targets[i],siva_target_one_flag)) {
        bck_reset_flag(targets[i],siva_target_one_flag);
        siva_num_one_target--;
      }else if (bck_is_flag_set(targets[i],siva_target_zero_flag)) {
        bck_reset_flag(targets[i],siva_target_zero_flag);
        siva_num_zero_target--;
      }
    }
    if (fixed_point) {
      if (goal_value) {
        bck_set_flag(targets[i],siva_constant_zero_flag);
      }else {
        bck_set_flag(targets[i],siva_constant_one_flag);        
      }          
    }
  }

  FREE(targets);
  siva_target_stat();
  siva_quit_atpg_solver();
}

/*
  --------------------------------------------------------------------------
  Symbolic simulation from a given state.
  If conjunction_mode = 1, then the targets are checked in conjunction;
  the return value is  equal to reachable.
  If conjunction_mode = 0, then the targets are solved one at a time;
  the return value is how many remained. the goal_values is changed to 2,
  to reflect the target is reached.
  The symbolic values are represented in structure:
  + Structures are not order dependent
  - Might have redundancy, but bck might reduce some     
  Use of SAT to work on the rest to check if they are constant
  TODO:   Use of rand cycle simulation to find non-constant nodes.
  --------------------------------------------------------------------------
*/
int
siva_symbolic_sim1(SIVA_NODE* targets, int *target_values,
                   int num_target, SIVA_STATE_WRAPPER *state_wrapper,
                   int conjunction_mode,
                   void *para,
                   STORE_WITNESS_SRV store_witness_srv,
                   PRINT_WITNESS_SRV print_witness_srv)                   
{
  register int i,k,j;
  SIVA_NODE *in_h;
  SIVA_NODE *in_v;
  SIVA_NODE *srcs;
  SIVA_NODE *dests;
  SIVA_NODE *input_nodes;
  int reachable;
  int num_remained;
  int *goal_values;
  
  unsigned num_bck = 0;
  time_t start_time;
  time_t end_time;
  int current_depth;
  long bck_built_time;

  SIVA_NODE pre_images[(siva_symbolic_sim_depth+1)];

  SIVA_NODE goals[num_target];
  
  /* next state + target nodes */
  srcs = ALLOC(SIVA_NODE,siva_num_latch+num_target);
  /* symbolic values of next state + target nodes */
  dests = ALLOC(SIVA_NODE,siva_num_latch+num_target);
  /* latch and input nodes */
  in_v = ALLOC(SIVA_NODE,siva_num_latch+siva_num_primary_input);
  /* symbolic values of latch and input nodes */
  in_h = ALLOC(SIVA_NODE,siva_num_latch+siva_num_primary_input);


  /* solve in disjunction mode */
  if (!conjunction_mode) {
    /* goals and goal_values might change */
    memcpy(goals,targets,num_target*sizeof(SIVA_NODE));
    goal_values = ALLOC(int, num_target);
    memcpy(goal_values,target_values,num_target*sizeof(int));    
    num_remained = num_target;
  }
  for(i=0;i<siva_num_latch;i++){
    in_v[i] = bck_get_regular(siva_latch_nodes[i].latch);
    srcs[i] = siva_latch_nodes[i].input;
    if (var_set_get_elt(state_wrapper->state_vector,i)){
      in_h[i] = bck_create_one();
    }else {
      in_h[i] = bck_create_zero();
    }
  }
  for(k=0,i=siva_num_latch;i<siva_num_latch+num_target;i++,k++){
    SIVA_ASSERT(!bck_is_inverted(targets[k]),
                "Target should not be inverted");       
      srcs[i] = (target_values[k] == 0) ?
        bck_create_not(targets[k]) : targets[k];    
  }

  for(i=siva_num_latch,k=0;i<siva_num_latch+siva_num_primary_input;i++,k++){
    in_v[i] = siva_primary_inputs[k];
  }
  
  bck_compose_vec(dests,srcs,siva_num_latch+num_target,
                  in_v,in_h,siva_num_latch,NULL);

  if (conjunction_mode) {
    if (num_target == 1) {
      pre_images[0] = dests[siva_num_latch];
    }    
    reachable =
      siva_check_target_reached(dests+siva_num_latch,
                                targets,
                                target_values,num_target,
                                siva_input_nodes_array,
                                0,state_wrapper,para,
                                store_witness_srv,
                                print_witness_srv);
    if (reachable == 1 || reachable == -1) {
      goto end;
    }
    num_remained = num_target;
  }else {
    num_remained =
      siva_check_target_reached_disj(dests+siva_num_latch,
                                     goals,
                                     goal_values,num_remained,
                                     siva_input_nodes_array,
                                     0,state_wrapper,para,
                                     store_witness_srv,
                                     print_witness_srv);
    if (num_remained == 0) {
      goto end;
    }
    for(k=0;k<num_remained;k++){
      srcs[k+siva_num_latch] = (goal_values[k] == 0) ?
        bck_create_not(goals[k]) : goals[k];    
    }
    
  }

  for(i=1;i<=siva_symbolic_sim_depth;i++){

    for(k=0;k<siva_num_latch;k++){
      in_h[k] = dests[k];      
    }
    if (i<array_n(siva_input_nodes_array)) {
      input_nodes = array_fetch(SIVA_NODE*,siva_input_nodes_array,i);
    }else {
      SIVA_ASSERT(i==array_n(siva_input_nodes_array),"");
      input_nodes = ALLOC(SIVA_NODE,siva_num_primary_input);
      array_insert_last(SIVA_NODE*,siva_input_nodes_array,input_nodes);

      for(k=0;k<siva_num_primary_input;k++) {
        input_nodes[k] = bck_create_var();
      }      
    }
    /* partial search can be performed by fixing some
       inputs to constant; value can be chosen randomly
    */
    for(k=siva_num_latch,j=0;k<siva_num_primary_input+siva_num_latch;k++,j++) {
      in_h[k] = input_nodes[j];
    }
    
    SIVA_VERB2_ARG2("Symbolic simulation ..%d\n",i);
    bck_compose_vec(dests,srcs,siva_num_latch+num_remained,
                    in_v,in_h,siva_num_latch+siva_num_primary_input,NULL);
    
    SIVA_VERB2(bck_stat(siva_output_file));
    SIVA_VERB2(siva_target_stat());
    
    num_bck = siva_get_fanin_cone_size(dests,siva_num_latch+num_remained);
    SIVA_VERB2_ARG2("Fanin cone size = %d\n",num_bck);
    if (conjunction_mode) {
      if (num_target == 1) {
        SIVA_NODE shell;
        int j;
        pre_images[i] = shell = dests[siva_num_latch];
        for(j=0;j<i;j++){
          shell = siva_bck_and(shell,pre_images[j],1,0);
        }
        dests[siva_num_latch] = shell;        
      }      
      if (i < siva_symbolic_sim_lower_depth) continue;    
      reachable =
        siva_check_target_reached(dests+siva_num_latch,
                                  targets,
                                  target_values,num_target,
                                  siva_input_nodes_array,
                                  i,state_wrapper,para,
                                  store_witness_srv,
                                  print_witness_srv);
      if (reachable == 1 || reachable == -1) {
        goto end;
      }
    }else {
      num_remained =
        siva_check_target_reached_disj(dests+siva_num_latch,
                                       goals,
                                       goal_values,num_remained,
                                       siva_input_nodes_array,
                                       i,state_wrapper,para,
                                       store_witness_srv,
                                       print_witness_srv);
      if (num_remained == 0) {
        goto end;
      }
      for(k=0;k<num_remained;k++){
        srcs[k+siva_num_latch] = (goal_values[k] == 0) ?
          bck_create_not(goals[k]) : goals[k];    
      }
      
    }
        
  }
 end:

  if (!conjunction_mode) {
    SIVA_NODE *unresolved_targets = goals;
    for(i=0;i<num_target;i++){
      SIVA_VERB2_ARG3("target %d value %d\n",targets[i],target_values[i]);
    }
    i=0;
    for(j=0;j<num_remained;j++) {
      while(i<num_target) {
        if (targets[i] == unresolved_targets[j]) {
          i++;
          break;
        }else {
          target_values[i] = 2;
          i++;
        }
      }
    }
    for(;i<num_target;i++) {
      target_values[i] = 2;
    } 
    FREE(goal_values);      
  }
        
  FREE(dests);
  FREE(srcs);
  FREE(in_v);
  FREE(in_h);
  SIVA_VERB2(bck_stat(siva_output_file));
  SIVA_VERB2(siva_target_stat());
  if (conjunction_mode) {
    if (reachable == 1) {
      return 1;
    }else {
      return 0;
    }
  }else {
    SIVA_VERB2_ARG2("Number remained = %d\n",num_remained);
    return num_remained;
  }
}


/*
  --------------------------------------------------------------------------
  Wrapper to check target in disjunction i.e., to check them
  independently.
  --------------------------------------------------------------------------
*/
static int
siva_check_target_reached_disj(SIVA_NODE *symbolic_values,
                               SIVA_NODE *nodes,
                               int *goal_values,
                               int num_symbolic_value,
                               array_t *input_nodes_array,
                               int sym_depth,
                               SIVA_STATE_WRAPPER *state_wrapper,
                               void *para,
                               STORE_WITNESS_SRV store_witness_srv,
                               PRINT_WITNESS_SRV print_witness_srv)
{
  int t = 0;
  int reachable;
  int num_remained = num_symbolic_value;
  while( t < num_remained) {
    reachable =
      siva_check_target_reached(symbolic_values+t,
                                nodes+t,
                                goal_values+t,1,
                                siva_input_nodes_array,
                                sym_depth,state_wrapper,
                                para,store_witness_srv,print_witness_srv);
    if (reachable != 1) {
      t++;
      continue;
    }    
    num_remained--;
    if (num_remained == 0) {
      return 0;
    }
    memmove(symbolic_values+t,symbolic_values+t+1,
            (num_remained-t) * sizeof(SIVA_NODE));
    memmove(nodes+t,nodes+t+1,
            (num_remained-t) * sizeof(SIVA_NODE));
    memmove(goal_values+t,goal_values+t+1,
            (num_remained-t) * sizeof(int));
  }
  return num_remained;  
}



/*
  --------------------------------------------------------------------------
  Wrapper to check target in conjunction.
  --------------------------------------------------------------------------
*/
static int
siva_check_target_reached(SIVA_NODE *symbolic_values,
                          SIVA_NODE *nodes,
                          int *goal_values,
                          int num_symbolic_value,
                          array_t *input_nodes_array,
                          int sym_depth,
                          SIVA_STATE_WRAPPER *state_wrapper,
                          void *para,
                          STORE_WITNESS_SRV store_witness_srv,
                          PRINT_WITNESS_SRV print_witness_srv)
{

  int reachable;
  SIVA_NODE targets[num_symbolic_value];
  SIVA_NODE goals[num_symbolic_value];
  int target_values[num_symbolic_value];
  int num_target = 0;
  int i;

  for(i=0;i<num_symbolic_value;i++) {
    if (bck_is_one(symbolic_values[i])) {
      SIVA_ERR_CHK1(siva_sym_verify_tautology(nodes[i],goal_values[i],
                                              state_wrapper->state_vector,
                                              sym_depth));
      continue;
    }
    if (bck_is_zero(symbolic_values[i])) {
      SIVA_ERR_CHK1(siva_sym_verify_tautology(nodes[i],1^goal_values[i],
                                              state_wrapper->state_vector,
                                              sym_depth));
      SIVA_VERB2_ARG2("Depth %d: Unreachable by symbolic construction\n",
                      sym_depth);
      return 0;
    }
    targets[num_target] = nodes[i];
    goals[num_target] = symbolic_values[i];
    target_values[num_target] = goal_values[i];
    num_target++;
  }

  if (num_target == 0) {
    SIVA_VERB2_ARG2("Depth %d: Trivially reachable. Checking constraints..\n",
                    sym_depth);
  }else {
    SIVA_VERB2_ARG4("Depth %d: Symbolic value %d for target %d\n",
                    sym_depth,goals[0],targets[0]);

  }
  
  reachable = siva_atpg_sym_solve(goals,
                                  targets,
                                  target_values,
                                  num_target,
                                  input_nodes_array,
                                  sym_depth,
                                  state_wrapper,
                                  para,store_witness_srv,
                                  print_witness_srv);
  SIVA_VERB3(util_print_cpu_stats(siva_output_file));


  return reachable;    
}

/*
  ---------------------------------------------------------------------------
  Verify that target is tautology as discovered by bck graph construction
  ---------------------------------------------------------------------------
*/

static void
siva_sym_verify_tautology(SIVA_NODE target,
                          int taut_value,
                          var_set_t *seed_state_vector,
                          int depth)
{
  int i;
  int value;
  unsigned curr_data[siva_latch_word_size];
  unsigned next_data[siva_latch_word_size];
  var_set_t curr_state_vector;
  var_set_t next_state_vector;

  siva_set_state_vector(&curr_state_vector,curr_data);
  siva_set_state_vector(&next_state_vector,next_data);

  var_set_clear(&curr_state_vector);
  var_set_clear(&next_state_vector);
  
  var_set_assign(&curr_state_vector,seed_state_vector);
  for(i=0;i<depth;i++){
    siva_sim_next_state_vector(&next_state_vector,NIL(var_set_t),
                               &curr_state_vector);
    var_set_assign(&curr_state_vector,&next_state_vector);
  }
  siva_evaluate_node(&target,&value,1,NIL(var_set_t),&curr_state_vector);
  SIVA_ASSERT(value==taut_value,
              "Taut value does not match");
  SIVA_VERB2_ARG1("Taut value verified\n");
}


/*
  --------------------------------------------------------------------------
  Return 1 if target is not reachable \until siva_symbolic_sim_depth.
  The target should not be reachable from initial state
  upto siva_symbolic_depth.
  --------------------------------------------------------------------------
*/

static int
siva_symbolic_fixed_point(SIVA_NODE target, int target_value)
{
  register int i,k,j;
  SIVA_NODE *in_h;
  SIVA_NODE *in_v;
  SIVA_NODE *srcs;
  SIVA_NODE *dests;
  SIVA_NODE *input_nodes;
  int reachable;
  int num_remained;
  int goal_value;
  SIVA_NODE goal;
  SIVA_NODE shell;
  unsigned num_bck = 0;
  time_t start_time;
  time_t end_time;
  int current_depth;
  SIVA_NODE pre_images[siva_symbolic_sim_depth+1];
  
  
  /* next state + the target node */
  srcs = ALLOC(SIVA_NODE,siva_num_latch+1);
  /* symbolic values of next state + the target node */
  dests = ALLOC(SIVA_NODE,siva_num_latch+1);
  /* latch and input nodes */
  in_v = ALLOC(SIVA_NODE,siva_num_latch+
               siva_num_primary_input);
  /* symbolic values of latch and input nodes */
  in_h = ALLOC(SIVA_NODE,siva_num_latch+
               siva_num_primary_input);

  SIVA_ASSERT(!bck_is_inverted(target),
              "Target should not be inverted");

  shell = target;  
  if (!target_value) {
    shell = bck_create_not(shell);
  }
  pre_images[0] = shell;

  srcs[siva_num_latch] = shell;  
  goal = bck_get_regular(shell);
  goal_value = 1^bck_is_inverted(shell);
  reachable =
    siva_atpg_is_satisfiable(&goal,&goal_value,1);
  if (reachable == 0) {
    current_depth = 0;
    goto end;
  }
  
  for(i=0;i<siva_num_latch;i++){
    in_v[i] = bck_get_regular(siva_latch_nodes[i].latch);
    srcs[i] = siva_latch_nodes[i].input;
    dests[i] = siva_latch_nodes[i].input;
  }
  for(i=siva_num_latch,k=0;i<siva_num_latch+siva_num_primary_input;i++,k++){
    in_v[i] = siva_primary_inputs[k];
  }
  for(i=1;i<=siva_symbolic_sim_depth;i++){
    SIVA_NODE *ptr1;

    for(k=0;k<siva_num_latch;k++){
      in_h[k] = dests[k];      
    }
    if (i<array_n(siva_input_nodes_array)) {
      input_nodes = array_fetch(SIVA_NODE*,siva_input_nodes_array,i);
    }else {
      SIVA_ASSERT(i==array_n(siva_input_nodes_array),"");
      input_nodes = ALLOC(SIVA_NODE,siva_num_primary_input);
      array_insert_last(SIVA_NODE*,siva_input_nodes_array,input_nodes);

      for(k=0;k<siva_num_primary_input;k++) {
        input_nodes[k] = bck_create_var();
      }      
    }
    ptr1 = in_h + siva_num_latch;
    for(k=0;k<siva_num_primary_input;k++){
      ptr1[k] = input_nodes[k];
    }
    
    SIVA_VERB2_ARG2("Symbolic simulation ..%d\n",i);
    bck_compose_vec(dests,srcs,siva_num_latch+1,
                    in_v,in_h,siva_num_latch+siva_num_primary_input,NULL);
    num_bck = siva_get_fanin_cone_size(dests,siva_num_latch+1);
    SIVA_VERB2_ARG2("Fanin cone size = %d\n",num_bck);

    
    SIVA_VERB2(bck_stat(siva_output_file));
    SIVA_VERB2(siva_target_stat());

    pre_images[i] = shell = dests[siva_num_latch];
    shell = pre_images[i];
    for(k=0;k<i;k++){
      SIVA_ASSERT(!bck_is_one(pre_images[k]),
                  "Target already reached?\n");
      shell = siva_bck_and(shell,pre_images[k],1,0);
    }
    goal = bck_get_regular(shell);
    goal_value = 1 ^ bck_is_inverted(shell);
    reachable =
      siva_atpg_is_satisfiable(&goal,&goal_value,1);
    if (reachable == 0) {
      current_depth = i;
        goto end;
    }
  }  
 end:        
  FREE(dests);
  FREE(srcs);
  FREE(in_v);
  FREE(in_h);
  SIVA_VERB2(bck_stat(siva_output_file));
  SIVA_VERB2(siva_target_stat());
  if (reachable == 0) {
    return 1;
  }else {
    return 0;
  }
}
