/*CFile**********************************************************************

  FileName                   [siva_util.c]

  PackageName                [siva_core]

  Synopsis                   [misc commands]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]
******************************************************************************/
#include "siva.h"
#include "math.h"


static void **siva_stack;
static unsigned siva_num_stack_allocated;
static unsigned siva_num_stack;
static unsigned siva_static_aux;

static int siva_latch_cmp(const void *latch_node_p1, const void *latch_node_p2);
static void siva_assign_level(SIVA_NODE latch);
static int siva_count_flag_srv(SIVA_NODE node, void *para);

/*---------------------------------------------------------------------------*/
int
siva_target_count_srv(SIVA_NODE node, void *para)
{
  SIVA_CONV conv;
  if (!bck_is_canonical(node)){
    return 1;
  }

  if (bck_is_flag_set(node,siva_target_one_flag)){
     siva_num_one_target++;
  }
  if (bck_is_flag_set(node,siva_target_zero_flag)){
     siva_num_zero_target++;
  }
  return 1;
}

/*---------------------------------------------------------------------------*/
void
siva_fill_fanin_cone(array_t *node_array, SIVA_COMPLEX_NODE *complex_node)
{  
  
  register int i;

  if (complex_node->visited)                               return;
  if (complex_node->tbl.type == CONSTANT ||
      complex_node->tbl.type == INPUT || complex_node->tbl.type == LATCH){
    array_insert_last(SIVA_COMPLEX_NODE*,node_array,complex_node);
    complex_node->visited = 1;
    return;
  }                   
  for(i=0;i<complex_node->tbl.num_input;i++){
    siva_fill_fanin_cone(node_array,complex_node->tbl.input[i]);
  }
  array_insert_last(SIVA_COMPLEX_NODE*,node_array,complex_node);
  complex_node->visited = 1;
  return;
}

/*---------------------------------------------------------------------------*/
void
siva_init_witness(void)
{
  int i;
  siva_input_vector = var_set_new(siva_num_primary_input);
  /* var_set: code hacking */
  siva_input_vector_data = siva_input_vector->data;
  siva_state_vector = var_set_new(siva_num_latch);
  siva_latch_word_size = ((siva_num_latch-1)/32) + 1;
  /* not sure of the position of the following code */
  siva_state_word_vector = var_set_word_new(siva_num_latch);
  siva_latch_inversion_bytes = ALLOC(unsigned char,siva_num_latch);

  for(i=0;i<siva_num_latch;i++){
    if(bck_is_inverted(siva_latch_nodes[i].input)){
      siva_latch_inversion_bytes[i] = 0xff;
    }else{
      siva_latch_inversion_bytes[i] = 0;
    }
  }
}

/*---------------------------------------------------------------------------*/
void
siva_quit_witness(void)
{
  FREE(siva_latch_inversion_bytes);
  var_set_word_free(siva_state_word_vector);
  var_set_free(siva_state_vector);
  /* var_set: code hacking */
  siva_input_vector->data = siva_input_vector_data;
  var_set_free(siva_input_vector);
}



/*---------------------------------------------------------------------------*/
SIVA_NODE *
siva_alloc_node(int size)
{
  return ALLOC(SIVA_NODE,size);

}
/*---------------------------------------------------------------------------*/
void                                   
siva_free_node(SIVA_NODE *nodes)
{
  FREE(nodes);

}

/*---------------------------------------------------------------------------*/
/* allocates latch nodes ; initializes                                       */
SIVA_LATCH_NODE *
siva_alloc_latch_node(int size)
{
  SIVA_LATCH_NODE  *latch_nodes = ALLOC(SIVA_LATCH_NODE,size);
  register int i;
  for(i=0;i<size;i++){
    latch_nodes[i].latch = bck_create_null();
    latch_nodes[i].input = bck_create_null();
  }
  return latch_nodes;
}


/*---------------------------------------------------------------------------*/
/* allocate string                                                          */
char *
siva_alloc_copy_string(char *string)
{
  char *new_string = ALLOC(char , strlen(string)+1);
  strcpy(new_string,string);
  *(new_string+strlen(string)) = '\0';
  return new_string;
}

/*---------------------------------------------------------------------------*/
/* allocates string of size n                                                */
char *
siva_alloc_n_copy_string(char *string,int size)
{
  char *new_string = ALLOC(char , size+1);
  strncpy(new_string,string,size);
  *(new_string+size) = '\0';
  return new_string;
}

/*---------------------------------------------------------------------------*/
void
siva_free_string(char *string)
{
  FREE(string);
}

/*---------------------------------------------------------------------------*/
void
siva_free_latch_node(SIVA_LATCH_NODE *latch_nodes)
{
  FREE(latch_nodes);
}

/*
  ---------------------------------------------------------------------------
  Partial Order of latches: 
  if L1 is latch input to L2, then L2 < L1.
  Caveat: Latch cycle is not allowed. A Latch cycle exists if there
  is a path from latch to its input that goes through only latches and inverters.
  ---------------------------------------------------------------------------
*/
void
siva_order_latch_nodes(void)
{
  register int i;
  SIVA_CONV conv;

  siva_global_aux = bck_alloc_aux(siva_init_aux);
  siva_static_aux = bck_alloc_aux(siva_init_aux);
#ifndef GPROF
  if (siva_verbosity_level>=3){
    fprintf(siva_output_file,"latch nodes before order\n");
    for(i=0;i<siva_num_latch;i++){
      fprintf(siva_output_file,"%d: bck = %d(%d)\n",i,
	      bck_get_index(siva_latch_nodes[i].latch),
	      bck_get_index(siva_latch_nodes[i].input));
    }
  }
#endif
  /*
    Catch: if latches are in loop (with inverters possibly) then
    by default they get 0 level. They are positioned in the begining.
  */
  for(i=0;i<siva_num_latch;i++){
    if (bck_get_index(siva_latch_nodes[i].latch) ==
	bck_get_index(siva_latch_nodes[i].input)){
      continue;
    }

    if ( bck_is_flag_set(siva_latch_nodes[i].input,siva_latch_flag)){      
      conv.ptr = bck_get_aux(siva_latch_nodes[i].input,
                             siva_static_aux);
      if (conv.array == NIL(array_t)){
        conv.array = array_alloc(SIVA_NODE,1);
        bck_set_aux(siva_latch_nodes[i].input,siva_static_aux,conv.ptr);
      }
      array_insert_last(SIVA_NODE,conv.array,siva_latch_nodes[i].latch);
    }
  }

  /* The requirement of first condition:
     1) latch 'a' has its input 'a' (with inverter)
     2) latch 'b' has its input 'a'
     Then b should be ordered before a.

     Caveat: latch in loops will be create cycle. This
     routine will loop infintely. However, design with malicious
     loop can be simplified easily. We don't deal such cases
  */
     
  for(i=0;i<siva_num_latch;i++){
    if ((bck_get_index(siva_latch_nodes[i].latch) ==
	bck_get_index(siva_latch_nodes[i].input)) ||
       (!bck_is_flag_set(siva_latch_nodes[i].input,siva_latch_flag))){

       siva_assign_level(siva_latch_nodes[i].latch);
    }
  }
      
  qsort(siva_latch_nodes,siva_num_latch,sizeof(SIVA_LATCH_NODE),
        siva_latch_cmp);

  for(i=0;i<siva_num_latch;i++){
    SIVA_CONV conv;
    conv.ptr = bck_get_aux(siva_latch_nodes[i].latch,siva_global_aux);
    if (!conv.int_value){
      bck_set_flag(siva_latch_nodes[i].latch,siva_loop_latch_flag);
      fprintf(siva_output_file,"Loop latch %d\n",
              bck_get_regular(siva_latch_nodes[i].latch));
    }
  }
#ifndef GPROF
  if (siva_verbosity_level>=3){
    fprintf(siva_output_file,"latch nodes after order\n");
    for(i=0;i<siva_num_latch;i++){
      fprintf(siva_output_file,"%d: bck = %d(%d)\n",i,
	      bck_get_index(siva_latch_nodes[i].latch),
	      bck_get_index(siva_latch_nodes[i].input));
    }
    getchar();
  }
#endif
  for(i=0;i<siva_num_latch;i++){
    if (bck_get_index(siva_latch_nodes[i].latch) ==
	bck_get_index(siva_latch_nodes[i].input)){
      continue;
    }
    if ( bck_is_flag_set(siva_latch_nodes[i].input,siva_latch_flag)){
      conv.ptr = bck_get_aux(siva_latch_nodes[i].input,
                               siva_static_aux);
      if (conv.array != NIL(array_t)){ 
        array_free(conv.array);
        conv.array = NIL(array_t);
        bck_set_aux(siva_latch_nodes[i].input,siva_static_aux,conv.ptr);
      }
    }
  }
  bck_release_aux(siva_static_aux,NULL);
  bck_release_aux(siva_global_aux,NULL);


}
/*---------------------------------------------------------------------------*/
static void
siva_assign_level(SIVA_NODE latch)
{

  /* latch has a unique input; hence it won't be revisted */
  /* Caveat: latch in loops will be create cycle. This
     routine will loop infintely. However, design with malicious
     loop can be simplified easily. We don't deal such cases
  */
  SIVA_CONV conv_1;
  int level = 0;  
  register int i;

  conv_1.ptr = bck_get_aux(latch,siva_static_aux);

  if (conv_1.array == NIL(array_t)){
    conv_1.value = 1;
    bck_set_aux(latch,siva_global_aux,conv_1.ptr);
    return;
  }

  for(i = 0; i < array_n(conv_1.array);i++){
    SIVA_NODE latch_out = array_fetch(SIVA_NODE,conv_1.array,i);
    SIVA_CONV conv_2;
    siva_assign_level(latch_out);
    conv_2.ptr = bck_get_aux(latch_out,siva_global_aux);
    if ( conv_2.value > level ) {
       level = conv_2.value;
    }    
  }
  conv_1.value = level+1;
  bck_set_aux(latch,siva_global_aux,conv_1.ptr);
}
/*---------------------------------------------------------------------------*/
static int
siva_latch_cmp(const void *latch_node_p1, const void *latch_node_p2)
{
  SIVA_LATCH_NODE latch_node_1 = *(SIVA_LATCH_NODE *)latch_node_p1;       
  SIVA_LATCH_NODE latch_node_2 = *(SIVA_LATCH_NODE *)latch_node_p2;       
  SIVA_CONV conv_1,conv_2;
  
  conv_1.ptr = bck_get_aux(latch_node_1.latch,siva_global_aux);
  conv_2.ptr = bck_get_aux(latch_node_2.latch,siva_global_aux);
  
  return (conv_1.value - conv_2.value);
}       

/*---------------------------------------------------------------------------*/
void
siva_init_stack(void)
{
  siva_stack = NULL;
  siva_num_stack_allocated = 512;
  siva_num_stack = 0;
  siva_stack = REALLOC(void *,siva_stack,
                       siva_num_stack_allocated);
}
/*---------------------------------------------------------------------------*/
void
siva_quit_stack(void)
{ 
  FREE(siva_stack);
  siva_stack = NULL;
  siva_num_stack_allocated = 0;
  siva_num_stack = 0;
}
/*---------------------------------------------------------------------------*/
void
siva_stack_push(void *entry)
{
  if (siva_num_stack >=   siva_num_stack_allocated ) {
    siva_num_stack_allocated += siva_num_stack_allocated >> 2;
    siva_stack = REALLOC(void *,siva_stack,
                         siva_num_stack_allocated);
  }
  siva_stack[siva_num_stack] = entry;
  siva_num_stack++;
}
/*---------------------------------------------------------------------------*/
void *
siva_stack_pop(void)
{
  SIVA_ASSERT((siva_num_stack > 0),"poping empty stack");
  
  return siva_stack[--siva_num_stack];

}
/*---------------------------------------------------------------------------*/
int
siva_stack_empty(void)
{  
  return (siva_num_stack == 0);
}

/*---------------------------------------------------------------------------*/
void
siva_init_aux(void **aux_field)
{
  (*aux_field) = NIL(void *);
}
/*---------------------------------------------------------------------------*/
int
siva_clear_aux_srv(SIVA_NODE in,void *aux)
{
  SIVA_CONV conv_1,conv_2;
  conv_1.ptr = aux;
  conv_2.value = 0;
  bck_set_aux(in,conv_1.value,conv_2.ptr);
  return 1;
}


/*---------------------------------------------------------------------------*/
char *
siva_strrchr(char *string, char c, int nth_last)
{
  register int i;
  register int j=-1;
  char *ptr= NIL(char);

  SIVA_ASSERT((nth_last>=0),"nth_last should be non-negative");

  for(i=strlen(string)-1;i>=0;i--){
    if (string[i] == c){
      j++;
      if (j== nth_last){
        ptr = &string[i];
      }
    }
  }
  return ptr;
}


/*---------------------------------------------------------------------------*/
unsigned
siva_string_hash(char *string)
{
  register unsigned hash_value=0;
  register unsigned c;

  while ((c = *string++) != '\0') {
    hash_value = hash_value*997 + c;
  }
  return hash_value;
}



/*---------------------------------------------------------------------------*/
void
siva_fill_targets(SIVA_NODE *targets)
{
  SIVA_NODE *bck_ptr = targets;
  SIVA_CONV conv;
  
  conv.bck_ptr_ptr      = &bck_ptr;
  bck_do_in_topological_order(conv.ptr,siva_fill_target_srv);
}

/*--------------------------------------------------------------------------*/
int
siva_fill_target_srv(SIVA_NODE node, void *para)
{
  if (!bck_is_canonical(node)){
    return 1;
  }
  if (bck_is_constant(node)){
    return 1;
  }
  if(bck_is_flag_set(node,siva_target_one_flag)){
    SIVA_CONV conv;
    conv.ptr = para;
    **conv.bck_ptr_ptr = node;  
    (*conv.bck_ptr_ptr)++;
  }
  if(bck_is_flag_set(node,siva_target_zero_flag)){
    SIVA_CONV conv;
    conv.ptr = para;
    **conv.bck_ptr_ptr = node;  
    (*conv.bck_ptr_ptr)++;
  }
  return 1;
}


/*--------------------------------------------------------------------------*/
static int
siva_count_flag_srv(SIVA_NODE node, void *para)
{
  unsigned flag = *(unsigned *)para;
  unsigned *size_p = (((unsigned *)para) + 1);
  if (bck_is_flag_set(node,flag)) {
    (*size_p)++;
  }
  return 1;
}

/*--------------------------------------------------------------------------*/
int
siva_get_fanin_cone_size(SIVA_NODE *nodes, int num_node)
{
  int k;
  unsigned para_lists[2];
  unsigned coi_flag = bck_alloc_flag();
  bck_reset_all_flag(coi_flag);
  for(k=0;k<num_node;k++){
    bck_set_fanin_cone_flag(nodes[k],coi_flag);
  }
  para_lists[0] = coi_flag;
  para_lists[1] = 0;
  bck_do_in_topological_order(para_lists,siva_count_flag_srv);
  bck_release_flag(coi_flag);
  return para_lists[1];
}

/*--------------------------------------------------------------------------*/
void
siva_set_state_vector(var_set_t *vector, unsigned *data)
{
  vector->n_elts = siva_num_latch;
  vector->n_words = siva_latch_word_size;
  vector->data = data;
}

/*--------------------------------------------------------------------------*/
int
siva_get_latch_id(SIVA_NODE latch)
{
  SIVA_CONV conv;
  if (bck_is_flag_set(latch,siva_latch_flag)){
    conv.ptr = bck_get_aux(latch,siva_latch_id_aux);
  }else {
    SIVA_VERB0_ARG2("%d is not latch\n",latch);
  }
  return conv.bck;
}

/*--------------------------------------------------------------------------*/
SIVA_NODE
siva_get_latch_from_id(int id)
{
  assert(id<siva_num_latch);
  return bck_get_regular(siva_latch_nodes[id].latch);
}

/*--------------------------------------------------------------------------*/
SIVA_NODE
siva_get_latch_input_from_id(int id)
{
  assert(id<siva_num_latch);
  return siva_latch_nodes[id].input;
}
  

/*--------------------------------------------------------------------------*/
void
siva_set_input_vector(var_set_t *vector, unsigned *data)
{
  vector->n_elts = siva_num_primary_input;
  vector->n_words = siva_primary_input_word_size;
  vector->data = data;
}

/*-------------------------------------------------------------------------*/
void
siva_print_hotness(SIVA_HOT_CONDITION *hot_conds)
{
  register int i;    
  for(i=0;i<siva_num_latch;i++){
    if (hot_conds[i].strong_one){
      fprintf(siva_output_file,"latch %d: strong_one\n",
              bck_get_index(siva_latch_nodes[i].latch));
    }else if(hot_conds[i].strong_zero){
      fprintf(siva_output_file,"latch %d: strong_zero\n",
              bck_get_index(siva_latch_nodes[i].latch));
    }else if(hot_conds[i].weak_one){
     fprintf(siva_output_file,"latch %d: weak_one\n",
              bck_get_index(siva_latch_nodes[i].latch));
    }else if(hot_conds[i].weak_zero){
      fprintf(siva_output_file,"latch %d: weak_zero\n",
              bck_get_index(siva_latch_nodes[i].latch));
    }else if(hot_conds[i].weak_conflict){
      fprintf(siva_output_file,"latch %d: weak_conflict\n",
              bck_get_index(siva_latch_nodes[i].latch));
    }
  }      
}


